<?php

namespace NF\Tickets\Admin\Controller;

use XF\Admin\Controller\AbstractController;
use XF\Mvc\Entity\Finder;
use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\AbstractReply;
use XF\Repository\User as UserRepo;
use XF\Util\Arr as ArrUtil;

use function count, is_array, strlen;

class BanEmail extends AbstractController
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    /**
     * @param Finder $finder
     * @param array $filters
     */
    protected function applyFilters(Finder $finder, array $filters)
    {
        $bannedEmails = $filters['banned_email'] ?? '';
        if (strlen($bannedEmails))
        {
            $finder->where('banned_email', 'LIKE', $bannedEmails, '%?%');
        }

        $creatorIds = $filters['create_user_ids'] ?? [];
        if (is_array($creatorIds) && count($creatorIds))
        {
            $finder->where('create_user_id', $creatorIds);
        }

        $reason = $filters['reason'] ?? '';
        if (strlen($reason))
        {
            $finder->where('reason', 'LIKE', $finder->escapeLike($reason, '%?%'));
        }

        $createDate = $filters['create_date_start'] ?? 0;
        if ($createDate)
        {
            $finder->where('create_date', '>=', $createDate);
        }
    }

    /**
     * @param bool $forRedirect
     * @return array
     */
    protected function getFilterInput(bool $forRedirect) : array
    {
        $filters = [];

        $input = $this->filter([
            'banned_email' => 'str',
            'creators' => 'str',
            'reason' => 'str',
            'create_date_start' => 'datetime'
        ]);

        if (strlen($input['banned_email']))
        {
            $filters['banned_email'] = $input['banned_email'];
        }

        if (strlen($input['creators']))
        {
            /** @var UserRepo $userRepo */
            $userRepo = $this->repository('XF:User');
            $users = $userRepo->getUsersByNames(ArrUtil::stringToArray($input['creators'], '/,\s*/'));

            if ($users->count())
            {
                $filters['creators'] = \implode(', ', \array_keys($users->groupBy('username'))) . ', ';

                if (!$forRedirect)
                {
                    $filters['create_user_ids'] = $users->keys();
                }
            }
        }

        if (strlen($input['reason']))
        {
            $filters['reason'] = $input['reason'];
        }

        if (!empty($input['create_date_start']))
        {
            $filters['create_date_start'] = $input['create_date_start'];
        }

        return $filters;
    }


    public function actionIndex(): AbstractReply
    {
        if ($this->filter('apply', 'bool'))
        {
            $filters = $this->getFilterInput(true);
            return $this->redirect($this->buildLink('tickets/ban-email', null, $filters));
        }
        $filters = $this->getFilterInput(false);

        $page = $this->filterPage();
        $perPage = 20;

        $order = $this->filter('order', 'str', 'create_date');
        $direction = $this->filter('direction', 'str', 'desc');

        $orderFields = [
            [$order, $direction]
        ];
        if ($order !== 'banned_email')
        {
            // If not already set, add this as a secondary sort because
            // majority of fields may be blank (especially legacy data)
            $orderFields[] = ['banned_email', 'asc'];
        }

        $emailBanFinder = $this->getBanningRepo()->findEmailBans()
                               ->with('User')
                               ->order($orderFields)
                               ->limitByPage($page, $perPage);
        $this->applyFilters($emailBanFinder, $filters);
        $total = $emailBanFinder->total();

        $this->assertValidPage($page, $perPage, $total, 'tickets/ban-email');

        $viewParams = [
            'emailBans' => $emailBanFinder->fetch(),
            'page' => $page,
            'perPage' => $perPage,
            'total' => $total,
            'filters' => $filters,
            'order' => $order,
            'direction' => $direction,
            'newEmail' => $this->em()->create('NF\Tickets:BanEmail')
        ];
        return $this->view('XF:Banning\Email\Listing', 'nf_tickets_ban_email_list', $viewParams);
    }

    public function actionAdd(): AbstractReply
    {
        $this->assertPostOnly();

        $this->getBanningRepo()->banEmail(
            $this->filter('email', 'str'),
            $this->filter('shadow_ban', 'bool'),
            $this->filter('reason', 'str')
        );
        return $this->redirect($this->buildLink('tickets/ban-email'));
    }

    public function actionDelete(): AbstractReply
    {
        $this->assertPostOnly();

        $deletes = $this->filter('delete', 'array-str');

        $emailBans = $this->em()->findByIds('NF\Tickets:BanEmail', $deletes);
        foreach ($emailBans AS $emailBan)
        {
            $emailBan->delete();
        }

        return $this->redirect($this->buildLink('tickets/ban-email'));
    }

    public function actionExport(): AbstractReply
    {
        $bannedEmails = $this->getBanningRepo()->findEmailBans();
        /** @var \XF\ControllerPlugin\Xml $xmlPlugin */
        $xmlPlugin = $this->plugin('XF:Xml');

        return $xmlPlugin->actionExport($bannedEmails, 'NF\Tickets:BannedEmails\Export');
    }

    public function actionImport(): AbstractReply
    {
        /** @var \XF\ControllerPlugin\Xml $xmlPlugin */
        $xmlPlugin = $this->plugin('XF:Xml');

        return $xmlPlugin->actionImport('tickets/ban-email', 'banned_emails', 'NF\Tickets:BannedEmails\Import');
    }

    protected function getBanningRepo(): \NF\Tickets\Repository\Banning
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Banning');
    }
}