<?php

namespace NF\Tickets\Admin\Controller;

use XF\Mvc\FormAction;
use XF\Mvc\ParameterBag;
use XF\Admin\Controller\AbstractController;

use function in_array;

/**
 * Class Category
 *
 * @package NF\Tickets\Admin\Controller
 */
class Category extends AbstractController
{
    /**
     * @param              $action
     * @param ParameterBag $params
     *
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    /**
     * @return \XF\ControllerPlugin\AbstractPlugin|\NF\Tickets\ControllerPlugin\CategoryTree
     */
    protected function getCategoryTreePlugin()
    {
        return $this->plugin('NF\Tickets:CategoryTree');
    }

    /**
     * @return \XF\Mvc\Reply\View
     */
    public function actionIndex()
    {
        return $this->getCategoryTreePlugin()->actionList([
            'permissionContentType' => 'nf_tickets_category'
        ]);
    }

    /**
     * @param \NF\Tickets\Entity\Category $category
     *
     * @return \XF\Mvc\Reply\View
     */
    protected function categoryAddEdit(\NF\Tickets\Entity\Category $category)
    {
        $categoryRepo = $this->getCategoryRepo();
        $categories = $categoryRepo->findCategoryList()->fetch();
        $categoryTree = $categoryRepo->createCategoryTree($categories);

        if ($category->thread_prefix_id && $category->ThreadForum)
        {
            $threadPrefixes = $category->ThreadForum->getPrefixesGrouped();
        }
        else
        {
            $threadPrefixes = [];
        }

        /** @var \NF\Tickets\Repository\TicketPrefix $prefixRepo */
        $prefixRepo = $this->repository('NF\Tickets:TicketPrefix');
        $availablePrefixes = $prefixRepo->findPrefixesForList()->fetch();
        $availablePrefixes = $availablePrefixes->pluckNamed('title', 'prefix_id');
        $prefixListData = $prefixRepo->getPrefixListData();

        /** @var \NF\Tickets\Repository\TicketField $fieldRepo */
        $fieldRepo = $this->repository('NF\Tickets:TicketField');
        $availableFields = $fieldRepo->findFieldsForList()->fetch();
        $availableFields = $availableFields->pluckNamed('title', 'field_id');

        /** @var \XF\Repository\Node $nodeRepo */
        $nodeRepo = $this->repository('XF:Node');
        $forumOptions = $nodeRepo->getNodeOptionsData(false, 'Forum');

        $viewParams = [
            'forumOptions' => $forumOptions,
            'threadPrefixes' => $threadPrefixes,
            'category' => $category,
            'categoryTree' => $categoryTree,

            'availableFields' => $availableFields,
            'availablePrefixes' => $availablePrefixes,

            'prefixGroups' => $prefixListData['prefixGroups'],
            'prefixesGrouped' => $prefixListData['prefixesGrouped'],
        ];
        return $this->view('NF\Tickets:Category\Edit', 'nf_tickets_category_edit', $viewParams);
    }

    /**
     * @param ParameterBag $parameterBag
     *
     * @return \XF\Mvc\Reply\View
     * @throws \XF\Mvc\Reply\Exception
     */
    public function actionEdit(ParameterBag $parameterBag)
    {
        $category = $this->assertCategoryExists($parameterBag->ticket_category_id);
        return $this->categoryAddEdit($category);
    }

    /**
     * @return \XF\Mvc\Reply\View
     */
    public function actionAdd()
    {
        /** @var \NF\Tickets\Entity\Category $category */
        $category = $this->em()->create('NF\Tickets:Category');
        $category->parent_category_id = $this->filter('parent_category', 'uint');

        return $this->categoryAddEdit($category);
    }

    /**
     * @param \NF\Tickets\Entity\Category $category
     *
     * @return FormAction
     */
    protected function categorySaveProcess(\NF\Tickets\Entity\Category $category)
    {
        $entityInput = $this->filter([
            'title' => 'str',
            'description' => 'str',
            'parent_category_id' => 'uint',
            'display_order' => 'uint',
            'thread_node_id' => 'uint',
            'thread_prefix_id' => 'uint',
            'require_prefix' => 'bool',
            'default_prefix_id' => 'uint',
            'priority_selection' => 'bool',
            'count_messages' => 'bool',
	        'allow_opening' => 'bool',
        ]);

        $username = $this->filter('anonymized_username', 'str');
        if ($username)
        {
            /** @var \XF\Entity\User $user */
            $user = $this->app()->em()->findOne('XF:User', ['username' => $username]);
            $entityInput['anonymous_user_id'] = $user ? $user->user_id : null;
        }

        if ($this->filter('enable_incoming_email', 'bool'))
        {
            $entityInput += $this->filter([
                'incoming_ticket_email_settings' => [
                    'address'    => 'str',
                    'connection' => [
                        'type'       => 'str',
                        'host'       => 'str',
                        'port'       => 'uint',
                        'username'   => 'str',
                        'password'   => 'str',
                        'encryption' => 'str'
                    ]
                ]
            ]);
        }
        else
        {
            $entityInput['incoming_ticket_email_settings'] = null;
        }

        $form = $this->formAction();
        $form->basicEntitySave($category, $entityInput);

        $prefixIds = $this->filter('available_prefixes', 'array-uint');
        $form->complete(function() use ($category, $prefixIds)
        {
            /** @var \NF\Tickets\Repository\CategoryPrefix $repo */
            $repo = $this->repository('NF\Tickets:CategoryPrefix');
            $repo->updateContentAssociations($category->ticket_category_id, $prefixIds);
        });

        if (!in_array($category->default_prefix_id, $prefixIds, true))
        {
            $category->default_prefix_id = 0;
        }

        $fieldIds = $this->filter('available_fields', 'array-str');
        $form->complete(function() use ($category, $fieldIds)
        {
            /** @var \NF\Tickets\Repository\CategoryField $repo */
            $repo = $this->repository('NF\Tickets:CategoryField');
            $repo->updateContentAssociations($category->ticket_category_id, $fieldIds);
        });

        return $form;
    }

    /**
     * @param ParameterBag $parameterBag
     *
     * @return \XF\Mvc\Reply\Redirect
     * @throws \XF\Mvc\Reply\Exception
     * @throws \XF\PrintableException
     */
    public function actionSave(ParameterBag $parameterBag)
    {
        $this->assertPostOnly();

        if ($parameterBag->ticket_category_id)
        {
            $category = $this->assertCategoryExists($parameterBag->ticket_category_id);
        }
        else
        {
            /** @var \NF\Tickets\Entity\Category $category **/
            $category = $this->em()->create('NF\Tickets:Category');
        }

        $this->categorySaveProcess($category)->run();

        return $this->redirect(
            $this->buildLink('tickets/categories') . $this->buildLinkHash($category->getEntityId())
        );
    }

    /**
     * @param ParameterBag $parameterBag
     *
     * @return \XF\Mvc\Reply\Redirect|\XF\Mvc\Reply\View
     */
    public function actionDelete(ParameterBag $parameterBag)
    {
        return $this->getCategoryTreePlugin()->actionDelete($parameterBag);
    }

    /**
     * @return \XF\Mvc\Reply\Redirect|\XF\Mvc\Reply\View
     */
    public function actionSort()
    {
        return $this->getCategoryTreePlugin()->actionSort();
    }

    /**
     * @return \NF\Tickets\ControllerPlugin\CategoryPermission
     */
    protected function getCategoryPermissionPlugin()
    {
        /** @var \NF\Tickets\ControllerPlugin\CategoryPermission $plugin */
        $plugin = $this->plugin('NF\Tickets:CategoryPermission');
        $plugin->setFormatters('NF\Tickets:Category\Permission%s', 'nf_tickets_category_permission_%s');
        $plugin->setRoutePrefix('tickets/categories/permissions');

        return $plugin;
    }

    /**
     * @param ParameterBag $params
     *
     * @return \XF\Mvc\Reply\View
     */
    public function actionPermissions(ParameterBag $params)
    {
        return $this->getCategoryPermissionPlugin()->actionList($params);
    }

    /**
     * @param ParameterBag $params
     *
     * @return \XF\Mvc\Reply\AbstractReply|\XF\Mvc\Reply\Redirect|\XF\Mvc\Reply\View
     */
    public function actionPermissionsEdit(ParameterBag $params)
    {
        return $this->getCategoryPermissionPlugin()->actionEdit($params);
    }

    /**
     * @param ParameterBag $params
     *
     * @return \XF\Mvc\Reply\Redirect
     */
    public function actionPermissionsSave(ParameterBag $params)
    {
        return $this->getCategoryPermissionPlugin()->actionSave($params);
    }

    /**
     * @param int     $id
     * @param array   $with
     * @param ?string $phraseKey
     * @throws \XF\Mvc\Reply\Exception
     * @return \NF\Tickets\Entity\Category
     */
    protected function assertCategoryExists(int $id, array $with = [], string $phraseKey = null): \NF\Tickets\Entity\Category
    {
        $with[] = 'AnonymousUser';
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->assertRecordExists('NF\Tickets:Category', $id, $with, $phraseKey);
    }

    protected function getCategoryRepo(): \NF\Tickets\Repository\Category
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Category');
    }

    protected function getTicketRepo(): \NF\Tickets\Repository\Ticket
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Ticket');
    }
}