<?php

namespace NF\Tickets\Admin\Controller;

use XF\Admin\Controller\AbstractController;
use XF\Mvc\ParameterBag;

class Escalation extends AbstractController
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    public function actionIndex()
    {
        $escalationRepo = $this->getEscalationRepo();
        $escalations = $escalationRepo->findEscalationsForList(false)->fetch();

        $viewParams = [
            'escalations' => $escalations,
        ];

        return $this->view('NF\Tickets:Escalation\List', 'nf_tickets_escalation_list', $viewParams);
    }

    public function escalationAddEdit(\NF\Tickets\Entity\Escalation $escalation)
    {
        $ticketCriteria = $this->app->criteria('NF\Tickets:Ticket', $escalation->ticket_criteria);

        $viewParams = [
            'escalation' => $escalation,

            'categories' => $this->getCategoryRepo()->getCategoryOptionsData(),
            'statuses'   => $this->getStatusRepo()->getStatusTitlePairs(),
            'priorities' => $this->getPriorityRepo()->getPriorityKeyValues(),

            'ticketCriteria' => $ticketCriteria,
        ];

        return $this->view('NF\Tickets:Escalation\Edit', 'nf_tickets_escalation_edit', $viewParams);
    }

    public function actionEdit(ParameterBag $params)
    {
        $condition = $this->assertEscalationExists($params->escalation_id);

        return $this->escalationAddEdit($condition);
    }

    public function actionAdd()
    {
        /** @var \NF\Tickets\Entity\Escalation $escalation */
        $escalation = $this->em()->create('NF\Tickets:Escalation');

        return $this->escalationAddEdit($escalation);
    }

    protected function escalationSaveProcess(\NF\Tickets\Entity\Escalation $escalation)
    {
        $entityInput = $this->filter([
            'escalation_title'       => 'str',
            'escalation_description' => 'str',
            'active'                 => 'bool',
            'execution_order'        => 'uint',

            'actions'         => 'array',
            'ticket_criteria' => 'array',
        ]);

        $entityInput['actions'] = \array_filter($entityInput['actions']);

        foreach (['message', 'alert'] AS $fieldSuffix)
        {
            if (!empty($entityInput['actions']["send_{$fieldSuffix}"])) // don't make changes to the stored body and from username
            {
                $entityInput = \array_merge($entityInput, $this->filter([
                    "{$fieldSuffix}_body" => 'str',
                    "{$fieldSuffix}_from_username" => 'str'
                ]));
            }
        }

        $form = $this->formAction();
        $form->basicEntitySave($escalation, $entityInput);

        return $form;
    }

    public function actionSave(ParameterBag $params)
    {
        $this->assertPostOnly();

        if ($params->escalation_id)
        {
            $condition = $this->assertEscalationExists($params->escalation_id);
        }
        else
        {
            /** @var \NF\Tickets\Entity\Escalation $condition */
            $condition = $this->em()->create('NF\Tickets:Escalation');
        }

        $this->escalationSaveProcess($condition)->run();

        return $this->redirect($this->buildLink('tickets/escalations'));
    }

    public function actionDelete(ParameterBag $params)
    {
        $escalation = $this->assertEscalationExists($params->escalation_id);

        /** @var \XF\ControllerPlugin\Delete $plugin */
        $plugin = $this->plugin('XF:Delete');

        return $plugin->actionDelete(
            $escalation,
            $this->buildLink('tickets/escalations/delete', $escalation),
            $this->buildLink('tickets/escalations/edit', $escalation),
            $this->buildLink('tickets/escalations'),
            $this->app->stringFormatter()->snippetString($escalation->escalation_title, 150)
        );
    }

    public function actionToggle()
    {
        /** @var \XF\ControllerPlugin\Toggle $plugin */
        $plugin = $this->plugin('XF:Toggle');

        return $plugin->actionToggle('NF\Tickets:Escalation');
    }

    /**
     * @param int     $id
     * @param array   $with
     * @param ?string $phraseKey
     * @throws \XF\Mvc\Reply\Exception
     * @return \NF\Tickets\Entity\Escalation
     */
    protected function assertEscalationExists(int $id, array $with = [], string $phraseKey = null): \NF\Tickets\Entity\Escalation
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->assertRecordExists('NF\Tickets:Escalation', $id, $with, $phraseKey);
    }

    protected function getCategoryRepo(): \NF\Tickets\Repository\Category
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Category');
    }

    protected function getEscalationRepo(): \NF\Tickets\Repository\Escalation
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Escalation');
    }

    protected function getStatusRepo(): \NF\Tickets\Repository\Status
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Status');
    }

    protected function getPriorityRepo(): \NF\Tickets\Repository\Priority
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Priority');
    }
}