<?php

namespace NF\Tickets\Admin\Controller;

use XF\Admin\Controller\AbstractController;
use XF\Mvc\FormAction;
use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\AbstractReply;

class Priority extends AbstractController
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    public function actionIndex()
    {
        $priorityRepo = $this->getPriorityRepo();
        $priorities = $priorityRepo->findPrioritiesForList(false)->fetch();

        $viewParams = [
            'priorities' => $priorities,
        ];

        return $this->view('NF\Tickets:Priority\List', 'nf_tickets_priority_list', $viewParams);
    }

    public function priorityAddEdit(\NF\Tickets\Entity\Priority $priority)
    {
        $viewParams = [
            'priority' => $priority,
        ];

        return $this->view('NF\Tickets:Priority\Edit', 'nf_tickets_priority_edit', $viewParams);
    }

    public function actionEdit(ParameterBag $params)
    {
        /** @noinspection PhpUndefinedFieldInspection */
        $condition = $this->assertPriorityExists($params->priority_id);

        return $this->priorityAddEdit($condition);
    }

    public function actionAdd()
    {
        /** @var \NF\Tickets\Entity\Priority $priority */
        $priority = $this->em()->create('NF\Tickets:Priority');

        return $this->priorityAddEdit($priority);
    }

    protected function prioritySaveProcess(\NF\Tickets\Entity\Priority $priority)
    {
        $entityInput = $this->filter([
            'display_order' => 'uint',
        ]);

        $form = $this->formAction();
        $form->basicEntitySave($priority, $entityInput);

        $titlePhrase = $this->filter('title', 'str');

        $form->validate(function (FormAction $form) use ($titlePhrase) {
            if ($titlePhrase === '')
            {
                $form->logError(\XF::phrase('please_enter_valid_title'), 'title');
            }
        });
        $form->apply(function () use ($titlePhrase, $priority) {
            $masterTitle = $priority->getMasterPhrase();
            $masterTitle->phrase_text = $titlePhrase;
            $masterTitle->save();
        });

        return $form;
    }

    public function actionSave(ParameterBag $params)
    {
        $this->assertPostOnly();
        /** @noinspection PhpUndefinedFieldInspection */
        $priorityId = $params->priority_id;
        if ($priorityId)
        {
            $condition = $this->assertPriorityExists($priorityId);
        }
        else
        {
            /** @var \NF\Tickets\Entity\Priority $condition **/
            $condition = $this->em()->create('NF\Tickets:Priority');
        }

        $this->prioritySaveProcess($condition)->run();

        return $this->redirect($this->buildLink('tickets/priorities'));
    }

    public function actionDelete(ParameterBag $params)
    {
        /** @noinspection PhpUndefinedFieldInspection */
        $priority = $this->assertPriorityExists($params->priority_id);
        /** @var \XF\ControllerPlugin\Delete $plugin */
        $plugin = $this->plugin('XF:Delete');

        return $plugin->actionDelete(
            $priority,
            $this->buildLink('tickets/priorities/delete', $priority),
            $this->buildLink('tickets/priorities/edit', $priority),
            $this->buildLink('tickets/priorities'),
            $this->app->stringFormatter()->snippetString($priority->title, 150)
        );
    }

    public function actionToggle()
    {
        /** @var \XF\ControllerPlugin\Toggle $plugin */
        $plugin = $this->plugin('XF:Toggle');

        return $plugin->actionToggle('NF\Tickets:Priority');
    }

    public function actionSort(ParameterBag $params): AbstractReply
    {
        if ($this->isPost())
        {
            $statuses = $this->getPriorityRepo()
                             ->findPrioritiesForList(false)
                             ->fetch();

            foreach ($this->filter('priorities', 'array-json-array') as $statusesData)
            {
                $lastOrder = 0;
                foreach ($statusesData as $statusesDatum)
                {
                    /** @var \NF\Tickets\Entity\Status $status */
                    $status = $statuses[$statusesDatum['id']] ?? null;
                    if ($status === null)
                    {
                        continue;
                    }

                    $lastOrder += 10;
                    $status->display_order = $lastOrder;
                    $status->saveIfChanged();
                }
            }

            return $this->redirect($this->buildLink('tickets/priorities'));
        }

        $viewParams = [
            'priorities' => $this->getPriorityRepo()->getPriorityKeyValues(),
        ];

        return $this->view('NF\Tickets:Priority\\Sort', 'nf_tickets_priority_sort', $viewParams);
    }

    /**
     * @param int     $id
     * @param array   $with
     * @param ?string $phraseKey
     * @throws \XF\Mvc\Reply\Exception
     * @return \NF\Tickets\Entity\Priority
     */
    protected function assertPriorityExists(int $id, array $with = [], ?string $phraseKey = 'nf_tickets_priority_not_found')
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->assertRecordExists('NF\Tickets:Priority', $id, $with, $phraseKey);
    }

    protected function getPriorityRepo(): \NF\Tickets\Repository\Priority
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Priority');
    }
}