<?php

namespace NF\Tickets\Admin\Controller;

use NF\Tickets\Entity\ResponseCategory;
use XF\Admin\Controller\AbstractController;
use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Repository;
use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\Exception;

class Response extends AbstractController
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    public function actionIndex()
    {
        $viewParams = $this->getRepo()->getResponseListData();

        return $this->view('NF\Tickets:Response\Listing', 'nf_tickets_response_list', $viewParams);
    }

    protected function responseAddEdit(\NF\Tickets\Entity\Response $response)
    {
        $repo = $this->getRepo();

        $viewParams = [
            'response'           => $response,
            'responseCategories' => $repo->findResponsesCategorized()->fetch()->pluckNamed('category_name', 'response_category_id'),
        ];

        return $this->view('NF\Tickets:Response\Edit', 'nf_tickets_response_edit', $viewParams);
    }

    public function actionEdit(ParameterBag $params)
    {
        $response = $this->assertResponseExists($params->response_id);

        return $this->responseAddEdit($response);
    }

    public function actionAdd()
    {
        $response = $this->getEntity();

        return $this->responseAddEdit($response);
    }

    public function actionSave(ParameterBag $params)
    {
        $this->assertPostOnly();

        if ($params->response_id)
        {
            $response = $this->assertResponseExists($params->response_id);
        }
        else
        {
            $response = $this->getEntity();
        }

        $this->responseSaveProcess($response)->run();

        return $this->redirect($this->buildLink('tickets/responses'));
    }

    public function actionDelete(ParameterBag $params)
    {
        $response = $this->assertResponseExists($params->response_id);

        /** @var \XF\ControllerPlugin\Delete $plugin */
        $plugin = $this->plugin('XF:Delete');

        return $plugin->actionDelete(
            $response,
            $this->buildLink('tickets/responses/delete', $response),
            $this->buildLink('tickets/responses/edit', $response),
            $this->buildLink('tickets/responses'),
            $response->response_title
        );
    }

    protected function responseSaveProcess(\NF\Tickets\Entity\Response $response)
    {
        $form = $this->formAction();

        $input = $this->filter([
            'response_category_id' => 'uint',
            'display_order'        => 'uint',
            'response_title'       => 'str',
            'message'              => 'str',
        ]);

        $input['user_id'] = 0;

        $form->basicEntitySave($response, $input);

        return $form;
    }

    public function actionCategoryEdit(ParameterBag $params)
    {
        $responseCategory = $this->assertResponseCategoryExists($params->response_category_id);

        return $this->categoryAddEdit($responseCategory);
    }

    public function actionCategoryAdd()
    {
        $responseCategory = $this->getCategoryEntity();

        return $this->categoryAddEdit($responseCategory);
    }

    public function actionCategoryDelete(ParameterBag $params)
    {
        $responseCategory = $this->assertResponseCategoryExists($params->response_category_id);

        /** @var \XF\ControllerPlugin\Delete $plugin */
        $plugin = $this->plugin('XF:Delete');

        return $plugin->actionDelete(
            $responseCategory,
            $this->buildLink('tickets/responses/categories/delete', $responseCategory),
            $this->buildLink('tickets/responses/categories/edit', $responseCategory),
            $this->buildLink('tickets/responses'),
            $responseCategory->category_name
        );
    }

    protected function categoryAddEdit(ResponseCategory $responseCategory)
    {
        $viewParams = [
            'responseCategory' => $responseCategory,
        ];

        return $this->view('NF\Tickets:ResponseCategory\Edit', 'nf_tickets_response_category_edit', $viewParams);
    }

    public function actionCategorySave(ParameterBag $params)
    {
        $this->assertPostOnly();

        if ($params->response_category_id)
        {
            $responseCategory = $this->assertResponseCategoryExists($params->response_category_id);
        }
        else
        {
            $responseCategory = $this->getCategoryEntity();
        }

        $this->responseCategorySaveProcess($responseCategory)->run();

        return $this->redirect($this->buildLink('tickets/responses'));
    }

    protected function responseCategorySaveProcess(ResponseCategory $responseCategory)
    {
        $form = $this->formAction();

        $input = $this->filter([
            'category_name' => 'str',
            'display_order' => 'uint',
        ]);

        $form->basicEntitySave($responseCategory, $input);

        return $form;
    }

    protected function getRepo(): \NF\Tickets\Repository\Response
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Response');
    }

    /**
     * @param int     $id
     * @param array   $with
     * @param ?string $phraseKey
     * @return \NF\Tickets\Entity\Response
     * @throws Exception
     */
    protected function assertResponseExists(int $id, array $with = [], string $phraseKey = null): \NF\Tickets\Entity\Response
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->assertRecordExists('NF\Tickets:Response', $id, $with, $phraseKey);
    }

    /**
     * @param int     $id
     * @param array   $with
     * @param ?string $phraseKey
     * @throws \XF\Mvc\Reply\Exception
     * @return ResponseCategory
     */
    protected function assertResponseCategoryExists(int $id, array $with = [], string $phraseKey = null): ResponseCategory
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->assertRecordExists('NF\Tickets:ResponseCategory', $id, $with, $phraseKey);
    }

    protected function getEntity(): \NF\Tickets\Entity\Response
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->em()->create('NF\Tickets:Response');
    }

    protected function getCategoryEntity(): ResponseCategory
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->em()->create('NF\Tickets:ResponseCategory');
    }
}