<?php

namespace NF\Tickets\Admin\Controller;

use XF\Admin\Controller\AbstractController;
use XF\Mvc\FormAction;
use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\AbstractReply;

class Status extends AbstractController
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    public function actionIndex()
    {
        $statusRepo = $this->getStatusRepo();
        $statuses = $statusRepo->findStatusesForList(false)->fetch();

        $viewParams = [
            'statuses' => $statuses,
        ];

        return $this->view('NF\Tickets:Status\List', 'nf_tickets_status_list', $viewParams);
    }

    public function statusAddEdit(\NF\Tickets\Entity\Status $status)
    {
        $viewParams = [
            'status' => $status,
            'displayStyles' => $this->getStatusRepo()->getDefaultDisplayStyles(),
        ];

        return $this->view('NF\Tickets:Status\Edit', 'nf_tickets_status_edit', $viewParams);
    }

    public function actionEdit(ParameterBag $params)
    {
        $condition = $this->assertStatusExists($params->status_id);

        return $this->statusAddEdit($condition);
    }

    public function actionAdd()
    {
        /** @var \NF\Tickets\Entity\Status $status */
        $status = $this->em()->create('NF\Tickets:Status');

        return $this->statusAddEdit($status);
    }

    protected function statusSaveProcess(\NF\Tickets\Entity\Status $status)
    {
        $entityInput = $this->filter([
            'active' => 'bool',
            'include_in_counts' => 'bool',
            'include_in_defaults' => 'bool',
            'is_close' => 'bool',
            'alert_on_transition' => 'bool',
            'css_class' => 'str',
            'display_order' => 'uint',
        ]);

        $form = $this->formAction();
        $form->basicEntitySave($status, $entityInput);

        $titlePhrase = $this->filter('title', 'str');

        $form->validate(function(FormAction $form) use ($titlePhrase)
        {
            if ($titlePhrase === '')
            {
                $form->logError(\XF::phrase('please_enter_valid_title'), 'title');
            }
        });
        $form->apply(function() use ($titlePhrase, $status)
        {
            $masterTitle = $status->getMasterPhrase();
            $masterTitle->phrase_text = $titlePhrase;
            $masterTitle->save();
        });

        return $form;
    }

    public function actionSave(ParameterBag $params)
    {
        $this->assertPostOnly();

        if ($params->status_id)
        {
            $condition = $this->assertStatusExists($params->status_id);
        }
        else
        {
            /** @var \NF\Tickets\Entity\Status $condition **/
            $condition = $this->em()->create('NF\Tickets:Status');
        }

        $this->statusSaveProcess($condition)->run();

        return $this->redirect($this->buildLink('tickets/statuses'));
    }

    public function actionDelete(ParameterBag $params)
    {
        $status = $this->assertStatusExists($params->status_id);

        if (!$status->preDelete())
        {
            return $this->error($status->getErrors());
        }

        if ($this->isPost())
        {
            $newStatus = $this->assertStatusExists($this->filter('new_status_id', 'uint'));

            \XF::db()->beginTransaction();

            $this->getStatusRepo()->updateTicketsStatus($status, $newStatus);
            $status->delete();

            \XF::db()->commit();

            return $this->redirect($this->buildLink('tickets/statuses'));
        }
        else
        {
            $viewParams = [
                'oldStatus' => $status,
                'statuses' => $this->getStatusRepo()->findStatusesForList()
                    ->where('status_id', '<>', $status->status_id)->fetch()
            ];

            return $this->view('XF:Delete\Delete', 'nf_tickets_status_delete', $viewParams);
        }
    }

    public function actionToggle()
    {
        /** @var \XF\ControllerPlugin\Toggle $plugin */
        $plugin = $this->plugin('XF:Toggle');

        return $plugin->actionToggle('NF\Tickets:Status');
    }

    public function actionSort(ParameterBag $params): AbstractReply
    {
        if ($this->isPost())
        {
            $statuses = $this->getStatusRepo()
                             ->findStatusesForList(false)
                             ->fetch();

            foreach ($this->filter('statuses', 'array-json-array') as $statusesData)
            {
                $lastOrder = 0;
                foreach ($statusesData as $statusesDatum)
                {
                    /** @var \NF\Tickets\Entity\Status $status */
                    $status = $statuses[$statusesDatum['id']] ?? null;
                    if ($status === null)
                    {
                        continue;
                    }

                    $lastOrder += 10;
                    $status->display_order = $lastOrder;
                    $status->saveIfChanged();
                }
            }

            return $this->redirect($this->buildLink('tickets/statuses'));
        }

        $viewParams = [
            'statuses' => $this->getStatusRepo()->getStatusTitlePairs(),
        ];

        return $this->view('NF\Tickets:Status\\Sort', 'nf_tickets_status_sort', $viewParams);
    }

    /**
     * @param int     $id
     * @param array   $with
     * @param ?string $phraseKey
     * @return \NF\Tickets\Entity\Status
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function assertStatusExists(int $id, array $with = [], string $phraseKey = null): \NF\Tickets\Entity\Status
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->assertRecordExists('NF\Tickets:Status', $id, $with, $phraseKey);
    }

    protected function getStatusRepo(): \NF\Tickets\Repository\Status
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Status');
    }
}