<?php

namespace NF\Tickets\Admin\Controller;

use XF\Admin\Controller\AbstractController;
use XF\Mvc\ParameterBag;

/**
 * Class Ticket
 *
 * @package NF\Tickets\Admin\Controller
 */
class Ticket extends AbstractController
{
    /**
     * @param              $action
     * @param ParameterBag $params
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    /**
     * @return \XF\Mvc\Reply\View
     */
    public function actionIndex()
    {
        return $this->view('NF\Tickets:Ticket', 'nf_tickets');
    }

    /**
     * @param string $id
     * @param array|string|null $with
     * @param null|string $phraseKey
     *
     * @return \XF\Entity\Node
     */
    protected function assertForumNodeExists($id, $with = null, $phraseKey = null)
    {
        /** @var \XF\Entity\Node $node */
        $node = $this->assertRecordExists('XF:Node', $id, $with, $phraseKey);
        if ($node->node_type_id !== 'Forum')
        {
            throw $this->exception($this->error(\XF::phrase('requested_node_not_found'), 404));
        }
        return $node;
    }

    public function actionForumsPrefixes()
    {
        $this->assertPostOnly();

        $viewParams = [];

        $nodeId = $this->filter('val', 'uint');
        if ($nodeId)
        {
            /** @var \XF\Entity\Forum $forum */
            $node = $this->assertForumNodeExists($nodeId);
            $forum = $node->getDataRelationOrDefault(false);

            $viewParams['forum'] = $forum;
            $viewParams['prefixes'] = $forum->getPrefixesGrouped();
        }

        return $this->view('XF:Forum\Prefixes', 'public:forum_prefixes', $viewParams);
    }

	public function actionList()
	{
		$this->setSectionContext('nfBatchUpdateTickets');

		$criteria = $this->filter('criteria', 'array');
		$order = $this->filter('order', 'str');
		$direction = $this->filter('direction', 'str');

		$page = $this->filterPage();
		$perPage = 50;

		$showingAll = $this->filter('all', 'bool');
		if ($showingAll)
		{
			$perPage = 500;
		}

		/** @var \NF\Tickets\Searcher\Ticket $searcher */
		$searcher = $this->searcher('NF\Tickets:Ticket', $criteria);
		$searcher->setOrder($order, $direction);

		$finder = $searcher->getFinder();
		$finder->limitByPage($page, $perPage);

		$total = $finder->total();
		$tickets = $finder->fetch();

		$viewParams = [
			'tickets' => $tickets,

			'total' => $total,
			'page' => $page,
			'perPage' => $perPage,

			'showingAll' => $showingAll,
			'showAll' => (!$showingAll && $total <= 500),

			'criteria' => $searcher->getFilteredCriteria(),
			'order' => $order,
			'direction' => $direction
		];
		return $this->view('NF\Tickets:Ticket\Listing', 'nf_tickets_ticket_list', $viewParams);
	}

	public function actionBatchUpdate()
	{
		$this->setSectionContext('nfBatchUpdateTickets');

		$searcher = $this->searcher('NF\Tickets:Ticket');

		$viewParams = [
			'criteria' => $searcher->getFormCriteria(),
			'success' => $this->filter('success', 'bool')
		] + $searcher->getFormData();

		return $this->view('NF\Tickets:Ticket\BatchUpdate', 'nf_tickets_ticket_batch_update', $viewParams);
	}

	public function actionBatchUpdateConfirm()
	{
		$this->setSectionContext('nfBatchUpdateTickets');

		$this->assertPostOnly();

		$input = $this->filterFormJson([
			'criteria' => 'array',
			'ticket_ids' => 'array-uint'
		]);

		$searcher = $this->searcher('NF\Tickets:Ticket', $input['criteria']);
		$ticketIds = $input['ticket_ids'];

		$total = count($ticketIds) ?: $searcher->getFinder()->total();
		if (!$total)
		{
			throw $this->exception($this->error(\XF::phraseDeferred('no_items_matched_your_filter')));
		}

		if ($ticketIds)
		{
			$ticketFinder = $this->finder('NF\Tickets:Ticket');
			$ticketFinder->where('ticket_id', $ticketIds);
		}
		else
		{
			$ticketFinder = clone $searcher->getFinder();
		}
		$hasPrefixes = (bool)$ticketFinder->where('prefix_id', '>', 0)->total();

		/** @var \NF\Tickets\Repository\TicketPrefix $prefixRepo */
		$prefixRepo = $this->repository('NF\Tickets:TicketPrefix');
		$prefixes = $prefixRepo->getPrefixListData();

		/** @var \NF\Tickets\Repository\Category $categoryRepo */
		$categoryRepo = $this->repository('NF\Tickets:Category');
		$categories = $categoryRepo->getCategoryOptionsData(false);

		/** @var \NF\Tickets\Repository\Status $statusRepo */
		$statusRepo = $this->repository('NF\Tickets:Status');
		$statuses = $statusRepo->getStatusTitlePairs(true);

		/** @var \NF\Tickets\Repository\Priority $priorityRepo */
		$priorityRepo = $this->repository('NF\Tickets:Priority');
		$priorities = $priorityRepo->getPriorityKeyValues(false);

		$viewParams = [
			'total' => $total,
			'ticketIds' => $ticketIds,
			'hasPrefixes' => $hasPrefixes,
			'criteria' => $searcher->getFilteredCriteria(),

			'prefixes' => $prefixes,
			'categories' => $categories,
			'statuses' => $statuses,
			'priorities' => $priorities
		];
		return $this->view('NF\Tickets:Ticket\BatchUpdate\Confirm', 'nf_tickets_ticket_batch_update_confirm', $viewParams);
	}

	public function actionBatchUpdateAction()
	{
		$this->setSectionContext('nfTicketsBatchUpdateTickets');

		$this->assertPostOnly();

		if ($this->request->exists('ticket_ids'))
		{
			$ticketIds = $this->filter('ticket_ids', 'json-array');
			$total = count($ticketIds);
			$jobCriteria = null;
		}
		else
		{
			$criteria = $this->filter('criteria', 'json-array');
			$searcher = $this->searcher('NF\Tickets:Ticket', $criteria);
			$total = $searcher->getFinder()->total();
			$jobCriteria = $searcher->getFilteredCriteria();

			$ticketIds = null;
		}

		if (!$total)
		{
			throw $this->exception($this->error(\XF::phraseDeferred('no_items_matched_your_filter')));
		}

		$actions = $this->filter('actions', 'array');

		if ($this->request->exists('confirm_delete') && empty($actions['delete']))
		{
			return $this->error(\XF::phrase('you_must_confirm_deletion_to_proceed'));
		}

		$this->app->jobManager()->enqueueUnique('nfTicketAction', 'NF\Tickets:TicketAction', [
			'total' => $total,
			'actions' => $actions,
			'ticketIds' => $ticketIds,
			'criteria' => $jobCriteria
		]);

		return $this->redirect($this->buildLink('tickets/batch-update', null, ['success' => true]));
	}
}