<?php

namespace NF\Tickets\Admin\Controller;

use XF\Admin\Controller\AbstractPrefix;
use XF\Mvc\Entity\AbstractCollection;
use XF\Mvc\Entity\ArrayCollection;
use XF\Mvc\ParameterBag;
use XF\Mvc\FormAction;

class TicketPrefix extends AbstractPrefix
{
    protected function preDispatchController($action, ParameterBag $params)
    {
        $this->assertAdminPermission('nfTickets');
    }

    protected function getClassIdentifier()
    {
        return 'NF\Tickets:TicketPrefix';
    }

    protected function getLinkPrefix()
    {
        return 'tickets/prefixes';
    }

    protected function getTemplatePrefix()
    {
        return 'nf_tickets_ticket_prefix';
    }

    protected function getCategoryParams(\NF\Tickets\Entity\TicketPrefix $prefix)
    {
        /** @var \NF\Tickets\Repository\Category $categoryRepo */
        $categoryRepo = \XF::repository('NF\Tickets:Category');
        $categoryTree = $categoryRepo->createCategoryTree($categoryRepo->findCategoryList()->fetch());

        /** @var AbstractCollection $categoryPrefixAssociations */
        $categoryPrefixAssociations = $prefix->getRelationOrDefault('CategoryPrefixes', false);

        return [
            'categoryTree' => $categoryTree,
            'categoryIds' => $categoryPrefixAssociations->pluckNamed('ticket_category_id')
        ];
    }

    protected function prefixAddEditResponse(\XF\Entity\AbstractPrefix $prefix)
    {
        $reply = parent::prefixAddEditResponse($prefix);

        if ($reply instanceof \XF\Mvc\Reply\View)
        {
            /** @var \NF\Tickets\Entity\TicketPrefix $prefix */
            $reply->setParams($this->getCategoryParams($prefix));
        }

        return $reply;
    }

    protected function quickSetAdditionalData(FormAction $form, ArrayCollection $prefixes)
    {
        $input = $this->filter([
            'apply_ticket_category_ids' => 'bool',
            'ticket_category_ids' => 'array-uint'
        ]);

        if ($input['apply_ticket_category_ids'])
        {
            $form->complete(function() use ($prefixes, $input)
            {
                $mapRepo = $this->getCategoryPrefixRepo();

                foreach ($prefixes AS $prefix)
                {
                    $mapRepo->updatePrefixAssociations($prefix, $input['ticket_category_ids']);
                }
            });
        }

        return $form;
    }

    public function actionQuickSet()
    {
        $reply = parent::actionQuickSet();

        if ($reply instanceof \XF\Mvc\Reply\View)
        {
            if ($reply->getTemplateName() === $this->getTemplatePrefix() . '_quickset_editor')
            {
                $reply->setParams($this->getCategoryParams($reply->getParam('prefix')));
            }
        }

        return $reply;
    }

    protected function saveAdditionalData(FormAction $form, \XF\Entity\AbstractPrefix $prefix)
    {
        $categoryIds = $this->filter('ticket_category_ids', 'array-uint');

        $form->complete(function() use ($prefix, $categoryIds)
        {
            $this->getCategoryPrefixRepo()->updatePrefixAssociations($prefix, $categoryIds);
        });

        return $form;
    }

    public function actionPrefixes(ParameterBag $params)
    {
        $this->assertPostOnly();

        $categoryId = $this->filter('val', 'uint');

        /** @var \NF\Tickets\Entity\Category $category */
        $category = $this->em()->find('NF\Tickets:Category', $categoryId,
            'Permissions|' . \XF::visitor()->permission_combination_id
        );
        if (!$category)
        {
            return $this->notFound(\XF::phrase('requested_category_not_found'));
        }

        if (!$category->canView($error))
        {
            return $this->noPermission($error);
        }

        $viewParams = [
            'category' => $category,
            'prefixes' => $category->getUsablePrefixes()
        ];

        return $this->view('NF\Tickets:Category\Prefixes', 'public:nf_tickets_category_prefixes', $viewParams);
    }

    protected function getCategoryPrefixRepo(): \NF\Tickets\Repository\CategoryPrefix
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:CategoryPrefix');
    }
}