<?php

namespace NF\Tickets\Api\Controller;

use XF\Api\Controller\AbstractController;
use XF\Mvc\Entity\Entity;
use XF\Mvc\ParameterBag;

class Tickets extends AbstractController
{
	protected function preDispatchController($action, ParameterBag $params)
	{
		$this->assertApiScopeByRequestMethod('nf_ticket');
	}

	public function actionGet()
	{
		$this->assertRegisteredUser(); // TODO: soon

		$page = $this->filterPage();
		$perPage = 20;

		$finder = $this->setupTicketFinder()->limitByPage($page, $perPage);
		$total = $finder->total();

		$this->assertValidApiPage($page, $perPage, $total);

		$tickets = $finder->fetch();

		if (\XF::isApiCheckingPermissions())
		{
			$tickets = $tickets->filterViewable();
		}

		return $this->apiResult([
			'tickets' => $tickets->toApiResults(),
			'pagination' => $this->getPaginationData($tickets, $page, $perPage, $total)
		]);
	}

	protected function setupTicketFinder(&$filters = [], &$sort = null)
	{
		/** @var \NF\Tickets\Repository\Ticket $repo */
		$repo = $this->repository('NF\Tickets:Ticket');
		$finder = $repo->findTicketsForApi();

		/** @var \NF\Tickets\Api\ControllerPlugin\Ticket $plugin */
		$plugin = $this->plugin('NF\Tickets:Api:Ticket');

		$filters = $plugin->applyTicketFilters($finder);
		$sort = $plugin->applyTicketSort($finder);

		return $finder;
	}

	public function actionPost()
	{
		$this->assertRequiredApiInput(['ticket_category_id', 'title', 'message']);
		$this->assertRegisteredUser(); // TODO: soon

		$categoryId = $this->filter('ticket_category_id', 'uint');

		/** @var \NF\Tickets\Entity\Category $category */
		$category = $this->assertViewableApiRecord('NF\Tickets:Category', $categoryId);

		if (\XF::isApiCheckingPermissions() && !$category->canCreateTicket($error))
		{
			return $this->noPermission($error);
		}

		$creator = $this->setupTicketCreate($category);

		if (\XF::isApiCheckingPermissions())
		{
			$creator->checkForSpam();
		}

		if (!$creator->validate($errors))
		{
			return $this->error($errors);
		}

		/** @var \NF\Tickets\Entity\Ticket $ticket */
		$ticket = $creator->save();
		$this->finalizeTicketCreate($creator);

		return $this->apiSuccess([
			'ticket' => $ticket->toApiResult(Entity::VERBOSITY_VERBOSE)
		]);
	}

	protected function setupTicketCreate(\NF\Tickets\Entity\Category $category)
	{
		$input = $this->filter([
			'title' => 'str',
			'message' => 'str',
			'priority' => 'uint',
			'prefix_id' => 'uint',
			'custom_fields' => 'array',
			'attachment_key' => 'str',
			'ticket_locked' => '?bool'
		]);

		$isBypassingPermissions = \XF::isApiBypassingPermissions();

		/** @var \NF\Tickets\Service\Ticket\Creator $creator */
		$creator = $this->service('NF\Tickets:Ticket\Creator', $category);

		$creator->setContent($input['title'], $input['message']);
		$creator->setCustomFields($input['custom_fields']);

		if ($input['prefix_id'] && ($isBypassingPermissions || $category->isPrefixUsable($input['prefix_id'])))
		{
			$creator->setPrefix($input['prefix_id']);
		}

        $priority = $this->getPriorityRepo()->getActivePriority($input['priority']);
        if ($priority)
        {
            $creator->setPriority($priority->priority_id);
        }

		if ($isBypassingPermissions || $category->canUploadAndManageAttachments())
		{
			$hash = $this->getAttachmentTempHashFromKey($input['attachment_key'], 'nf_tickets_message', ['ticket_category_id' => $category->ticket_category_id]);
			$creator->setAttachmentHash($hash);
		}

		$ticket = $creator->getTicket();

		if (isset($input['ticket_locked']) && ($isBypassingPermissions || $ticket->canLockUnlockTicket()))
		{
			$creator->setLocked($input['ticket_locked']);
		}

		return $creator;
	}

	protected function finalizeTicketCreate(\NF\Tickets\Service\Ticket\Creator $creator)
	{
		$creator->sendNotifications();

		$ticket = $creator->getTicket();
		$visitor = \XF::visitor();

		if ($visitor->user_id)
		{
			$this->getTicketRepo()->markTicketReadByVisitor($ticket, $ticket->start_date);
		}
	}


	protected function getTicketRepo(): \NF\Tickets\Repository\Ticket
	{
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Ticket');
	}

	protected function getMessageRepo(): \NF\Tickets\Repository\Ticket
	{
        /** @noinspection PhpIncompatibleReturnTypeInspection */
		return $this->repository('NF\Tickets:Ticket');
	}

    protected function getPriorityRepo(): \NF\Tickets\Repository\Priority
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Priority');
    }
}