<?php

namespace NF\Tickets\Attachment;

use XF\Attachment\AbstractHandler;
use XF\Entity\Attachment;
use XF\Mvc\Entity\Entity;

use function intval;

/**
 * Class Message
 *
 * @package NF\Tickets\Attachment
 */
class Message extends AbstractHandler
{
    /**
     * @return array
     */
    public function getContainerWith()
    {
        $visitor = \XF::visitor();

        return ['Ticket', 'Ticket.Category', 'Ticket.Category.Permissions|' . $visitor->permission_combination_id];
    }

    /**
     * @param Attachment $attachment
     * @param Entity     $container
     * @param null       $error
     * @return bool
     */
    public function canView(Attachment $attachment, Entity $container, &$error = null)
    {
        /** @var \NF\Tickets\Entity\Message $container */
        $ticket = $container->Ticket;
        if (!$ticket || !$container->canView())
        {
            return false;
        }

        return $ticket->canViewAttachments($error);
    }

    /**
     * @param array $context
     * @param null  $error
     * @return bool
     */
    public function canManageAttachments(array $context, &$error = null)
    {
        $em = \XF::em();

        if (!empty($context['message_id']))
        {
            /** @var \NF\Tickets\Entity\Message $message */
            $message = $em->find('NF\Tickets:Message', intval($context['message_id']), ['Ticket', 'Ticket.Category']);
            if (!$message || !$message->canView() || !$message->canEdit())
            {
                return false;
            }

            $category = $message->Ticket->Category;
        }
        else if (!empty($context['ticket_id']))
        {
            /** @var \NF\Tickets\Entity\Ticket $ticket */
            $ticket = $em->find('NF\Tickets:Ticket', intval($context['ticket_id']), ['Category']);
            if (!$ticket || !$ticket->canView())
            {
                return false;
            }

            $category = $ticket->Category;
        }
        else if (!empty($context['ticket_category_id']))
        {
            /** @var \NF\Tickets\Entity\Category $category */
            $category = $em->find('NF\Tickets:Category', intval($context['ticket_category_id']));
            if (!$category || !$category->canView())
            {
                return false;
            }
        }
        else
        {
            return false;
        }

        return $category->canUploadAndManageAttachments();
    }

    /**
     * @param Attachment                             $attachment
     * @param Entity|\NF\Tickets\Entity\Message|null $container
     * @throws \XF\PrintableException
     */
    public function onAttachmentDelete(Attachment $attachment, Entity $container = null)
    {
        if (!$container)
        {
            return;
        }

        $container->attach_count--;
        $container->save();
    }

    /**
     * @param array $context
     * @return array
     */
    public function getConstraints(array $context)
    {
        /** @var \XF\Repository\Attachment $attachmentRepo */
        $attachmentRepo = \XF::repository('XF:Attachment');

        return $attachmentRepo->getDefaultAttachmentConstraints();
    }

    /**
     * @param array $context
     * @return int|null
     */
    public function getContainerIdFromContext(array $context)
    {
        return isset($context['message_id']) ? intval($context['message_id']) : null;
    }

    /**
     * @param Entity|\NF\Tickets\Entity\Message $container
     * @param array                             $extraParams
     * @return mixed|string
     */
    public function getContainerLink(Entity $container, array $extraParams = [])
    {
        return \XF::app()->router('public')->buildLink('tickets/messages', $container, $extraParams);
    }

    /**
     * @param Entity|\NF\Tickets\Entity\Message|null $entity
     * @param array                                  $extraContext
     * @return array
     */
    public function getContext(Entity $entity = null, array $extraContext = [])
    {
        if ($entity instanceof \NF\Tickets\Entity\Message)
        {
            $extraContext['message_id'] = $entity->message_id;
        }
        else if ($entity instanceof \NF\Tickets\Entity\Ticket)
        {
            $extraContext['ticket_id'] = $entity->ticket_id;
        }
        else if ($entity instanceof \NF\Tickets\Entity\Category)
        {
            $extraContext['ticket_category_id'] = $entity->ticket_category_id;
        }
        else
        {
            throw new \InvalidArgumentException('Entity must be message, ticket or category');
        }

        return $extraContext;
    }
}