<?php

namespace NF\Tickets\ControllerPlugin;

use NF\Tickets\Utils\FeatureFlags;
use XF\ControllerPlugin\AbstractPlugin;
use function array_key_exists;
use function in_array;
use function intval;
use function is_array;

/**
 * Class Overview
 *
 * @package NF\Tickets\ControllerPlugin
 */
class TicketList extends AbstractPlugin
{
    /**
     * @param \NF\Tickets\Entity\Category|null $category
     *
     * @return array
     */
    public function getCategoryListData(\NF\Tickets\Entity\Category $category = null)
    {
        $categoryRepo = $this->getCategoryRepo();
        $categories = $categoryRepo->getViewableCategories($category);
        $categoryTree = $categoryRepo->createCategoryTree($categories);
        $categoryExtras = $categoryRepo->getCategoryListExtras($categoryTree);

        return [
            //'priorities' => $priorityRepo->findPrioritiesForList()->fetch(),
            'statuses' => $this->getStatusRepo()->findStatusesForList()->fetch(),
            'categories' => $categories,
            'categoryTree' => $categoryTree,
            'categoryExtras' => $categoryExtras
        ];
    }

    /**
     * @param array $sourceCategoryIds
     * @param \NF\Tickets\Entity\Category|null $category
     * @param \XF\Entity\User|null $user
     * @param bool $allowOwnPending
     *
     * @return array
     */
    public function getCoreListData(
        array $sourceCategoryIds, \NF\Tickets\Entity\Category $category = null, \XF\Entity\User $user = null, $allowOwnPending = false)
    {
        $ticketRepo = $this->getTicketRepo();

        $ticketFinder = $ticketRepo->findTicketsForOverviewList($sourceCategoryIds, $category === null, $allowOwnPending);

        $filters = $this->getDefaultFilters($user === null ? null : []);
        $filters = $this->getTicketFilterInput($filters);
        if ($user)
        {
            $filters['creator_id'] = $user->user_id;
        }
        $this->applyTicketFilters($ticketFinder, $filters);

        // TODO: if no filters and can't view deleted, can total from category info
        $totalTickets = $ticketFinder->total();

        $page = (int)$this->filterPage();
        $perPage = (int)$this->options()->nftCatTicketsPerPage;

        $ticketFinder->limitByPage($page, $perPage);

        if ($page === 1)
        {
            $assignedTickets = clone $ticketFinder;
            $assignedTickets = $assignedTickets->isSelfAssigned()->fetch()->filterViewable();
        }
        else
        {
            $assignedTickets = [];
        }

        $tickets = $ticketFinder->isNotSelfAssigned()->fetch()->filterViewable();

        if (!$user && !empty($filters['creator_id']))
        {
            /** @var \XF\Entity\User $creatorFilter */
            $creatorFilter = $this->em()->find('XF:User', $filters['creator_id']);
        }
        else
        {
            $creatorFilter = null;
        }

        $canInlineMod = false;
        /** @var \NF\Tickets\Entity\Ticket $ticket */
        foreach ($tickets AS $ticket)
        {
            if ($ticket->canUseInlineModeration())
            {
                $canInlineMod = true;
                break;
            }
        }

        /** @var \NF\Tickets\Entity\Ticket $ticket */
        foreach ($assignedTickets AS $ticket)
        {
            if ($ticket->canUseInlineModeration())
            {
                $canInlineMod = true;
                break;
            }
        }

        return [
            'user' => $user,
            'category' => $category,

            'tickets' => $tickets,
            'assignedTickets' => $assignedTickets,
            'filters' => $filters,
            'creatorFilter' => $creatorFilter,
            'canInlineMod' => $canInlineMod,

            'total' => $totalTickets,
            'page' => $page,
            'perPage' => $perPage,

			'newUi' => FeatureFlags::isFeatureFlagEnabled('newUi')
        ];
    }

    /**
     * @param \NF\Tickets\Finder\Ticket $ticketFinder
     * @param array $filters
     */
    public function applyTicketFilters(\NF\Tickets\Finder\Ticket $ticketFinder, array $filters)
    {
        if (!\XF::visitor()->user_id)
        {
            $ticketFinder->where('ticket_id', $this->getGuestTicketIds());
        }

        if (!empty($filters['unassigned']))
        {
        	$ticketFinder->isUnassigned();
        }
        else if (!empty($filters['assigned_self']))
        {
        	$ticketFinder->isSelfAssigned();
        }
        else if (!empty($filters['assigned_others']))
        {
        	$ticketFinder->isAssignedOthers();
        }

        if (!empty($filters['status']) && !in_array(-1, $filters['status'], true))
        {
            $ticketFinder->where('status_id', '=', $filters['status']);
        }

        if (!empty($filters['status_not']) && !in_array(-1, $filters['status_not'], true))
        {
            $ticketFinder->where('status_id', '<>', $filters['status_not']);
        }

        if (!empty($filters['last_days']) && empty($filters['no_date_limit']))
        {
            if ($filters['last_days'] > 0)
            {
                $ticketFinder->where('last_message_date', '>=', \XF::$time - ($filters['last_days'] * 86400));
            }
        }

        if (!empty($filters['participant_id']))
        {
            $ticketFinder->whereOr([
                'user_id',
                intval($filters['participant_id'])
            ], ["Participants|{$filters['participant_id']}.ticket_id", '!=', null]);
        }

        if (!empty($filters['prefix_id']))
        {
            $ticketFinder->where('prefix_id', intval($filters['prefix_id']));
        }

        if (!empty($filters['creator_id']))
        {
            $ticketFinder->where('user_id', intval($filters['creator_id']));
        }

        $sorts = $this->getAvailableTicketSorts();

        if (!empty($filters['order']) && isset($sorts[$filters['order']]))
        {
            $ticketFinder->order($sorts[$filters['order']], $filters['direction']);
        }
        // else the default order has already been applied
    }

    protected function getAvailableDateLimits()
    {
        return [-1, 7, 14, 30, 60, 90, 182, 365];
    }

    protected function getDefaultFilters($statusOverride = null)
    {
        $defaults = [];

        if ($statusOverride !== null)
        {
            foreach ($statusOverride AS $statusId)
            {
                $defaults['status_not'][] = $statusId;
            }
        }
        else
        {
            /** @var \NF\Tickets\Repository\Status $repo */
            $repo = \XF::repository('NF\Tickets:Status');
            $statuses = $repo->findStatusesForList()
                ->isIncludedInDefaultTicketList(false)
                ->fetch();
            foreach ($statuses as $status)
            {
                $defaults['status_not'][] = $status->status_id;
            }
        }

        return $defaults;
    }

    /**
     * @param array $defaultFilters
     *
     * @return array
     */
    public function getTicketFilterInput(array $defaultFilters)
    {
        $filters = $defaultFilters;

        $input = $this->filter([
        	'unassigned' => 'bool',
            'assigned_self' => 'bool',
            'assigned_others' => 'bool',

            'creator' => 'str',
            'creator_id' => 'int',
            'order' => 'str',
            'direction' => 'str',

            'last_days' => 'int',
            'no_date_limit' => 'bool',

            'status' => 'array-int',
            'status_not' => 'array-int',

            'prefix_id' => 'uint'
        ]);

        if ($input['unassigned'])
        {
        	$filters['unassigned'] = true;
        }
        else if ($input['assigned_self'])
        {
        	$filters['assigned_self'] = true;
        }
        else if ($input['assigned_others'])
        {
        	$filters['assigned_others'] = true;
        }

        if ($input['prefix_id'])
        {
            // TODO: Category prefix validity check?
            $filters['prefix_id'] = $input['prefix_id'];
        }

        if ($input['no_date_limit'])
        {
            $filters['no_date_limit'] = $input['no_date_limit'];
        }
        else if (isset($defaultFilters['no_date_limit']))
        {
            $filters['no_date_limit'] = false;
        }

        if (empty($filters['no_date_limit']))
        {
            if ($input['last_days'] === -1 || in_array($input['last_days'], $this->getAvailableDateLimits(), true))
            {
                $filters['last_days'] = $input['last_days'];
            }
        }

        if ($input['status'])
        {
            $filters['status'] = \array_filter($input['status']);
        }

        if ($input['status_not'])
        {
            $filters['status_not'] = \array_filter($input['status_not']);
        }

        if ($input['creator_id'])
        {
            if ($input['creator_id'] >= 0)
            {
                $filters['creator_id'] = $input['creator_id'];
            }
            else if (isset($defaultFilters['creator_id']))
            {
                unset($filters['creator_id']);
            }
        }
        else if ($input['creator'])
        {
            /** @var \XF\Entity\User $user */
            $user = $this->em()->findOne('XF:User', ['username' => $input['creator']]);
            if ($user)
            {
                $filters['creator_id'] = $user->user_id;
            }
        }

        $sorts = $this->getAvailableTicketSorts();

        if ($input['order'] && isset($sorts[$input['order']]))
        {
            if (!in_array($input['direction'], ['asc', 'desc']))
            {
                $input['direction'] = 'desc';
            }

            $defaultOrder = $this->options()->nfTicketsListDefaultOrder ?: 'last_message_date';
            $defaultDir = $defaultOrder === 'title' ? 'asc' : 'desc';

            if ($input['order'] !== $defaultOrder || $input['direction'] !== $defaultDir)
            {
                $filters['order'] = $input['order'];
                $filters['direction'] = $input['direction'];
            }
        }

        return $filters;
    }

    /**
     * @return array
     */
    public function getAvailableTicketSorts()
    {
        // maps [name of sort] => field in/relative to NF\Tickets:Ticket entity
        return [
            'priority' => '',
            'last_message_date' => 'last_message_date',
            'last_modified_date' => 'last_modified_date',
            'start_date' => 'start_date',
            'title' => 'title'
        ];
    }

    /**
     * @param \NF\Tickets\Entity\Category|null $category
     * @param \XF\Entity\User|null $user
     *
     * @return \XF\Mvc\Reply\Redirect|\XF\Mvc\Reply\View
     */
    public function actionFilters(\NF\Tickets\Entity\Category $category = null, \XF\Entity\User $user = null)
    {
        $defaultFilters = $this->getDefaultFilters();
        $filters = $this->getTicketFilterInput([]);

        if ($user)
        {
            $route = 'tickets/members';
            $routeData = $user;
        }
        else if ($category)
        {
            $route = 'tickets/categories';
            $routeData = $category;
        }
        else
        {
            $route = 'tickets/queue';
            $routeData = $category;
        }

        if ($this->filter('apply', 'bool'))
        {
            foreach ($defaultFilters as $key => $value)
            {
                if (isset($filters[$key]))
                {
                    if ($filters[$key] === $value)
                    {
                        unset($filters[$key]);
                    }
                }
                else
                {
                    // no value submitted, so instead use -1 to signal 'none'
                    $filters[$key] = is_array($value) ? [-1] : -1;
                }
            }

            return $this->redirect($this->buildLink($route, $routeData, $filters));
        }

        if (!$user && !empty($filters['creator_id']))
        {
            /** @var \XF\Entity\User $creatorFilter */
            $creatorFilter = $this->em()->find('XF:User', $filters['creator_id']);
        }
        else
        {
            $creatorFilter = null;
        }

        $defaultOrder = $this->options()->nfTicketsListDefaultOrder ?: 'last_message_date';
        $defaultDir = $defaultOrder === 'title' ? 'asc' : 'desc';

        if (empty($filters['order']))
        {
            $filters['order'] = $defaultOrder;
        }
        if (empty($filters['direction']))
        {
            $filters['direction'] = $defaultDir;
        }

        $viewParams = [
            'statuses' => $this->getStatusRepo()->findStatusesForList()->fetch(),
            'user' => $user,
            'category' => $category,
            'prefixes' => $category ? $category->prefixes->groupBy('prefix_group_id') : $this->finder('NF\Tickets:TicketPrefix')
                ->fetch()->groupBy('prefix_group_id'),

            'filters' => $filters,
            'creatorFilter' => $creatorFilter,

            'baseLinkPath' => $route,
            'linkData' => $routeData,
        ];

        return $this->view($user ? 'NF\Tickets:Member\Filters' : 'NF\Tickets:Filters', 'nf_tickets_filters', $viewParams);
    }

    /**
     * @param null $ticketId
     *
     * @return array|bool|mixed
     */
    public function getGuestTicketIds($ticketId = null)
    {
        $ticketAuth = $this->controller->session()->get('nfGuestTicketAuth') ?: [];

        if (is_array($ticketAuth) && array_key_exists($ticketId, $ticketAuth))
        {
            return $ticketAuth[$ticketId];
        }

        if ($ticketId !== null)
        {
            return false;
        }

        return is_array($ticketAuth) ? \array_keys($ticketAuth) : [];
    }

    protected function getStatusRepo(): \NF\Tickets\Repository\Status
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Status');
    }

    protected function getPriorityRepo(): \NF\Tickets\Repository\Priority
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Priority');
    }

    protected function getTicketRepo(): \NF\Tickets\Repository\Ticket
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Ticket');
    }

    protected function getCategoryRepo(): \NF\Tickets\Repository\Category
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Category');
    }
}