<?php

namespace NF\Tickets\Criteria;

use XF\Criteria\AbstractCriteria;
use XF\Util\Arr;

use function in_array;

/**
 * Class Ticket
 *
 * @package NF\Tickets\Criteria
 */
class Ticket extends AbstractCriteria
{
    /**
     * @var \NF\Tickets\Entity\Ticket
     */
    protected $ticket;

    /**
     * Ticket constructor.
     *
     * @param \XF\App                   $app
     * @param array                     $criteria
     * @param \NF\Tickets\Entity\Ticket $ticket
     */
    public function __construct(\XF\App $app, array $criteria, \NF\Tickets\Entity\Ticket $ticket = null)
    {
        parent::__construct($app, $criteria);

        $this->ticket = $ticket;
    }

    /**
     * @param                 $rule
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function isUnknownMatched($rule, array $data, \XF\Entity\User $user)
    {
        $eventReturnValue = false;
        $this->app->fire('nf_tickets_criteria_ticket', [$rule, $data, $user, &$eventReturnValue]);

        return $eventReturnValue;
    }

    // region Ticket

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchTicketStatus(array $data, \XF\Entity\User $user)
    {
        return in_array($this->ticket->status_id, $data, true);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchTicketPriority(array $data, \XF\Entity\User $user)
    {
        return $this->ticket->priority !== null &&
               in_array((int)$this->ticket->priority, $data, true);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchTicketCategory(array $data, \XF\Entity\User $user)
    {
        return in_array($this->ticket->ticket_category_id, $data, true);
    }

    protected function invertCheck(array $data, bool $value): bool
    {
        return empty($data['not']) ? $value : !$value;
    }

    // region Starter

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchStarterIsGuest(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data, !$this->ticket->starter_user_id);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchStarterIsStaff(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data, $this->ticket->User && $this->ticket->User->is_staff);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchStarterIsModerator(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data, $this->ticket->User && $this->ticket->User->is_moderator);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchStarterIsAdmin(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data, $this->ticket->User && $this->ticket->User->is_admin);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchStarterUsername(array $data, \XF\Entity\User $user)
    {
        $names = Arr::stringToArray(\utf8_strtolower($data['names']), '/\s*,\s*/');

        return in_array(\utf8_strtolower($this->ticket->starter_username), $names, true);
    }

	protected function _matchTimeSinceStart(array $data, \XF\Entity\User $user)
	{
		$cutoff = \strtotime('+' . $data['amount'] . ' ' . $data['timescale'], $this->ticket->first_message_date);
		return (\XF::$time >= $cutoff);
    }

	protected function _matchTimeSinceStartLess(array $data, \XF\Entity\User $user)
	{
        return $this->invertCheck(['not' => true], $this->_matchTimeSinceStart($data, $user));
	}

    // endregion
    // region First Reply

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchFirstReplyIsGuest(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->FirstMessage && !$this->ticket->FirstMessage->user_id);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchFirstReplyIsStaff(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->FirstMessage &&
               $this->ticket->FirstMessage->User &&
               $this->ticket->FirstMessage->User->is_staff);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchFirstReplyIsModerator(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->FirstMessage &&
               $this->ticket->FirstMessage->User &&
                                          $this->ticket->FirstMessage->User->is_moderator);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchFirstReplyIsAdmin(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->FirstMessage &&
               $this->ticket->FirstMessage->User &&
                                          $this->ticket->FirstMessage->User->is_admin);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchFirstReplyUsername(array $data, \XF\Entity\User $user)
    {
        $names = Arr::stringToArray(\utf8_strtolower($data['names']), '/\s*,\s*/');

        return $this->ticket->FirstMessage &&
               in_array(\utf8_strtolower($this->ticket->FirstMessage->username), $names, true);
    }

    // time_since_first_reply_is_greater_than
    // time_since_first_reply_is_less_than

    // endregion
    // region Last Reply

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchLastReplyIsGuest(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->LastMessage &&
               !$this->ticket->LastMessage->user_id);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchLastReplyIsStarter(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->LastMessage &&
               ($this->ticket->LastMessage->user_id === $this->ticket->starter_user_id));
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchLastReplyIsStaff(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,  $this->ticket->LastMessage &&
               $this->ticket->LastMessage->User &&
               $this->ticket->LastMessage->User->is_staff);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchLastReplyIsModerator(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,   $this->ticket->LastMessage &&
               $this->ticket->LastMessage->User &&
                                           $this->ticket->LastMessage->User->is_moderator);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchLastReplyIsAdmin(array $data, \XF\Entity\User $user)
    {
        return $this->invertCheck($data,   $this->ticket->LastMessage &&
               $this->ticket->LastMessage->User &&
                                           $this->ticket->LastMessage->User->is_admin);
    }

    /**
     * @param array           $data
     * @param \XF\Entity\User $user
     * @return bool
     */
    protected function _matchLastReplyUsername(array $data, \XF\Entity\User $user)
    {
        $names = Arr::stringToArray(\utf8_strtolower($data['names']), '/\s*,\s*/');

        return $this->ticket->LastMessage &&
               in_array(\utf8_strtolower($this->ticket->LastMessage->username), $names, true);
    }

	protected function _matchTimeSinceLastReply(array $data, \XF\Entity\User $user)
	{
		$cutoff = \strtotime('+' . $data['amount'] . ' ' . $data['timescale'], $this->ticket->last_message_date);
		return (\XF::$time >= $cutoff);
	}

	protected function _matchTimeSinceLastReplyLess(array $data, \XF\Entity\User $user)
	{
		return $this->invertCheck(['not' => true], $this->_matchTimeSinceLastReply($data, $user));
	}

    // endregion
}