<?php

namespace NF\Tickets\EmailTickets;

use BadMethodCallException;
use DateTime;
use Exception;
use NF\Tickets\Mail\Mail;
use XF;
use Laminas\Mail\Storage\Message;

use function intval, is_string, strlen;

class Parser22
{
    protected $verpBase;
    protected $verpHmacKey;

    /**
     * Parser22 constructor.
     * @param $verpBase
     * @param $verpHmacKey
     */
    public function __construct($verpBase, $verpHmacKey)
    {
        $this->verpBase = $verpBase;
        $this->verpHmacKey = $verpHmacKey;
    }

    public function parseMessage($message)
    {
        if (is_string($message))
        {
            $message = new Message(['raw' => $message]);
        }
        else if (!($message instanceof Message))
        {
            throw new \InvalidArgumentException("Message must be a string containing headers and message or Message object");
        }

        $result = new ParseResult();
        $result->messageType = ParseResult::TYPE_TICKET;

        $this->extractDate($message, $result);
        $this->extractContent($message, $result);
        $this->extractTicketAndSender($message, $result);

        return $result;
    }

    protected function extractDate(Message $message, ParseResult $result)
    {
        $timestamp = XF::$time;

        if (isset($message->date))
        {
            try
            {
                $dt = new DateTime($message->date);
                $timestamp = intval($dt->format('U'));
            } catch (Exception $e)
            {
            }
        }

        $result->date = $timestamp;
    }

    protected function extractContent(Message $message, ParseResult $result)
    {
        $subject = $message->subject ?? null;
        $textContent = null;
        $htmlContent = null;

        if ($message->isMultipart())
        {
            foreach ($message AS $part)
            {
                if (!isset($part->contentType))
                {
                    if (!$textContent)
                    {
                        $content = \trim($part->getContent());
                        if (strlen($content) && \preg_match('/./su', $content))
                        {
                            // if it's valid UTF-8, it's probably textual
                            $textContent = $this->decodeMimeContent($part, $content);
                        }
                    }

                    continue;
                }

                $contentType = $part->contentType;
                if (\preg_match('#^text/html#i', $contentType))
                {
                    $htmlContent = $this->decodeMimeContent($part, $part->getContent());
                }
                else if (\preg_match('#^text/plain#i', $contentType))
                {
                    $textContent = $this->decodeMimeContent($part, $part->getContent());
                }
            }
        }

        if ($textContent === null)
        {
            $textContent = $this->decodeMimeContent($message, $message->getContent());
        }

        $result->subject = \trim($subject);
        $result->htmlContent = \trim($htmlContent);
        $result->textContent = \trim($textContent);
    }

    protected function extractTicketAndSender(Message $message, ParseResult $result)
    {
        if (!$this->verpBase)
        {
            throw new BadMethodCallException("Cannot extract ticket from email without verpBase");
        }

        $ticketId = null;

        if (isset($message->to))
        {
            $matchRegex = \str_replace('@', '\+([0-9]+)\+([a-z0-9]+)\+([^@=]+=[^@=]+)@', \preg_quote($this->verpBase, '#'));
            if (\preg_match("#$matchRegex#i", $message->to, $verpMatch))
            {
                $ticketId = $verpMatch[1];
                $verpEmail = \str_replace('=', '@', $verpMatch[3]);
                $hmac = \hash_hmac('md5', \sprintf(Mail::TICKET_HMAC_FORMAT, $verpEmail, $ticketId), $this->verpHmacKey);

                $result->senderTrusted = (\substr($hmac, 0, strlen($verpMatch[2])) === $verpMatch[2]);
                $result->senderEmail = $verpEmail;
            }
        }

        if ($ticketId)
        {
            /** @var \NF\Tickets\Entity\Ticket $ticket */
            $ticket = XF::app()->em()->find('NF\Tickets:Ticket', $ticketId);
            if ($ticket)
            {
                $result->ticket = $ticket;
                $result->messageType = ParseResult::TYPE_REPLY;
            }
            else
            {
                // Ticket has been deleted. Create a new one?
                // TODO: Is this the best route?
                $result->messageType = ParseResult::TYPE_TICKET;
            }
        }

        if (isset($message->from) && \preg_match('#([^\n<]*?)<([^>@]+@[^>]+)>(\s|,|$)#i', $message->from, $match))
        {
            $result->senderName = \trim($match[1]);

            if (!$result->senderEmail)
            {
                $result->senderEmail = \trim($match[2]);
            }
        }
        else if (isset($message->from) && \preg_match('#([^>@]+@[^\s]+)(\r?\n|,|$)#i', $message->from, $match))
        {
            // No name, just use their email.
            $result->senderName = \trim($match[1]);
            if (!$result->senderEmail)
            {
                $result->senderEmail = $match[1];
            }
        }
    }

    /**
     * @param \Laminas\Mail\Storage\Part $part
     * @param string                  $content
     * @return false|string
     */
    protected function decodeMimeContent(\Laminas\Mail\Storage\Part $part, string $content)
    {
        if ($part->getHeaders()->has('Content-Transfer-Encoding'))
        {
            switch ($part->getHeader('Content-Transfer-Encoding', 'string'))
            {
                case 'quoted-printable':
                    $content = quoted_printable_decode($content);
                    break;
                case 'base64':
                    $content = base64_decode($content);
                    break;
            }
        }

        return $content;
    }
}