<?php

namespace NF\Tickets\EmailTickets;

use NF\Tickets\Entity\Category;
use NF\Tickets\Entity\TicketEmailLog;
use NF\Tickets\Repository\Ticket;
use NF\Tickets\Service\Ticket\Creator;
use NF\Tickets\Service\Ticket\Replier;
use XF\App;
use XF\Entity\User;
use XF\Html\Renderer\BbCode;
use Laminas\Mail\Storage;

use function intval;

class Processor22
{
	/**
	 * @var App
	 */
	protected $app;

	/**
	 * @var Parser22
	 */
	protected $ticketParser;

	/**
	 * @var Ticket
	 */
	protected $ticketRepo;

	/**
	 * @var Category
	 */
	protected $category;

	public function __construct(App $app, $parser, $category)
	{
		$this->app = $app;
		$this->ticketParser = $parser;
		$this->category = $category;
		$this->ticketRepo = $app->repository('NF\Tickets:Ticket');
	}

	public function processFromStorage(Storage\AbstractStorage $storage, $maxRunTime = 0)
	{
		$s = microtime(true);

		$total = $storage->countMessages();
		if (!$total)
		{
			return true;
		}

		$finished = true;

		for ($messageId = $total; $messageId > 0; $messageId--)
		{
			$contentValid = false;

			try
			{
				$headers = $storage->getRawHeader($messageId);
				$content = $storage->getRawContent($messageId);
				$contentValid = true;
			} catch (\InvalidArgumentException $e)
			{
                $headers = '';
                $content = '';
				$contentValid = false;
			} finally
			{
				$storage->removeMessage($messageId);
			}

			$rawMessage = \trim($headers) . "\r\n\r\n" . \trim($content);

			if ($contentValid)
			{
				$this->processMessage($rawMessage);
			}
			else
			{
				$result = new ParseResult();
				$result->date = \XF::$time;

				$this->logTicketMessage($rawMessage, null, $result);
			}

			if ($maxRunTime && microtime(true) - $s > $maxRunTime)
			{
				$finished = false;
				break;
			}
		}

		return $finished;
	}

	public function processMessage($rawMessage, &$result = null)
	{
		try
		{
			$result = $this->ticketParser->parseMessage($rawMessage);
		} catch (\Exception $e)
		{
			\XF::logException($e, false, "Ticket message processing failed: ");

			$result = new ParseResult();
			$result->date = \XF::$time;

			return $this->logTicketMessage($rawMessage, null, $result);
		}

		if ($result->senderEmail)
		{
			/** @var User $user */
			$user = $this->app->em()->findOne('XF:User', ['email' => $result->senderEmail]);
		}
		else
		{
			$user = null;
		}

		$action = null;

		if ($result->messageType === ParseResult::TYPE_TICKET)
		{
			// TODO: Create ticket
			if ($user)
			{
				\XF::asVisitor($user, function() use ($result)
				{
					/** @var Creator $ticketCreateService */
					$ticketCreateService = $this->app->service('NF\Tickets:Ticket\Creator', $this->category);
					$ticketCreateService->logIp(false);
					$content = \XF::cleanString($result->htmlContent ? BbCode::renderFromHtml($result->htmlContent) : $result->textContent);
					$ticketCreateService->setContent($result->subject ?: "New ticket", $content, false);
					$ticketCreateService->save();
					$ticketCreateService->sendNotifications();
				});
			}
			else
			{
				// Guest ticket
                /** @var \XF\Repository\User $userRepo */
                $userRepo = $this->app->repository('XF:User');

				$senderName = substr($result->senderName, 0, 50);

                $guestUser = $userRepo->getGuestUser($senderName);
				\XF::asVisitor($guestUser, function() use ($result)
				{
					/** @var Creator $ticketCreateService */
					$ticketCreateService = $this->app->service('NF\Tickets:Ticket\Creator', $this->category);
					$ticketCreateService->logIp(false);
					$content = \XF::cleanString($result->htmlContent ? BbCode::renderFromHtml($result->htmlContent) : $result->textContent);
					$ticketCreateService->setContent($result->subject ?: "New ticket", $content, false);
					$ticketCreateService->setGuestUser($result->senderName, $result->senderEmail);

					if ($ticketCreateService->validate())
					{
						$ticketCreateService->save();
						$ticketCreateService->sendNotifications();
					}
					else
					{
						\XF::logError('[NF] Tickets - An incoming ticket email could not be parsed. Please check the email inbox manually.');
					}
				});
			}
		}
		else if ($user && $result->messageType === ParseResult::TYPE_REPLY)
		{
			if ($result->senderTrusted)
			{
				\XF::asVisitor($user, function() use ($result)
				{
					/** @var Replier $ticketReplyService */
					$ticketReplyService = $this->app->service('NF\Tickets:Ticket\Replier', $result->ticket);
					$ticketReplyService->logIp(false);
					$content = \XF::cleanString($result->htmlContent ? BbCode::renderFromHtml($result->htmlContent) : $result->textContent);
					$ticketReplyService->setMessage($content, false);
					$ticketReplyService->save();
					$ticketReplyService->sendNotifications();
				});
			}

			// Deliberately ignoring untrusted ticket reply
		}

		return $this->logTicketMessage($rawMessage, $action, $result, $user);
	}

	protected function logTicketMessage($rawMessage, $action, ParseResult $result, User $user = null)
	{
		/** @var TicketEmailLog $ticketEmail */
		$ticketEmail = $this->app->em()->create('NF\Tickets:TicketEmailLog');
		$ticketEmail->bulkSet([
			'email_date'      => $result->date,
			'message_type'    => $result->messageType ?: ParseResult::TYPE_UNKNOWN,
			'action_taken'    => $action ?: '',
			'user_id'         => $user->user_id ?? null,
			'sender'          => $result->senderEmail,
			'raw_message'     => \substr($rawMessage, 0, 1024 * 1024), // limit logging to 1MB to prevent potential insert issues
			'diagnostic_info' => $result->remoteDiagnostics
		]);
		$ticketEmail->save();

		return $ticketEmail;
	}

	/**
	 * @return null|Storage\AbstractStorage
	 * @throws \Exception
	 */
	public function getTicketCategoryEmailHandlerStorage()
	{
        $app = $this->app;
        $category = $this->category;
		if (!$category->incoming_ticket_email_settings)
		{
			return null;
		}

		$handler = $category->incoming_ticket_email_settings['connection'];
		if (!$handler)
		{
			return null;
		}

		$config = [
			'host'     => $handler['host'],
			'user'     => $handler['username'],
			'password' => $handler['password']
		];
		if ($handler['port'])
		{
			$config['port'] = intval($handler['port']);
		}
		if ($handler['encryption'])
		{
			$config['ssl'] = \strtoupper($handler['encryption']);
		}

		try
		{
			if ($handler['type'] === 'pop3')
			{
				$connection = new Storage\Pop3($config);
			}
			else if ($handler['type'] === 'imap')
			{
				$connection = new Storage\Imap($config);
			}
			else
			{
				throw new \Exception("Unknown ticket category email handler $handler[type]");
			}
		} catch (\Laminas\Mail\Exception\ExceptionInterface $e)
		{
			$app->logException($e, false, "Ticket category email connection error: ");
			return null;
		}

		return $connection;
	}
}