<?php

namespace NF\Tickets\Entity;

use XF\Entity\LinkableInterface;
use XF\Mvc\Entity\Structure;

use XF\Entity\AbstractCategoryTree;
use XF\Phrase;

/**
 * COLUMNS
 * @property int $ticket_category_id
 * @property string $title
 * @property string $description
 * @property int $ticket_count
 * @property int $last_ticket
 * @property string $last_ticket_title
 * @property int $last_ticket_id
 * @property int $thread_node_id
 * @property int $thread_prefix_id
 * @property array $field_cache
 * @property bool $count_messages_
 * @property bool $allow_opening
 * @property int $priority_selection
 * @property int $empty_message_allowed
 * @property int $hide_message_field
 * @property string $category_emails
 * @property array|null $incoming_ticket_email_settings
 * @property array $prefix_cache
 * @property int $default_prefix_id
 * @property bool $require_prefix
 * @property int|null $anonymous_user_id
 * @property int $parent_category_id
 * @property int $display_order
 * @property int $lft
 * @property int $rgt
 * @property int $depth
 * @property array $breadcrumb_data
 *
 * GETTERS
 * @property mixed $count_messages
 * @property \XF\Draft $draft_ticket
 * @property \XF\Mvc\Entity\AbstractCollection $prefixes
 *
 * RELATIONS
 * @property \XF\Entity\Forum $ThreadForum
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\CategoryRead[] $Read
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\CategoryWatch[] $Watch
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\Draft[] $DraftTickets
 * @property \XF\Entity\User $AnonymousUser
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\PermissionCacheContent[] $Permissions
 */
class Category extends AbstractCategoryTree implements LinkableInterface
{
    protected $_viewableDescendants = [];

    /**
     * @param null $error
     *
     * @return bool
     */
    public function canCreateTicket(&$error = null)
    {
		if (!$this->allow_opening)
		{
			$error = \XF::phraseDeferred(\XF::phrase('nf_tickets_you_may_not_perform_this_action_because_this_category_does_not_permit'));
			return false;
		}

        if (!$this->hasPermission('create'))
        {
            return false;
        }

        return true;
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canOpenTicketForAnyone(&$error = null)
    {
        if (!\XF::visitor()->user_id)
        {
            return false;
        }

        return (bool)$this->hasPermission('openAny');
    }

    /**
     * @param Phrase|string|null $error
     * @return bool
     * @noinspection PhpUnusedParameterInspection
     */
    public function canActionAsAnonymized(&$error = null)
    {
        if (!\XF::visitor()->user_id)
        {
            return false;
        }

        if (!$this->AnonymousUser)
        {
            return false;
        }

        return (bool)$this->hasPermission('replyAsAnonymized');
    }

    /**
     * @param null $error
     *
     * @return bool
     */
    public function canView(&$error = null)
    {
        return (bool)$this->hasPermission('view');
    }

    /**
     * @param null $error
     *
     * @return bool
     */
    public function canWatch(&$error = null)
    {
        $visitor = \XF::visitor();

        if (!$visitor->user_id)
        {
            return false;
        }

        return true;
    }

    /**
     * @param null $error
     *
     * @return bool
     */
    public function canViewOthers(&$error = null)
    {
        return (bool)$this->hasPermission('viewOthers');
    }

    /**
     * @return bool
     */
    public function canViewDeletedTickets()
    {
        return (bool)$this->hasPermission('viewDeleted');
    }

    /**
     * @return bool
     */
    public function canViewModeratedTickets()
    {
        return (bool)$this->hasPermission('viewDeleted');
    }

    /**
     * @return bool
     */
    public function canUploadAndManageAttachments()
    {
        return (bool)$this->hasPermission('uploadAttachment');
    }

    public function canSetPriorityOnTicketCreation(Phrase &$error = null) : bool
    {
        if (!$this->priority_selection)
        {
            return false;
        }

        return (bool)$this->hasPermission('viewOwnPriority');
    }

    /**
     * @param string $permission
     * @return bool|int
     */
    public function hasPermission(string $permission)
    {
        /** @var \NF\Tickets\XF\Entity\User $visitor */
        $visitor = \XF::visitor();

        return $visitor->hasTicketCategoryPermission($this->ticket_category_id, $permission);
    }

    public function getViewableDescendants()
    {
        $userId = \XF::visitor()->user_id;
        if (!isset($this->_viewableDescendants[$userId]))
        {
            /** @var \NF\Tickets\Repository\Category $repo */
            $repo = $this->repository('NF\Tickets:Category');
            $viewable = $repo->getViewableCategories($this);
            $this->_viewableDescendants[$userId] = $viewable->toArray();
        }

        return $this->_viewableDescendants[$userId];
    }

    public function cacheViewableDescendants(array $descendents, $userId = null)
    {
        if ($userId === null)
        {
            $userId = \XF::visitor()->user_id;
        }

        $this->_viewableDescendants[$userId] = $descendents;
    }

    public function getUsablePrefixes($forcePrefix = null)
    {
        $prefixes = $this->prefixes;

        if ($forcePrefix instanceof TicketPrefix)
        {
            $forcePrefix = $forcePrefix->prefix_id;
        }

        $prefixes = $prefixes->filter(function($prefix) use ($forcePrefix)
        {
            if ($forcePrefix && $forcePrefix === $prefix->prefix_id)
            {
                return true;
            }
            return $this->isPrefixUsable($prefix);
        });

        return $prefixes->groupBy('prefix_group_id');
    }

    public function getPrefixesGrouped()
    {
        return $this->prefixes->groupBy('prefix_group_id');
    }

    /**
     * @return \XF\Mvc\Entity\AbstractCollection
     */
    public function getPrefixes()
    {
        if (!$this->prefix_cache)
        {
            return $this->_em->getEmptyCollection();
        }

        $prefixes = $this->finder('NF\Tickets:TicketPrefix')
            ->where('prefix_id', $this->prefix_cache)
            ->order('materialized_order')
            ->fetch();

        return $prefixes;
    }

    public function isPrefixUsable($prefix, \XF\Entity\User $user = null)
    {
        if (!$this->isPrefixValid($prefix))
        {
            return false;
        }

        if (!($prefix instanceof TicketPrefix))
        {
            /** @var TicketPrefix $prefix */
            $prefix = $this->em()->find('NF\Tickets:TicketPrefix', $prefix);
            if (!$prefix)
            {
                return false;
            }
        }

        return $prefix->isUsableByUser($user);
    }

    public function isPrefixValid($prefix)
    {
        if ($prefix instanceof TicketPrefix)
        {
            $prefix = $prefix->prefix_id;
        }

        return (!$prefix || isset($this->prefix_cache[$prefix]));
    }

    /**
     * @return Ticket
     */
    public function getNewTicket()
    {
        /** @var Ticket $ticket */
        $ticket = $this->_em->create('NF\Tickets:Ticket');
        $ticket->ticket_category_id = $this->ticket_category_id;

        return $ticket;
    }

    /**
     * @param bool   $includeSelf
     * @param string $linkType
     *
     * @return array
     */
    public function getBreadcrumbs($includeSelf = true, $linkType = 'public')
    {
        return $this->_getBreadcrumbs($includeSelf, $linkType, 'tickets/categories');
    }

    /**
     * @return array
     */
    public function getCategoryListExtras()
    {
        return [
            'ticket_count' => $this->ticket_count,
            'last_ticket_date' => $this->last_ticket,
            'last_ticket_title' => $this->last_ticket_title,
            'last_ticket_id' => $this->last_ticket_id
        ];
    }

    /**
     * @param Ticket $ticket
     */
    public function ticketAdded(Ticket $ticket)
    {
        $this->ticket_count++;

        if ($ticket->last_message_date >= $this->last_ticket)
        {
            $this->last_ticket = $ticket->last_message_date;
            $this->last_ticket_title = $ticket->title;
            $this->last_ticket_id = $ticket->ticket_id;
        }
    }

    public function ticketDataChanged(Ticket $ticket)
    {
        if ($ticket->isChanged(['last_message_date', 'title']))
        {
            if ($ticket->last_message_date >= $this->last_ticket)
            {
                $this->last_ticket = $ticket->last_message_date;
                $this->last_ticket_title = $ticket->title;
                $this->last_ticket_id = $ticket->ticket_id;
            }
            elseif ($ticket->getExistingValue('last_message_date') === $this->last_ticket)
            {
                $this->rebuildLastTicket();
            }
        }
    }

    public function ticketRemoved(Ticket $ticket)
    {
        $this->ticket_count--;

        if ($ticket->last_message_date === $this->last_ticket)
        {
            $this->rebuildLastTicket();
        }
    }

    public function getCountMessages()
    {
        return $this->count_messages_;
    }

    /**
     * @return \XF\Draft
     */
    public function getDraftTicket()
    {
        return \XF\Draft::createFromEntity($this, 'DraftTickets');
    }

    public function rebuildCounters()
    {
        $this->ticket_count = $this->db()->fetchOne("
			SELECT COUNT(*)
			FROM xf_nf_tickets_ticket
			WHERE ticket_category_id = ?
				AND discussion_state = 'visible'
		", $this->ticket_category_id);

        $this->rebuildLastTicket();
    }

    public function rebuildLastTicket()
    {
        $ticket = $this->db()->fetchRow("
			SELECT *
			FROM xf_nf_tickets_ticket
			WHERE ticket_category_id = ?
				AND discussion_state = 'visible'
			ORDER BY last_modified_date DESC
			LIMIT 1
		", $this->ticket_category_id);
        if ($ticket)
        {
            $this->last_ticket = $ticket['last_modified_date'];
            $this->last_ticket_title = $ticket['title'];
            $this->last_ticket_id = $ticket['ticket_id'];
        }
        else
        {
            $this->last_ticket = 0;
            $this->last_ticket_title = '';
            $this->last_ticket_id = 0;
        }
    }

    public function getContentUrl(bool $canonical = false, array $extraParams = [], $hash = null): string
    {
        return \XF::app()->router('public')->buildLink(($canonical ? 'canonical:' : '') . 'tickets/categories', $this);
    }

    public function getContentPublicRoute(): string
    {
        return 'tickets/categories';
    }

    public function getContentTitle(string $context = ''): string
    {
        return $this->title;
    }

    protected function _postDelete()
    {
        $db = $this->db();

        $db->delete('xf_nf_tickets_category_field', 'ticket_category_id = ?', $this->ticket_category_id);
        $db->delete('xf_nf_tickets_category_watch', 'ticket_category_id = ?', $this->ticket_category_id);

        if ($this->getOption('delete_tickets'))
        {
            $this->app()->jobManager()->enqueueUnique('nfTicketsCategoryDelete' . $this->ticket_category_id, 'NF\Tickets:CategoryDelete', [
                'ticket_category_id' => $this->ticket_category_id
            ]);
        }
    }

	/**
	 * @param \XF\Api\Result\EntityResult $result
	 * @param int $verbosity
	 * @param array $options
	 */
	protected function setupApiResultData(
		\XF\Api\Result\EntityResult $result, $verbosity = self::VERBOSITY_NORMAL, array $options = []
	)
	{
		if ($verbosity > self::VERBOSITY_NORMAL)
		{
			$result->prefixes = $this->prefixes->toApiResults();

			$fields = [];
			if ($this->field_cache)
			{
                /** @var \NF\Tickets\Repository\TicketField $repo */
                $repo = $this->repository('NF\Tickets:TicketField');
				$fieldEntities = $repo->findFieldsForList()
					->whereIds($this->field_cache)
					->fetch();
				$fields = $fieldEntities->toApiResults();
			}

			$result->custom_fields = $fields;
		}

		$result->can_create = $this->canCreateTicket();
	}

    /**
     * @param Structure $structure
     *
     * @return Structure
     */
	public static function getStructure(Structure $structure)
	{
		$structure->table = 'xf_nf_tickets_category';
		$structure->shortName = 'NF\Tickets:Category';
		$structure->primaryKey = 'ticket_category_id';
		$structure->contentType = 'nf_tickets_category';
		$structure->columns = [
			'ticket_category_id' => ['type' => self::UINT, 'autoIncrement' => true],
			'title' => ['type' => self::STR,
				'required' => 'please_enter_valid_title', 'maxLength' => 100
			],
			'description' => ['type' => self::STR, 'default' => ''],
			'ticket_count' => ['type' => self::UINT, 'forced' => true, 'default' => 0],
			'last_ticket' => ['type' => self::UINT, 'default' => 0],
			'last_ticket_title' => ['type' => self::STR, 'default' => '', 'maxLength' => 150],
			'last_ticket_id' => ['type' => self::UINT, 'default' => 0],
			'thread_node_id' => ['type' => self::UINT, 'default' => 0],
			'thread_prefix_id' => ['type' => self::UINT, 'default' => 0],
			'field_cache' => ['type' => self::JSON_ARRAY, 'default' => []],
			'count_messages' => ['type' => self::BOOL, 'default' => true],
			'allow_opening' => ['type' => self::BOOL, 'default' => true],
			'priority_selection' => ['type' => self::UINT, 'default' => 1],
			'empty_message_allowed' => ['type' => self::UINT, 'default' => 0],
			'hide_message_field' => ['type' => self::UINT, 'default' => 0],
            'category_emails' => ['type' => self::STR, 'default' => ''],
            'incoming_ticket_email_settings' => ['type' => self::JSON_ARRAY, 'nullable' => true],
            'prefix_cache' => ['type' => self::JSON_ARRAY, 'default' => []],
            'default_prefix_id' => ['type' => self::UINT, 'default' => 0],
            'require_prefix' => ['type' => self::BOOL, 'default' => false],
            'anonymous_user_id' => ['type' => self::UINT, 'default' => null, 'nullable' => true],
        ];
        $structure->getters = [
            'count_messages' => ['getter' => true, 'cache' => false],
            'draft_ticket' => true,
            'prefixes' => true,
        ];
		$structure->relations = [
            'ThreadForum' => [
                'entity' => 'XF:Forum',
                'type' => self::TO_ONE,
                'conditions' => [['node_id', '=', '$thread_node_id']],
                'primary' => true,
                'with' => 'Node'
            ],
            'Read' => [
                'entity' => 'NF\Tickets:CategoryRead',
                'type' => self::TO_MANY,
                'conditions' => 'ticket_category_id',
                'key' => 'user_id'
            ],
            'Watch' => [
                'entity' => 'NF\Tickets:CategoryWatch',
                'type' => self::TO_MANY,
                'conditions' => 'ticket_category_id',
                'key' => 'user_id'
            ],
            'DraftTickets' => [
                'entity'     => 'XF:Draft',
                'type'       => self::TO_MANY,
                'conditions' => [
                    ['draft_key', '=', 'nfTickets-category-', '$ticket_category_id']
                ],
                'key'        => 'user_id'
            ],
            'AnonymousUser' => [
                'entity' => 'XF:User',
                'type' => self::TO_ONE,
                'conditions' => [['user_id','=','$anonymous_user_id']],
                'primary' => true,
                'api' => true
            ],
        ];
        $structure->options = [
            'delete_tickets' => true
        ];
		$structure->withAliases = [
			'api' => []
		];

        static::addCategoryTreeStructureElements($structure);

		return $structure;
	}
}