<?php

namespace NF\Tickets\Entity;

use XF\BbCode\RenderableContentInterface;
use XF\Entity\LinkableInterface;
use XF\Entity\Attachment;
use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Structure;

/**
 * COLUMNS
 * @property int $kb_article_id
 * @property int $kb_category_id
 * @property string $title
 * @property string $content
 * @property int $attach_count
 * @property int $views
 * @property int $created_at
 * @property int $updated_at
 * @property array|null $embed_metadata
 *
 * RELATIONS
 * @property \NF\Tickets\Entity\KBCategory $Category
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\Attachment[] $Attachments
 */
class KBArticle extends Entity implements RenderableContentInterface, LinkableInterface
{
	/**
	 * @param bool $includeSelf
	 *
	 * @return array
	 */
	public function getBreadcrumbs($includeSelf = true)
	{
		$breadcrumbs = $this->Category ? $this->Category->getBreadcrumbs() : [];
		if ($includeSelf)
		{
			$breadcrumbs[] = [
				'href' => $this->app()->router()->buildLink('knowledge-base', $this),
				'value' => $this->title
			];
		}

		return $breadcrumbs;
	}

	/**
	 * @param null $error
	 * @return mixed
	 */
	public function canView(&$error = null)
	{
	    $perm = $this->hasPermission('view');

		return $perm;
	}

	/**
	 * @param null $error
	 * @return bool
	 */
	public function canEdit(&$error = null)
	{
		$visitor = \XF::visitor();
		if (!$visitor->user_id)
		{
			return false;
		}

		return (bool)$this->hasPermission('edit');
	}

	public function isVisible()
	{
		return true;
	}

    /**
     * @param string $permission
     * @return false|int
     */
	public function hasPermission(string $permission)
	{
		/** @var \NF\Tickets\XF\Entity\User $visitor */
		$visitor = \XF::visitor();
		return $visitor->hasPermission('nfTicketsKB', $permission);
	}

	public function isAttachmentEmbedded($attachmentId)
	{
		if (!$this->embed_metadata)
		{
			return false;
		}

		if ($attachmentId instanceof Attachment)
		{
			$attachmentId = $attachmentId->attachment_id;
		}

		return isset($this->embed_metadata['attachments'][$attachmentId]);
	}

	public function getBbCodeRenderOptions($context, $type)
	{
		return [
			'entity' => $this,
			'attachments' => $this->attach_count ? $this->Attachments : [],
			'viewAttachments' => true,
		];
	}

	public function getContentUrl(bool $canonical = false, array $extraParams = [], $hash = null)
	{
		$route = $canonical ? 'canonical:knowledge-base' : 'knowledge-base';
		return $this->app()->router('public')->buildLink($route, $this, $extraParams, $hash);
	}

	public function getContentPublicRoute()
	{
		return 'knowledge-base';
	}

	public function getContentTitle(string $context = '')
	{
		return \XF::phrase('nf_tickets_kb_article_x', ['title' => $this->title]);
	}

	public static function getStructure(Structure $structure)
    {
        $structure->table = 'xf_nf_tickets_kb_article';
        $structure->shortName = 'NF\Tickets:KBArticle';
        $structure->contentType = 'nf_tickets_kb_article';
        $structure->primaryKey = 'kb_article_id';

        $structure->columns = [
            'kb_article_id' => ['type' => self::UINT, 'autoIncrement' => true],
            'kb_category_id' => ['type' => self::INT, 'required' => true],
            'title' => [
                'type' => self::STR,
                'required' => 'please_enter_valid_title', 'maxLength' => 100,
            ],
            'content' => ['type' => self::STR, 'censor' => true, 'required' => true],
            'attach_count' => ['type' => self::UINT, 'max' => 65535, 'forced' => true, 'default' => 0],
            'views' => ['type' => self::UINT, 'default' => 0],
            'created_at' => ['type' => self::UINT, 'default' => \XF::$time],
            'updated_at' => ['type' => self::UINT, 'default' => \XF::$time],
            'embed_metadata' => ['type' => self::JSON_ARRAY, 'nullable' => true, 'default' => null],
        ];

        $structure->behaviors=  [
            'XF:Indexable' => [
                'checkForUpdates' => ['title', 'content', 'kb_category_id', 'updated_at']
            ],
        ];

        $structure->relations = [
            'Category' => [
                'entity' => 'NF\Tickets:KBCategory',
                'type' => self::TO_ONE,
                'conditions' => 'kb_category_id',
                'primary' => true,
            ],
            'Attachments'   => [
                'entity'     => 'XF:Attachment',
                'type'       => self::TO_MANY,
                'conditions' => [
                    ['content_type', '=', 'nf_tickets_kb_article'],
                    ['content_id', '=', '$kb_article_id'],
                ],
                'with'       => 'Data',
                'order'      => 'attach_date',
            ],
        ];

        return $structure;
    }
}