<?php

namespace NF\Tickets\Entity;

use XF\Entity\AbstractCategoryTree;
use XF\Mvc\Entity\Structure;

/**
 * COLUMNS
 * @property int $kb_category_id
 * @property string $title
 * @property string $description
 * @property int $parent_category_id
 * @property int $display_order
 * @property int $lft
 * @property int $rgt
 * @property int $depth
 * @property array $breadcrumb_data
 *
 * GETTERS
 * @property \XF\Draft $draft_article
 *
 * RELATIONS
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\Draft[] $DraftArticles
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\KBArticle[] $Articles
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\PermissionCacheContent[] $Permissions
 */
class KBCategory extends AbstractCategoryTree
{
    protected $_viewableDescendants = [];

    /**
     * @param null $error
     * @return bool
     */
    public function canView(&$error = null)
    {
        return \XF::visitor()->hasPermission('nfTicketsKB', 'view');
    }

    /**
     * @return bool
     */
    public function canViewDeletedArticles()
    {
        return (bool)$this->hasPermission('viewDeleted');
    }

    /**
     * @return bool
     */
    public function canUploadAndManageAttachments()
    {
        return \XF::visitor()->hasPermission('nfTicketsKB', 'uploadAttachment');
    }

	/**
	 * @return bool
	 */
    public function canViewAttachments()
    {
        return \XF::visitor()->hasPermission('nfTicketsKB', 'viewAttachment');
    }

    /**
     * @param string|$permission
     * @return bool|int
     */
    public function hasPermission(string $permission)
    {
        /** @var \NF\Tickets\XF\Entity\User $visitor */
        $visitor = \XF::visitor();

        return $visitor->hasTicketKBCategoryPermission($this->kb_category_id, $permission);
    }

    public function getViewableDescendants()
    {
        $userId = \XF::visitor()->user_id;
        if (!isset($this->_viewableDescendants[$userId]))
        {
            /** @var \NF\Tickets\Repository\KBCategory $repo */
            $repo = $this->repository('NF\Tickets:KBCategory');
            $viewable = $repo->getViewableCategories($this);
            $this->_viewableDescendants[$userId] = $viewable->toArray();
        }

        return $this->_viewableDescendants[$userId];
    }

    public function cacheViewableDescendants(array $descendents, $userId = null)
    {
        if ($userId === null)
        {
            $userId = \XF::visitor()->user_id;
        }

        $this->_viewableDescendants[$userId] = $descendents;
    }

    /**
     * @return KBArticle
     */
    public function getNewArticle()
    {
        /** @var KBArticle $article */
        $article = $this->_em->create('NF\Tickets:KBArticle');
        $article->kb_category_id = $this->kb_category_id;

        return $article;
    }

    /**
     * @param bool   $includeSelf
     * @param string $linkType
     * @return array
     */
    public function getBreadcrumbs($includeSelf = true, $linkType = 'public')
    {
        return $this->_getBreadcrumbs($includeSelf, $linkType, 'knowledge-base/categories');
    }

    /**
     * @return array
     */
    public function getCategoryListExtras()
    {
        return [
            //'article_count' => $this->article_count,
        ];
    }

    public function articleAdded(KBArticle $article)
    {
        $this->article_count++;
    }

    public function articleRemoved(KBArticle $article)
    {
        $this->article_count--;
    }

    /**
     * @return \XF\Draft
     */
    public function getDraftArticle()
    {
        return \XF\Draft::createFromEntity($this, 'DraftArticles');
    }

    public function rebuildCounters()
    {
        $this->article_count = $this->db()->fetchOne("
            SELECT COUNT(*)
            FROM xf_nf_tickets_kb_article
            WHERE kb_category_id = ?
        ", $this->kb_category_id);
    }

    protected function _postDelete()
    {
        if ($this->getOption('delete_articles'))
        {
            $this->app()->jobManager()->enqueueUnique('nfTicketsKBCategoryDelete' . $this->kb_category_id, 'NF\Tickets:KBCategoryDelete', [
                'kb_category_id' => $this->kb_category_id,
            ]);
        }
    }

    /**
     * @param Structure $structure
     * @return Structure
     */
    public static function getStructure(Structure $structure)
    {
        $structure->table = 'xf_nf_tickets_kb_category';
        $structure->shortName = 'NF\Tickets:KBCategory';
        $structure->primaryKey = 'kb_category_id';
        $structure->contentType = 'nf_tickets_kb_category';
        $structure->columns = [
            'kb_category_id' => ['type' => self::UINT, 'autoIncrement' => true],
            'title'          => [
                'type'     => self::STR,
                'required' => 'please_enter_valid_title', 'maxLength' => 100,
            ],
            'description'    => ['type' => self::STR, 'default' => ''],
        ];
        $structure->getters = [
            'draft_article' => true,
        ];
        $structure->relations = [
            'DraftArticles' => [
                'entity'     => 'XF:Draft',
                'type'       => self::TO_MANY,
                'conditions' => [
                    ['draft_key', '=', 'nfTickets-kb-category-', '$kb_category_id'],
                ],
                'key'        => 'user_id',
            ],
            'Articles' => [
                'entity' => 'NF\Tickets:KBArticle',
                'type' => self::TO_MANY,
                'conditions' => 'kb_category_id'
            ]
        ];
        $structure->options = [
            'delete_articles' => true,
        ];

        static::addCategoryTreeStructureElements($structure);

        return $structure;
    }

    public function getPopularArticles($limit = 3)
    {
        return $this->getRelationFinder('Articles')->limit($limit)->order('views', 'desc')->fetch();
    }
}