<?php

namespace NF\Tickets\Entity;

use XF\Entity\Attachment;
use XF\Entity\BookmarkTrait;
use XF\Entity\QuotableInterface;
use XF\Entity\LinkableInterface;
use XF\Entity\ReactionTrait;
use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Structure;

/**
 * COLUMNS
 * @property int|null $message_id
 * @property int $ticket_id
 * @property int $user_id
 * @property string $username
 * @property string $user_name
 * @property string $user_email
 * @property int $message_date
 * @property string $message
 * @property int $ip_id
 * @property string $message_state
 * @property int $attach_count
 * @property int $warning_id
 * @property string $warning_message
 * @property int $position
 * @property int $last_edit_date
 * @property int $last_edit_user_id
 * @property int $edit_count
 * @property array|null $embed_metadata
 * @property bool $anonymized
 * @property string|null $user_agent
 * @property int $reaction_score
 * @property array $reactions_
 * @property array $reaction_users_
 *
 * GETTERS
 * @property \XF\Entity\User $AnonymousUser
 * @property mixed $Unfurls
 * @property mixed $reactions
 * @property mixed $reaction_users
 *
 * RELATIONS
 * @property \NF\Tickets\Entity\Ticket $Ticket
 * @property \XF\Entity\User $User
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\Attachment[] $Attachments
 * @property \XF\Entity\DeletionLog $DeletionLog
 * @property \XF\Entity\ApprovalQueue $ApprovalQueue
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\ReactionContent[] $Reactions
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\BookmarkItem[] $Bookmarks
 */
class Message extends Entity implements QuotableInterface, LinkableInterface, \XF\BbCode\RenderableContentInterface
{
    use ReactionTrait;
    use BookmarkTrait;

    /**
     * @param null $error
     * @return bool
     */
    public function canView(&$error = null)
    {
        $ticket = $this->Ticket;
        if ($ticket === null)
        {
            return false;
        }

        if (!$ticket->canView($error))
        {
            return false;
        }

        if ($this->message_state === 'deleted' && !$ticket->Category->canViewDeletedTickets())
        {
            return false;
        }
        if ($this->message_state === 'hidden' && (\XF::visitor()->user_id === $this->user_id || !$ticket->Category->canViewModeratedTickets()))
        {
            return false;
        }

        return true;
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canEdit(&$error = null)
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
            return false;
        }

        $ticket = $this->Ticket;

        if ($ticket->hasPermission('editAnyMessage'))
        {
            return true;
        }

        if ($this->user_id === $visitor->user_id && $ticket->hasPermission('editOwnMessage'))
        {
            $editLimit = $ticket->hasPermission('editOwnMessageTimeLimit');
            if ($editLimit !== -1 && (!$editLimit || $this->message_date < \XF::$time - 60 * $editLimit))
            {
                $error = \XF::phrase('message_edit_time_limit_expired', ['editLimit' => $editLimit]);

                return false;
            }

	        if (!$this->Ticket->Category->allow_opening)
	        {
		        $error = \XF::phraseDeferred(\XF::phrase('nf_tickets_you_may_not_perform_this_action_because_this_category_does_not_permit'));
		        return false;
	        }

            return true;
        }

        return false;
    }

    public function canEditSilently(&$error = null)
    {
        $ticket = $this->Ticket;
        $visitor = \XF::visitor();
        if (!$visitor->user_id || !$ticket)
        {
            return false;
        }

        if ($ticket->hasPermission('editAnyMessage'))
        {
            return true;
        }

        return false;
    }

    public function canViewHistory(&$error = null)
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
            return false;
        }

        if (!$this->app()->options()->editHistory['enabled'])
        {
            return false;
        }

        if ($this->Ticket->hasPermission('editAnyMessage'))
        {
            return true;
        }

        return false;
    }

    /**
     * @param string $type
     * @param null   $error
     * @return bool
     */
    public function canDelete($type = 'soft', &$error = null)
    {
        $visitor = \XF::visitor();

        $ticket = $this->Ticket;

        if ($type !== 'soft' && !$ticket->hasPermission('hardDeleteAnyMessage'))
        {
            return false;
        }

        if ($ticket->hasPermission('deleteAnyMessage'))
        {
            return true;
        }

        if ($this->user_id === $visitor->user_id && $ticket->hasPermission('deleteOwnMessage'))
        {
            $editLimit = $ticket->hasPermission('editOwnMessageTimeLimit');
            if ($editLimit !== -1 && (!$editLimit || $this->message_date < \XF::$time - 60 * $editLimit))
            {
                $error = \XF::phrase('message_edit_time_limit_expired', ['editLimit' => $editLimit]);

                return false;
            }

	        if (!$this->Ticket->Category->allow_opening)
	        {
		        $error = \XF::phraseDeferred(\XF::phrase('nf_tickets_you_may_not_perform_this_action_because_this_category_does_not_permit'));
		        return false;
	        }

            return true;
        }

        return false;
    }

    public function canUndelete(&$error = null)
    {
        $visitor = \XF::visitor();
        $ticket = $this->Ticket;
        if (!$visitor->user_id || !$ticket)
        {
            return false;
        }

        return (bool)$ticket->hasPermission('undelete');
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canReact(&$error = null)
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
            return false;
        }

        if ($this->message_state !== 'visible')
        {
            return false;
        }

        if ($this->user_id === $visitor->user_id)
        {
            $error = \XF::phraseDeferred('reacting_to_your_own_content_is_considered_cheating');

            return false;
        }

        $ticket = $this->Ticket;
        if (!$ticket)
        {
            return false;
        }

        return (bool)$ticket->hasPermission('react');
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canBookmarkContent(&$error = null)
    {
        $visitor = \XF::visitor();
        if ($visitor->hasOption('nf_is_banned') && $visitor->hasOption('nf_is_banned'))
        {
            return false;
        }

        return $this->isVisible();
    }

    /**
     * @param null                 $error
     * @param \XF\Entity\User|null $asUser
     * @return bool
     */
    public function canReport(&$error = null, \XF\Entity\User $asUser = null)
    {
        $asUser = $asUser ?: \XF::visitor();

        return $asUser->canReport($error);
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canWarn(&$error = null)
    {
        $visitor = \XF::visitor();

        if ($this->warning_id
            || !$this->user_id
            || !$visitor->user_id
            || $this->user_id === $visitor->user_id
            || !$this->Ticket->hasPermission('warn')
        )
        {
            return false;
        }

        return ($this->User && $this->User->isWarnable());
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canCleanSpam(&$error = null)
    {
        return (\XF::visitor()->canCleanSpam() && $this->User && $this->User->isPossibleSpammer());
    }

    public function canSendModeratorActionAlert()
    {
        $visitor = \XF::visitor();

        return (
            $visitor->user_id
            && $visitor->user_id !== $this->user_id
            && $this->message_state === 'visible'
        );
    }

    public function canApproveUnapprove(&$error = null)
    {
        $visitor = \XF::visitor();

        return $visitor->user_id && $this->Ticket && $this->Ticket->hasPermission('approveUnapprove');
    }

    /**
     * @return bool
     */
    public function isIgnored()
    {
        return \XF::visitor()->isIgnoring($this->user_id);
    }

    /**
     * @return bool
     */
    public function isLiked()
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
            return false;
        }

        return isset($this->Likes[$visitor->user_id]);
    }

    /**
     * @return bool
     */
    public function isFirstMessage()
    {
        $ticket = $this->Ticket;
        if (!$ticket)
        {
            return false;
        }

        if ($this->message_id === $ticket->first_message_id)
        {
            return true;
        }

        // this can be called during an insert where the ticket hasn't actually been updated yet
        // just assume it's the first post
        if (!$ticket->ticket_id)
        {
            return true;
        }

        if (!$ticket->first_message_id && $this->message_date === $ticket->start_date)
        {
            return true;
        }

        return false;
    }

    /**
     * @return bool
     */
    public function isLastMessage()
    {
        $ticket = $this->Ticket;
        if (!$ticket)
        {
            return false;
        }

        if ($this->message_id === $ticket->last_message_id)
        {
            return true;
        }

        // this can be called during an insert where the ticket hasn't actually been updated yet
        // just assume it's the first post
        if (!$ticket->ticket_id)
        {
            return true;
        }

        if (!$ticket->last_message_id && $this->message_date === $ticket->last_message_date)
        {
            return true;
        }

        return false;
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canUseInlineModeration(&$error = null)
    {
        // Inline mod actions for ticket messages are not yet implemented
        //return $this->Ticket->canUseInlineModeration($error);
        return false;
    }

    /**
     * @return bool
     */
    public function isVisible()
    {
        return $this->message_state === 'visible';
    }

    /**
     * @return bool
     */
    public function isUnread()
    {
        if (!$this->Ticket)
        {
            return false;
        }

        $readDate = $this->Ticket->getVisitorReadDate();
        if ($readDate === null)
        {
            return false;
        }

        return $readDate < $this->message_date;
    }

    public function isAttachmentEmbedded($attachmentId)
    {
        if (!$this->embed_metadata)
        {
            return false;
        }

        if ($attachmentId instanceof Attachment)
        {
            $attachmentId = $attachmentId->attachment_id;
        }

        return isset($this->embed_metadata['attachments'][$attachmentId]);
    }

    public function getQuoteWrapper($inner)
    {
        $user = $this->getAnonymousUser(false);

        return '[QUOTE="'
               . ($user->username)
               . ', ticketMessage: ' . $this->message_id
               . ($user->user_id ? ", member: $user->user_id" : '')
               . '"]'
               . "\n" . $inner . "\n"
               . "[/QUOTE]\n";
    }

    public function getBbCodeRenderOptions($context, $type)
    {
        return [
            'entity'          => $this,
            'user'            => $this->User,
            'attachments'     => $this->attach_count ? $this->Attachments : [],
            'viewAttachments' => true,
            'unfurls'         => $this->Unfurls ?: [],
        ];
    }

    public function getUnfurls()
    {
        return $this->_getterCache['Unfurls'] ?? [];
    }

    public function setUnfurls($unfurls)
    {
        $this->_getterCache['Unfurls'] = $unfurls;
    }

    /**
     * @return \XF\Entity\User
     */
    protected function getPseudoUser()
    {
        $user = $this->User;
        if ($user)
        {
            return $user;
        }

        /** @var \XF\Repository\User $userRepo */
        $userRepo = $this->repository('XF:User');

        return $userRepo->getGuestUser($this->username);
    }

    /**
     * @return \XF\Entity\User
     */
    protected function getAnonymousUserGetter()
    {
        return $this->getAnonymousUser(false);
    }

    /**
     * @param bool $withPermissionCheck
     * @return \XF\Entity\User
     */
    public function getAnonymousUser($withPermissionCheck = true)
    {
        if (!$this->anonymized)
        {
            return $this->getPseudoUser();
        }

        $ticket = $this->Ticket;
        if (!$ticket)
        {
            return $this->getPseudoUser();
        }

        return $ticket->getAnonymousUser($withPermissionCheck);
    }

    public function softDelete($reason = '', \XF\Entity\User $byUser = null)
    {
        $byUser = $byUser ?: \XF::visitor();

        if ($this->message_state === 'deleted')
        {
            return false;
        }

        $this->message_state = 'deleted';

        /** @var \XF\Entity\DeletionLog $deletionLog */
        $deletionLog = $this->getRelationOrDefault('DeletionLog');
        $deletionLog->setFromUser($byUser);
        $deletionLog->delete_reason = $reason;

        $this->save();

        return true;
    }

	protected function verifyUserAgent(&$userAgent): bool
	{
		$userAgent = substr($userAgent, 0, 255);
		return true;
	}

    protected function _postSave()
    {
        $visibilityChange = $this->isStateChanged('message_state', 'visible');
        $approvalChange = $this->isStateChanged('message_state', 'hidden');
        $deletionChange = $this->isStateChanged('message_state', 'deleted');

        if ($this->isUpdate())
        {
            if ($visibilityChange === 'enter')
            {
                $this->messageMadeVisible();

                if ($approvalChange)
                {
                    $this->submitHamData();
                }
            }
            else if ($visibilityChange === 'leave')
            {
                $this->messageHidden();
            }

            if ($deletionChange === 'leave' && $this->DeletionLog)
            {
                $this->DeletionLog->delete();
            }

            if ($approvalChange === 'leave' && $this->ApprovalQueue)
            {
                $this->ApprovalQueue->delete();
            }
        }
        else
        {
            // insert
            if ($this->message_state === 'visible')
            {
                $this->messageInsertedVisible();
            }
        }

        if ($approvalChange === 'enter')
        {
            /** @var \XF\Entity\ApprovalQueue $approvalQueue */
            $approvalQueue = $this->getRelationOrDefault('ApprovalQueue', false);
            $approvalQueue->content_date = $this->message_date;
            $approvalQueue->save();
        }
        else if ($deletionChange === 'enter' && !$this->DeletionLog)
        {
            /** @var \XF\Entity\DeletionLog $delLog */
            $delLog = $this->getRelationOrDefault('DeletionLog', false);
            $delLog->setFromVisitor();
            $delLog->save();
        }

        $this->updateTicketRecord();

        if ($this->isUpdate() && $this->getOption('log_moderator'))
        {
            $this->app()->logger()->logModeratorChanges('nf_tickets_message', $this);
        }
    }

    protected function messageInsertedVisible()
    {
        $this->adjustUserMessageCountIfNeeded(1);
    }

    protected function messageMadeVisible()
    {

        $this->adjustUserMessageCountIfNeeded(1);
    }

    protected function messageHidden($hardDelete = false)
    {
        $this->adjustUserMessageCountIfNeeded(-1);

        /** @var \XF\Repository\UserAlert $alertRepo */
        $alertRepo = $this->repository('XF:UserAlert');
        $alertRepo->fastDeleteAlertsForContent('nf_tickets_message', $this->message_id);
    }

    protected function submitHamData()
    {
        /** @var \XF\Spam\ContentChecker $submitter */
        $submitter = $this->app()->container('spam.contentHamSubmitter');
        $submitter->submitHam('nf_tickets_message', $this->message_id);
    }

    protected function adjustUserMessageCountIfNeeded($amount)
    {
        if ($this->user_id
            && !$this->isFirstMessage()
            && $this->Ticket
            && $this->Ticket->Category
            && $this->Ticket->Category->count_messages
            && $this->Ticket->discussion_state === 'visible'
        )
        {
            $this->User->fastUpdate('message_count', \max(0, $this->User->message_count + $amount));
        }
    }

    protected function updateTicketRecord()
    {
        if (!$this->Ticket || !$this->Ticket->exists())
        {
            // inserting a ticket, don't try to write to it
            return;
        }

        $visibilityChange = $this->isStateChanged('message_state', 'visible');
        if ($visibilityChange === 'enter')
        {
            $this->Ticket->messageAdded($this);
            $this->Ticket->save();
        }
        else if ($visibilityChange === 'leave')
        {
            $this->Ticket->messageRemoved($this);
            $this->Ticket->save();
        }
    }

    protected function _postDelete()
    {
        if ($this->message_state === 'visible')
        {
            $this->messageHidden(true);
        }

        if ($this->Ticket && $this->message_state === 'visible')
        {
            $this->Ticket->messageRemoved($this);
            $this->Ticket->save();
        }

        if ($this->DeletionLog)
        {
            $this->DeletionLog->delete();
        }

        if ($this->ApprovalQueue)
        {
            $this->ApprovalQueue->delete();
        }


        if ($this->getOption('log_moderator'))
        {
            $this->app()->logger()->logModeratorAction('nf_tickets_message', $this, 'delete_hard');
        }

        $this->db()->delete('xf_edit_history', 'content_type = ? AND content_id = ?', ['nf_tickets_message', $this->message_id]);

        /** @var \XF\Repository\UserAlert $alertRepo */
        $alertRepo = $this->repository('XF:UserAlert');
        $alertRepo->fastDeleteAlertsForContent('nf_tickets_message', $this->message_id);

        /** @var \XF\Repository\Attachment $attachRepo */
        $attachRepo = $this->repository('XF:Attachment');
        $attachRepo->fastDeleteContentAttachments('nf_tickets_message', $this->message_id);
    }

	public function getContentUrl(bool $canonical = false, array $extraParams = [], $hash = null)
	{
		$route = $canonical ? 'canonical:ticket-message' : 'ticket-message';
		return $this->app()->router('public')->buildLink($route, $this, $extraParams, $hash);
	}

	public function getContentPublicRoute()
	{
		return 'ticket-message';
	}

	public function getContentTitle(string $context = '')
	{
		if ($this->isFirstMessage())
		{
			return $this->Ticket->getContentTitle($context);
		}

		return \XF::phrase('nf_tickets_message_in_ticket_x', ['title' => $this->Ticket->title]);
	}

    /**
     * @param Structure $structure
     * @return Structure
     */
    public static function getStructure(Structure $structure)
    {
        $structure->table = 'xf_nf_tickets_message';
        $structure->shortName = 'NF\Tickets:Message';
        $structure->primaryKey = 'message_id';
        $structure->contentType = 'nf_tickets_message';
        $structure->columns = [
            'message_id'        => ['type' => self::UINT, 'autoIncrement' => true, 'nullable' => true],
            'ticket_id'         => ['type' => self::UINT, 'required' => true],
            'user_id'           => ['type' => self::UINT, 'required' => true],
            'username'          => [
                'type'     => self::STR, 'maxLength' => 50,
                'required' => 'please_enter_valid_name',
            ],
            'user_name'         => [
                'type'     => self::STR, 'maxLength' => 50,
                'required' => 'please_enter_valid_name',
            ],
            'user_email'        => ['type' => self::STR, 'default' => '', 'maxLength' => 120],
            'message_date'      => ['type' => self::UINT, 'required' => true, 'default' => \XF::$time],
            'message'           => [
                'type'     => self::STR,
                'required' => 'please_enter_valid_message',
            ],
            // Pre-XF2.2.9 bug, where UINT assumes 32bit value, so explicitly set 'max'
            'ip_id'             => ['type' => self::UINT, 'max' => \PHP_INT_MAX, 'default' => 0],
            'message_state'     => [
                'type'          => self::STR, 'default' => 'visible',
                'allowedValues' => ['visible', 'moderated', 'hidden', 'deleted'],
            ],
            'attach_count'      => ['type' => self::UINT, 'max' => 65535, 'forced' => true, 'default' => 0],
            'warning_id'        => ['type' => self::UINT, 'default' => 0],
            'warning_message'   => ['type' => self::STR, 'default' => ''],
            'position'          => ['type' => self::UINT, 'forced' => true],
            'last_edit_date'    => ['type' => self::UINT, 'default' => 0],
            'last_edit_user_id' => ['type' => self::UINT, 'default' => 0],
            'edit_count'        => ['type' => self::UINT, 'forced' => true, 'default' => 0],
            'embed_metadata'    => ['type' => self::JSON_ARRAY, 'nullable' => true, 'default' => null],
            'anonymized'        => ['type' => self::BOOL, 'default' => false],
			'user_agent'        => ['type' => self::STR, 'maxLength' => 255, 'default' => null, 'nullable' => true],
        ];
        $structure->behaviors = [
            'XF:Reactable' => ['stateField' => 'message_state'],
            'XF:Indexable' => [
                'checkForUpdates' => ['message', 'user_id', 'ticket_id', 'message_date', 'message_state', 'anonymized'],
            ],
        ];
        $structure->getters = [
            'AnonymousUser' =>  ['getter' => 'getAnonymousUserGetter', 'cache' => false],
            'Unfurls' => true,
        ];
        $structure->relations = [
            'Ticket'        => [
                'entity'     => 'NF\Tickets:Ticket',
                'type'       => self::TO_ONE,
                'conditions' => 'ticket_id',
                'primary'    => true,
                'with'       => 'Category',
            ],
            'User'          => [
                'entity'     => 'XF:User',
                'type'       => self::TO_ONE,
                'conditions' => 'user_id',
                'primary'    => true,
            ],
            'Attachments'   => [
                'entity'     => 'XF:Attachment',
                'type'       => self::TO_MANY,
                'conditions' => [
                    ['content_type', '=', 'nf_tickets_message'],
                    ['content_id', '=', '$message_id'],
                ],
                'with'       => 'Data',
                'order'      => 'attach_date',
            ],
            'DeletionLog'   => [
                'entity'     => 'XF:DeletionLog',
                'type'       => self::TO_ONE,
                'conditions' => [
                    ['content_type', '=', 'nf_tickets_message'],
                    ['content_id', '=', '$message_id'],
                ],
                'primary'    => true,
            ],
            'ApprovalQueue' => [
                'entity'     => 'XF:ApprovalQueue',
                'type'       => self::TO_ONE,
                'conditions' => [
                    ['content_type', '=', 'nf_tickets_message'],
                    ['content_id', '=', '$message_id'],
                ],
                'primary'    => true,
            ],
        ];
        $structure->defaultWith = [
            'Ticket',
        ];
        $structure->options = [
            'log_moderator' => true,
        ];

        static::addReactableStructureElements($structure);
        static::addBookmarkableStructureElements($structure);

        return $structure;
    }
}