<?php

namespace NF\Tickets\Entity;

use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Structure;

/**
 * COLUMNS
 * @property int|null $status_id
 * @property bool $include_in_counts
 * @property bool $include_in_defaults
 * @property bool $is_close
 * @property bool $alert_on_transition
 * @property int $display_order
 * @property bool $active
 * @property string $css_class
 *
 * GETTERS
 * @property \XF\Phrase $title
 *
 * RELATIONS
 * @property \XF\Entity\Phrase $MasterTitle
 */
class Status extends Entity
{
    /**
     * @return \XF\Phrase
     */
    public function getTitle()
    {
        return \XF::phrase($this->getPhraseName());
    }

    public function getPhraseName()
    {
        return 'nf_tickets_status.' . $this->status_id;
    }

    public function canDelete(): bool
    {
        return !$this->isSystemStatus();
    }

    public function isSystemStatus(): bool
    {
        return in_array($this->status_id, [
            (int)(\XF::options()->nftOpenStatus ?? 1),
            (int)(\XF::options()->nftResolvedStatus ?? 0),
            (int)(\XF::options()->nftDeletedStatus ?? 0),
        ], true);
    }

    public function getMasterPhrase()
    {
        $phrase = $this->MasterTitle;

        if (!$phrase)
        {
            /** @var \XF\Entity\Phrase $phrase */
            $phrase = $this->_em->create('XF:Phrase');
            $phrase->title = $this->_getDeferredValue(function () { return $this->getPhraseName(); }, 'save');
            $phrase->language_id = 0;
            $phrase->addon_id = '';
        }

        return $phrase;
    }

    protected function _preDelete()
    {
        if ($this->isSystemStatus())
        {
            $this->error(\XF::phrase('nfTickets_cannot_delete_system_statuses'));
        }
    }

    protected function _postSave()
    {
        if ($this->isChanged('include_in_counts'))
        {
            $this->rebuildTicketCounts();
        }
    }

    protected function _postDelete()
    {
        $this->rebuildTicketCounts();

        if ($this->MasterTitle)
        {
            $this->MasterTitle->delete();
        }
    }

    protected function rebuildTicketCounts()
    {
        \XF::runOnce('nfTicketCountsRebuild', function () {
            $this->getTicketRepo()->rebuildTicketCounts();
        });
    }

    public static function getStructure(Structure $structure)
    {
        $structure->table = 'xf_nf_tickets_status';
        $structure->shortName = 'NF\Tickets:Status';
        $structure->primaryKey = 'status_id';
        $structure->columns = [
            'status_id'           => ['type' => self::UINT, 'autoIncrement' => true, 'nullable' => true],
            'include_in_counts'   => ['type' => self::BOOL, 'default' => 1],
            'include_in_defaults' => ['type' => self::BOOL, 'default' => 1],
            'is_close'            => ['type' => self::BOOL, 'default' => 0],
            'alert_on_transition' => ['type' => self::BOOL, 'default' => 1],
            'display_order'       => ['type' => self::UINT, 'default' => 0],
            'active'              => ['type' => self::BOOL, 'default' => true],
            'css_class'           => ['type' => self::STR, 'maxLength' => 50, 'default' => 'label label--primary'],
        ];
        $structure->behaviors = [];
        $structure->getters = [
            'title' => true,
        ];
        $structure->relations = [
            'MasterTitle' => [
                'entity'     => 'XF:Phrase',
                'type'       => self::TO_ONE,
                'conditions' => [
                    ['language_id', '=', 0],
                    ['title', '=', 'nf_tickets_status.', '$status_id'],
                ],
            ],
        ];

        return $structure;
    }

    protected function getStatusRepo(): \NF\Tickets\Repository\Status
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Status');
    }

    protected function getTicketRepo(): \NF\Tickets\Repository\Ticket
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Ticket');
    }
}