<?php

namespace NF\Tickets\Entity;

use XF\Entity\BookmarkTrait;
use XF\Entity\ReactionTrait;
use XF\Entity\User;
use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Structure;
use XF\Entity\LinkableInterface;

use function count;

/**
 * COLUMNS
 * @property int|null $ticket_id
 * @property string $ticket_ref
 * @property string $title
 * @property int $user_id
 * @property string $username
 * @property string $user_name
 * @property string $user_email
 * @property string|null $password
 * @property int $starter_user_id
 * @property string $starter_username
 * @property int $start_date
 * @property int $first_message_id
 * @property int $first_message_date
 * @property int $first_message_reaction_score
 * @property array $first_message_reactions
 * @property int|null $priority
 * @property int $status_id
 * @property string $ticket_state
 * @property bool $ticket_locked
 * @property string $discussion_state
 * @property string $discussion_type
 * @property int $assigner_user_id
 * @property string $assigner_username
 * @property int $assigned_user_id
 * @property string $assigned_username
 * @property int $ticket_category_id
 * @property int $last_message_id
 * @property int $last_message_date
 * @property int $last_message_user_id
 * @property string $last_message_username
 * @property int $last_modified_date
 * @property int $last_modified_user_id
 * @property string $last_modified_username
 * @property int $last_status_change_date
 * @property int $reply_count
 * @property array $custom_fields_
 * @property int $prefix_id
 * @property int $thread_id
 *
 * GETTERS
 * @property int[] $message_ids
 * @property \XF\Draft $draft_message
 * @property \XF\CustomField\Set $custom_fields
 * @property array $last_message_cache
 *
 * RELATIONS
 * @property \NF\Tickets\Entity\Category $Category
 * @property \XF\Entity\User $User
 * @property \XF\Entity\User $Assignee
 * @property \NF\Tickets\Entity\Message $FirstMessage
 * @property \NF\Tickets\Entity\Message $LastMessage
 * @property \XF\Entity\User $LastMessageUser
 * @property \NF\Tickets\Entity\TicketPrefix $Prefix
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\TicketRead[] $Read
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\TicketWatch[] $Watch
 * @property \XF\Entity\DeletionLog $DeletionLog
 * @property \XF\Entity\ApprovalQueue $ApprovalQueue
 * @property \XF\Mvc\Entity\AbstractCollection|\XF\Entity\Draft[] $DraftMessages
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\TicketFieldValue[] $CustomFields
 * @property \XF\Entity\Thread $Thread
 * @property \NF\Tickets\Entity\Priority $Priority
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\TicketSpamLog[] $SpamLogs
 * @property \NF\Tickets\Entity\Status $Status
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\TicketParticipant[] $Participants
 * @property \XF\Mvc\Entity\AbstractCollection|\NF\Tickets\Entity\TicketGuestParticipant[] $GuestParticipants
 */
class Ticket extends Entity implements LinkableInterface
{
	use BookmarkTrait, ReactionTrait;

	/**
	 * @param bool $includeSelf
	 *
	 * @return array
	 */
	public function getBreadcrumbs($includeSelf = true)
	{
		$breadcrumbs = $this->Category ? $this->Category->getBreadcrumbs() : [];
		if ($includeSelf)
		{
			$breadcrumbs[] = [
				'href' => $this->app()->router()->buildLink('tickets', $this),
				'value' => $this->title
			];
		}

		return $breadcrumbs;
	}

	protected function checkGuestVisibility()
	{
		if (\XF::visitor()->user_id)
		{
		    return false;
		}

		if ($this->discussion_state === 'deleted')
		{
			return false;
		}

		$guestTicketAuth = \XF::session()->get('nfGuestTicketAuth') ?: [];

		if (!isset($guestTicketAuth[$this->ticket_id]))
		{
			return false;
		}

		$ticketAuth = $guestTicketAuth[$this->ticket_id];
		$nonce = $ticketAuth['nonce'];

		$parts = \explode(',', $nonce);
		if (count($parts) === 2)
		{
			list($nonceTime, $nonceValue) = $parts;

			if ($nonceTime + 60 * 60 < \XF::$time)
			{
				$this->getTicketRepo()->removeGuestTicketAccess($this);

				return false;
			}

			/** @var \Closure $validator */
			$validator = $this->app()['csrf.validator'];

			if ($validator($this->password, $nonceTime) === $nonceValue)
			{
				return true;
			}
			else
            {
                return $this->checkGuestParticipantVisibility($nonceTime, $nonceValue);
            }
		}

		return false;
	}

    protected function checkGuestParticipantVisibility($nonceTime, $nonceValue)
    {
        $guestParticipants = $this->GuestParticipants;

        foreach ($guestParticipants AS $guestParticipant)
        {
            /** @var \Closure $validator */
            $validator = $this->app()['csrf.validator'];

            if ($validator($guestParticipant->password, $nonceTime) === $nonceValue)
            {
                // TODO: Some sort of indication as to which participant is viewing?

                return true;
            }
        }

        return false;
	}

	protected function userIsParticipant(User $visitor)
	{
		return isset($this->Participants[$visitor->user_id]);
	}

	protected function hasGuestParticipants()
    {
        return $this->getRelationFinder('GuestParticipants')->total() > 0;
    }

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canView(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		if (!$this->Category || !$this->Category->canView())
		{
			return false;
		}

		$visitor = \XF::visitor();

		if (!$visitor->user_id)
		{
			return $this->checkGuestVisibility();
		}

		if (!$this->hasPermission('view'))
		{
			return false;
		}

		if ($this->discussion_state === 'deleted' && !$this->Category->canViewDeletedTickets())
        {
            return false;
        }

		if ($visitor->user_id !== $this->user_id && $this->hasPermission('viewOthers'))
		{
			return true;
		}
		elseif ($this->hasPermission('viewSelf') && ($visitor->user_id === $this->user_id || $this->userIsParticipant($visitor)))
		{
			return true;
		}

		return false;
	}

    public function viewableToGuests()
    {
        return ((!$this->user_id || $this->hasGuestParticipants()) && ($this->discussion_state !== 'deleted'));
    }

    public function confirmPassword($password)
    {
        $validPasswords = \array_merge([$this->password], $this->GuestParticipants->pluckNamed('password'));
        $passwordAttempts = \array_unique([$password, \utf8_trim($password)]);

        $hasValidPassword = false;
        foreach ($passwordAttempts as $passwordAttempt)
        {
            foreach ($validPasswords as $validPassword)
            {
                $hasValidPassword = \password_verify($passwordAttempt, $validPassword) || $hasValidPassword;
            }
        }

        return $hasValidPassword;
    }

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canReact(&$error = null)
	{
		$visitor = \XF::visitor();
		if (!$visitor->user_id)
		{
			return false;
		}

		if ($this->discussion_state !== 'visible')
		{
			return false;
		}

		if ($this->user_id === $visitor->user_id)
		{
			$error = \XF::phraseDeferred('reacting_to_your_own_content_is_considered_cheating');
			return false;
		}

		if (!$this->Category)
		{
			return false;
		}

		return (bool)$this->hasPermission('react');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canWatch(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		$visitor = \XF::visitor();

		if (!$visitor->user_id)
		{
			return false;
		}

		return true;
	}

	/**
	 * @return bool
	 */
	public function canViewDeletedMessages()
	{
		return (bool)$this->hasPermission('viewDeleted');
	}

    /**
     * @return bool
     */
    public function canViewModeratedPosts()
    {
        return (bool)$this->hasPermission('viewModerated');
    }

    public function canApproveUnapprove(&$error = null)
    {
        $visitor = \XF::visitor();
        return $visitor->user_id && $this->hasPermission('approveUnapprove');
    }

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canViewAttachments(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('viewAttachment');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canUndelete(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('undelete');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canPreview(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		// assumes view check has already been run

		/** @var \NF\Tickets\XF\Entity\User $visitor */
		$visitor = \XF::visitor();
		$categoryId = $this->ticket_category_id;

		return (
			$this->discussion_type !== 'redirect'
			&& $this->first_message_id
			&& $this->app()->options()->discussionPreview
			&& $visitor->hasTicketCategoryPermission($categoryId, 'view')
		);
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canEdit(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
	    if ($this->hasPermission('editAnyTicket'))
	    {
	        return true;
        }

	    // TODO: Edit own ticket permission?
	    if (\XF::visitor()->user_id && $this->hasPermission('manageOwnPriority'))
        {
            return true;
        }

	    return false;
	}

    /**
     * @param null $error
     *
     * @return bool
     * @internal May be removed in a future version.
     */
    public function canEditTitle(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
    {
        return (bool)$this->hasPermission('editAnyTicket');
    }

    /**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canMove(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('manageAnyTicket');
	}

	public function canMerge(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('manageAnyTicket');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canBookmarkContent(&$error = null)
	{
        $visitor = \XF::visitor();
	    if ($visitor->hasOption('nf_is_banned') && $visitor->hasOption('nf_is_banned'))
        {
            return false;
        }

        return $this->isVisible();
	}

	/**
	 * @param string $type
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canDelete($type = 'soft', &$error = null)
	{
        if ($type !== 'soft' && !$this->hasPermission('hardDeleteAnyTicket'))
        {
            return false;
        }

        if ($this->hasPermission('deleteAnyTicket'))
        {
            return true;
        }

        return false;
	}

	public function canSendModeratorActionAlert()
	{
		$visitor = \XF::visitor();

		return (
			$visitor->user_id
			&& $visitor->user_id !== $this->user_id
			&& $this->discussion_state === 'visible'
		);
	}

    /**
     * @param string|null $error
     * @return bool
     * @noinspection PhpUnusedParameterInspection
     */
	public function canViewPriority(&$error = null)
    {
        $category = $this->Category;
        if (!$category || !$category->priority_selection)
        {
            return false;
        }

        if ($this->hasPermission('managePriority'))
        {
            return true;
        }

        if (\XF::visitor()->user_id === $this->user_id)
        {
            if (!$this->hasPermission('viewOwnPriority'))
            {
                return false;
            }

            return true;
        }

        return false;
    }

	/**
	 * @param string|null $error
	 * @return bool
	 * @noinspection PhpUnusedParameterInspection
     */
    public function canUpdatePriority(&$error = null)
	{
	    $category = $this->Category;
	    if (!$category || !$category->priority_selection)
        {
            return false;
        }

	    if ($this->hasPermission('managePriority'))
		{
		    return true;
        }

		if (\XF::visitor()->user_id === $this->user_id)
        {
            if (!$this->hasPermission('viewOwnPriority'))
            {
                return false;
            }

            if (!$this->hasPermission('manageOwnPriority'))
            {
                return false;
            }

            return true;
        }

		return false;
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canLockUnlockTicket(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('lockUnlockTicket');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canUpdateStatus(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('manageAnyTicket');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canViewEmails(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		$visitor = \XF::visitor();

		return (
			$visitor->user_id
			&& $this->hasPermission('viewEmail')
		);
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canManage(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		$visitor = \XF::visitor();

		return (
			$visitor->user_id
			&& $this->hasPermission('manageAnyTicket')
		);
	}

    /**
     * @param null $error
     *
     * @return bool
     */
    public function canResolve(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
    {
        return !$this->Status->is_close && ($this->canManage() || (\XF::visitor()->user_id === $this->user_id && $this->hasPermission('resolveSelf')));
	}

	public function canConvertToThread(&$error = null)
	{
		return \XF::visitor()->user_id && $this->hasPermission('convertToThread');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canChangeDiscussionThread(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		$visitor = \XF::visitor();

		return (
			$visitor->user_id
			&& $this->hasPermission('editThreadId')
		);
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canUseInlineModeration(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		$visitor = \XF::visitor();
		return ($visitor->user_id && $this->hasPermission('inlineMod'));
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canViewModeratorLogs(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		$visitor = \XF::visitor();
		return $visitor->user_id && ($this->hasPermission('editAnyTicket') || $this->hasPermission('deleteAnyTicket'));
	}

    /**
     * @param null $error
     *
     * @return bool
     */
    public function canViewSpamLogs(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
    {
        $visitor = \XF::visitor();
        return $visitor->canApproveRejectUser() && $this->canManage();
    }

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canReassign(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		return (bool)$this->hasPermission('reassign');
	}

	/**
	 * @param null $error
	 *
	 * @return bool
	 */
	public function canReply(/** @noinspection PhpUnusedParameterInspection */ &$error = null)
	{
		if ($this->ticket_locked && !$this->canLockUnlockTicket())
		{
			return false;
		}

		if (!$this->Category->allow_opening)
		{
			$error = \XF::phraseDeferred(\XF::phrase('nf_tickets_you_may_not_perform_this_action_because_this_category_does_not_permit'));
			return false;
		}

		return (bool)$this->hasPermission('postMessage');
	}

    public function getExpectedThreadTitle($currentValues = true)
    {
        $title = $currentValues ? $this->getValue('title') : $this->getExistingValue('title');

        $template = '{title}';
        $threadTitle = \str_replace('{title}', $title, $template);

        return $this->app()->stringFormatter()->wholeWordTrim($threadTitle, 100);
    }

	/**
	 * @return \XF\Draft
	 */
	public function getDraftMessage()
	{
		return \XF\Draft::createFromEntity($this, 'DraftMessages');
	}

	public function isAssignedTo($userId = null)
	{
		if ($userId === null)
		{
			$userId = \XF::visitor()->user_id;
		}
		if ($userId instanceof User)
		{
			$userId = $userId->user_id;
		}

		return $this->assigned_user_id && $this->assigned_user_id === $userId;
	}

	public function isUnread()
	{
		if ($this->discussion_state === 'deleted')
		{
			return false;
		}

		if ($this->discussion_type === 'redirect')
		{
			return false;
		}

		$readDate = $this->getVisitorReadDate();
		if ($readDate === null)
		{
			return false;
		}

		return $readDate < $this->last_message_date;
	}

	/**
	 * @return \XF\CustomField\Set
	 */
	public function getCustomFields()
	{
		/** @var \XF\CustomField\DefinitionSet $fieldDefinitions */
		$fieldDefinitions = $this->app()->container('customFields.tickets');

		return new \XF\CustomField\Set($fieldDefinitions, $this);
	}

	/**
	 * @return string
	 */
	public function getFieldEditMode()
	{
		$visitor = \XF::visitor();

		$isSelf = ($visitor->user_id === $this->user_id || !$this->ticket_id);
		$isMod = ($visitor->user_id && $this->hasPermission('editAnyTicket'));

		if ($isMod || !$isSelf)
		{
			return $isSelf ? 'moderator_user' : 'moderator';
		}

		return 'user';
	}

	/**
	 * @return bool
	 */
	public function hasViewableThread()
	{
		if (!$this->thread_id)
		{
			return false;
		}

		$thread = $this->Thread;
		if (!$thread)
		{
			return false;
		}

		return $thread->canView();
	}

	/**
	 * @return bool
	 */
	public function hasExtraInfoTab()
	{
		if (!$this->getValue('custom_fields'))
		{
			// if they haven't set anything, we can bail out quickly
			return false;
		}

        $fieldSet = $this->custom_fields;
		$definitionSet = $fieldSet->getDefinitionSet()
			->filterOnly($this->Category->field_cache)
			->filterGroup('extra_tab')
			->filterWithValue($fieldSet);

		return ($definitionSet->count() > 0);
	}

	/**
	 * @return array
	 */
	public function getExtraFieldTabs()
	{
		if (!$this->getValue('custom_fields'))
		{
			// if they haven't set anything, we can bail out quickly
			return [];
		}

        $fieldSet = $this->custom_fields;
		$definitionSet = $fieldSet->getDefinitionSet()
			->filterOnly($this->Category->field_cache)
			->filterGroup('new_tab')
			->filterWithValue($fieldSet);

		$output = [];
		foreach ($definitionSet AS $fieldId => $definition)
		{
			$output[$fieldId] = $definition->title;
		}

		return $output;
	}

	/**
	 * @param string $permission
	 *
	 * @return bool|int
	 */
	public function hasPermission(string $permission)
	{
		/** @var \NF\Tickets\XF\Entity\User $visitor */
		$visitor = \XF::visitor();
		return $visitor->hasTicketCategoryPermission($this->ticket_category_id, $permission);
	}

	public function getNewParticipant(User $user)
	{
		/** @var TicketParticipant $participant */
		$participant = $this->em()->create('NF\Tickets:TicketParticipant');
		$participant->ticket_id = $this->ticket_id;
		$participant->user_id = $user->user_id;

		return $participant;
	}

    public function getNewGuestParticipant($email, $username, &$password = null)
    {
        /** @var TicketGuestParticipant $guestParticipant */
        $guestParticipant = $this->em()->create('NF\Tickets:TicketGuestParticipant');
        $guestParticipant->ticket_id = $this->ticket_id;
        $guestParticipant->email = $email;
        $guestParticipant->username = $username;

        if ($password === null)
        {
            $password = $this->getTicketRepo()->generatePassword();
            $guestParticipant->password = password_hash($password, PASSWORD_DEFAULT);
        }

        return $guestParticipant;
	}

	/**
	 * @return Message
	 */
	public function getNewMessage()
	{
		/** @var Message $message */
		$message = $this->em()->create('NF\Tickets:Message');
		$message->ticket_id = $this->_getDeferredValue(function()
		{
			return $this->ticket_id;
		}, 'save');
		return $message;
	}

	/**
	 * @return int[]
	 */
	public function getMessageIds()
	{
		return $this->db()->fetchAllColumn("
            SELECT message_id
			FROM xf_nf_tickets_message
			WHERE ticket_id = ?
			ORDER BY message_date
        ", $this->ticket_id);
	}

	/**
	 * @return null|int
	 */
	public function getVisitorReadDate()
	{
		$visitor = \XF::visitor();
		if (!$visitor->user_id)
		{
			return null;
		}

		$ticketRead = $this->Read[$visitor->user_id];
		$categoryRead = $this->Category ? $this->Category->Read[$visitor->user_id] : null;

		$dates = [\XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400];
		if ($ticketRead)
		{
			$dates[] = $ticketRead->ticket_read_date;
		}
		if ($categoryRead)
		{
			$dates[] = $categoryRead->category_read_date;
		}

		return \max($dates);
	}

	/**
	 * @return array
	 */
	public function getLastMessageCache()
	{
		return [
			'message_id' => $this->last_message_id,
			'user_id' => $this->last_message_user_id,
			'username' => $this->last_message_username,
			'message_date' => $this->last_message_date
		];
	}

	public function isVisible()
	{
		return ($this->discussion_state === 'visible');
	}

	public function rebuildCounters()
	{
		$this->rebuildFirstMessageInfo();
		$this->rebuildLastMessageInfo();
		$this->rebuildReplyCount();

		return true;
	}

	public function rebuildFirstMessageInfo()
	{
		$firstMessage = $this->db()->fetchRow("
			SELECT message_id, message_date, user_id, username, reaction_score, reactions
			FROM xf_nf_tickets_message
			WHERE ticket_id = ?
			ORDER BY message_date
			LIMIT 1
		", $this->ticket_id);
		if (!$firstMessage)
		{
			return false;
		}

        $this->start_date = $firstMessage['message_date'];
		$this->first_message_id = $firstMessage['message_id'];
		$this->first_message_date = $firstMessage['message_date'];
		$this->first_message_reaction_score = $firstMessage['reaction_score'];
		$this->first_message_reactions = json_decode($firstMessage['reactions'], true) ?: [];

		return true;
	}

	public function rebuildLastMessageInfo()
	{
		$lastMessage = $this->db()->fetchRow("
			SELECT message_id, message_date, user_id, username, anonymized
			FROM xf_nf_tickets_message
			WHERE ticket_id = ?
				AND message_state = 'visible'
			ORDER BY message_date DESC
			LIMIT 1
		", $this->ticket_id);
		if (!$lastMessage)
		{
			return false;
		}

		$this->last_message_id = $lastMessage['message_id'];
		$this->last_message_date = $lastMessage['message_date'];
		if ($lastMessage['anonymized'])
        {
            $category = $this->Category;
            $anonymousUser = $category ? $category->AnonymousUser : null;

            if ($anonymousUser)
            {
                $lastMessage['user_id'] = $anonymousUser->user_id;
                $lastMessage['username'] = $anonymousUser->username;
            }
            else
            {
                $lastMessage['user_id'] = 0;
                $lastMessage['username'] = '';
            }
        }
        $lastMessage['username'] = $lastMessage['username'] ?: '-';
        $this->last_message_user_id = $lastMessage['user_id'];
        $this->last_message_username = $lastMessage['username'];

		// These can't be rebuild accurately, however they may be missing if data is imported,
        // so use the last message as a best-effort.
		if (!$this->last_modified_date)
        {
            $this->last_modified_date = $lastMessage['message_date'];
        }

		if (!$this->last_modified_user_id || !$this->last_modified_username)
        {
            $this->last_modified_user_id = $lastMessage['user_id'];
            $this->last_modified_username = $lastMessage['username'];
        }

		if (!$this->last_status_change_date)
        {
            $this->last_status_change_date = $lastMessage['message_date'];
        }

		return true;
	}

	public function rebuildReplyCount()
	{
		$visibleMessages = $this->db()->fetchOne("
			SELECT COUNT(*)
			FROM xf_nf_tickets_message
			WHERE ticket_id = ?
				AND message_state = 'visible'
		", $this->ticket_id);
		$this->reply_count = \max(0, $visibleMessages - 1);

		return $this->reply_count;
	}

	public function messageAdded(Message $message)
	{
		if (!$this->first_message_id)
		{
			$this->first_message_id = $message->message_id;
		}
		else
		{
			$this->reply_count++;
		}

		if ($message->message_date >= $this->last_message_date)
		{
			$this->last_message_date = $message->message_date;
			$this->last_message_id = $message->message_id;
            $user = $message->getAnonymousUser(false);
            $this->last_message_user_id = $user->user_id;
            $this->last_message_username = $user->username;
		}

		unset($this->_getterCache['message_ids']);
	}

	public function messageRemoved(Message $message)
	{
		$this->reply_count--;

		if ($message->message_id === $this->last_message_id)
		{
			$this->rebuildLastMessageInfo();
		}

        unset($this->_getterCache['message_ids']);
	}

	public function canInvite()
	{
		return (bool)$this->hasPermission('invite');
	}

    public function isShadowBanned(): bool
    {
        return $this->discussion_state === 'shadow_ban';
    }

    /**
     * @param \XF\Phrase|string|null $error
     * @return bool
     */
    public function canViewNonanonymized(&$error = null)
    {
        $visitor = \XF::visitor();

        return $visitor->user_id && $this->hasPermission('viewNonanonymizedUser');
    }


    public function canActionAsAnonymized(&$error = null)
    {
        if (!$this->Category)
        {
            return false;
        }

        return $this->Category->canActionAsAnonymized($error);

    }

    /**
     * @return User
     */
    protected function getPseudoUser()
    {
        $user = $this->User;
        if ($user)
        {
            return $user;
        }

        /** @var \XF\Repository\User $userRepo */
        $userRepo = $this->repository('XF:User');

        return $userRepo->getGuestUser($this->username);
    }

    /**
     * @return User
     */
    protected function getAnonymousUserGetter()
    {
        return $this->getAnonymousUser(false);
    }

    /**
     * @param bool $withPermissionCheck
     * @return User
     */
    public function getAnonymousUser($withPermissionCheck = true)
    {

        if ($withPermissionCheck && $this->canViewNonanonymized())
        {
            return $this->getPseudoUser();
        }

        $category = $this->Category;
        if (!$category)
        {
            return $this->getPseudoUser();
        }

        $anonymousUser = $category->AnonymousUser;
        if (!$anonymousUser)
        {
            return $this->getPseudoUser();
        }

        return $anonymousUser;
    }

    protected function _preSave()
	{
        if ($this->prefix_id && ($this->isChanged(['prefix_id', 'ticket_category_id'])))
        {
            if (!$this->Category->isPrefixValid($this->prefix_id))
            {
                $this->prefix_id = 0;
            }
        }

		if ($this->isInsert())
		{
			/** @var \NF\Tickets\Repository\Ticket $ticketRepo */
			$ticketRepo = $this->repository('NF\Tickets:Ticket');

			if (!$this->ticket_ref)
			{
				$this->ticket_ref = $ticketRepo->generateReference(8);
			}
		}

		if ($this->isChanged('thread_id') && $this->thread_id)
        {
            $thread = $this->Thread;
            if ($thread && $thread->discussion_type !== 'discussion')
            {
                $this->error(\XF::phrase('nf_tickets_new_ticket_discussion_thread_must_be_standard_thread'));
            }
        }
	}

	protected function _postSave()
	{
        // on shadow-banned, disable most notification functionality
        if ($this->discussion_state === 'shadow_ban')
        {
            $shadowBanRetention = (int)(\XF::options()->nftPurgeShadowBanned ?? -1);
            if (!$shadowBanRetention)
            {
                \XF::runLater(function () {
                    if ($this->exists())
                    {
                        $this->delete(false);
                    }
                });
            }

            return;
        }

        $visibilityChange = $this->isStateChanged('discussion_state', 'visible');
        $approvalChange = $this->isStateChanged('discussion_state', 'moderated');
        $deletionChange = $this->isStateChanged('discussion_state', 'deleted');

        if ($this->isUpdate())
        {
            if ($visibilityChange === 'enter')
            {
                $this->ticketMadeVisible();

                if ($approvalChange)
                {
                    $this->submitHamData();
                }
            }
            else if ($visibilityChange === 'leave')
            {
                $this->ticketHidden();
            }

            if ($this->isChanged('ticket_category_id'))
            {
                /** @var Category $oldCategory */
                $oldCategory = $this->getExistingRelation('Category');
                if ($oldCategory && $this->Category)
                {
                    $this->ticketMoved($oldCategory, $this->Category);
                }
            }

            if ($deletionChange === 'leave' && $this->DeletionLog)
            {
                $this->DeletionLog->delete();
            }

            if ($approvalChange === 'leave' && $this->ApprovalQueue)
            {
                $this->ApprovalQueue->delete();
            }
        }
        else // insert
        {
            if ($this->discussion_state === 'visible')
            {
                $this->ticketInsertedVisible();
            }
        }

        if ($approvalChange === 'enter')
        {
            /** @var \XF\Entity\ApprovalQueue $approvalQueue */
            $approvalQueue = $this->getRelationOrDefault('ApprovalQueue', false);
            $approvalQueue->content_date = $this->start_date;
            $approvalQueue->save();
        }
        else if ($deletionChange === 'enter' && !$this->DeletionLog)
        {
            /** @var \XF\Entity\DeletionLog $delLog */
            $delLog = $this->getRelationOrDefault('DeletionLog', false);
            $delLog->setFromVisitor();
            $delLog->save();
        }

        if ($this->isChanged('thread_id'))
        {
            $oldThreadId = $this->getPreviousValue('thread_id');
            /** @var \XF\Entity\Thread $oldThread */
            $oldThread = $oldThreadId ? \XF::app()->find('XF:Thread', $oldThreadId) : null;
            if ($oldThread && $oldThread->discussion_type === 'nf_tickets_ticket')
            {
                $oldThread->discussion_type = 'discussion';
                $oldThread->saveIfChanged($saved, false, false);
            }

            /** @var \XF\Entity\Thread $oldThread */
            $thread = $this->Thread;
            if ($thread && $thread->discussion_type === 'discussion')
            {
                $thread->discussion_type = 'nf_tickets_ticket';
                $thread->saveIfChanged($saved, false, false);
            }
        }

		$this->updateCategoryRecord();

		if ($this->isUpdate() && $this->getOption('log_moderator'))
		{
			$this->app()->logger()->logModeratorChanges('nf_tickets_ticket', $this);
		}

		$this->_postSaveBookmarks();

		$this->rebuildTicketCounts();
	}

    protected function ticketMadeVisible()
    {
        $this->adjustUserTicketCountIfNeeded(1);

        /** @var \XF\Repository\Reaction $reactionRepo */
        $reactionRepo = $this->repository('XF:Reaction');
        $reactionRepo->recalculateReactionIsCounted('nf_tickets_message', $this->message_ids);

        $deletedStatusId = (int)(\XF::options()->nftDeletedStatus ?? 0);
        $openStatusId = (int)(\XF::options()->nftOpenStatus ?? 1);
        if ($this->status_id === $deletedStatusId)
        {
            $this->fastUpdate('status_id', $openStatusId);
        }
    }

    protected function ticketInsertedVisible()
    {
        $this->adjustUserTicketCountIfNeeded(1);
    }

    protected function submitHamData()
    {
        /** @var \XF\Spam\ContentChecker $submitter */
        $submitter = $this->app()->container('spam.contentHamSubmitter');
        $submitter->submitHam('nf_tickets_ticket', $this->ticket_id);
    }

    protected function ticketMoved(Category $from, Category $to)
    {
        if (!$this->isStateChanged('discussion_state', 'visible'))
        {
            $newCounts = $to->count_messages;
            $oldCounts = $from->count_messages;
            if ($newCounts !== $oldCounts)
            {
                $this->adjustUserTicketCountIfNeeded($newCounts ? 1 : -1);
            }
        }
    }

    protected function adjustUserMessageCountIfNeeded($direction, $forceChange = false)
    {
        if ($forceChange || !empty($this->Category->count_messages))
        {
            $updates = $this->db()->fetchPairs("
				SELECT user_id, COUNT(*)
				FROM xf_nf_tickets_message
				WHERE ticket_id = ?
					AND user_id > 0
					AND message_state = 'visible'
				GROUP BY user_id
			", $this->ticket_id);

            $operator = $direction > 0 ? '+' : '-';
            foreach ($updates AS $userId => $adjust)
            {
                $this->db()->query("
					UPDATE xf_user
					SET message_count = GREATEST(0, CAST(message_count AS SIGNED) {$operator} ?)
					WHERE user_id = ?
				", [$adjust, $userId]);

                /** @var User $userEntity */
                $userEntity = $this->em()->findCached('XF:User', $userId);
                if ($userEntity)
                {
                    $userEntity->setAsSaved('message_count', \max(0, $userEntity->message_count + ($direction > 0 ? $adjust : -$adjust) ));
                }
            }
        }
    }

	public function quickEscalate()
	{
		/** @var \NF\Tickets\Repository\Escalation $ticketRepo */
		$ticketRepo = $this->repository('NF\Tickets:Escalation');
		$ticketRepo->quickEscalate($this);
	}

    /**
     * @param int $spamTriggerDays
     * @param int $spamTriggerLimit
     */
    public function logSpamTriggerRecords($spamTriggerDays = null, $spamTriggerLimit = null)
    {
        $options = $this->app()->options();
        if ($spamTriggerDays === null)
        {
            $spamTriggerDays = (int)$options->nfTicketSpamTriggerDays;
        }
        if ($spamTriggerLimit === null)
        {
            $spamTriggerLimit = (int)$options->nfTicketSpamTriggerLimit;
        }

        if (!$spamTriggerDays || !$spamTriggerLimit)
        {
            return;
        }

        $binaryIp = null;
        $firstMessage = $this->FirstMessage;
        if ($firstMessage && $firstMessage->ip_id)
        {
            $binaryIp = $this->db()->fetchOne('SELECT ip FROM xf_ip WHERE ip_id = ?', $firstMessage->ip_id);
        }

        /** @var \NF\Tickets\Repository\Ticket $ticketsRepo */
        $ticketsRepo = $this->repository('NF\Tickets:Ticket');
        $ticketsRepo->copySpamTriggerLogs($this->ticket_id, $this->start_date, $binaryIp, $this->user_email, $spamTriggerDays, $spamTriggerLimit);
    }

    /**
	 * @param bool $hardDelete
	 */
	protected function ticketHidden($hardDelete = false)
	{
	    $this->adjustUserTicketCountIfNeeded(-1);

        if (!$hardDelete)
        {
            // on hard delete the reactions will be removed which will do this
            /** @var \XF\Repository\Reaction $reactionRepo */
            $reactionRepo = $this->repository('XF:Reaction');
            $reactionRepo->recalculateReactionIsCounted('nf_tickets_message', $this->message_ids, false);

            $resolvedStatusId = (int)(\XF::options()->nftResolvedStatus ?? 0);
            $deletedStatusId = (int)(\XF::options()->nftDeletedStatus ?? 0);
            if ($this->status_id !== $resolvedStatusId && $this->status_id !== $deletedStatusId)
            {
                $this->fastUpdate('status_id', $deletedStatusId);
            }
        }

        /** @var \XF\Repository\UserAlert $alertRepo */
        $alertRepo = $this->repository('XF:UserAlert');
        $alertRepo->fastDeleteAlertsForContent('nf_tickets_message', $this->message_ids);
    }

	protected function adjustUserTicketCountIfNeeded($amount, $userId = null)
    {
        if ($userId === null)
        {
            $userId = $this->user_id;
        }

        if ($userId && $amount)
        {
            $this->db()->query("
                UPDATE xf_user
                SET nf_tickets_count = GREATEST(0, nf_tickets_count + ?)
                WHERE user_id = ?
            ", [$amount, $userId]);

            $this->adjustUserMessageCountIfNeeded($amount);
        }
    }

    protected function moveLinkThreadToMatchCategory()
    {
        $category = $this->Category;

        if ($this->thread_id && $this->Thread && ($this->Thread->discussion_type === 'nf_tickets_ticket'))
        {
            $thread = $this->Thread;
            if ($category->thread_node_id)
            {
                $thread->node_id = $category->thread_node_id;
                $thread->prefix_id = $category->thread_prefix_id;
                if ($this->discussion_state === 'visible' && $thread->discussion_state === 'deleted')
                {
                    // presumably the thread was soft deleted by being moved to a category without a thread
                    $thread->discussion_state = 'visible';
                }
            }
            else
            {
                // this category doesn't have a thread
                $thread->discussion_state = 'deleted';
            }

            $thread->saveIfChanged($saved, false, false);
        }
    }

	protected function updateCategoryRecord()
	{
		if (!$this->Category)
		{
			return;
		}

        $category = $this->Category;

		if ($this->isUpdate() && $this->isChanged('ticket_category_id'))
		{
			// moved, trumps the rest
			if ($this->discussion_state === 'visible')
			{
				$category->ticketAdded($this);
				$category->save();
			}

			if ($this->getExistingValue('discussion_state') === 'visible')
			{
				/** @var Category $oldCategory */
				$oldCategory = $this->getExistingRelation('Category');
				if ($oldCategory)
				{
					$oldCategory->ticketRemoved($this);
					$oldCategory->save();
				}
			}

            $this->moveLinkThreadToMatchCategory();

			return;
		}
		else if ($this->isChanged('thread_id'))
        {
            $this->moveLinkThreadToMatchCategory();
        }

        $visibilityChange = $this->isStateChanged('discussion_state', 'visible');
        if ($visibilityChange === 'enter' && $category)
        {
            $category->ticketAdded($this);
            $category->save();
        }
        elseif ($visibilityChange === 'leave' && $category)
        {
            $category->ticketRemoved($this);
            $category->save();
        }
        elseif ($this->isUpdate() && $this->discussion_state === 'visible')
        {
            $category->ticketDataChanged($this);
            $category->save();
        }
	}

	protected function _postDelete()
	{
        if ($this->thread_id && $this->Thread && ($this->Thread->discussion_type === 'nf_tickets_ticket'))
        {
            $this->Thread->discussion_type = 'discussion';
            $this->Thread->saveIfChanged($saved, true, false);
        }

		if ($this->discussion_state === 'visible')
		{
			$this->ticketHidden(true);
		}

		if ($this->Category && $this->discussion_state === 'visible')
        {
            $this->Category->ticketRemoved($this);
            $this->Category->save();
        }

		if ($this->discussion_state === 'deleted' && $this->DeletionLog)
		{
			$this->DeletionLog->delete();
		}

		if ($this->getOption('log_moderator'))
		{
			$this->app()->logger()->logModeratorAction(
				'nf_tickets_ticket',
				$this,
				'delete_hard'
			);
		}

		$db = $this->db();

        $db->delete('xf_nf_tickets_spam_log', 'ticket_id = ?', $this->ticket_id);
        $db->delete('xf_nf_tickets_ticket_read', 'ticket_id = ?', $this->ticket_id);
        $db->delete('xf_nf_tickets_ticket_watch', 'ticket_id = ?', $this->ticket_id);
        $db->delete('xf_nf_tickets_ticket_field_value', 'ticket_id = ?', $this->ticket_id);

        $db->delete('xf_edit_history', 'content_id = ? AND content_type = ?', [$this->ticket_id, 'nf_tickets_ticket']);
        $db->delete('xf_deletion_log', 'content_id = ? AND content_type = ?', [$this->ticket_id, 'nf_tickets_ticket']);

		/** @var \XF\Repository\Attachment $attachRepo */
		$attachRepo = $this->repository('XF:Attachment');
		$attachRepo->fastDeleteContentAttachments(
			'nf_tickets_ticket',
			$this->ticket_id
		);

		$this->_postDeleteBookmarks();

		$this->rebuildTicketCounts();
	}

	protected function rebuildTicketCounts()
	{
		\XF::runOnce('nfTicketCountsRebuild', function()
		{
			$this->getTicketRepo()->rebuildTicketCounts();
		});
	}

	/**
	 * @param string $reason
	 * @param User $byUser
	 *
	 * @return bool
	 */
	public function softDelete($reason = '', User $byUser = null)
	{
		$byUser = $byUser ?: \XF::visitor();

		if ($this->discussion_state === 'deleted')
		{
			return false;
		}

		$this->discussion_state = 'deleted';
		/** @var \XF\Entity\DeletionLog $deletionLog */
		$deletionLog = $this->getRelationOrDefault('DeletionLog');
		$deletionLog->setFromUser($byUser);
		$deletionLog->delete_reason = $reason;

		$this->save();

		return true;
	}

	/**
	 * @param \XF\Api\Result\EntityResult $result
	 * @param int $verbosity
	 * @param array $options
	 */
	protected function setupApiResultData(
		\XF\Api\Result\EntityResult $result, $verbosity = self::VERBOSITY_NORMAL, array $options = []
	)
	{
		$result->username = $this->User ? $this->User->username : $this->username;

		$visitor = \XF::visitor();
		if ($visitor->user_id)
		{
			$result->is_watching = isset($this->Watch[$visitor->user_id]);
			$result->is_unread = $this->isUnread();
		}

		if (!empty($options['skip_category']))
		{
			$result->skipRelation('Category');
		}

		$result->custom_fields = (object)$this->custom_fields->getNamedFieldValues($this->Category->field_cache);

		if ($this->prefix_id)
		{
			$result->prefix = \XF::phrase('nf_tickets_ticket_prefix.' . $this->prefix_id);
		}

		$result->can_edit = $this->canEdit();
		$result->can_reply = $this->canReply();
		$result->can_soft_delete = $this->canDelete();
		$result->can_hard_delete = $this->canDelete('hard');
		$result->can_view_attachments = $this->canViewAttachments();

		$result->view_url = $this->getContentUrl(true);
	}

	public function getContentUrl(bool $canonical = false, array $extraParams = [], $hash = null)
	{
		$route = $canonical ? 'canonical:tickets' : 'tickets';
		return $this->app()->router('public')->buildLink($route, $this, $extraParams, $hash);
	}

	public function getContentPublicRoute()
	{
		return 'tickets';
	}

	public function getContentTitle(string $context = '')
	{
		return \XF::phrase('nf_tickets_ticket_x', ['title' => $this->title]);
	}

	/**
	 * @param Structure $structure
	 *
	 * @return Structure
	 */
	public static function getStructure(Structure $structure)
	{
		$structure->table = 'xf_nf_tickets_ticket';
		$structure->shortName = 'NF\Tickets:Ticket';
		$structure->primaryKey = 'ticket_id';
		$structure->contentType = 'nf_tickets_ticket';
		$structure->columns = [
			'ticket_id' => ['type' => self::UINT, 'autoIncrement' => true, 'nullable' => true],
			'ticket_ref' => ['type' => self::STR, 'maxLength' => 100],
			'title' => ['type' => self::STR, 'maxLength' => 150,
				'required' => 'please_enter_valid_title',
				'censor' => true
			],
			'user_id' => ['type' => self::UINT, 'required' => true],
			'username' => ['type' => self::STR, 'maxLength' => 50,
				'required' => 'please_enter_valid_name'
			],
			'user_name' => ['type' => self::STR, 'maxLength' => 50,
				'required' => 'please_enter_valid_name'
			],
			'user_email' => ['type' => self::STR, 'default' => '', 'maxLength' => 120], // TODO: additional verification?
			'password' => ['type' => self::STR, 'default' => null, 'maxLength' => 255, 'nullable' => true],
			'starter_user_id' => ['type' => self::UINT, 'default' => 0],
			'starter_username' => ['type' => self::STR, 'default' => ''],
			'start_date' => ['type' => self::UINT, 'default' => \XF::$time],
			'first_message_id' => ['type' => self::UINT, 'default' => 0],
			'first_message_date' => ['type' => self::UINT, 'default' => 0],
			'first_message_reaction_score' => ['type' => self::INT, 'default' => 0, 'api' => true],
			'first_message_reactions' => ['type' => self::JSON_ARRAY, 'default' => []],
			'priority' => ['type' => self::UINT, 'nullable' => true, 'default' => null],
			'status_id' => ['type' => self::UINT, 'default' => (int)(\XF::options()->nftOpenStatus ?? 1)],
			'ticket_state' => ['type' => self::STR, 'default' => 'open',
				'allowedValues' => ['open', 'pending', 'locked', 'assigned', 'resolved']
			],
			'ticket_locked' => ['type' => self::BOOL, 'default' => false],
			'discussion_state' => ['type' => self::STR, 'default' => 'visible',
				'allowedValues' => ['visible', 'moderated', 'deleted', 'shadow_ban']
			],
			'discussion_type' => ['type' => self::STR, 'default' => ''],
			'assigner_user_id' => ['type' => self::UINT, 'default' => 0],
			'assigner_username' => ['type' => self::STR, 'default' => ''],
			'assigned_user_id' => ['type' => self::UINT, 'default' => 0],
			'assigned_username' => ['type' => self::STR, 'default' => ''],
			'ticket_category_id' => ['type' => self::UINT, 'required' => 'nf_tickets_please_select_valid_ticket_category'],
			'last_message_id' => ['type' => self::UINT, 'default' => 0],
			'last_message_date' => ['type' => self::UINT, 'default' => 0],
			'last_message_user_id' => ['type' => self::UINT, 'default' => 0],
			'last_message_username' => ['type' => self::STR, 'default' => ''],
			'last_modified_date' => ['type' => self::UINT, 'default' => 0],
			'last_modified_user_id' => ['type' => self::UINT, 'default' => 0],
            'last_modified_username' => ['type' => self::STR, 'default' => ''],
			'last_status_change_date' => ['type' => self::UINT, 'default' => 0],
			'reply_count' => ['type' => self::UINT, 'default' => 0],
			'custom_fields' => ['type' => self::JSON_ARRAY, 'default' => []],
            'prefix_id' => ['type' => self::UINT, 'default' => 0, 'api' => true],
			'thread_id' => ['type' => self::UINT, 'default' => 0]
		];
		$structure->getters = [
		    'message_ids' => true,
			'draft_message' => true,
			'custom_fields' => true,
			'last_message_cache' => true
		];
		$structure->behaviors = [
            'XF:IndexableContainer' => [
                'childContentType' => 'nf_tickets_message',
                'childIds' => function($ticket) { return $ticket->message_ids; },
                'checkForUpdates' => ['ticket_category_id', 'discussion_state', 'status_id', 'assigner_user_id', 'assigned_user_id']
            ],
			'XF:Taggable' => ['stateField' => 'discussion_state'],
			'XF:CustomFieldsHolder' => [
				'valueTable' => 'xf_nf_tickets_ticket_field_value',
				'checkForUpdates' => ['ticket_category_id'],
				'getAllowedFields' => function($ticket)
				{
					return $ticket->Category ? $ticket->Category->field_cache : [];
				}
			]
		];
		$structure->relations = [
			'Category' => [
				'entity' => 'NF\Tickets:Category',
				'type' => self::TO_ONE,
				'conditions' => 'ticket_category_id',
				'primary' => true,
				'api' => true
			],
			'User' => [
				'entity' => 'XF:User',
				'type' => self::TO_ONE,
				'conditions' => 'user_id',
				'primary' => true,
				'api' => true
			],
			'Assignee' => [
				'entity' => 'XF:User',
				'type' => self::TO_ONE,
				'conditions' => 'assigned_user_id',
				'primary' => true
			],
			'FirstMessage' => [
				'entity' => 'NF\Tickets:Message',
				'type' => self::TO_ONE,
				'conditions' => [['message_id', '=', '$first_message_id']],
				'primary' => true
			],
			'LastMessage' => [
				'entity' => 'NF\Tickets:Message',
				'type' => self::TO_ONE,
				'conditions' => [['message_id', '=', '$last_message_id']],
				'primary' => true
			],
			'LastMessageUser' => [
				'entity' => 'XF:User',
				'type' => self::TO_ONE,
				'conditions' => [['user_id', '=', '$last_message_user_id']],
				'primary' => true
			],
            'Prefix' => [
                'entity' => 'NF\Tickets:TicketPrefix',
                'type' => self::TO_ONE,
                'conditions' => 'prefix_id',
                'primary' => true
            ],
			'Read' => [
				'entity' => 'NF\Tickets:TicketRead',
				'type' => self::TO_MANY,
				'conditions' => 'ticket_id',
				'key' => 'user_id'
			],
			'Watch' => [
				'entity' => 'NF\Tickets:TicketWatch',
				'type' => self::TO_MANY,
				'conditions' => 'ticket_id',
				'key' => 'user_id'
			],
			'DeletionLog' => [
				'entity' => 'XF:DeletionLog',
				'type' => self::TO_ONE,
				'conditions' => [
					['content_type', '=', 'nf_tickets_ticket'],
					['content_id', '=', '$ticket_id']
				],
				'primary' => true
			],
            'ApprovalQueue' => [
                'entity' => 'XF:ApprovalQueue',
                'type' => self::TO_ONE,
                'conditions' => [
                    ['content_type', '=', 'nf_tickets_ticket'],
                    ['content_id', '=', '$ticket_id']
                ],
                'primary' => true
            ],
			'DraftMessages' => [
				'entity' => 'XF:Draft',
				'type' => self::TO_MANY,
				'conditions' => [
					['draft_key', '=', 'ticket-', '$ticket_id']
				],
				'key' => 'user_id'
			],
			'CustomFields' => [
				'entity' => 'NF\Tickets:TicketFieldValue',
				'type' => self::TO_MANY,
				'conditions' => 'ticket_id',
				'key' => 'field_id'
			],
			'Thread' => [
				'entity' => 'XF:Thread',
				'type' => self::TO_ONE,
				'conditions' => 'thread_id',
				'primary' => true
			],
			'Priority' => [
				'entity' => 'NF\Tickets:Priority',
				'type' => self::TO_ONE,
				'conditions' => 'priority',
				'primary' => true
			],
            'SpamLogs' => [
                'entity' => 'NF\Tickets:TicketSpamLog',
                'type' => self::TO_MANY,
                'conditions' => 'ticket_id',
                'order' => ['log_date', 'desc'],
            ],
			'Status' => [
				'entity' => 'NF\Tickets:Status',
				'type' => self::TO_ONE,
				'conditions' => 'status_id',
				'primary' => true
			],
			'Participants' => [
				'entity' => 'NF\Tickets:TicketParticipant',
				'type' => self::TO_MANY,
				'conditions' => 'ticket_id',
                'key' => 'user_id'
			],
            'GuestParticipants' => [
                'entity' => 'NF\Tickets:TicketGuestParticipant',
                'type' => self::TO_MANY,
                'conditions' => 'ticket_id'
            ]
		];
		$structure->options = [
		    'shadow_banned' => false,
			'log_moderator' => true
		];

		$structure->withAliases = [
			'api' => [
				'User.api',
				'Category.api'
			]
		];

		return $structure;
	}

	protected function getTicketRepo(): \NF\Tickets\Repository\Ticket
	{
        /** @noinspection PhpIncompatibleReturnTypeInspection */
		return $this->repository('NF\Tickets:Ticket');
	}
}
