<?php

namespace NF\Tickets\Entity;

use XF\Mvc\Entity\Entity;
use XF\Mvc\Entity\Structure;

use function is_array;

/**
 * COLUMNS
 * @property int|null $trigger_log_id
 * @property int $ticket_id
 * @property int $log_date
 * @property int $user_id
 * @property string $username
 * @property string $ip_address
 * @property string $result
 * @property array $details_
 * @property array $request_state
 *
 * GETTERS
 * @property \XF\Phrase $status
 * @property string[] $details
 *
 * RELATIONS
 * @property \XF\Entity\User $User
 * @property \NF\Tickets\Entity\Ticket $Ticket
 */
class TicketSpamLog extends Entity
{
    /**
     * @return \XF\Phrase
     */
    public function getStatus()
    {
        return \XF::phrase('nf_tickets_spam_status.' . ($this->result ?: 'allowed'));
    }

    /**
     * @return string[]
     */
    public function getDetails()
    {
        $output = [];

        $app = \XF::app();
        $language = \XF::app()->language();
        $isAdmin = $app instanceof \XF\Admin\App;
        $router = \XF::app()->router($isAdmin ? 'admin' : 'public');
        $linkUsername = $this->getOption('linkUsernameInDetails');

        foreach ($this->details_ AS $detail)
        {
            if (!isset($detail['phrase']))
            {
                continue;
            }
            $data = $detail['data'] ?? [];

            // ensure score is consistently formatted.
            if (isset($data['score']) && \is_numeric($data['score']) && $data['score'] > 0)
            {
                $data['score'] = "+" . $language->numberFormat((int)$data['score'], 0);
            }

            // flatten, workaround bug in earlier Signup Abuse Blocking
            foreach ($data as $key => &$value)
            {
                if (is_array($value))
                {
                    $value = \reset($value);
                }
            }

            switch ($detail['phrase'])
            {
                case 'sv_reg_log.multi_account_accept':
                case 'sv_reg_log.multi_account_accept_score':
                case 'sv_reg_log.multi_account_moderate':
                case 'sv_reg_log.multi_account_reject':
                    if ($linkUsername && isset($data['username']) && isset($data['user_id']))
                    {
                        $data['username'] = new \XF\PreEscaped('<a href="' . $router->buildLink($isAdmin ? 'canonical:users/edit' : 'canonical:members', $data) . '">' . \XF::escapeString($data['username']) . '</a>');
                    }
                    break;
            }

            $output[] = \XF::phrase($detail['phrase'], $data);
        }

        return $output;
    }

    /**
     * @param Structure $structure
     * @return Structure
     */
    public static function getStructure(Structure $structure)
    {
        $structure->table = 'xf_nf_tickets_spam_log';
        $structure->shortName = 'NF\Tickets:TicketSpamLog';
        $structure->primaryKey = 'trigger_log_id';
        $structure->columns = [
            'trigger_log_id' => ['type' => self::UINT, 'autoIncrement' => true, 'nullable' => true],
            'ticket_id'      => ['type' => self::UINT, 'required' => true],
            'log_date'       => ['type' => self::UINT, 'default' => \XF::$time],
            'user_id'        => ['type' => self::UINT, 'required' => true],
            'username'       => ['type' => self::STR, 'maxLength' => 100],
            'ip_address'     => ['type' => self::BINARY, 'maxLength' => 16],
            'result'         => ['type' => self::STR, 'required' => true],
            'details'        => ['type' => self::JSON_ARRAY, 'required' => true],
            'request_state'  => ['type' => self::JSON_ARRAY, 'required' => true],
        ];
        $structure->getters = [
            'status'  => true,
            'details' => true,
        ];
        $structure->relations = [
            'User'   => [
                'entity'     => 'XF:User',
                'type'       => self::TO_ONE,
                'conditions' => 'user_id',
                'primary'    => true,
            ],
            'Ticket' => [
                'entity'     => 'NF\Tickets:Ticket',
                'type'       => self::TO_ONE,
                'conditions' => 'ticket_id',
                'primary'    => true,
            ],
        ];
        $structure->options['linkUsernameInDetails'] = true;

        return $structure;
    }
}