<?php

namespace NF\Tickets\Finder;

use XF\Mvc\Entity\Finder;

use function intval;

/**
 * Class Message
 *
 * @package NF\Tickets\Finder
 */
class Message extends Finder
{
    /**
     * @param \NF\Tickets\Entity\Ticket $ticket
     * @param array                     $limits
     * @return $this
     */
    public function inTicket(\NF\Tickets\Entity\Ticket $ticket, array $limits = [])
    {
        $limits = \array_replace([
            'visibility'      => true,
            'allowOwnPending' => true,
        ], $limits);

        $this->where('ticket_id', $ticket->ticket_id);

        if ($limits['visibility'])
        {
            $this->applyVisibilityChecksInTicket($ticket, $limits['allowOwnPending']);
        }

        return $this;
    }

    /**
     * @param \NF\Tickets\Entity\Ticket $ticket
     * @param bool                      $allowOwnPending
     * @return $this
     */
    public function applyVisibilityChecksInTicket(\NF\Tickets\Entity\Ticket $ticket, $allowOwnPending = true)
    {
        $conditions = [];
        $viewableStates = ['visible'];

        if ($ticket->canViewDeletedMessages())
        {
            $viewableStates[] = 'deleted';

            $this->with('DeletionLog');
        }
        $visitor = \XF::visitor();
        if ($ticket->canViewModeratedPosts())
        {
            $viewableStates[] = 'hidden';
        }
        else if ($allowOwnPending)
        {
            $conditions[] = [
                'message_state' => 'hidden',
                'user_id'       => $visitor->user_id,
            ];
        }

        $conditions[] = ['message_state', $viewableStates];

        $this->whereOr($conditions);

        return $this;
    }

    /**
     * @param int      $page
     * @param int|null $perPage
     * @return $this
     */
    public function onPage($page, $perPage = null)
    {
        $page = \max(1, intval($page));
        if ($perPage === null)
        {
            $perPage = $this->app()->options()->messagesPerPage;
        }
        $perPage = \max(1, intval($perPage));

        $start = ($page - 1) * $perPage;
        $end = $start + $perPage;

        $this->where('position', '>=', $start)->where('position', '<', $end);

        return $this;
    }

    /**
     * @param string $direction
     * @return $this
     */
    public function orderByDate($direction = 'ASC')
    {
        $this->order('position', $direction);
        $this->order('message_date', $direction);

        return $this;
    }

    /**
     * @param $date
     * @return $this
     */
    public function newerThan($date)
    {
        $this->where('message_date', '>', $date);

        return $this;
    }

    /**
     * @return $this
     */
    public function forFullView()
    {
        $this->with(['User', 'User.Profile', 'User.Privacy']);

        if ($this->app()->options()->showMessageOnlineStatus)
        {
            $this->with('User.Activity');
        }

        $userId = \XF::visitor()->user_id;
        if ($userId)
        {
            $this->with('Reactions|' . $userId);
            $this->with('Bookmarks|' . $userId);
        }

        return $this;
    }

    /**
     * @param \XF\Entity\User|null $user
     * @return $this
     */
    public function skipIgnored(\XF\Entity\User $user = null)
    {
        if (!$user)
        {
            $user = \XF::visitor();
        }

        if (!$user->user_id)
        {
            return $this;
        }

        if ($user->Profile && $user->Profile->ignored)
        {
            $this->where('user_id', '<>', \array_keys($user->Profile->ignored));
        }

        return $this;
    }

    public function isNotFirstMessage()
    {
        $this->where($this->expression(
            '%s <> %s',
            'Ticket.first_message_id',
            'message_id'
        ));

        return $this;
    }
}