<?php

namespace NF\Tickets\Finder;

use XF\Mvc\Entity\Finder;

/**
 * Class Ticket
 *
 * @package NF\Tickets\Finder
 */
class Ticket extends Finder
{
    /**
     * @param bool $allowOwnPending
     * @return $this
     */
    public function applyGlobalVisibilityChecks($allowOwnPending = false)
    {
        $visitor = \XF::visitor();
        $viewableStates = ['visible'];

        if ($visitor->hasPermission('ticket', 'viewDeleted'))
        {
            $viewableStates[] = 'deleted';
            $this->with('DeletionLog');
        }
        if ($visitor->hasPermission('ticket', 'viewModerated'))
        {
            $viewableStates[] = 'moderated';
        }
        else if ($visitor->user_id && $allowOwnPending)
        {
            $conditions[] = [
                'discussion_state' => 'moderated',
                'user_id'          => $visitor->user_id,
            ];
            $conditions[] = [
                'discussion_state' => 'shadow_ban',
                'user_id'          => $visitor->user_id,
            ];
        }
        $conditions[] = ['discussion_state', $viewableStates];
        $this->whereOr($conditions);

        return $this;
    }

    /**
     * @param array|null $viewableCategoryIds
     * @param bool       $allowOwnPending
     * @return $this
     */
    public function applyCategoryVisibilityChecks(array $viewableCategoryIds = null, $allowOwnPending = false)
    {
        $visitor = \XF::visitor();
        $userId = $visitor->user_id;
        $conditions = [];

        /** @var \NF\Tickets\Entity\Category[] $categories */
        $categories = [];
        $categoryIds = [];
        $em = \XF::em();
        foreach ($viewableCategoryIds as $categoryId)
        {
            $category = $em->findCached('NF\Tickets:Category', $categoryId);
            if ($category)
            {
                $categories[$categoryId] = $category;
            }
            else
            {
                $categoryIds[] = $categoryId;
            }
        }
        if ($categoryIds)
        {
            $categories = \array_merge($em->find('NF\Tickets:Category', \array_unique($categoryIds))->toArray(), $categories);
        }

        // check for per-category permissions
        foreach ($categories as $category)
        {
            $viewableStates = ['visible'];
            if ($category->canViewDeletedTickets())
            {
                $viewableStates[] = 'deleted';
                $this->with('DeletionLog');
            }
            if ($category->canViewModeratedTickets())
            {
                $viewableStates[] = 'moderated';
            }
            else if ($userId && $allowOwnPending)
            {
                $viewableStates[] = 'moderated';
                $conditions[] = [
                    'ticket_category_id' => $category->ticket_category_id,
                    'user_id'          => $userId,
                    'discussion_state' => 'shadow_ban',
                ];
            }

            $condition = [
                'ticket_category_id' => $category->ticket_category_id,
                'discussion_state'   => $viewableStates,
            ];
            if (!$category->canViewOthers())
            {
                if (!$userId)
                {
                    continue;
                }
                $condition['user_id'] = $userId;
            }

            $conditions[] = $condition;
        }

        if ($conditions)
        {
            $this->whereOr($conditions);
        }

        return $this;
    }

    public function isCounted()
    {
        /** @var \NF\Tickets\Repository\Status $repo */
        $repo = \XF::repository('NF\Tickets:Status');
        $activeStatuses = $repo->findStatusesForList()
                               ->isCounted()
                               ->fetch()
                               ->keys();
        if ($activeStatuses)
        {
            $this->where('status_id', $activeStatuses);
        }
        else
        {
            $this->whereImpossible();
        }

        return $this;
    }

    public function isUnassigned()
    {
    	$this->where('assigned_user_id', 0);

    	return $this;
    }

    public function isAssignedOthers()
    {
    	$this->where('assigned_user_id', '>', 0);
    	$this->where('assigned_user_id', '!=', \XF::visitor()->user_id);

    	return $this;
    }

    public function isSelfAssigned()
    {
        if (\XF::visitor()->user_id)
        {
            $this->where('assigned_user_id', \XF::visitor()->user_id);
        }

        return $this;
    }

    public function isNotSelfAssigned()
    {
        if (\XF::visitor()->user_id)
        {
            $this->where('assigned_user_id', '!=', \XF::visitor()->user_id);
        }

        return $this;
    }

    public function withReadData($userId = null)
    {
        if ($userId === null)
        {
            $userId = \XF::visitor()->user_id;
        }

        if ($userId)
        {
            $this->with([
                'Read|' . $userId,
            ]);
        }

        return $this;
    }

    public function unreadOnly($userId = null)
    {
        if ($userId === null)
        {
            $userId = \XF::visitor()->user_id;
        }
        if (!$userId)
        {
            // no user, no read tracking
            return $this;
        }

        $ticketReadExpression = $this->expression(
            '%s > COALESCE(%s, 0)',
            'last_message_date',
            'Read|' . $userId . '.ticket_read_date'
        );

        $this
            ->where('last_message_date', '>', (
                \XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400)
            )
            ->where($ticketReadExpression);

        return $this;
    }

    public function watchedOnly($userId = null)
    {
        if ($userId === null)
        {
            $userId = \XF::visitor()->user_id;
        }
        if (!$userId)
        {
            // no user, just ignore
            return $this;
        }

        $this->whereOr(
            ['Watch|' . $userId . '.user_id', '!=', null],
            ['Category.Watch|' . $userId . '.user_id', '!=', null]
        );

        return $this;
    }

    /**
     * @param bool $includeCategory
     * @return $this
     */
    public function forFullView($includeCategory = true)
    {
        $visitor = \XF::visitor();

        $this->with('User');

        if ($visitor->user_id)
        {
            $this->with('Read|' . $visitor->user_id);
            $this->with('Watch|' . $visitor->user_id);
            $this->with('Participants|' . $visitor->user_id);
        }

        if ($includeCategory)
        {
            $this->with(['Category']);

            if ($visitor->user_id)
            {
                $this->with('Category.Read|' . $visitor->user_id);
                $this->with('Category.Watch|' . $visitor->user_id);
            }
        }

        return $this;
    }

    /**
     * @return $this
     */
    public function useDefaultOrder()
    {
        $defaultOrder = $this->app()->options()->nfTicketsListDefaultOrder ?: 'last_message_date';
        $defaultDir = $defaultOrder === 'title' ? 'asc' : 'desc';

        $this->setDefaultOrder($defaultOrder, $defaultDir);

        return $this;
    }

    /**
     * @param \XF\Entity\User|null $user
     * @return $this
     */
    public function skipIgnored(\XF\Entity\User $user = null)
    {
        if (!$user)
        {
            $user = \XF::visitor();
        }

        if (!$user->user_id)
        {
            return $this;
        }

        if ($user->Profile && $user->Profile->ignored)
        {
            $this->where('user_id', '<>', \array_keys($user->Profile->ignored));
        }

        return $this;
    }
}