<?php

namespace NF\Tickets\Import\Importer;

use NF\Tickets\Import\Data\Category;
use NF\Tickets\Import\Data\Message;
use NF\Tickets\Import\Data\Status;
use NF\Tickets\Import\Data\Ticket;
use NF\Tickets\Import\Data\TicketField;
use NF\Tickets\Import\Data\TicketPrefix;
use NF\Tickets\Import\Data\TicketPrefixGroup;
use XF\Db\AbstractAdapter;
use XF\Db\Mysqli\Adapter as MysqlAdapter;
use XF\Import\Data\EditHistory;
use XF\Import\Importer\AbstractImporter;
use XF\Import\Importer\XenForoSourceTrait;
use XF\Import\StepState;

class BriviumTickets extends AbstractImporter
{
    use XenForoSourceTrait;

    /**
     * @var MysqlAdapter
     */
    protected $sourceDb;

    public static function getListInfo()
    {
        return [
            'target' => 'Tickets by NixFifty',
            'source' => 'Brivium - Support Ticket System',
            'beta' => true
        ];
    }

    protected function validateVersion(AbstractAdapter $db, &$error)
    {
        return true;
    }

    protected function getStepConfigDefault()
    {
        return null;
    }

    public function renderStepConfigOptions(array $vars)
    {
        return null;
    }

    public function validateStepConfig(array $steps, array &$stepConfig, array &$errors)
    {
        return true;
    }

    public function canRetainIds()
    {
        $maxTicketId = $this->db()->fetchOne("SELECT MAX(ticket_id) FROM xf_nf_tickets_ticket");
        if ($maxTicketId)
        {
            return false;
        }

        $maxMessageId = $this->db()->fetchOne("SELECT MAX(message_id) FROM xf_nf_tickets_message");
        if ($maxMessageId)
        {
            return false;
        }

        $maxCategoryId = $this->db()->fetchOne("SELECT MAX(ticket_category_id) FROM xf_nf_tickets_category");
        if ($maxCategoryId > 1)
        {
            return false;
        }

        $maxStatusId = $this->db()->fetchOne("SELECT MAX(status_id) FROM xf_nf_tickets_status");
        if ($maxStatusId > 3)
        {
            return false;
        }

        return true;
    }

    public function resetDataForRetainIds()
    {
        /** @var \NF\Tickets\Entity\Category $category */
        $category = $this->em()->find('NF\Tickets:Category', 1);
        if ($category)
        {
            $category->delete();
        }

        for ($i = 1; $i <= 3; $i++)
        {
            /** @var \NF\Tickets\Entity\Status $status */
            $status = $this->em()->find('NF\Tickets:Status', $i);
            if ($status)
            {
                $status->delete();
            }
        }
    }

    public function getSteps()
    {
        return [
            'departments' => [
                'title' => \XF::phrase('nf_tickets_departments'),
            ],
            'watchedDepartments' => [
                'title' => \XF::phrase('nf_tickets_watched_departments')
            ],
            'ticketPrefixes' => [
                'title' => \XF::phrase('nf_tickets_ticket_prefixes'),
                'depends' => ['departments']
            ],
            'ticketFields' => [
                'title' => \XF::phrase('nf_tickets_custom_ticket_fields'),
                'depends' => ['departments']
            ],
            'ticketStatuses' => [
                'title' => \XF::phrase('nf_tickets_ticket_statuses'),
                'depends' => ['departments']
            ],
            'tickets' => [
                'title' => \XF::phrase('nf_tickets_tickets'),
                'depends' => ['departments']
            ],
            'messages' => [
                'title' => \XF::phrase('nf_tickets_messages'),
                'depends' => ['tickets']
            ],
            'messageEditHistory' => [
                'title' => \XF::phrase('nf_tickets_message_edit_history'),
                'depends' => ['messages']
            ],
            'messageAttachments' => [
                'title' => \XF::phrase('nf_tickets_message_attachments'),
                'depends' => ['messages']
            ]
        ];
    }

    protected function doInitializeSource()
    {
        $this->sourceDb = new MysqlAdapter(
            $this->baseConfig['db'],
            $this->app->config('fullUnicode')
        );
    }

    // ############################## STEP: DEPARTMENTS #########################

    public function stepDepartments(StepState $state)
    {
        $departments = $this->sourceDb->fetchAll("
            SELECT * FROM xf_mjst_department ORDER BY department_id
        ");

        foreach ($departments AS $department)
        {
            $oldId = $department['department_id'];

            $data = $this->mapXfKeys($department, [
                'title',
                'description' => 'short_description',
                'ticket_count',
                'display_order'
            ]);

            /** @var Category $import */
            $import = $this->newHandler('NF\Tickets:Category');
            $import->bulkSet($data);
            $import->save($oldId);

            $state->imported++;
        }

        return $state->complete();
    }

    // ############################## STEP: WATCHED DEPARTMENTS #########################

    public function getStepEndWatchedDepartments()
    {
        return $this->sourceDb->fetchOne("SELECT MAX(department_id) FROM xf_mjst_department") ?? 0;
    }

    public function stepWatchedDepartments(StepState $state, array $stepConfig, $maxTime)
    {
        $timer = new \XF\Timer($maxTime);

        $this->typeMap('category');

        $departmentIds = $this->sourceDb->fetchAllColumn("
			SELECT department_id
			FROM xf_mjst_department
			WHERE department_id > ?
			ORDER BY department_id
		", $state->startAfter);
        if (!$departmentIds)
        {
            return $state->complete();
        }

        /** @var \NF\Tickets\Import\DataHelper\Category $categoryHelper */
        $categoryHelper = $this->getDataHelper('NF\Tickets:Category');

        foreach ($departmentIds AS $oldId)
        {
            $state->startAfter = $oldId;

            $categoryId = $this->lookupId('category', $oldId);
            if (!$categoryId)
            {
                continue;
            }

            $watches = $this->sourceDb->fetchAllKeyed("
				SELECT *
				FROM xf_mjst_department_watch
				WHERE department_id = ?
			", 'user_id', $oldId);

            $watchData = [];
            foreach ($watches AS $userId => $watch)
            {
                $watchData[$userId] = [
                    'notify_on' => $watch['notify_on'],
                    'send_alert' => $watch['send_alert'],
                    'send_email' => $watch['send_email']
                ];

                $state->imported++;
            }

            $categoryHelper->importCategoryWatchBulk($categoryId, $watchData);

            if ($timer->limitExceeded())
            {
                break;
            }
        }

        return $state->resumeIfNeeded();
    }

    // ############################## STEP: TICKET PREFIXES #########################

    public function stepTicketPrefixes(StepState $state)
    {
        $this->typeMap('category');

        $prefixGroups = $this->sourceDb->fetchAllKeyed("
            SELECT tpg.*,
                p.phrase_text
            FROM xf_mjst_ticket_prefix_group AS tpg
            INNER JOIN xf_phrase AS p ON (p.language_id = 0 AND p.title = CONCAT('support_ticket_prefix_group.', tpg.prefix_group_id))
        ", 'prefix_group_id');
        $mappedGroupIds = [];

        foreach ($prefixGroups AS $oldGroupId => $group)
        {
            /** @var TicketPrefixGroup $importGroup */
            $importGroup = $this->newHandler('NF\Tickets:TicketPrefixGroup');
            $importGroup->display_order = $group['display_order'];
            $importGroup->setTitle($group['phrase_text']);

            $newGroupId = $importGroup->save($oldGroupId);
            if ($newGroupId)
            {
                $mappedGroupIds[$oldGroupId] = $newGroupId;
            }
        }

        $prefixes = $this->sourceDb->fetchAllKeyed("
            SELECT tp.*,
                p.phrase_text
            FROM xf_mjst_ticket_prefix AS tp
            INNER JOIN xf_phrase AS p ON (p.language_id = 0 AND p.title = CONCAT('support_ticket_prefix.', tp.prefix_id))
        ", 'prefix_id');

        $prefixCategoryMap = [];
        $departmentPrefixes = $this->sourceDb->fetchAll("
            SELECT *
            FROM xf_mjst_department_prefix
        ");
        foreach ($departmentPrefixes AS $departmentPrefix)
        {
            $categoryId = $this->lookupId('category', $departmentPrefix['department_id']);
            if ($categoryId)
            {
                $prefixCategoryMap[$departmentPrefix['prefix_id']][] = $categoryId;
            }
        }

        foreach ($prefixes AS $oldPrefixId => $prefix)
        {
            if ($prefix['allowed_user_group_ids'] === '-1')
            {
                $prefix['allowed_user_group_ids'] = [-1];
            }
            else
            {
                $prefix['allowed_user_group_ids'] = \explode(',', $prefix['allowed_user_group_ids']);
            }

            /** @var TicketPrefix $importPrefix */
            $importPrefix = $this->newHandler('NF\Tickets:TicketPrefix');
            $importPrefix->bulkSet($this->mapKeys($prefix, [
                'display_order',
                'css_class',
                'allowed_user_group_ids'
            ]));
            $importPrefix->prefix_group_id = $mappedGroupIds[$prefix['prefix_group_id']] ?? 0;
            $importPrefix->setTitle($prefix['phrase_text']);

            if (!empty($prefixCategoryMap[$oldPrefixId]))
            {
                $importPrefix->setCategories($prefixCategoryMap[$oldPrefixId]);
            }

            if ($importPrefix->save($oldPrefixId))
            {
                $state->imported++;
            }
        }

        return $state->complete();
    }

    // ########################### STEP: TICKET FIELDS ###############################

    protected function mapTicketFieldDisplayGroup($originalDisplayGroup)
    {
        // original => nf
        return [
            'before' => 'above_message',
            'after' => 'below_message',
            'ticket_info' => 'sidebar_ticket_info',
            'submitter_info' => 'sidebar_user_info'
        ][$originalDisplayGroup];
    }

    public function stepTicketFields(StepState $state)
    {
        $this->typeMap('category');

        $fields = $this->sourceDb->fetchAllKeyed("
            SELECT field.*,
                ptitle.phrase_text AS title,
                pdesc.phrase_text AS description
            FROM xf_mjst_ticket_field AS field
            INNER JOIN xf_phrase AS ptitle ON
                (ptitle.language_id = 0 AND ptitle.title = CONCAT('thread_field_title.', field.field_id))
            INNER JOIN xf_phrase AS pdesc ON
                (pdesc.language_id = 0 AND pdesc.title = CONCAT('thread_field_desc.', field.field_id))
        ", 'field_id');

        $existingFields = $this->db()->fetchPairs("SELECT field_id, field_id FROM xf_nf_tickets_ticket_field");

        $fieldCategoryMap = [];
        $departmentFields = $this->sourceDb->fetchAll("
            SELECT *
            FROM xf_mjst_department_field
        ");
        foreach ($departmentFields AS $departmentField)
        {
            $categoryId = $this->lookupId('category', $departmentField['department_id']);
            if ($categoryId)
            {
                $fieldCategoryMap[$departmentField['field_id']][] = $categoryId;
            }
        }

        foreach ($fields AS $oldId => $field)
        {
            if (!empty($existingFields[$oldId]))
            {
                $this->logHandler('NF\Tickets:TicketField', $oldId, $oldId);
            }
            else
            {
                $userGroupIds = $field['editable_user_group_ids'];
                $displayGroup = $field['display_group'];
                unset($field['editable_user_group_ids']);
                unset($field['display_group']);

                /** @var TicketField $import */
                $import = $this->setupCustomFieldImport('NF\Tickets:TicketField', $field);
                $import->bulkSet([
                    'editable_user_group_ids' => $userGroupIds,
                    'display_group' => $this->mapTicketFieldDisplayGroup($displayGroup)
                ]);

                if (!empty($fieldCategoryMap[$oldId]))
                {
                    $import->setCategories($fieldCategoryMap[$oldId]);
                }

                $import->save($oldId);
            }

            $state->imported++;
        }

        return $state->complete();
    }

    // ############################## STEP: TICKET STATUSES #########################

    public function stepTicketStatuses(StepState $state)
    {
        $statuses = $this->sourceDb->fetchAllKeyed("
            SELECT ts.*,
                p.phrase_text
            FROM xf_mjst_ticket_status AS ts
            INNER JOIN xf_phrase AS p ON (p.language_id = 0 AND p.title = CONCAT('mjst_status.', ts.ticket_status_id))
            ORDER BY ticket_status_id
        ", 'ticket_status_id');

        foreach ($statuses AS $oldId => $status)
        {
            /** @var Status $import */
            $import = $this->newHandler('NF\Tickets:Status');
            $import->display_order = $oldId * 10;
            $import->is_close = !$status['group_active'];
            $import->setTitle($status['phrase_text']);
            // TODO: Map color to CSS class?

            if ($import->save($oldId))
            {
                $state->imported++;
            }
        }

        return $state->complete();
    }

    // ############################## STEP: TICKETS #########################

    public function getStepEndTickets()
    {
        return $this->sourceDb->fetchOne("SELECT MAX(support_ticket_id) FROM xf_mjst_ticket") ?? 0;
    }

    public function stepTickets(StepState $state, array $stepConfig, $maxTime)
    {
        $limit = 1000;
        $timer = new \XF\Timer($maxTime);

        $tickets = $this->sourceDb->fetchAllKeyed("
            SELECT t.*
            FROM xf_mjst_ticket AS t
            WHERE t.support_ticket_id > ? AND t.support_ticket_id <= ?
            ORDER BY t.support_ticket_id
            LIMIT {$limit}
        ", 'support_ticket_id', [$state->startAfter, $state->end]);
        if (!$tickets)
        {
            return $state->complete();
        }

        $this->typeMap('category');
        $this->typeMap('ticket_prefix');
        $this->typeMap('status');

        foreach ($tickets AS $oldId => $ticket)
        {
            $state->startAfter = $oldId;

            $categoryId = $this->lookupId('category', $ticket['department_id']);
            if (!$categoryId)
            {
                continue;
            }

            $import = $this->setupTicketImport($ticket, $categoryId);

            $watchers = $this->sourceDb->fetchPairs("
                SELECT user_id, email_subscribe
                FROM xf_mjst_ticket_watch
                WHERE support_ticket_id = ?
            ", $ticket['support_ticket_id']);
            if ($watchers)
            {
                foreach ($watchers AS $watcherUserId => $emailSubscribe)
                {
                    if (!empty($watcherUserIdMap[$watcherUserId]))
                    {
                        $import->addTicketWatcher($watcherUserIdMap[$watcherUserId], $emailSubscribe ? true : false);
                    }
                }
            }

            $newId = $import->save($oldId);
            if ($newId)
            {
                $state->imported++;
            }

            if ($timer->limitExceeded())
            {
                break;
            }
        }

        return $state->resumeIfNeeded();
    }

    protected function getTicketUrgencyMapping($urgency)
    {
        // Brivium urgency => NF priority_id
        return [
            1 => 3,
            2 => 2,
            3 => 1
        ][$urgency];
    }

    protected function getTicketMapping()
    {
        // NF field => Brivium field
        return [
            'ticket_ref' => 'ticket_id',
            'title',
            'user_id',
            'username',
            'user_name',
            'user_email',
            'starter_user_id' => 'creator_user_id',
            'starter_username' => 'creator_username',
            'start_date' => 'open_date',
            'last_message_id',
            'last_message_date',
            'last_message_user_id',
            'last_message_username',
            'last_modified_date' => 'last_update',
            'reply_count'
        ];
    }

    protected function setupTicketImport(array $ticket, $categoryId)
    {
        /** @var Ticket $import */
        $import = $this->newHandler('NF\Tickets:Ticket');
        $import->bulkSet($this->mapXfKeys($ticket, $this->getTicketMapping()));
        $import->bulkSet([
            'ticket_category_id' => $categoryId,
            'prefix_id' => $ticket['prefix_id'] ? $this->lookupId('ticket_prefix', $ticket['prefix_id'], 0) : 0,
            'status_id' => $ticket['ticket_status_id'] ? $this->lookupId('status', $ticket['ticket_status_id'], 0) : 0,
            'priority' => $this->getTicketUrgencyMapping($ticket['urgency']),
        ]);
        if (!empty($ticket['guest_password']))
        {
            $import->setTicketPassword($ticket['guest_password']);
        }

        $customFields = $this->decodeValue($ticket['custom_fields'], 'serialized-json-array');
        if ($customFields)
        {
            $import->setCustomFields($this->mapCustomFields('ticket_field', $customFields));
        }

        return $import;
    }

    // ############################## STEP: MESSAGES #########################

    public function getStepEndMessages()
    {
        return $this->sourceDb->fetchOne("SELECT MAX(message_id) FROM xf_mjst_ticket_message") ?? 0;
    }

    public function stepMessages(StepState $state, array $stepConfig, $maxTime)
    {
        $limit = 1000;
        $timer = new \XF\Timer($maxTime);

        $messages = $this->sourceDb->fetchAllKeyed("
            SELECT m.*,
                ip.ip,
                d.delete_date, d.delete_user_id, d.delete_username, d.delete_reason
            FROM xf_mjst_ticket_message AS m
            LEFT JOIN xf_ip AS ip ON (ip.ip_id = m.ip_id)
            LEFT JOIN xf_deletion_log AS d ON (d.content_type = 'support_ticket_message' AND d.content_id = m.message_id)
            WHERE m.message_id > ? AND m.message_id <= ?
            ORDER BY m.message_id
            LIMIT {$limit}
        ", 'message_id', [$state->startAfter, $state->end]);
        if (!$messages)
        {
            return $state->complete();
        }

        $mapTicketIds = [];
        foreach ($messages AS $message)
        {
            $mapTicketIds = $message['support_ticket_id'];
        }

        $this->lookup('ticket', $mapTicketIds);

        foreach ($messages AS $oldId => $message)
        {
            $state->startAfter = $oldId;

            $ticketId = $this->lookupId('ticket', $message['support_ticket_id']);
            if (!$ticketId)
            {
                continue;
            }

            if (!$message['message_state'])
            {
                continue;
            }

            /** @var Message $import */
            $import = $this->newHandler('NF\Tickets:Message');
            $import->bulkSet($this->mapXfKeys($message, [
                'user_id',
                'username',
                'user_name',
                'user_email',
                'message_date',
                'message',
                'message_state',
                'position',
                'last_edit_date',
                'last_edit_user_id',
                'edit_count'
            ]));
            $import->ticket_id = $ticketId;
            $import->setLoggedIp($message['ip']);
            $import->setDeletionLogData($this->extractDeletionLogData($message));

            $newId = $import->save($oldId);
            if ($newId)
            {
                $state->imported++;
            }

            if ($timer->limitExceeded())
            {
                break;
            }
        }

        return $state->resumeIfNeeded();
    }

    // ########################### STEP: MESSAGE EDIT HISTORY ###############################

    public function getStepEndMessageEditHistory()
    {
        return $this->sourceDb->fetchOne("SELECT MAX(edit_history_id) FROM xf_edit_history WHERE content_type = 'support_ticket_message'") ?? 0;
    }

    public function stepMessageEditHistory(StepState $state, array $stepConfig, $maxTime)
    {
        $limit = 1000;
        $timer = new \XF\Timer($maxTime);

        $edits = $this->sourceDb->fetchAllKeyed("
            SELECT * FROM
            xf_edit_history
            WHERE edit_history_id > ? AND edit_history_id <= ?
                AND content_type = 'support_ticket_message'
            ORDER BY edit_history_id
            LIMIT {$limit}
        ", 'edit_history_id', [$state->startAfter, $state->end]);
        if (!$edits)
        {
            return $state->complete();
        }

        $mapMessageIds = [];
        foreach ($edits AS $edit)
        {
            $mapMessageIds[] = $edit['content_id'];
        }

        $this->lookup('message', $mapMessageIds);

        foreach ($edits AS $oldId => $edit)
        {
            $state->startAfter = $oldId;

            $messageId = $this->lookupId('message', $oldId);
            if (!$messageId)
            {
                continue;
            }

            /** @var EditHistory $import */
            $import = $this->newHandler('XF:EditHistory');
            $import->preventRetainIds();
            $import->bulkSet($this->mapKeys($edit, [
                'edit_date',
                'old_text',
                'edit_user_id'
            ]));
            $import->content_type = 'nf_tickets_message';
            $import->content_id = $messageId;

            $newId = $import->save($oldId);
            if ($newId)
            {
                $state->imported++;
            }

            if ($timer->limitExceeded())
            {
                break;
            }
        }

        return $state->resumeIfNeeded();
    }

    // ############################## STEP: ATTACHMENTS #########################

    public function getStepEndMessageAttachments()
    {
        return $this->sourceDb->fetchOne("SELECT MAX(attachment_id) FROM xf_attachment") ?? 0;
    }

    public function stepMessageAttachments(StepState $state, array $stepConfig, $maxTime)
    {
        $limit = 500;
        $timer = new \XF\Timer($maxTime);

        $attachments = $this->sourceDb->fetchAllKeyed("
			SELECT a.*,
				ad.*
			FROM xf_attachment AS a
			INNER JOIN xf_attachment_data AS ad ON (a.data_id = ad.data_id)
			WHERE a.attachment_id > ? AND a.attachment_id <= ?
				AND a.content_type IN ('support_ticket_message')
			ORDER BY a.attachment_id
			LIMIT {$limit}
		", 'attachment_id', [$state->startAfter, $state->end]);
        if (!$attachments)
        {
            return $state->complete();
        }

        $mapContentIds = [];
        foreach ($attachments AS $attachment)
        {
            if ($attachment['content_type'] === 'support_ticket_message')
            {
                $attachment['content_type'] = 'message';
            }

            $mapContentIds[$attachment['content_type']][] = $attachment['content_id'];
        }

        foreach ($mapContentIds AS $contentType => $contentIds)
        {
            $this->lookup($contentType, $contentIds);
        }

        foreach ($attachments AS $oldId => $attachment)
        {
            $state->startAfter = $oldId;

            if ($attachment['content_type'] === 'support_ticket_message')
            {
                $attachment['content_type'] = 'message';
            }

            $contentId = $this->lookupId($attachment['content_type'], $attachment['content_id']);
            if (!$contentId)
            {
                continue;
            }

            $sourceFile = $this->getSourceAttachmentDataPath(
                $attachment['data_id'], $attachment['file_path'], $attachment['file_hash']
            );
            if (!\file_exists($sourceFile) || !\is_readable($sourceFile))
            {
                continue;
            }

            /** @var \XF\Import\Data\Attachment $import */
            $import = $this->newHandler('XF:Attachment');
            $import->preventRetainIds();
            $import->bulkSet($this->mapKeys($attachment, [
                'attach_date',
                'temp_hash',
                'unassociated',
                'view_count'
            ]));
            $import->content_type = 'nf_tickets_message';
            $import->content_id = $contentId;
            $import->setDataExtra('upload_date', $attachment['upload_date']);
            $import->setDataExtra('file_path', $attachment['file_path']);
            $import->setDataUserId($attachment['user_id']);
            $import->setSourceFile($sourceFile, $attachment['filename']);
            $import->setContainerCallback([$this, 'rewriteEmbeddedAttachments']);

            $newId = $import->save($oldId);
            if ($newId)
            {
                $state->imported++;
            }

            if ($timer->limitExceeded())
            {
                break;
            }
        }

        return $state->resumeIfNeeded();
    }

    public function getFinalizeJobs(array $stepsRun)
    {
        return [
            'NF\Tickets:Category',
            'NF\Tickets:Ticket',
            'NF\Tickets:UserTicketCount'
        ];
    }
}