<?php

namespace NF\Tickets\Job;

use NF\Tickets\Entity\Escalation as EscalationEntity;
use NF\Tickets\Repository\Response as ResponseRepo;
use XF\Db\AbstractAdapter as DbAdapter;
use XF\Entity\User as UserEntity;
use NF\Tickets\XF\Entity\User as ExtendedUserEntity;
use XF\Job\AbstractRebuildJob;
use XF\Mvc\Entity\Manager as EntityManager;
use XF\Mvc\Entity\Repository;
use XF\Repository\UserAlert as UserAlertRepo;
use XF\App as BaseApp;
use NF\Tickets\Entity\Ticket as TicketEntity;

use function array_slice;

class EscalationUserAlert extends AbstractRebuildJob
{
    protected $defaultData = [
        'ticket_id' => 0,
        'escalation_id' => 0,
        'actor_id' => 0
    ];

    protected $ticket;

    protected $escalation;

    protected $actor;

    protected function setupData(array $data)
    {
        $data = parent::setupData($data);

        $this->ticket = $this->findTicketById($data['ticket_id']);
        $this->escalation = $this->findEscalationById($data['escalation_id']);
        $this->actor = $this->findUserById($data['actor_id']);

        return $data;
    }

    /**
     * @return TicketEntity|null
     */
    public function getTicket()
    {
        return $this->ticket;
    }

    /**
     * @return EscalationEntity|null
     */
    public function getEscalation()
    {
        return $this->escalation;
    }

    /**
     * @return UserEntity|ExtendedUserEntity|null
     */
    public function getActor()
    {
        return $this->actor;
    }

    protected function getNextIds($start, $batch) : array
    {
        $ticket = $this->getTicket();
        $actor = $this->getActor();
        $escalation = $this->getEscalation();

        if (!$ticket || !$actor || !$escalation || empty($escalation->actions['send_alert']))
        {
            return [];
        }

        $db = $this->db();
        $nextIds = $db->fetchAllColumn($db->limit(
            "
				SELECT participant.user_id
				FROM xf_nf_tickets_ticket_participant as participant
				JOIN xf_nf_tickets_ticket xntt ON participant.ticket_id = xntt.ticket_id
				WHERE participant.user_id > ?
				  AND participant.ticket_id = ?
				  AND participant.user_id <> ?
				  AND xntt.discussion_state = 'visible'
				ORDER BY participant.user_id ASC
			", $batch
        ), [$start, $ticket->ticket_id, $actor->user_id]);

        $firstUserId = \reset($nextIds);
        $lastUserId = \end($nextIds);
        if ($firstUserId === false && $lastUserId === false)
        {
            $nextIds[] = $ticket->user_id;
        }
        else if ($firstUserId < $ticket->user_id && $lastUserId > $ticket->user_id)
        {
            $nextIds[] = $ticket->user_id;
            $nextIds = array_slice($nextIds, 0, $batch);
        }

        return $nextIds;
    }

    protected function getAlertText(UserEntity $receiver) : string
    {
        return \strtr(
            $this->getEscalation()->alert_body,
            $this->getResponseRepo()->getReplacementValues($this->getTicket())
        );
    }

    protected function rebuildById($id)
    {
        $ticket = $this->getTicket();
        $sender = $this->getActor();
        $receiver = $this->findUserById($id);

        if (!$sender || !$receiver)
        {
            return;
        }

        $alertText = $this->getAlertText($receiver);
        $extraParams = ['alert_text' => $alertText];

        $this->getUserAlertRepo()->alertFromUser(
            $receiver,
            $sender,
            $ticket->getEntityContentType(),
            $ticket->getEntityId(),
            'escalation',
            $extraParams
        );
    }

    protected function getStatusType()
    {
        return null;
    }

    public function canTriggerByChoice() : bool
    {
        return false;
    }

    public function canCancel() : bool
    {
        return false;
    }

    protected function app() : BaseApp
    {
        return $this->app;
    }

    protected function db() : DbAdapter
    {
        return $this->app()->db();
    }

    protected function em() : EntityManager
    {
        return $this->app()->em();
    }

    protected function repository(string $identifier) : Repository
    {
        return $this->app()->repository($identifier);
    }

    protected function findTicketById(int $ticketId, array $with = []): ?TicketEntity
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->em()->find('NF\Tickets:Ticket', $ticketId, $with);
    }

    /**
     * @return UserEntity|ExtendedUserEntity|null
     */
    protected function findUserById(int $userId, array $with = [])
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->em()->find('XF:User', $userId, $with);
    }

    protected function findEscalationById(int $escalationId, array $with = []): ?EscalationEntity
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->em()->find('NF\Tickets:Escalation', $escalationId, $with);
    }

    protected function getUserAlertRepo() : UserAlertRepo
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('XF:UserAlert');
    }

    protected function getResponseRepo() : ResponseRepo
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Response');
    }
}