<?php

namespace NF\Tickets\Job;

use XF\Job\AbstractJob;

class TicketAction extends AbstractJob
{
	protected $defaultData = [
		'start' => 0,
		'count' => 0,
		'total' => null,
		'criteria' => null,
		'ticketIds' => null,
		'actions' => []
	];

	public function run($maxRunTime): \XF\Job\JobResult
	{
		if (is_array($this->data['criteria']) && is_array($this->data['ticketIds']))
		{
			throw new \LogicException("Cannot have both criteria and ticketIds values; one must be null");
		}

		$startTime = microtime(true);
		$em = $this->app->em();

		$ids = $this->prepareTicketIds();
		if (!$ids)
		{
			return $this->complete();
		}

		$db = $this->app->db();
		$db->beginTransaction();

		$limitTime = ($maxRunTime > 0);
		foreach ($ids AS $key => $id)
		{
			$this->data['count']++;
			$this->data['start'] = $id;
			unset($ids[$key]);

			/** @var \NF\Tickets\Entity\Ticket $ticket */
			$ticket = $em->find('NF\Tickets:Ticket', $id);
			if ($ticket)
			{
				if ($this->getActionValue('delete'))
				{
					$ticket->delete(false, false);
					continue; // no further action required
				}

				$this->applyInternalTicketChange($ticket);
				$ticket->save(false, false);

				$this->applyExternalTicketChange($ticket);
			}

			if ($limitTime && microtime(true) - $startTime > $maxRunTime)
			{
				break;
			}
		}

		if (is_array($this->data['ticketIds']))
		{
			$this->data['ticketIds'] = $ids;
		}

		$db->commit();

		return $this->resume();
	}

	protected function getActionValue($action)
	{
		$value = null;
		if (!empty($this->data['actions'][$action]))
		{
			$value = $this->data['actions'][$action];
		}
		return $value;
	}

	protected function prepareTicketIds(): array
	{
		if (is_array($this->data['criteria']))
		{
			$searcher = $this->app->searcher('NF\Tickets:Ticket', $this->data['criteria']);
			$results = $searcher->getFinder()
				->where('ticket_id', '>', $this->data['start'])
				->order('ticket_id')
				->limit(1000)
				->fetchColumns('ticket_id');
			$ids = array_column($results, 'ticket_id');
		}
		else if (is_array($this->data['ticketIds']))
		{
			$ids = $this->data['ticketIds'];
		}
		else
		{
			$ids = [];
		}
		sort($ids, SORT_NUMERIC);
		return $ids;
	}

	protected function applyInternalTicketChange(\NF\Tickets\Entity\Ticket $ticket): void
	{
		if ($categoryId = $this->getActionValue('ticket_category_id'))
		{
			$ticket->ticket_category_id = $categoryId;
		}

		if ($this->getActionValue('apply_ticket_prefix'))
		{
			$ticket->prefix_id = (int)$this->getActionValue('prefix_id');
		}

		if ($statusId = $this->getActionValue('apply_ticket_status'))
		{
			$ticket->status_id = $statusId;
		}

		if ($priorityId = $this->getActionValue('apply_ticket_priority'))
		{
			$ticket->priority = $priorityId;
		}

		if ($this->getActionValue('lock'))
		{
			$ticket->ticket_locked = true;
		}
		if ($this->getActionValue('unlock'))
		{
			$ticket->ticket_locked = false;
		}

		if ($this->getActionValue('approve'))
		{
			$ticket->discussion_state = 'visible';
		}
		if ($this->getActionValue('unapprove'))
		{
			$ticket->discussion_state = 'moderated';
		}

		if ($this->getActionValue('soft_delete'))
		{
			$ticket->discussion_state = 'deleted';
		}
	}

	protected function applyExternalTicketChange(\NF\Tickets\Entity\Ticket $ticket): void
	{
	}

	public function getStatusMessage(): string
	{
		$actionPhrase = \XF::phrase('updating');
		$typePhrase = \XF::phrase('nf_tickets_tickets');

		if ($this->data['total'] !== null)
		{
			return sprintf('%s... %s (%d/%d)', $actionPhrase, $typePhrase, $this->data['count'], $this->data['total']);
		}

		return sprintf('%s... %s (%d)', $actionPhrase, $typePhrase, $this->data['start']);
	}

	public function canCancel(): bool
	{
		return true;
	}

	public function canTriggerByChoice(): bool
	{
		return false;
	}
}