<?php

namespace NF\Tickets\Legacy\DataWriter;

use function is_array;

/**
 * Data writer for categories.
 */
class Category extends \XF\Legacy\DataWriter
{
    /**
     * Title of the phrase that will be created when a call to set the
     * existing data fails (when the data doesn't exist).
     *
     * @var string
     */
    protected $_existingDataErrorPhrase = 'nf_tickets_requested_ticket_category_not_found';

    /**
     * Option that represents whether associated caches will be automatically
     * rebuilt. Defaults to true.
     *
     * @var string
     */
    const OPTION_REBUILD_CACHE = 'rebuildCache';

    const DATA_FIELD_IDS = 'fieldIds';

    /**
     * Gets the fields that are defined for the table. See parent for explanation.
     *
     * @return array
     */
    protected function _getFields()
    {
        return [
            'xf_nixfifty_ticket_category' => [
                'category_id'           => ['type' => self::TYPE_UINT, 'autoIncrement' => true],
                'category_title'        => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'maxLength' => 100,
                    'requiredError' => 'please_enter_valid_title',
                ],
                'category_description'  => ['type' => self::TYPE_STRING, 'default' => ''],
                'parent_category_id'    => [
                    'type'         => self::TYPE_UINT, 'default' => 0,
                    'verification' => ['$this', '_validateParentCategoryId'],
                ],
                'depth'                 => ['type' => self::TYPE_UINT, 'default' => 0],
                'lft'                   => ['type' => self::TYPE_UINT, 'default' => 0],
                'rgt'                   => ['type' => self::TYPE_UINT, 'default' => 0],
                'display_order'         => ['type' => self::TYPE_UINT, 'default' => 1],
                'ticket_count'          => ['type' => self::TYPE_UINT_FORCED, 'default' => 0],
                'last_ticket'           => ['type' => self::TYPE_UINT, 'default' => 0],
                'last_ticket_title'     => ['type' => self::TYPE_STRING, 'default' => '', 'maxLength' => 100],
                'last_ticket_id'        => ['type' => self::TYPE_UINT, 'default' => 0],
                'category_breadcrumb'   => ['type' => self::TYPE_SERIALIZED, 'default' => 'a:0:{}'],
                'thread_node_id'        => ['type' => self::TYPE_UINT, 'default' => 0],
                'thread_prefix_id'      => ['type' => self::TYPE_UINT, 'default' => 0],
                'field_cache'           => ['type' => self::TYPE_SERIALIZED, 'default' => ''],
                'priority_selection'    => ['type' => self::TYPE_UINT, 'default' => 1],
                'empty_message_allowed' => ['type' => self::TYPE_UINT, 'default' => 0],
                'hide_message_field'    => ['type' => self::TYPE_UINT, 'default' => 0],
            ],
        ];
    }

    /**
     * Gets the actual existing data out of data that was passed in. See parent for explanation.
     *
     * @param mixed
     * @param mixed $data
     * @return array|bool
     */
    protected function _getExistingData($data)
    {
        if (!$id = $this->_getExistingPrimaryKey($data))
        {
            return false;
        }

        return ['xf_nixfifty_ticket_category' => $this->_getCategoryModel()->getCategoryById($id)];
    }

    /**
     * Gets SQL condition to update the existing record.
     *
     * @param $tableName
     * @return string
     */
    protected function _getUpdateCondition($tableName)
    {
        return 'category_id = ' . $this->_db->quote($this->getExisting('category_id'));
    }

    /**
     * Gets the default set of options for this data writer.
     *
     * @return array
     */
    protected function _getDefaultOptions()
    {
        return [
            self::OPTION_REBUILD_CACHE => true,
        ];
    }

    protected function _validateParentCategoryId(&$parentId)
    {
        if ($this->isUpdate() && $parentId != 0 && $parentId != $this->getExisting('parent_category_id'))
        {
            $possibleParents = $this->_getCategoryModel()->getPossibleParentCategories($this->getMergedExistingData());
            if (!isset($possibleParents[$parentId]))
            {
                $this->error(new XenForo_Phrase('nf_tickets_please_select_valid_parent_ticket_category'), 'parent_category_id');

                return false;
            }
        }

        return true;
    }

    protected function _preSave()
    {
        if ($this->isChanged('thread_node_id') || $this->isChanged('thread_prefix_id'))
        {
            if (!$this->get('thread_node_id'))
            {
                $this->set('thread_prefix_id', 0);
            }
            else
            {
                $forum = $this->_getForumModel()->getForumById($this->get('thread_node_id'));
                if (!$forum)
                {
                    $this->set('thread_node_id', 0);
                    $this->set('thread_prefix_id', 0);
                }
                else if ($this->get('thread_prefix_id'))
                {
                    $prefix = $this->_getThreadPrefixModel()->getPrefixIfInForum(
                        $this->get('thread_prefix_id'), $forum['node_id']
                    );
                    if (!$prefix)
                    {
                        $this->set('thread_prefix_id', 0);
                    }
                }
            }
        }
    }

    protected function _postSave()
    {
        if ($this->isInsert()
            || $this->isChanged('display_order')
            || $this->isChanged('parent_category_id')
            || $this->isChanged('category_title')
        )
        {
            $this->_getCategoryModel()->rebuildCategoryStructure();
        }

        $newFieldIds = $this->getExtraData(self::DATA_FIELD_IDS);
        if (is_array($newFieldIds))
        {
            $this->_updateFieldAssociations($newFieldIds);
            $this->_getFieldModel()->rebuildFieldCategoryAssociationCache([$this->get('category_id')]);
        }

        if ($this->isInsert() || $this->isChanged('parent_category_id'))
        {
            if ($this->getOption(self::OPTION_REBUILD_CACHE))
            {
                XenForo_Application::defer('Permission', [], 'Permission', true);
            }
        }
    }

    protected function _updateFieldAssociations(array $fieldIds)
    {
        $fieldIds = \array_unique($fieldIds);

        $db = $this->_db;
        $categoryId = $this->get('category_id');

        $db->delete('xf_nixfifty_ticket_field_category', 'category_id = ' . $db->quote($categoryId));

        foreach ($fieldIds AS $fieldId)
        {
            $db->insert('xf_nixfifty_ticket_field_category', [
                'field_id'    => $fieldId,
                'category_id' => $categoryId,
            ]);
        }

        return $fieldIds;
    }

    /**
     * Post-delete handling.
     */
    protected function _postDelete()
    {
        $categoryId = $this->get('category_id');
        $db = $this->_db;

        $db->update('xf_nixfifty_ticket_category',
            ['parent_category_id' => $this->get('parent_category_id')],
            'parent_category_id = ' . $this->_db->quote($categoryId)
        );

        $this->_getCategoryModel()->rebuildCategoryStructure();

        if ($this->getOption(self::OPTION_REBUILD_CACHE))
        {
            XenForo_Application::defer('Permission', [], 'Permission', true);
        }
    }

    /**
     * Called when a ticket is updated in this category.
     *
     * @param NixFifty_Tickets_DataWriter_Ticket $ticket
     */
    public function ticketUpdate(NixFifty_Tickets_DataWriter_Ticket $ticket)
    {
        if ($ticket->isInsert())
        {
            $this->updateTicketCount(1);
        }

        if ($ticket->isUpdate() && $ticket->isChanged('category_id'))
        {
            $this->updateTicketCount(1);

            /** @var NixFifty_Tickets_DataWriter_Category $oldCat */
            $oldCat = XenForo_DataWriter::create('NixFifty_Tickets_DataWriter_Category', XenForo_DataWriter::ERROR_SILENT);
            if ($oldCat->setExistingData($ticket->getExisting('category_id')))
            {
                $oldCat->ticketRemoved($ticket);
                $oldCat->save();
            }
        }

        if ($ticket->get('last_modified_date') >= $this->get('last_ticket'))
        {
            $this->set('last_ticket', $ticket->get('last_modified_date'));
            $this->set('last_ticket_title', $ticket->get('title'));
            $this->set('last_ticket_id', $ticket->get('ticket_id'));
        }
    }

    public function ticketRemoved(NixFifty_Tickets_DataWriter_Ticket $ticket)
    {
        $this->updateTicketCount(-1);

        if ($this->get('last_ticket_id') == $ticket->get('ticket_id'))
        {
            $this->updateLastUpdate();
        }
    }

    public function updateLastUpdate()
    {
        $ticket = $this->_db->fetchRow($this->_db->limit("
                SELECT *
                FROM xf_nixfifty_ticket
                WHERE category_id = ?
                ORDER BY last_modified_date DESC
            ", 1
        ), $this->get('category_id'));
        if (!$ticket)
        {
            $this->set('ticket_count', 0);
            $this->set('last_ticket', 0);
            $this->set('last_ticket_title', '');
            $this->set('last_ticket_id', 0);
        }
        else
        {
            $this->set('last_ticket', $ticket['last_modified_date']);
            $this->set('last_ticket_title', $ticket['title']);
            $this->set('last_ticket_id', $ticket['ticket_id']);
        }
    }

    public function updateTicketCount($adjust = null)
    {
        if ($adjust === null)
        {
            $this->set('ticket_count', $this->_db->fetchOne("
                SELECT COUNT(*)
                FROM xf_nixfifty_ticket
                WHERE category_id = ? AND discussion_state = 'visible'
            ", $this->get('category_id')));
        }
        else
        {
            $this->set('ticket_count', $this->get('ticket_count') + $adjust);
        }
    }

    public function rebuildCounters()
    {
        $this->updateLastUpdate();
        $this->updateTicketCount();
    }

    /**
     * @return NixFifty_Tickets_Model_Ticket
     */
    protected function _getTicketModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Ticket');
    }

    /**
     * @return NixFifty_Tickets_Model_Category
     */
    protected function _getCategoryModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Category');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketField
     */
    protected function _getFieldModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketField');
    }

    /**
     * @return XenForo_Model_Forum
     */
    protected function _getForumModel()
    {
        return $this->getModelFromCache('XenForo_Model_Forum');
    }

    /**
     * @return XenForo_Model_ThreadPrefix
     */
    protected function _getThreadPrefixModel()
    {
        return $this->getModelFromCache('XenForo_Model_ThreadPrefix');
    }
}
