<?php

namespace NF\Tickets\Legacy\DataWriter;

class Escalation extends \XF\Legacy\DataWriter
{
    const DATA_MESSAGE_TEXT = 'phraseMessageText';
    const DATA_ALERT_TEXT   = 'phraseAlertText';
    const DATA_EMAIL_TEXT   = 'phraseEmailText';

    /**
     * Phrase that will be created when data creation fails.
     *
     * @var string
     */
    protected $_existingDataErrorPhrase = 'nf_tickets_requested_escalation_not_found';

    /**
     * Gets fields defined for this particular table.
     *
     * @return array
     */
    protected function _getFields()
    {
        return [
            'xf_nixfifty_ticket_escalation' => [
                'escalation_id'          => ['type' => self::TYPE_UINT, 'autoIncrement' => true],
                'escalation_title'       => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'maxLength' => 100,
                    'requiredError' => 'please_enter_valid_title',
                ],
                'escalation_description' => ['type' => self::TYPE_STRING, 'default' => ''],
                'active'                 => ['type' => self::TYPE_BOOLEAN, 'default' => 1],
                'execution_order'        => ['type' => self::TYPE_UINT, 'default' => 1],
                'user_criteria'          => [
                    'type'         => self::TYPE_UNKNOWN, 'required' => true,
                    'verification' => ['$this', '_verifyCriteria'],
                ],
                'ticket_criteria'        => [
                    'type'         => self::TYPE_UNKNOWN, 'required' => true,
                    'verification' => ['$this', '_verifyCriteria'],
                ],
                'message_from_username'  => ['type' => self::TYPE_STRING, 'default' => ''],
                'alert_from_username'    => ['type' => self::TYPE_STRING, 'default' => ''],
                'actions'                => [
                    'type'         => self::TYPE_UNKNOWN, 'required' => true,
                    'verification' => ['$this', '_verifyActions'],
                ],
            ],
        ];
    }

    /**
     * Gets the actual existing data out of data that was passed in.
     *
     * @param mixed
     * @param mixed $data
     * @return array|false
     */
    protected function _getExistingData($data)
    {
        if (!$id = $this->_getExistingPrimaryKey($data))
        {
            return false;
        }

        return ['xf_nixfifty_ticket_escalation' => $this->_getEscalationModel()->getEscalationById($id)];
    }

    /**
     * Gets a SQL condition to update the existing record.
     *
     * @param $tableName
     * @return string
     */
    protected function _getUpdateCondition($tableName)
    {
        return 'escalation_id = ' . $this->_db->quote($this->getExisting('escalation_id'));
    }

    /**
     * Post-save handling.
     */
    protected function _postSave()
    {
        $id = $this->get('escalation_id');

        $phraseData = [
            self::DATA_MESSAGE_TEXT => $this->_getEscalationMessagePhraseName($id),
            self::DATA_ALERT_TEXT   => $this->_getEscalationAlertPhraseName($id),
            self::DATA_EMAIL_TEXT   => $this->_getEscalationEmailPhraseName($id),
        ];

        foreach ($phraseData AS $phraseDataElement => $phraseName)
        {
            $phraseText = $this->getExtraData($phraseDataElement);
            if ($phraseText !== null)
            {
                $this->_insertOrUpdateMasterPhrase($phraseName, $phraseText, '');
            }
        }

        parent::_postSave();
    }

    /**
     * Post-delete handling.
     */
    protected function _postDelete()
    {
        $id = $this->get('escalation_id');

        $this->_deleteMasterPhrase($this->_getEscalationMessagePhraseName($id));
    }

    /**
     * Gets the name of the ticket message phrase for this escalation.
     *
     * @param string $id
     * @return string
     */
    protected function _getEscalationMessagePhraseName($id)
    {
        return $this->_getEscalationModel()->getEscalationMessagePhraseName($id);
    }

    /**
     * Gets the name of the alert phrase for this escalation.
     *
     * @param string $id
     * @return string
     */
    protected function _getEscalationAlertPhraseName($id)
    {
        return $this->_getEscalationModel()->getEscalationAlertPhraseName($id);
    }

    /**
     * Gets the name of the email phrase for this escalation.
     *
     * @param string $id
     * @return string
     */
    protected function _getEscalationEmailPhraseName($id)
    {
        return $this->_getEscalationModel()->getEscalationEmailPhraseName($id);
    }

    protected function _verifyActions(&$actions)
    {
        $actions = XenForo_Helper_Php::safeSerialize($actions);

        return true;
    }

    /**
     * Verifies that the criteria is valid and formats is correctly.
     * Expected input format: [] with children: [rule] => name, [data] => info.
     *
     * @param array|string $criteria Criteria array or serialize string; see above for format. Modified by ref.
     * @return bool
     */
    protected function _verifyCriteria(&$criteria)
    {
        $criteriaFiltered = XenForo_Helper_Criteria::prepareCriteriaForSave($criteria);
        $criteria = XenForo_Helper_Php::safeSerialize($criteriaFiltered);

        return true;
    }

    /**
     * @return NixFifty_Tickets_Model_Escalation
     */
    protected function _getEscalationModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Escalation');
    }
}
