<?php

use function array_key_exists, count, intval, is_array, is_string, strval;

class NixFifty_Tickets_DataWriter_Ticket extends XenForo_DataWriter
{
    const DATA_MESSAGE  = 'messageBbCode';
    const DATA_CATEGORY = 'categoryInfo';

    /**
     * Holds the reason for soft deletion.
     *
     * @var string
     */
    const DATA_DELETE_REASON = 'deleteReason';

    const OPTION_REQUIRE_INSERT_FIRST_MESSAGE = 'requireFirstMessage';
    const OPTION_VERIFY_THREAD_ASSOCIATION    = 'verifyThreadAssociation';
    const OPTION_CREATE_THREAD_NODE_ID        = 'createThreadNodeId';
    const OPTION_CREATE_THREAD_PREFIX_ID      = 'createThreadNodeId';

    const OPTION_USERNAME_LENGTH_MIN       = 'usernameLengthMin';
    const OPTION_USERNAME_LENGTH_MAX       = 'usernameLengthMax';
    const OPTION_USERNAME_DISALLOWED_NAMES = 'usernameIllegalNames';
    const OPTION_USERNAME_REGEX            = 'usernameRegex';

    const OPTION_IS_FOREIGN_OPEN = 'foreignOpen';

    const OPTION_FIRE_ESCALATIONS   = 'fireEscalations';
    const OPTION_SEND_NOTIFICATIONS = 'sendNotifications';

    /**
     * Phrase that will be created when data creation fails.
     *
     * @var string
     */
    protected $_existingDataErrorPhrase = 'nf_tickets_requested_ticket_not_found';

    protected $_firstMessageDw = null;

    protected $_guestPassword = null;

    /**
     * The custom fields to be updated. Use setCustomFields to manage this.
     *
     * @var array
     */
    protected $_updateCustomFields = null;

    public static $categoryCache = [];

    /**
     * Cache of the messages in this discussion so they're only retrieved when needed
     *
     * @var array
     */
    protected $_messageCache = [];

    /**
     * Gets fields defined for this particular table.
     *
     * @return array
     */
    protected function _getFields()
    {
        return [
            'xf_nixfifty_ticket' => [
                'ticket_id'               => ['type' => self::TYPE_UINT, 'autoIncrement' => true],
                'ticket_ref'              => ['type' => self::TYPE_STRING, 'required' => true, 'maxLength' => 25],
                'title'                   => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'maxLength' => 100,
                    'requiredError' => 'please_enter_valid_title',
                ],
                'user_id'                 => ['type' => self::TYPE_UINT, 'required' => true, 'default' => '0'],
                'username'                => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'default' => '',
                    'requiredError' => 'nf_tickets_please_enter_valid_name',
                ],
                'user_name'               => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'default' => '', 'maxLength' => 50,
                    'requiredError' => 'nf_tickets_please_enter_valid_name',
                    'verification'  => ['$this', '_verifyUsernameGuest'],
                ],
                'user_email'              => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'maxLength' => 120,
                    'verification'  => ['$this', '_verifyEmail'],
                    'requiredError' => 'please_enter_valid_email',
                ],
                'password'                => ['type' => self::TYPE_STRING, 'default' => '', 'maxLength' => 255],
                'starter_user_id'         => ['type' => self::TYPE_UINT, 'required' => true, 'default' => '0'],
                'starter_username'        => ['type' => self::TYPE_STRING, 'required' => true, 'default' => ''],
                'start_date'              => ['type' => self::TYPE_UINT, 'default' => XenForo_Application::$time],
                'first_message_id'        => ['type' => self::TYPE_UINT, 'required' => true, 'default' => '0'],
                'first_message_date'      => ['type' => self::TYPE_UINT, 'default' => XenForo_Application::$time],
                'priority'                => [
                    'type'          => self::TYPE_UINT, 'default' => '2',
                    'allowedValues' => ['0', '1', '2', '3'],
                ],
                'ticket_status_id'        => ['type' => self::TYPE_UINT, 'required' => true],
                'ticket_state'            => [
                    'type'          => self::TYPE_STRING, 'default' => 'open',
                    'allowedValues' => ['open', 'pending', 'locked', 'assigned', 'resolved'],
                ],
                'ticket_locked'           => ['type' => self::TYPE_BOOLEAN, 'default' => 0],
                'discussion_state'        => [
                    'type'          => self::TYPE_STRING, 'default' => 'visible',
                    'allowedValues' => ['visible', 'moderated', 'deleted'],
                ],
                'discussion_type'         => ['type' => self::TYPE_STRING, 'default' => '', 'maxLength' => 25],
                'assigned_user_id'        => ['type' => self::TYPE_UINT, 'default' => 0],
                'assigned_username'       => ['type' => self::TYPE_STRING, 'default' => ''],
                'category_id'             => ['type' => self::TYPE_UINT, 'verification' => ['$this', '_verifyTicketCategory']],
                'last_message_id'         => ['type' => self::TYPE_UINT, 'required' => true, 'default' => '0'],
                'last_message_date'       => ['type' => self::TYPE_UINT, 'default' => XenForo_Application::$time],
                'last_message_user_id'    => ['type' => self::TYPE_UINT, 'default' => 0],
                'last_message_username'   => ['type' => self::TYPE_STRING, 'default' => '', 'maxLength' => 50],
                'last_modified_date'      => ['type' => self::TYPE_UINT, 'default' => XenForo_Application::$time],
                'last_modified_user_id'   => ['type' => self::TYPE_UINT, 'default' => 0],
                'last_modified_username'  => ['type' => self::TYPE_STRING, 'default' => '', 'maxLength' => 50],
                'last_status_change_date' => ['type' => self::TYPE_UINT, 'default' => 0],
                'reply_count'             => ['type' => self::TYPE_UINT_FORCED, 'default' => 0],
                'custom_ticket_fields'    => ['type' => self::TYPE_SERIALIZED, 'default' => ''],
                'thread_id'               => ['type' => self::TYPE_UINT, 'default' => 0],
                //'sensitive_data' => ['type'  => self::TYPE_STRING, 'default' => ''],
                //'nonce' => ['type'  => self::TYPE_STRING, 'default' => ''],
            ],
        ];
    }

    /**
     * Gets the actual existing data out of data that was passed in.
     *
     * @param mixed
     * @param mixed $data
     * @return array|false
     */
    protected function _getExistingData($data)
    {
        if (!$id = $this->_getExistingPrimaryKey($data))
        {
            return false;
        }

        return ['xf_nixfifty_ticket' => $this->_getTicketModel()->getTicketById($id)];
    }

    /**
     * Gets a SQL condition to update the existing record.
     *
     * @param $tableName
     * @return string
     */
    protected function _getUpdateCondition($tableName)
    {
        return 'ticket_id = ' . $this->_db->quote($this->getExisting('ticket_id'));
    }

    /**
     * Gets the default set of options for this data writer.
     *
     * @return array
     */
    protected function _getDefaultOptions()
    {
        $options = XenForo_Application::get('options');

        return [
            self::OPTION_USERNAME_LENGTH_MIN          => intval($options->get('usernameLength', 'min')),
            self::OPTION_USERNAME_LENGTH_MAX          => intval($options->get('usernameLength', 'max')),
            self::OPTION_USERNAME_DISALLOWED_NAMES    => \preg_split('/\r?\n/', $options->get('usernameValidation', 'disallowedNames')),
            self::OPTION_USERNAME_REGEX               => $options->get('usernameValidation', 'matchRegex'),
            self::OPTION_REQUIRE_INSERT_FIRST_MESSAGE => true,
            self::OPTION_VERIFY_THREAD_ASSOCIATION    => true,
            self::OPTION_CREATE_THREAD_NODE_ID        => null,
            self::OPTION_CREATE_THREAD_PREFIX_ID      => null,
            self::OPTION_IS_FOREIGN_OPEN              => false,
            self::OPTION_FIRE_ESCALATIONS             => true,
            self::OPTION_SEND_NOTIFICATIONS           => true,
        ];
    }

    public function addReply(array $reply)
    {
        if ($this->isUpdate())
        {
            $master = $this->_db->fetchRow('
                SELECT *
                FROM xf_nixfifty_ticket
                WHERE ticket_id = ?
                FOR UPDATE
            ', $this->get('ticket_id'));
            if ($master)
            {
                $this->_existingData['xf_nixfifty_ticket'] = $master;
            }
        }

        $this->set('reply_count', $this->updateCounts());
        if ($reply['message_date'] > $this->get('last_message_date'))
        {
            $this->set('last_message_id', $reply['message_id']);
            $this->set('last_message_date', $reply['message_date']);
            $this->set('last_message_user_id', $reply['user_id']);
            $this->set('last_message_username', $reply['username']);

            $this->set('last_modified_date', $reply['message_date']);
            $this->set('last_modified_user_id', $reply['user_id']);
            $this->set('last_modified_username', $reply['username']);
        }
    }

    /**
     * Updates denormalized counters, based on changes made to the provided
     * discussion message, after the message has been saved.
     *
     * @param NixFifty_Tickets_DataWriter_TicketMessage $messageDw
     */
    public function updateCountersAfterMessageSave(NixFifty_Tickets_DataWriter_TicketMessage $messageDw)
    {
        if ($messageDw->get('message_state') == 'visible' && $messageDw->get('message_date') > $this->get('last_message_date'))
        {
            $this->set('last_message_date', $messageDw->get('message_date'));
            $this->set('last_message_id', $messageDw->get('message_id'));
            $this->set('last_message_user_id', $messageDw->get('user_id'));
            $this->set('last_message_username', $messageDw->get('username'));

            $this->set('last_modified_date', $messageDw->get('message_date'));
            $this->set('last_modified_user_id', $messageDw->get('user_id'));
            $this->set('last_modified_username', $messageDw->get('username'));
        }

        if ($messageDw->get('message_state') == 'visible' && $messageDw->getExisting('message_state') != 'visible')
        {
            if (NixFifty_Tickets_Globals::$isFirstMessage)
            {
                $this->set('reply_count', 0);
            }
            else
            {
                $this->set('reply_count', $this->get('reply_count') + 1);
            }
        }
        else if ($messageDw->getExisting('message_state') == 'visible' && $messageDw->get('message_state') != 'visible')
        {
            $this->set('reply_count', $this->get('reply_count') - 1);

            if ($messageDw->get('message_id') == $this->get('last_message_id'))
            {
                $this->updateLastMessage();
            }
        }

        if ($replyActions = $messageDw->getExtraData(NixFifty_Tickets_DataWriter_TicketMessage::DATA_POST_REPLY_ACTIONS))
        {
            if (array_key_exists('ticket_state', $replyActions))
            {
                $this->set('ticket_state', $replyActions['ticket_state']);
            }

            if (array_key_exists('assigned_user_id', $replyActions))
            {
                $this->set('assigned_user_id', $replyActions['assigned_user_id']);
                $this->set('assigned_username', $replyActions['assigned_username']);
            }
        }
    }

    /**
     * Updates denormalized counters. Used after a message has been deleted.
     *
     * @param NixFifty_Tickets_DataWriter_TicketMessage $messageDw
     * @param bool                                      $deleteIfFirstMessage If true and message if first, delete discussion
     * @return string State changes to discussion: delete means remove discussion; firstDelete means first message was removed but still valid
     */
    public function updateCountersAfterMessageDelete(NixFifty_Tickets_DataWriter_TicketMessage $messageDw, $deleteIfFirstMessage = true)
    {
        $messageId = $messageDw->get('message_id');

        if ($messageId == $this->get('first_message_id'))
        {
            if (!$deleteIfFirstMessage && $this->rebuildTicket())
            {
                return 'firstDelete';
            }
            else
            {
                return 'delete';
            }
        }

        if ($messageId == $this->get('last_message_id'))
        {
            $this->updateLastMessage();
        }

        if ($messageDw->get('message_state') == 'visible')
        {
            $this->set('reply_count', $this->get('reply_count') - 1);
        }

        return '';
    }

    /**
     * Updates the value of the last post for this discussion.
     */
    public function updateLastMessage()
    {
        $lastMessage = $this->_getLastMessageInDiscussion();
        if ($lastMessage)
        {
            $this->set('last_message_id', $lastMessage['message_id']);
            $this->set('last_message_date', $lastMessage['message_date']);
            $this->set('last_message_user_id', $lastMessage['user_id']);
            $this->set('last_message_username', $lastMessage['username']);
        }
        else
        {
            $this->set('last_message_id', $this->get('first_post_id'));
            $this->set('last_message_date', $this->get('message_date'));
            $this->set('last_message_user_id', $this->get('user_id'));
            $this->set('last_message_username', $this->get('username'));
        }
    }

    protected function _preSave()
    {
        if ($this->getOption(self::OPTION_REQUIRE_INSERT_FIRST_MESSAGE) && $this->isInsert() && !$this->_firstMessageDw)
        {
            throw new XenForo_Exception('First message DataWriter is missing.');
        }

        if ($this->isUpdate() && $this->isChanged('ticket_state'))
        {
            $this->set('last_status_change_date', XenForo_Application::$time);
        }

        /*
        if ($this->isUpdate() || !$this->get('last_modified_date'))
        {
            $this->set('last_modified_date', XenForo_Application::$time);
        }

        if ($this->isUpdate() || !$this->get('last_message_date'))
        {
            $this->set('last_message_date', XenForo_Application::$time);
        }
        */

        if ($this->isChanged('category_id'))
        {
            if ($this->isUpdate() && !is_array($this->_updateCustomFields))
            {
                // need to filter the custom fields to only the fields that apply to the new category
                $fieldModel = $this->_getFieldModel();

                $this->_updateCustomFields = $this->_filterValidFields(
                    $fieldModel->getTicketFieldValues($this->get('ticket_id')),
                    $fieldModel->getTicketFieldsForEdit($this->get('category_id'))
                );
                $this->set('custom_ticket_fields', $this->_updateCustomFields);
            }
        }

        if ($this->isInsert())
        {
            if (!$this->get('password'))
            {
                $this->setPassword();
            }

            if (!$this->get('ticket_ref'))
            {
                $length = \max(XenForo_Application::get('options')->ticketRefLength, 8);
                $ticketRef = $this->_getTicketModel()->generateTicketRef($length);
                $this->set('ticket_ref', $ticketRef);
            }

            if (!$this->isChanged('start_date'))
            {
                $this->set('start_date', XenForo_Application::$time);
            }

            $fieldMap = [
                'last_modified_date'     => 'start_date',
                'last_modified_user_id'  => 'user_id',
                'last_modified_username' => 'username',

                'starter_user_id'  => 'user_id',
                'starter_username' => 'username',
            ];

            foreach ($fieldMap AS $childField => $parentField)
            {
                if (!$this->isChanged($childField))
                {
                    $this->set($childField, $this->get($parentField));
                }
            }

            if ($this->_firstMessageDw)
            {
                $messageDw = $this->_firstMessageDw;
                $messageDw->set('ticket_id', 0);
                $messageDw->set('position', 0);

                foreach ($this->_newData AS $table => $newData)
                {
                    foreach ($newData AS $field => $value)
                    {
                        if ($field == 'user_id')
                        {
                            $field = 'user_id';
                        }
                        else if ($field == 'username')
                        {
                            $field = 'username';
                        }

                        $messageDw->set($field, $value, '', ['ignoreInvalidFields' => true]);
                    }
                }

                $messageDw->preSave();
                $firstMessageErrors = $messageDw->getErrors();
                if ($firstMessageErrors)
                {
                    $this->_errors = \array_merge($this->_errors, $firstMessageErrors);
                }
            }
        }
    }

    /**
     * Gets all messages in this discussion, in order
     *
     * @param bool $includeMessage
     * @return array
     */
    protected function _getDiscussionMessages($includeMessage = false)
    {
        $cacheKey = 'ticket-messages' . ($includeMessage ? '-message' : '');
        if (!isset($this->_messageCache[$cacheKey]))
        {
            $this->_messageCache[$cacheKey] = $this->_getMessagesInDiscussionSimple($includeMessage);
        }

        return $this->_messageCache[$cacheKey];
    }

    protected function _getMessageIdsInDiscussion()
    {
        return \array_keys($this->_getDiscussionMessages(false));
    }

    protected function _getDiscussionMessageIds()
    {
        if (!isset($this->_messageCache['messageIds']))
        {
            if (isset($this->_messageCache['messages']))
            {
                $this->_messageCache['messageIds'] = \array_keys($this->_messageCache['messages']);
            }
            else
            {
                $this->_messageCache['messageIds'] = $this->_getMessageIdsInDiscussion();
            }
        }

        return $this->_messageCache['messageIds'];
    }

    protected function _postSave()
    {
        $ticketId = $this->get('ticket_id');

        $this->updateCustomFields();

        if ($this->isUpdate() && $this->isChanged('title'))
        {
            $threadDw = XenForo_DataWriter::create('XenForo_DataWriter_Discussion_Thread', XenForo_DataWriter::ERROR_SILENT);
            if ($threadDw->setExistingData($this->get('thread_id')) && $threadDw->get('discussion_type') == 'tickets')
            {
                $threadDw->set('title', $this->get('title'));
                $threadDw->save();
            }
        }

        $categoryDw = $this->_getCategoryDwForUpdate();
        if ($this->isUpdate() && $this->isChanged('category_id') && $this->get('thread_id'))
        {
            $nodeId = $this->getOption(self::OPTION_CREATE_THREAD_NODE_ID);
            if ($nodeId === null)
            {
                $nodeId = $categoryDw->get('thread_node_id');
            }
            $prefixId = $this->getOption(self::OPTION_CREATE_THREAD_PREFIX_ID);
            if ($prefixId === null)
            {
                $prefixId = $categoryDw->get('thread_prefix_id');
            }

            $threadDw = XenForo_DataWriter::create('XenForo_DataWriter_Discussion_Thread', XenForo_DataWriter::ERROR_SILENT);
            if ($threadDw->setExistingData($this->get('thread_id')) && $threadDw->get('discussion_type') == 'tickets' && $nodeId > 0)
            {
                $threadDw->set('node_id', $nodeId);
                $threadDw->set('prefix_id', $prefixId);
                $threadDw->save();
            }
        }

        if ($this->_firstMessageDw)
        {
            $this->_firstMessageDw->set('ticket_id', $ticketId, '', ['setAfterPreSave' => true]);
            $this->_firstMessageDw->save();
            $firstMessageId = $this->_firstMessageDw->get('message_id');

            $toUpdate = [
                'first_message_id'      => $firstMessageId,
                'last_message_id'       => $firstMessageId,
                'last_message_date'     => XenForo_Application::$time,
                'last_message_user_id'  => $this->_firstMessageDw->get('user_id'),
                'last_message_username' => $this->_firstMessageDw->get('username'),
            ];
            $this->_db->update('xf_nixfifty_ticket', $toUpdate, 'ticket_id = ' . $this->_db->quote($ticketId));
            $this->bulkSet($toUpdate, ['setAfterPreSave' => true]);
        }

        if ($categoryDw)
        {
            $categoryDw->ticketUpdate($this);
            $categoryDw->save();
        }

        $this->_updateDeletionLog();

        if ($this->isInsert())
        {
            if (XenForo_Application::isRegistered('fc'))
            {
                (new NixFifty_Tickets_Helper_GuestTicket(XenForo_Application::getFc()->getRequest()))->addTicketAuth($ticketId, $this->_guestPassword);
            }
            $this->getModelFromCache('NixFifty_Tickets_Model_NewsFeed')->publishItem($this->getMergedData(), 'ticket');
        }

        if ($this->isInsert() AND $this->get('user_id') AND $this->get('ticket_state') == 'open')
        {
            $this->_db->query('
                UPDATE xf_user
                SET nixfifty_ticket_count = nixfifty_ticket_count + 1
                WHERE user_id = ?
            ', $this->get('user_id'));
        }

        if ($this->isUpdate() && $this->isChanged('discussion_state') && $this->get('discussion_state') == 'deleted')
        {
            $this->_softDeleted();
        }

        $this->_indexForSearch();

        if ($this->isInsert())
        {
            $options = XenForo_Application::getOptions();
            /** @noinspection PhpUndefinedFieldInspection */
            $spamTriggerDays = $options->nfTicketSpamTriggerDays;
            /** @noinspection PhpUndefinedFieldInspection */
            $spamTriggerLimit = $options->nfTicketSpamTriggerLimit;
            $this->logSpamTriggerRecords($spamTriggerDays, $spamTriggerLimit);
        }
    }

    protected function logSpamTriggerRecords($spamTriggerDays, $spamTriggerLimit)
    {
        if (!$spamTriggerDays || !$spamTriggerLimit)
        {
            return;
        }

        $binaryIp = null;
        $firstMessageDw = $this->getFirstMessageDw();
        if ($firstMessageDw && $firstMessageDw->get('ip_id'))
        {
            $binaryIp = $this->_db->fetchOne('SELECT ip FROM xf_ip WHERE ip_id = ?', $firstMessageDw->get('ip_id'));
        }
        $ticketModel = $this->_getTicketModel();
        $ticketModel->copySpamTriggerLogs($this->get('ticket_id'), $this->get('start_date'), $binaryIp, $this->get('user_email'), $spamTriggerDays, $spamTriggerLimit);
    }

    /**
     * Updates the search index for this discussion.
     */
    protected function _indexForSearch()
    {
        if ($this->get('discussion_state') == 'visible')
        {
            if ($this->getExisting('discussion_state') != 'visible')
            {
                $this->_insertIntoSearchIndex();
            }
            else if ($this->_needsSearchIndexUpdate())
            {
                $this->_updateSearchIndexTitle();
            }
        }
        else if ($this->isUpdate() && $this->get('discussion_state') != 'visible' && $this->getExisting('discussion_state') == 'visible')
        {
            $this->_deleteFromSearchIndex();
        }
    }

    /**
     * Returns true if the changes made require the search index to be updated.
     *
     * @return boolean
     */
    protected function _needsSearchIndexUpdate()
    {
        if ($this->isChanged('title') || $this->isChanged($this->get('ticket_id')))
        {
            return true;
        }

        return false;
    }

    /**
     * Inserts a record in the search index for this discussion.
     */
    protected function _insertIntoSearchIndex()
    {
        $discussion = $this->getMergedData();
        $indexer = new XenForo_Search_Indexer();

        $discussionHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_Ticket');
        if ($discussionHandler)
        {
            $discussionHandler->insertIntoIndex($indexer, $discussion);
        }

        if ($this->isUpdate())
        {
            XenForo_Application::defer('SearchIndexPartial', [
                'contentType' => 'ticket_message',
                'contentIds'  => $this->_getDiscussionMessageIds(),
            ]);
        }
    }

    /**
     * Updates the title in the search index for this discussion.
     */
    protected function _updateSearchIndexTitle()
    {
        $indexer = new XenForo_Search_Indexer();
        $mergedData = $this->getMergedData();

        $discussionHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_Ticket');
        if ($discussionHandler)
        {
            $discussionHandler->insertIntoIndex($indexer, $mergedData);
        }

        if ($this->isUpdate())
        {
            $messageHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_TicketMessage');
            if ($messageHandler && $firstMessageDw = $this->getFirstMessageDw())
            {
                if ($firstMessageDw->getMergedData())
                {
                    $messageHandler->insertIntoIndex($indexer, $firstMessageDw->getMergedData(), $mergedData);
                }
            }
        }
    }

    /**
     * Deletes this discussion from the search index.
     */
    protected function _deleteFromSearchIndex()
    {
        $discussion = $this->getMergedData();
        $indexer = new XenForo_Search_Indexer();

        $discussionHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_Ticket');
        if ($discussionHandler)
        {
            $discussionHandler->deleteFromIndex($indexer, $discussion);
        }

        $messageHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_TicketMessage');
        if ($messageHandler)
        {
            $messageHandler->deleteFromIndex($indexer, $this->_getDiscussionMessageIds());
        }
    }

    protected function _softDeleted()
    {
        if ($categoryId = $this->get('category_id'))
        {
            $this->_db->query("UPDATE xf_nixfifty_ticket_category SET ticket_count=ticket_count-1 WHERE category_id=? AND ticket_count > 0", $categoryId);
        }
    }

    protected function _unDeleted()
    {
        if ($categoryId = $this->get('category_id'))
        {
            $this->_db->query("UPDATE xf_nixfifty_ticket_category SET ticket_count=ticket_count+1 WHERE category_id=?", $categoryId);
        }
    }

    /**
     * Updates the deletion log if necessary.
     */
    protected function _updateDeletionLog($hardDelete = false)
    {
        if ($hardDelete
            || ($this->isChanged('discussion_state') && $this->getExisting('discussion_state') == 'deleted')
        )
        {
            $this->getModelFromCache('XenForo_Model_DeletionLog')->removeDeletionLog(
                'ticket', $this->get('ticket_id')
            );
        }

        if ($this->isChanged('discussion_state') && $this->get('discussion_state') == 'deleted')
        {
            $reason = $this->getExtraData(self::DATA_DELETE_REASON);
            $this->getModelFromCache('XenForo_Model_DeletionLog')->logDeletion(
                'ticket', $this->get('ticket_id'), $reason
            );
        }
    }

    protected function _postSaveAfterTransaction()
    {
        if ($this->isInsert() ||
            $this->isChanged('assigned_user_id') ||
            $this->isChanged('last_message_date') ||
            $this->isChanged('last_modified_date') ||
            $this->isChanged('last_status_change_date'))
        {
            $this->_getTicketModel()->rebuildTicketCountCache();
        }

        if (!$this->getOption(self::OPTION_SEND_NOTIFICATIONS))
        {
            return;
        }

        $category = $this->_getCategoriesModel()->getCategoryById($this->get('category_id'));

        $ticket = $this->getMergedData();
        $ticket = $this->_getTicketModel()->prepareTicket($ticket, $category);
        $ticket = $this->_getTicketModel()->prepareTicketCustomFields($ticket, $category);

        if (!isset($ticket['messageText']))
        {
            $ticket['message'] = $this->_getTicketMessageModel()->getTicketMessageById($ticket['first_message_id']);

            $bbCodeParserText = XenForo_BbCode_Parser::create(XenForo_BbCode_Formatter_Base::create('Text'));
            $ticket['messageText'] = new XenForo_BbCode_TextWrapper($ticket['message']['message'], $bbCodeParserText);

            $bbCodeParserHtml = XenForo_BbCode_Parser::create(XenForo_BbCode_Formatter_Base::create('HtmlEmail'));
            $ticket['messageHtml'] = new XenForo_BbCode_TextWrapper($ticket['message']['message'], $bbCodeParserHtml);
        }

        if (!isset($ticket['titleCensored']))
        {
            $ticket['titleCensored'] = XenForo_Helper_String::censorString($ticket['title']);
        }

        if ($this->isInsert() AND $ticket['ticket_state'] == 'open')
        {
            /* @var $userModel XenForo_Model_User */
            $userModel = XenForo_Model::create('XenForo_Model_User');

            $ticketUser = $userModel->getUserById($this->get('user_id'));
            if (!$ticketUser)
            {
                $ticketUser = $userModel->getVisitingGuestUser();
                $ticketUser['username'] = $this->get('user_name');
                $ticketUser['email'] = $this->get('user_email');
                $ticketUser['guest'] = true;
            }
            else
            {
                $ticketUser['guest'] = false;
            }

            if ($ticket['starter_user_id'] != $ticket['user_id'])
            {
                $emailTemplate = 'ticket_insert_foreign';
            }
            else
            {
                $emailTemplate = ($ticketUser['guest'] ? 'ticket_insert_guest' : 'ticket_insert');
            }

            $mail = XenForo_Mail::create($emailTemplate, [
                'category'   => $category,
                'ticket'     => $ticket,
                'ticketUser' => $ticketUser,
                'password'   => $this->_guestPassword,
            ], $ticketUser['language_id']);

            $mail->enableAllLanguagePreCache();
            $mail->send($ticketUser['email'], $ticketUser['username']);

            $categoryWatch = $this->_getCategoryWatchModel()->sendNotificationToWatchUsersOnTicketInsert($ticket);

            if ($categoryWatch)
            {
                $this->_getTicketUserWatchModel()->sendNotificationToWatchUsersOnTicketInsert($ticket, $categoryWatch['alerted'], $categoryWatch['emailed']);
            }
        }
    }

    protected function _verifyUsernameGuest(&$username)
    {
        if (($this->isUpdate() && $username === $this->getExisting('user_name')) || $this->get('user_id'))
        {
            return true;
        }

        // standardize white space in names
        $username = \trim(\preg_replace('/\s+/', ' ', $username));

        $usernameLength = \utf8_strlen($username);
        $minLength = $this->getOption(self::OPTION_USERNAME_LENGTH_MIN);
        $maxLength = $this->getOption(self::OPTION_USERNAME_LENGTH_MAX);

        if ($minLength > 0 && $usernameLength < $minLength)
        {
            $this->error(new XenForo_Phrase('please_enter_name_that_is_at_least_x_characters_long', ['count' => $minLength]), 'user_name');

            return false;
        }
        if ($maxLength > 0 && $usernameLength > $maxLength)
        {
            $this->error(new XenForo_Phrase('please_enter_name_that_is_at_most_x_characters_long', ['count' => $maxLength]), 'user_name');

            return false;
        }

        $disallowedNames = $this->getOption(self::OPTION_USERNAME_DISALLOWED_NAMES);
        if ($disallowedNames)
        {
            foreach ($disallowedNames AS $name)
            {
                $name = \trim($name);
                if ($name === '')
                {
                    continue;
                }
                if (\stripos($username, $name) !== false)
                {
                    $this->error(new XenForo_Phrase('please_enter_another_name_disallowed_words'), 'user_name');

                    return false;
                }
            }
        }

        $matchRegex = $this->getOption(self::OPTION_USERNAME_REGEX);
        if ($matchRegex)
        {
            $matchRegex = \str_replace('#', '\\#', $matchRegex); // escape delim only
            if (!\preg_match('#' . $matchRegex . '#i', $username))
            {
                $this->error(new XenForo_Phrase('please_enter_another_name_required_format'), 'user_name');

                return false;
            }
        }

        $censoredUserName = XenForo_Helper_String::censorString($username);
        if ($censoredUserName !== $username)
        {
            $this->error(new XenForo_Phrase('please_enter_name_that_does_not_contain_any_censored_words'), 'user_name');

            return false;
        }

        // ignore check if unicode properties aren't compiled
        try
        {
            if (@\preg_match("/\p{C}/u", $username))
            {
                $this->error(new XenForo_Phrase('please_enter_name_without_using_control_characters'), 'user_name');

                return false;
            }
        }
        catch (Exception $e)
        {
        }

        if (\strpos($username, ',') !== false)
        {
            $this->error(new XenForo_Phrase('please_enter_name_that_does_not_contain_comma'), 'user_name');

            return false;
        }

        if (Zend_Validate::is($username, 'EmailAddress'))
        {
            $this->error(new XenForo_Phrase('please_enter_name_that_does_not_resemble_an_email_address'), 'user_name');

            return false;
        }

        return true;
    }

    /**
     * Verification callback to check the email address is in a valid form.
     *
     * @param string Email Address
     * @return bool
     */
    protected function _verifyEmail(&$email)
    {
        if ($email == '')
        {
            return true;
        }

        if ($this->isUpdate() && $email === $this->getExisting('email'))
        {
            return true;
        }

        if (!XenForo_Helper_Email::isEmailValid($email))
        {
            $this->error(new XenForo_Phrase('please_enter_valid_email'), 'email');

            return false;
        }

        if (XenForo_Helper_Email::emailHasTypo($email))
        {
            $this->error(new XenForo_Phrase('email_address_you_entered_appears_have_typo'), 'email');

            return false;
        }

        return true;
    }

    /**
     * Verifies that ticket category is valid.
     *
     * @param int $ticketCategoryId
     * @return bool
     */
    protected function _verifyTicketCategory(&$ticketCategoryId)
    {
        if (!empty($ticketCategoryId))
        {
            // verify that the category exists
            $ticketCategory = $this->_getCategoriesModel()->getCategoryById($ticketCategoryId);

            if (empty($ticketCategory))
            {
                $this->error(new XenForo_Phrase('nf_tickets_please_select_valid_ticket_category'), 'category_id');

                return false;
            }
        }

        return true;
    }

    public function updateCounts()
    {
        $count = $this->_db->fetchOne("
            SELECT COUNT(*) AS reply_count
            FROM xf_nixfifty_ticket_message
            WHERE ticket_id = ?
                AND message <> ''
                AND message_state = 'visible'
        ", $this->get('ticket_id'));

        return --$count;
    }

    public function setPassword($password = false)
    {
        if (!$password)
        {
            $password = \strtoupper(XenForo_Application::generateRandomString(8));
        }

        if (\function_exists('password_hash'))
        {
            $hash = \password_hash($password, PASSWORD_BCRYPT);
        }
        else
        {
            /** @var NixFifty_Tickets_Helper_Password $helperClass */
            $helperClass = XenForo_Application::resolveDynamicClass('NixFifty_Tickets_Helper_Password');
            $helper = new $helperClass();

            $hash = $helper->password_hash($password, PASSWORD_BCRYPT);
        }

        $this->_guestPassword = $password;

        return $this->set('password', $hash);
    }

    public function setCustomFields(array $fieldValues, array $fieldsShown = null)
    {
        $fieldModel = $this->_getFieldModel();
        $fields = $fieldModel->getTicketFieldsForEdit($this->get('category_id'));

        if (!is_array($fieldsShown))
        {
            $fieldsShown = \array_keys($fields);
        }

        if ($this->get('ticket_id') && !$this->_importMode)
        {
            $existingValues = $fieldModel->getTicketFieldValues($this->get('ticket_id'));
        }
        else
        {
            $existingValues = [];
        }

        $finalValues = [];

        foreach ($fieldsShown AS $fieldId)
        {
            if (!isset($fields[$fieldId]))
            {
                continue;
            }

            $field = $fields[$fieldId];
            $multiChoice = ($field['field_type'] == 'checkbox' || $field['field_type'] == 'multiselect');

            if ($multiChoice)
            {
                // multi selection - array
                $value = [];
                if (isset($fieldValues[$fieldId]))
                {
                    if (is_string($fieldValues[$fieldId]))
                    {
                        $value = [$fieldValues[$fieldId]];
                    }
                    else if (is_array($fieldValues[$fieldId]))
                    {
                        $value = $fieldValues[$fieldId];
                    }
                }
            }
            else
            {
                // single selection - string
                if (isset($fieldValues[$fieldId]))
                {
                    if (is_array($fieldValues[$fieldId]))
                    {
                        $value = count($fieldValues[$fieldId]) ? strval(\reset($fieldValues[$fieldId])) : '';
                    }
                    else
                    {
                        $value = strval($fieldValues[$fieldId]);
                    }
                }
                else
                {
                    $value = '';
                }
            }

            $existingValue = (isset($existingValues[$fieldId]) ? $existingValues[$fieldId] : null);

            if (!$this->_importMode)
            {
                $valid = $fieldModel->verifyTicketFieldValue($field, $value, $error);
                if (!$valid)
                {
                    $this->error($error, "custom_field_$fieldId");
                    continue;
                }

                if ($field['required'] && ($value === '' || $value === []))
                {
                    $this->error(new XenForo_Phrase('please_enter_value_for_all_required_fields'), "required");
                    continue;
                }
            }

            if ($value !== $existingValue)
            {
                $finalValues[$fieldId] = $value;
            }
        }

        $this->_updateCustomFields = $this->_filterValidFields($finalValues + $existingValues, $fields);
        $this->set('custom_ticket_fields', $this->_updateCustomFields);
    }

    public function updateCustomFields()
    {
        if (is_array($this->_updateCustomFields))
        {
            $ticketId = $this->get('ticket_id');

            $this->_db->query('DELETE FROM xf_nixfifty_ticket_field_value WHERE ticket_id = ?', $ticketId);

            foreach ($this->_updateCustomFields AS $fieldId => $value)
            {
                if (is_array($value))
                {
                    $value = \serialize($value);
                }
                $this->_db->query('
                    INSERT INTO xf_nixfifty_ticket_field_value
                        (ticket_id, field_id, field_value)
                    VALUES
                        (?, ?, ?)
                    ON DUPLICATE KEY UPDATE
                        field_value = VALUES(field_value)
                ', [$ticketId, $fieldId, $value]);
            }
        }
    }

    protected function _filterValidFields(array $values, array $fields)
    {
        $newValues = [];
        foreach ($fields AS $field)
        {
            if (isset($values[$field['field_id']]))
            {
                $newValues[$field['field_id']] = $values[$field['field_id']];
            }
        }

        return $newValues;
    }

    protected function _preDelete()
    {
        $this->_getTicketMessageModel()->deleteMessagesByTicketId($this->get('ticket_id'));
    }

    protected function _postDelete()
    {
        $ticketId = $this->_db->quote($this->get('ticket_id'));

        $this->getModelFromCache('XenForo_Model_Alert')->deleteAlerts(
            'ticket', $ticketId
        );

        $idQuoted = $this->_db->quote($this->get('ticket_id'));
        $this->_db->delete('xf_nixfifty_ticket_message', 'ticket_id = ' . $idQuoted);
        $this->_db->delete('xf_nixfifty_ticket_field_value', 'ticket_id = ' . $idQuoted);
        $this->_db->delete('xf_nixfifty_ticket_read', 'ticket_id = ' . $idQuoted);
        $this->_db->delete('xf_nixfifty_ticket_watch', 'ticket_id = ' . $idQuoted);
        $this->_db->query('DELETE FROM xf_nixfifty_spam_log WHERE ticket_id = ?', $this->get('ticket_id'));

        if ($categoryId = $this->get('category_id'))
        {
            $this->_db->query("UPDATE xf_nixfifty_ticket_category SET ticket_count=ticket_count-1 WHERE category_id=? AND ticket_count > 0", $categoryId);
        }

        $this->_ticketHardDeleted();

        $this->_updateDeletionLog(true);

        $this->getModelFromCache('XenForo_Model_ModerationQueue')->deleteFromModerationQueue(
            'ticket', $this->get('ticket_id')
        );

        $this->getModelFromCache('XenForo_Model_EditHistory')->deleteEditHistoryForContent(
            'ticket', $this->get('ticket_id')
        );

        $this->_getTicketModel()->rebuildTicketCountCache();
    }

    protected function _ticketHardDeleted()
    {
        if ($this->get('thread_id'))
        {
            $threadDw = XenForo_DataWriter::create('XenForo_DataWriter_Discussion_Thread', XenForo_DataWriter::ERROR_SILENT);
            if ($threadDw->setExistingData($this->get('thread_id')) && $threadDw->get('discussion_type') == 'tickets')
            {
                $threadDw->delete();
            }
        }

        $this->_db->query('
            UPDATE xf_user
            SET nixfifty_ticket_count = IF(nixfifty_ticket_count > 0, nixfifty_ticket_count - 1, 0)
            WHERE user_id = ?
        ', $this->get('user_id'));
    }

    public function getFirstMessageDw()
    {
        if (!$this->_firstMessageDw)
        {
            $this->_firstMessageDw = XenForo_DataWriter::create('NixFifty_Tickets_DataWriter_TicketMessage',
                $this->_errorHandler);
            $this->_firstMessageDw->setOption(NixFifty_Tickets_DataWriter_TicketMessage::OPTION_UPDATE_TICKET, false);
        }

        return $this->_firstMessageDw;
    }

    /**
     * Gets information about the last message in this discussion.
     *
     * @return array|false
     */
    protected function _getLastMessageInDiscussion()
    {
        return $this->_getTicketMessageModel()->getLastMessageInTicket($this->get('ticket_id'));
    }

    /**
     * Gets simple information about all messages in this discussion.
     *
     * @param bool $includeMessage If true, includes the message contents
     * @return array Format: [message id] => info
     */
    protected function _getMessagesInDiscussionSimple($includeMessage = false)
    {
        return $this->_getTicketMessageModel()->getTicketMessagesInTicketSimple($this->get('ticket_id'),
            $includeMessage);
    }

    /**
     * Rebuilds the ticket info.
     *
     * @return bool True if still valid
     */
    public function rebuildTicket()
    {
        $ticketId = $this->get('ticket_id');

        $newCounters = $this->_getTicketMessageModel()->recalculateMessagePositionsInTicket($ticketId);
        if (!$newCounters['firstMessageId'])
        {
            return false;
        }

        $this->rebuildTicketCounters($newCounters['visibleCount'] - 1, $newCounters['firstMessageId'],
            $newCounters['lastMessageId']);

        if ($this->get('ticket_state') == 'closed')
        {
            $this->set('ticket_state', 'resolved');
        }

        if ($this->get('ticket_state') == 'locked')
        {
            $this->set('ticket_state', 'resolved');
            $this->set('ticket_locked', 1);
        }

        return true;
    }

    /**
     * Rebuilds the counters of the ticket.
     *
     * @param int|bool $replyCount     Total reply count, if known
     * @param int|bool $firstMessageId First message ID, if known already
     * @param int|bool $lastMessageId  Last message ID, if known already
     */
    public function rebuildTicketCounters($replyCount = false, $firstMessageId = false, $lastMessageId = false)
    {
        $messageModel = $this->_getTicketMessageModel();
        $ticketId = $this->get('ticket_id');

        if ($firstMessageId && $lastMessageId)
        {
            $messages = $messageModel->getTicketMessagesByIds(
                [$firstMessageId, $lastMessageId],
                ['join' => NixFifty_Tickets_Model_TicketMessage::FETCH_USER]
            );
            $firstMessage = $messages[$firstMessageId];
            $lastMessage = $messages[$lastMessageId];
        }
        else
        {
            $messagesTemp = $messageModel->getTicketMessagesInTicket(
                $ticketId,
                ['join' => NixFifty_Tickets_Model_TicketMessage::FETCH_USER, 'limit' => 1]
            );
            $firstMessage = \reset($messagesTemp);

            $lastMessage = $messageModel->getLastMessageInTicket(
                $ticketId,
                ['join' => NixFifty_Tickets_Model_TicketMessage::FETCH_USER]
            );
        }

        if (!$firstMessage || !$lastMessage)
        {
            return;
        }

        if ($replyCount === false)
        {
            $replyCount = $messageModel->countVisibleMessagesInTicket($ticketId) - 1;
        }

        $this->set('first_message_id', $firstMessage['message_id']);
        $this->set('first_message_date', $firstMessage['message_date']);
        $this->set('start_date', $firstMessage['message_date']);
        $this->set('user_id', $this->get('user_id'));
        $this->set('username', $this->get('username'));
        $this->set('starter_user_id', $firstMessage['user_id']);
        $this->set('starter_username', $firstMessage['username'] !== '' ? $firstMessage['username'] : '-');
        //$this->set('first_message_likes', $firstMessage['likes']);
        $this->set('reply_count', $replyCount);

        $this->set('last_message_id', $lastMessage['message_id']);
        $this->set('last_message_date', $lastMessage['message_date']);
        $this->set('last_message_user_id', $lastMessage['user_id']);
        $this->set('last_message_username', $lastMessage['username']);

        $this->set('last_modified_date', $lastMessage['message_date']);
        $this->set('last_modified_user_id', $lastMessage['user_id']);
        $this->set('last_modified_username', $lastMessage['username']);
    }

    /**
     * Gets the discussion from the update marked with "for update" to ensure that position
     * counters are maintained correctly.
     *
     * @param Zend_Db_Adapter_Abstract $db
     * @param integer                  $id
     * @return array|false Discussion info or false to use what's in the DW already
     */
    public function getTicketForUpdate(Zend_Db_Adapter_Abstract $db, $id)
    {
        return $db->fetchRow('
            SELECT *
            FROM xf_nixfifty_ticket
            WHERE ticket_id = ?
            FOR UPDATE
        ', $id);
    }

    /**
     * @return NixFifty_Tickets_DataWriter_Category|bool
     */
    protected function _getCategoryDwForUpdate()
    {
        $dw = XenForo_DataWriter::create('NixFifty_Tickets_DataWriter_Category', XenForo_DataWriter::ERROR_SILENT);
        if ($dw->setExistingData($this->get('category_id')))
        {
            return $dw;
        }
        else
        {
            return false;
        }
    }

    /**
     * @return NixFifty_Tickets_Model_Ticket
     */
    protected function _getTicketModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Ticket');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketWatch
     */
    protected function _getTicketWatchModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketWatch');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketMessage
     */
    protected function _getTicketMessageModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketMessage');
    }

    /**
     * @return NixFifty_Tickets_Model_NewsFeed
     */
    protected function _getTicketNewsFeedModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_NewsFeed');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketField
     */
    protected function _getFieldModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketField');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketUserWatch
     */
    protected function _getTicketUserWatchModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketUserWatch');
    }

    /**
     * @return NixFifty_Tickets_Model_Category
     */
    protected function _getCategoriesModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Category');
    }

    /**
     * @return NixFifty_Tickets_Model_CategoryWatch
     */
    protected function _getCategoryWatchModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_CategoryWatch');
    }
}
