<?php

use function array_slice, strlen;

class NixFifty_Tickets_DataWriter_TicketMessage extends XenForo_DataWriter
{
    /**
     * Phrase that will be created when data creation fails.
     *
     * @var string
     */
    protected $_existingDataErrorPhrase = 'nf_tickets_requested_ticket_message_not_found';

    const DATA_CATEGORY = 'categoryInfo';

    const OPTION_UPDATE_TICKET    = 'updateTicket';
    const DATA_POST_REPLY_ACTIONS = 'postReplyActions';
    const DATA_MESSAGE_SENDER     = 'sendingUser';

    /**
     * Option that controls whether an IP address should be recorded for this message.
     * Defaults to true.
     *
     * @var string
     */
    const OPTION_SET_IP_ADDRESS = 'setIpAddress';

    /**
     * Holds the temporary hash used to pull attachments and associate them with this message.
     *
     * @var string
     */
    const DATA_ATTACHMENT_HASH = 'attachmentHash';

    /**
     * Holds the reason for soft deletion.
     *
     * @var string
     */
    const DATA_DELETE_REASON = 'deleteReason';

    /**
     * Option for the maximum number of tagged (mentioned) users.
     */
    const OPTION_MAX_TAGGED_USERS = 'maxTaggedUsers';

    /**
     * Option that controls whether data in the parent ticket should be updated,
     * including reply counts and last message info. Defaults to true.
     *
     * @var string
     */
    const OPTION_UPDATE_PARENT_DISCUSSION = 'updateParentDiscussion';

    /**
     * Option that controls whether the data in this message should be indexed for
     * search. If this value is set inconsistently for the same message, data might
     * be orphaned in the search index.
     *
     * @var string
     */
    const OPTION_INDEX_FOR_SEARCH = 'indexForSearch';

    /**
     * Option that controls whether a guest user name is checked to confirm that it
     * doesn't conflict with a valid user. Defaults to true.
     *
     * @var string
     */
    const OPTION_VERIFY_GUEST_USERNAME = 'verifyGuestUsername';

    /**
     * Option that controls whether edits are logged via the edit date, if supported. Defaults to true.
     *
     * @var string
     */
    const OPTION_UPDATE_EDIT_DATE = 'logEditDate';

    /**
     * Option that controls whether the discussion is removed if deleting the first message.
     * Note that if the discussion only has the first message and it's deleted, it will always be removed.
     *
     * @var string
     */
    const OPTION_DELETE_DISCUSSION_FIRST_MESSAGE = 'deleteDiscussionFirstMessage';

    /**
     * Option that controls how long (in seconds) the delay is before an edit log will be added (if supported/enabled).
     *
     * @var string
     */
    const OPTION_EDIT_DATE_DELAY = 'editDateDelay';

    const OPTION_LOG_EDIT = 'logEdit';

    const OPTION_MIN_MESSAGE_LENGTH     = 'minMessageLength';
    const OPTION_MAX_MESSAGE_LENGTH     = 'maxMessageLength';
    const OPTION_ENFORCE_MESSAGE_LENGTH = 'enforceMessageLength';
    const OPTION_SEND_NOTIFICATIONS     = 'sendNotifications';

    /**
     * Maximum number of images allowed in a message.
     *
     * @var string
     */
    const OPTION_MAX_IMAGES = 'maxImages';

    /**
     * Maximum pieces of media allowed in a message.
     *
     * @var string
     */
    const OPTION_MAX_MEDIA = 'maxMedia';

    /**
     * Identifies if a message has a parent discussion item.
     * Must overload {@see getDiscussionDataWriter()} if set to true.
     *
     * @var bool
     */
    protected $_hasParentDiscussion = true;

    /**
     * Reflects important changes to the parent discussion. Current values:
     *    * [empty] - no change of value
     *  * delete - discussion has been deleted
     *  * firstDelete - first message has been deleted, restructure has been carried out.
     *
     * @var string
     */
    protected $_discussionChange = '';

    /**
     * The discussion data writer.
     *
     * @var NixFifty_Tickets_DataWriter_Ticket
     */
    protected $_discussionDw = null;

    /**
     * The insert/update mode of the discussion data writer.
     *
     * @var string insert|update
     */
    protected $_discussionMode = null;

    protected $_taggedUsers = [];

    /**
     * Gets fields defined for this particular table.
     *
     * @return array
     */
    protected function _getFields()
    {
        return [
            'xf_nixfifty_ticket_message' => [
                'message_id'        => ['type' => self::TYPE_UINT, 'autoIncrement' => true],
                'ticket_id'         => ['type' => self::TYPE_UINT, 'required' => true],
                'user_id'           => ['type' => self::TYPE_UINT, 'required' => true],
                'username'          => [
                    'type'          => self::TYPE_STRING, 'required' => true, 'maxLength' => 50,
                    'requiredError' => 'please_enter_valid_name',
                ],
                'user_name'         => ['type' => self::TYPE_STRING, 'default' => '', 'maxLength' => 50],
                'user_email'        => ['type' => self::TYPE_STRING, 'maxLength' => 120],
                'message_date'      => ['type' => self::TYPE_UINT, 'default' => XenForo_Application::$time],
                'message'           => ['type' => self::TYPE_STRING],
                'ip_id'             => ['type' => self::TYPE_UINT, 'default' => 0],
                'likes'             => ['type' => self::TYPE_UINT_FORCED, 'default' => 0],
                'like_users'        => ['type' => self::TYPE_SERIALIZED, 'default' => 'a:0:{}'],
                'message_state'     => [
                    'type'          => self::TYPE_STRING, 'default' => 'visible',
                    'allowedValues' => ['visible', 'hidden', 'deleted'],
                ],
                'position'          => ['type' => self::TYPE_UINT_FORCED, 'default' => 0],
                'attach_count'      => ['type' => self::TYPE_UINT_FORCED, 'default' => 0, 'max' => 65535],
                'last_edit_date'    => ['type' => self::TYPE_UINT, 'default' => 0],
                'last_edit_user_id' => ['type' => self::TYPE_UINT, 'default' => 0],
                'edit_count'        => ['type' => self::TYPE_UINT_FORCED, 'default' => 0],
            ],
        ];
    }

    /**
     * Gets the actual existing data out of data that was passed in.
     *
     * @param mixed
     * @param mixed $data
     * @return array|false
     */
    protected function _getExistingData($data)
    {
        if (!$id = $this->_getExistingPrimaryKey($data))
        {
            return false;
        }

        return ['xf_nixfifty_ticket_message' => $this->_getTicketMessageModel()->getTicketMessageById($id)];
    }

    /**
     * Gets a SQL condition to update the existing record.
     *
     * @param $tableName
     * @return string
     */
    protected function _getUpdateCondition($tableName)
    {
        return 'message_id = ' . $this->_db->quote($this->getExisting('message_id'));
    }

    protected function _getDefaultOptions()
    {
        $options = XenForo_Application::get('options');

        return [
            self::OPTION_MIN_MESSAGE_LENGTH              => $options->nftMinMessageLength,
            self::OPTION_MAX_MESSAGE_LENGTH              => $options->nftMaxMessageLength,
            self::OPTION_ENFORCE_MESSAGE_LENGTH          => true,
            self::OPTION_MAX_IMAGES                      => $options->messageMaxImages,
            self::OPTION_MAX_MEDIA                       => $options->messageMaxMedia,
            self::OPTION_UPDATE_TICKET                   => true,
            self::OPTION_SET_IP_ADDRESS                  => true,
            self::OPTION_INDEX_FOR_SEARCH                => false,
            self::OPTION_UPDATE_PARENT_DISCUSSION        => true,
            self::OPTION_VERIFY_GUEST_USERNAME           => true,
            self::OPTION_DELETE_DISCUSSION_FIRST_MESSAGE => true,
            self::OPTION_UPDATE_EDIT_DATE                => true,
            self::OPTION_LOG_EDIT                        => $options->editHistory['enabled'],
            self::OPTION_EDIT_DATE_DELAY                 => $options->editLogDisplay['enabled'] ? $options->editLogDisplay['delay'] * 60 : -1,
            self::OPTION_MAX_TAGGED_USERS                => 0,
            self::OPTION_SEND_NOTIFICATIONS              => true,
        ];
    }

    protected function _preSave()
    {
        if ($this->isInsert() && !$this->isChanged('message_state'))
        {
            $this->set('message_state', 'visible');
        }
        if ($this->isInsert() && !$this->isChanged('message_date'))
        {
            $this->set('message_date', XenForo_Application::$time);
        }

        if ($this->isChanged('message'))
        {
            $this->_checkMessageValidity();
        }

        if ($this->_hasParentDiscussion)
        {
            $this->_checkFirstMessageState();

            if (!$this->isChanged('position'))
            {
                $this->_setPosition();
            }
        }

        if ($this->isUpdate() && $this->isChanged('message'))
        {
            if ($this->getOption(self::OPTION_UPDATE_EDIT_DATE)
                && !$this->isChanged('last_edit_date')
                && $this->getOption(self::OPTION_EDIT_DATE_DELAY) >= 0
            )
            {
                if ($this->get('date_added') + $this->getOption(self::OPTION_EDIT_DATE_DELAY) <= XenForo_Application::$time)
                {
                    $this->set('last_edit_date', XenForo_Application::$time);
                    if (!$this->isChanged('last_edit_user_id'))
                    {
                        $this->set('last_edit_user_id', XenForo_Visitor::getUserId());
                    }
                }
            }

            if (!$this->isChanged('edit_count'))
            {
                $this->set('edit_count', $this->get('edit_count') + 1);
            }
        }
        if ($this->isChanged('edit_count') && $this->get('edit_count') == 0)
        {
            $this->set('last_edit_date', 0);
        }
        if (!$this->get('last_edit_date'))
        {
            $this->set('last_edit_user_id', 0);
        }

        if ($this->isChanged('message'))
        {
            /** @var $taggingModel XenForo_Model_UserTagging */
            $taggingModel = $this->getModelFromCache('XenForo_Model_UserTagging');

            $this->_taggedUsers = $taggingModel->getTaggedUsersInMessage(
                $this->get('message'), $newMessage, 'bb'
            );
            $this->set('message', $newMessage);
        }
    }

    /**
     * Check that the contents of the message are valid, based on length, images, etc.
     */
    protected function _checkMessageValidity()
    {
        $message = $this->get('message');
        $category = $this->_getCategoryInfo();

        if (!$this->getError('message'))
        {
            $formatter = XenForo_BbCode_Formatter_Base::create('XenForo_BbCode_Formatter_Base');
            $parser = XenForo_BbCode_Parser::create($formatter);

            $output = $parser->render($this->get('message'));
            if (!strlen(\trim($output)) && !$category['empty_message_allowed'])
            {
                $this->error(new XenForo_Phrase('please_enter_valid_message'), 'message');
            }
        }

        $messageLength = \utf8_strlen($this->get('message'));
        $minMessageLength = $this->getOption(self::OPTION_MIN_MESSAGE_LENGTH);
        $maxMessageLength = $this->getOption(self::OPTION_MAX_MESSAGE_LENGTH);

        if ($messageLength < $minMessageLength && $this->getOption(self::OPTION_ENFORCE_MESSAGE_LENGTH))
        {
            $underAmount = $minMessageLength - $messageLength;
            $this->error(new XenForo_Phrase('nf_tickets_please_enter_message_with_at_least_x_characters',
                ['count' => $minMessageLength, 'under' => $underAmount]), 'message'
            );
        }

        if ($messageLength > $maxMessageLength && $this->getOption(self::OPTION_ENFORCE_MESSAGE_LENGTH))
        {
            $overAmount = $messageLength - $maxMessageLength;
            $this->error(new XenForo_Phrase('nf_tickets_please_enter_message_with_no_more_than_x_characters',
                ['count' => $maxMessageLength, 'over' => $overAmount]), 'message'
            );
        }
        else
        {
            $maxImages = $this->getOption(self::OPTION_MAX_IMAGES);
            $maxMedia = $this->getOption(self::OPTION_MAX_MEDIA);
            if ($maxImages || $maxMedia)
            {
                $formatter = XenForo_BbCode_Formatter_Base::create('ImageCount', false);
                $parser = XenForo_BbCode_Parser::create($formatter);
                $parser->render($message);

                if ($maxImages && $formatter->getImageCount() > $maxImages)
                {
                    $this->error(new XenForo_Phrase('please_enter_message_with_no_more_than_x_images', ['count' => $maxImages]), 'message');
                }
                if ($maxMedia && $formatter->getMediaCount() > $maxMedia)
                {
                    $this->error(new XenForo_Phrase('please_enter_message_with_no_more_than_x_media', ['count' => $maxMedia]), 'message');
                }
            }
        }
    }

    /**
     * Sets the message's position within the discussion.
     */
    protected function _setPosition()
    {
        if ($this->isInsert() || $this->isChanged('message_state'))
        {
            $discussionDw = $this->getDiscussionDataWriter();
            if ($this->isInsert())
            {
                if (!$discussionDw || $discussionDw->isInsert())
                {
                    // likely the first message
                    $this->set('position', 0);
                }
                else
                {
                    /*if ($this->get('message_date') < $discussionDw->get('last_message_date'))
                    {
                        throw new XenForo_Exception('Cannot insert a message in the middle of a discussion.');
                    }*/

                    $discussion = $discussionDw->getTicketForUpdate($this->_db, $this->get('ticket_id'));
                    $replyCount = ($discussion && isset($discussion['reply_count'])
                        ? $discussion['reply_count']
                        : $discussionDw->get('reply_count')
                    );

                    if ($this->get('message_state') == 'visible')
                    {
                        $this->set('position', $replyCount + 1);
                    }
                    else
                    {
                        $this->set('position', $replyCount);
                    }
                }
            }
            else
            {
                // updated the state on an existing message -- need to slot in
                if ($this->get('message_state') == 'visible' && $this->getExisting('message_state') != 'visible')
                {
                    $this->set('position', $this->get('position') + 1);
                }
                else if ($this->get('message_state') != 'visible' && $this->getExisting('message_state') == 'visible')
                {
                    $this->set('position', $this->get('position') - 1);
                }
            }
        }
    }

    /**
     * Checks that the first message has a valid state. Normally,
     * the first message must always be visible. If a non-visible
     * first message is required, change the discussion.
     */
    protected function _checkFirstMessageState()
    {
        if ($this->get('message_state') != 'visible')
        {
            if ($this->isUpdate() && $this->get('position') > 0)
            {
                // position > 0 means this isn't the first post
                return;
            }

            $discussionDw = $this->getDiscussionDataWriter();
            if (!$discussionDw)
            {
                // debugging message, no need for phrasing
                $this->error('The first message of a discussion must be visible.');
            }
            else if ($discussionDw->isInsert() || $discussionDw->get('first_message_id') == $this->get('message_id'))
            {
                // treat first message state as the discussion state
                $discussionDw->set('ticket_state', 'open');
                $this->set('message_state', 'visible');

                // in case we're deleting the discussion, push the reason up
                $discussionDw->setExtraData(NixFifty_Tickets_DataWriter_Ticket::DATA_DELETE_REASON,
                    $this->getExtraData(self::DATA_DELETE_REASON)
                );
            }
        }
    }

    protected function _postSave()
    {
        if ($this->isUpdate() && $this->get('message_state') == 'deleted' && $this->getExisting('message_state') == 'visible')
        {
            $this->getModelFromCache('XenForo_Model_Alert')->deleteAlerts('ticket_message', $this->get('message_id'));
        }

        if ($this->isInsert() && $this->getOption(self::OPTION_SET_IP_ADDRESS) && !$this->get('ip_id'))
        {
            $this->_updateIpData();
        }

        if ($this->isInsert())
        {
            $this->getModelFromCache('NixFifty_Tickets_Model_NewsFeed')->publishItem($this->getMergedData(), 'nf_tickets_message');
        }

        $attachmentHash = $this->getExtraData(self::DATA_ATTACHMENT_HASH);
        if ($attachmentHash)
        {
            $this->_associateAttachments($attachmentHash);
        }

        if ($this->_hasParentDiscussion && $this->getOption(self::OPTION_UPDATE_PARENT_DISCUSSION))
        {
            $this->_updateDiscussionPostSave();
        }

        if ($this->_hasParentDiscussion && $this->isUpdate() && $this->isChanged('message_state'))
        {
            $this->_updateMessagePositionList();
        }

        if ($this->get('user_id') && $this->isChanged('message_state'))
        {
            if ($this->get('likes'))
            {
                $this->_updateUserLikeCount();
            }
        }

        if ($this->getOption(self::OPTION_LOG_EDIT)
            && $this->isUpdate() && $this->isChanged('edit_count') && $this->get('edit_count')
        )
        {
            $this->_insertEditHistory();
        }

        if ($this->isUpdate() && $this->isChanged('message'))
        {
            /** @var XenForo_Model_BbCode $bbCodeModel */
            $bbCodeModel = $this->getModelFromCache('XenForo_Model_BbCode');
            $bbCodeModel->deleteBbCodeParseCacheForContent('nf_tickets_message', $this->get('message_id'));
        }

        $this->_indexForSearch();

        $this->_saveDiscussionDataWriter();
    }

    /**
     * Updates the IP data.
     */
    protected function _updateIpData()
    {
        if (!empty($this->_extraData['ipAddress']))
        {
            $ipAddress = $this->_extraData['ipAddress'];
        }
        else
        {
            $ipAddress = null;
        }

        $ipId = XenForo_Model_Ip::log(
            $this->get('user_id'), 'nf_tickets_message', $this->get('message_id'), 'insert', $ipAddress
        );
        $this->set('ip_id', $ipId, '', ['setAfterPreSave' => true]);

        $this->_db->update('xf_nixfifty_ticket_message', [
            'ip_id' => $ipId,
        ], 'message_id' . ' = ' . $this->_db->quote($this->get('message_id')));
    }

    /**
     * Associates attachments with this message.
     *
     * @param string $attachmentHash
     */
    protected function _associateAttachments($attachmentHash)
    {
        $rows = $this->_db->update('xf_attachment', [
            'content_type' => 'nf_tickets_message',
            'content_id'   => $this->get('message_id'),
            'temp_hash'    => '',
            'unassociated' => 0,
        ], 'temp_hash = ' . $this->_db->quote($attachmentHash));
        if ($rows)
        {
            $this->set('attach_count', $this->get('attach_count') + $rows, '', ['setAfterPreSave' => true]);

            $this->_db->update('xf_nixfifty_ticket_message', [
                'attach_count' => $this->get('attach_count'),
            ], 'message_id = ' . $this->_db->quote($this->get('message_id')));
        }
    }

    /**
     * Updates the search index for this message.
     */
    protected function _indexForSearch()
    {
        if ($this->get('message_state') == 'visible')
        {
            if ($this->getExisting('message_state') != 'visible' || $this->isChanged('message'))
            {
                $this->_insertOrUpdateSearchIndex();
            }
        }
        else if ($this->isUpdate() && $this->get('message_state') != 'visible' && $this->getExisting('message_state') == 'visible')
        {
            $this->_deleteFromSearchIndex();
        }
    }

    /**
     * Inserts or updates a record in the search index for this message.
     */
    protected function _insertOrUpdateSearchIndex()
    {
        $dataHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_TicketMessage');
        if (!$dataHandler)
        {
            return;
        }

        $indexer = new XenForo_Search_Indexer();
        $dataHandler->insertIntoIndex($indexer, $this->getMergedData(), $this->getDiscussionData());
    }

    /**
     * Inserts a record for the edit history.
     */
    protected function _insertEditHistory()
    {
        $historyDw = XenForo_DataWriter::create('XenForo_DataWriter_EditHistory', XenForo_DataWriter::ERROR_SILENT);
        $historyDw->bulkSet([
            'content_type' => 'nf_tickets_message',
            'content_id'   => $this->get('message_id'),
            'edit_user_id' => XenForo_Visitor::getUserId(),
            'old_text'     => $this->getExisting('message'),
        ]);
        $historyDw->save();
    }

    protected function _postSaveAfterTransaction()
    {
        $message = $this->getMergedData();
        $ticket = $this->getDiscussionData();
        $category = $this->_getCategoryInfo();

        $noEmail = [];

        if ($this->getOption(self::OPTION_SEND_NOTIFICATIONS) && $this->get('message_state') == 'visible')
        {
            if ($this->isInsert() || $this->getExisting('message_state') == 'moderated')
            {
                if ($message['position'] == 0)
                {
                    $noEmail[] = $ticket['user_id'];
                }

                $alertedMembers = $this->_getTicketMessageModel()->alertQuotedMembers($message, $ticket, $category);

                $maxTagged = $this->getOption(self::OPTION_MAX_TAGGED_USERS);
                if ($maxTagged && $this->_taggedUsers)
                {
                    if ($maxTagged > 0)
                    {
                        $alertTagged = array_slice($this->_taggedUsers, 0, $maxTagged, true);
                    }
                    else
                    {
                        $alertTagged = $this->_taggedUsers;
                    }

                    $alertedMembers = \array_merge(
                        $alertedMembers,
                        $this->_getTicketMessageModel()->alertMentionedUsers($message, $ticket, $category, $alertTagged, $alertedMembers)
                    );
                }

                $this->_getTicketWatchModel()->sendNotificationToWatchUsersOnComment($message, $alertedMembers, $noEmail);
            }
        }

        /** @var NixFifty_Tickets_Model_Escalation $escalationModel */
        $escalationModel = XenForo_Model::create('NixFifty_Tickets_Model_Escalation');
        $escalationModel->quickEscalate($ticket, $category);
    }

    /**
     * Updates the discussion container info.
     */
    protected function _updateDiscussionPostSave()
    {
        $discussionDw = $this->getDiscussionDataWriter();
        if ($discussionDw && $discussionDw->isUpdate())
        {
            $discussionDw->updateCountersAfterMessageSave($this);
        }
    }

    /**
     * Updates the position list based on state changes.
     */
    protected function _updateMessagePositionList()
    {
        if ($this->get('message_state') == 'visible' && $this->getExisting('message_state') != 'visible')
        {
            $this->_adjustPositionListForInsert();
        }
        else if ($this->get('message_state') != 'visible' && $this->getExisting('message_state') == 'visible')
        {
            $this->_adjustPositionListForRemoval();
        }
    }

    protected function _postDelete()
    {
        if ($this->_hasParentDiscussion && $this->getOption(self::OPTION_UPDATE_PARENT_DISCUSSION))
        {
            $this->_updateDiscussionPostDelete();
        }

        // firstDelete would trigger this
        if (!$this->discussionDeleted())
        {
            if ($this->_hasParentDiscussion)
            {
                $this->_adjustPositionListForRemoval();
            }
        }

        $this->getModelFromCache('XenForo_Model_DeletionLog')->removeDeletionLog(
            'ticket_message', $this->get('message_id')
        );
        $this->getModelFromCache('XenForo_Model_ModerationQueue')->deleteFromModerationQueue(
            'ticket_message', $this->get('message_id')
        );
        $this->getModelFromCache('XenForo_Model_EditHistory')->deleteEditHistoryForContent(
            'ticket_message', $this->get('message_id')
        );
        $this->getModelFromCache('XenForo_Model_BbCode')->deleteBbCodeParseCacheForContent(
            'ticket_message', $this->get('message_id')
        );

        if ($this->get('attach_count'))
        {
            $this->_deleteAttachments();
        }

        if ($this->getOption(self::OPTION_INDEX_FOR_SEARCH))
        {
            $this->_deleteFromSearchIndex();
        }

        if ($this->get('likes'))
        {
            $this->_deleteLikes();
        }

        if ($this->get('message_state') == 'visible')
        {
            $this->getModelFromCache('XenForo_Model_Alert')->deleteAlerts(
                'ticket_message', $this->get('message_id')
            );
        }
    }

    /**
     * Updates data in the discussion after the message is deleted.
     * This may cause the entire discussion to be deleted.
     */
    protected function _updateDiscussionPostDelete()
    {
        $discussionDw = $this->getDiscussionDataWriter();

        if ($discussionDw)
        {
            $deleteIfFirst = $this->getOption(self::OPTION_DELETE_DISCUSSION_FIRST_MESSAGE);
            $discussionChange = $discussionDw->updateCountersAfterMessageDelete($this, $deleteIfFirst);
            $this->_discussionChange = $discussionChange;

            if ($discussionChange == 'delete')
            {
                $discussionDw->delete();
            }
        }
    }

    /**
     * Deletes all attachments associated with this message.
     */
    protected function _deleteAttachments()
    {
        $this->getModelFromCache('XenForo_Model_Attachment')->deleteAttachmentsFromContentIds(
            'ticket_message',
            [$this->get('message_id')]
        );
    }

    /**
     * Updates the user like count.
     *
     * @param bool $isDelete True if hard deleting the message
     */
    protected function _updateUserLikeCount($isDelete = false)
    {
        $likes = $this->get('likes');
        if (!$likes)
        {
            return;
        }

        if ($this->_hasParentDiscussion)
        {
            $discussionDw = $this->getDiscussionDataWriter();
            if ($discussionDw && $discussionDw->get('discussion_state') != 'visible')
            {
                return;
            }
        }

        if ($this->getExisting('message_state') == 'visible'
            && ($this->get('message_state') != 'visible' || $isDelete)
        )
        {
            $this->_db->query('
                UPDATE xf_user
                SET like_count = IF(like_count > ?, like_count - ?, 0)
                WHERE user_id = ?
            ', [$likes, $likes, $this->get('user_id')]);
        }
        else if ($this->get('message_state') == 'visible' && $this->getExisting('message_state') != 'visible')
        {
            $this->_db->query('
                UPDATE xf_user
                SET like_count = like_count + ?
                WHERE user_id = ?
            ', [$likes, $this->get('user_id')]);
        }
    }

    /**
     * Deletes this message from the search index.
     */
    protected function _deleteFromSearchIndex()
    {
        /** @var NixFifty_Tickets_Search_DataHandler_TicketMessage $dataHandler */
        $dataHandler = XenForo_Search_DataHandler_Abstract::create('NixFifty_Tickets_Search_DataHandler_TicketMessage');
        if (!$dataHandler)
        {
            return;
        }

        $indexer = new XenForo_Search_Indexer();
        $dataHandler->deleteFromIndex($indexer, $this->getMergedData());
    }

    /**
     * Delete all like entries for content.
     */
    protected function _deleteLikes()
    {
        $updateUserLikeCounter = ($this->get('message_state') == 'visible');

        /*if ($updateUserLikeCounter && $this->_hasParentDiscussion)
        {
            $discussionDw = $this->getDiscussionDataWriter();
            if ($discussionDw && $discussionDw->get('discussion_state') != 'visible')
            {
                $updateUserLikeCounter = false;
            }
        }*/

        $this->getModelFromCache('XenForo_Model_Like')->deleteContentLikes(
            'ticket_message', $this->get('message_id'), $updateUserLikeCounter
        );
    }

    /**
     * Adjust the position list surrounding this message, when this message
     * has been put from a position that "counts" (removed or hidden).
     */
    protected function _adjustPositionListForInsert()
    {
        if ($this->get('message_state') != 'visible')
        {
            // only renumber if becoming visible
            return;
        }

        $containerKey = 'ticket_id';
        $containerCondition = "$containerKey = " . $this->get($containerKey);

        $positionQuoted = $this->_db->quote($this->getExisting('position'));
        $postDateQuoted = $this->_db->quote($this->get('message_date'));
        $messageKeyCondition = 'message_id <> ' . $this->_db->quote($this->get('message_id'));

        $this->_db->query("
            UPDATE xf_nixfifty_ticket_message
            SET position = position + 1
            WHERE $containerCondition
                AND (position > $positionQuoted
                    OR (position = $positionQuoted AND message_date > $postDateQuoted)
                )
                AND $messageKeyCondition
        ");
    }

    /**
     * Adjust the position list surrounding this message, when this message
     * has been removed from a position that "counts" (removed or hidden).
     */
    protected function _adjustPositionListForRemoval()
    {
        if ($this->getExisting('message_state') != 'visible')
        {
            // no need to renumber after removal something that didn't count
            return;
        }

        $containerKey = 'ticket_id';
        $containerCondition = "$containerKey = " . $this->get($containerKey);

        $messageKeyCondition = 'message_id <> ' . $this->_db->quote($this->get('message_id'));

        $this->_db->query('
            UPDATE xf_nixfifty_ticket_message
            SET position = IF(position > 0, position - 1, 0)
            WHERE ' . $containerCondition . '
                AND position >= ?
                AND ' . $messageKeyCondition . '
        ', $this->getExisting('position'));
    }

    /**
     * Gets the discussion data writer. Note that if the container value changes,
     * this cache will not be removed.
     *
     * @return NixFifty_Tickets_DataWriter_Ticket|false
     */
    public function getDiscussionDataWriter()
    {
        if (!$this->_hasParentDiscussion)
        {
            return false;
        }

        if ($this->_discussionDw === null)
        {
            $containerId = $this->get('ticket_id');

            if (!$containerId)
            {
                $this->_discussionDw = false;
            }
            else
            {
                $this->_discussionDw = XenForo_DataWriter::create('NixFifty_Tickets_DataWriter_Ticket',
                    $this->_errorHandler);
                $this->_discussionDw->setExistingData($containerId);

                if ($this->_discussionDw && $this->_discussionMode === null)
                {
                    $this->_discussionMode = 'update';
                }
            }
        }

        return $this->_discussionDw;
    }

    /**
     * Sets the data writer for the discussion this message is in--or will be in.
     *
     * @param XenForo_DataWriter_Discussion|null $discussionDw
     * @param bool True if                       $discussionDataWriter->isInsert()
     * @param null|mixed                         $isInsert
     */
    public function setDiscussionDataWriter(XenForo_DataWriter_Discussion $discussionDw = null, $isInsert = null)
    {
        $this->_discussionDw = $discussionDw;

        if ($isInsert !== null)
        {
            $this->_discussionMode = ($isInsert ? 'insert' : 'update');
        }
    }

    /**
     * Saves the discussion data writer if it exists and has changed.
     */
    protected function _saveDiscussionDataWriter()
    {
        if ($this->_discussionDw && $this->_discussionDw->hasChanges())
        {
            $this->_discussionDw->save();
        }
    }

    /**
     * Gets the data about the discussion this message is in. This may use the
     * discussion data writer, or some other source if needed.
     *
     * @return array|null
     */
    public function getDiscussionData()
    {
        if ($this->_hasParentDiscussion)
        {
            $discussionDw = $this->getDiscussionDataWriter();
            if ($discussionDw)
            {
                return $discussionDw->getMergedData();
            }
        }

        return null;
    }

    /**
     * Returns true if the parent discussion has been deleted (instead of this).
     *
     * @return bool
     */
    public function discussionDeleted()
    {
        return ($this->_discussionChange == 'delete');
    }

    /**
     * Gets the current value of the ticket message ID for this message.
     *
     * @return int
     */
    public function getMessageId()
    {
        return $this->get('ticket_message_id');
    }

    protected function _getCategoryInfo()
    {
        if (!$category = $this->getExtraData(self::DATA_CATEGORY))
        {
            $category = $this->_getCategoryModel()->getCategoryByTicketId($this->get('ticket_id'));

            $this->setExtraData(self::DATA_CATEGORY, $category ? $category : []);
        }

        return $this->getExtraData(self::DATA_CATEGORY);
    }

    /**
     * @return NixFifty_Tickets_Model_Ticket
     */
    protected function _getTicketModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Ticket');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketMessage
     */
    protected function _getTicketMessageModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketMessage');
    }

    /**
     * @return NixFifty_Tickets_Model_Category
     */
    protected function _getCategoryModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_Category');
    }

    /**
     * @return NixFifty_Tickets_Model_TicketWatch
     */
    protected function _getTicketWatchModel()
    {
        return $this->getModelFromCache('NixFifty_Tickets_Model_TicketWatch');
    }
}
