<?php

namespace NF\Tickets;

use NF\Tickets\XF\Entity\User;
use XF\Container;

/**
 * Class Listener
 *
 * @package NF\Tickets
 */
class Listener
{
    /**
     * @param \XF\App $app
     *
     * @throws \XF\Db\Exception
     */
    public static function appSetup(\XF\App $app)
    {
        $container = $app->container();

        $container['customFields.tickets'] = $app->fromRegistry('nfTicketFields',
            function(Container $c)
            {
                return $c['em']->getRepository('NF\Tickets:TicketField')->rebuildFieldCache();
            },
            function(array $fields)
            {
                return new \XF\CustomField\DefinitionSet($fields);
            }
        );

        $container['prefixes.nf_tickets_ticket'] = $app->fromRegistry('nf_ticketPrefixes',
            function(Container $c) { return $c['em']->getRepository('NF\Tickets:TicketPrefix')->rebuildPrefixCache(); }
        );

        $container['nfTicketCounts'] = $app->fromRegistry('nfTicketCounts',
            function(Container $c) { return $c['em']->getRepository('NF\Tickets:Ticket')->rebuildTicketCounts(); }
        );

        $container['nfTicketsEmail'] = function($c)
        {
            $class = \XF::app()->extendClass('NF\Tickets\SubContainer\TicketEmail');
            return new $class($c, \XF::app());
        };
    }

    public static function appPubStartEnd(\XF\App $app)
    {
        // no-op, left here for upgrades from pre v2.3.2
    }

    public static function criteriaPage($rule, array $data, \XF\Entity\User $user, array $params, &$returnValue)
    {
        if (isset($params['breadcrumbs']) && \is_array($params['breadcrumbs']) && !empty($data['ticket_category_ids']) && isset($params['nfTicketsCategory']))
        {
            $returnValue = false;

            if (empty($data['tickets_category_only']))
            {
                foreach ($params['breadcrumbs'] AS $i => $navItem)
                {
                    if (isset($navItem['attributes']['ticket_category_id']) && \in_array($navItem['attributes']['ticket_category_id'], $data['ticket_category_ids']))
                    {
                        $returnValue = true;
                    }
                }
            }

            if ($params['containerKey'])
            {
                list ($type, $id) = \explode('-', $params['containerKey'], 2);

                if ($type === 'nfTicketsCategory' && $id && \in_array($id, $data['ticket_category_ids']))
                {
                    $returnValue = true;
                }
            }
        }
    }

    /**
     * Adds categories, statuses and priorities to the criteria helper template parameters.
     *
     * @param array $templateData
     */
    public static function criteriaTemplateData(array &$templateData)
    {
        /** @var \NF\Tickets\Repository\Category $categoryRepo */
        $categoryRepo = \XF::repository('NF\Tickets:Category');
        $templateData['nfTicketCategories'] = $categoryRepo->getCategoryOptionsData(false);

        /** @var \NF\Tickets\Repository\Status $statusRepo */
        $statusRepo = \XF::repository('NF\Tickets:Status');
        $templateData['nfTicketStatuses'] = $statusRepo->findStatusesForList(false)->fetch();

        /** @var \NF\Tickets\Repository\Priority $priorityRepo */
        $priorityRepo = \XF::repository('NF\Tickets:Priority');
        $templateData['nfTicketPriorities'] = $priorityRepo->findPrioritiesForList(false)->fetch();
    }

    public static function userContentChangeInit(\XF\Service\User\ContentChange $changeService, array &$updates)
    {
        $updates['xf_nf_tickets_ticket'] = [
            ['user_id', 'username'],
            ['last_message_user_id', 'last_message_username'],
            ['last_modified_user_id', 'last_modified_username']
        ];
        $updates['xf_nf_tickets_message'] = ['user_id', 'username'];
        $updates['xf_nf_tickets_ticket_read'] = ['user_id', 'emptyable' => false];
        $updates['xf_nf_tickets_category_read'] = ['user_id', 'emptyable' => false];
        $updates['xf_nf_tickets_category_watch'] = ['user_id', 'emptyable' => false];
        $updates['xf_nf_tickets_user_watch'] = ['user_id', 'emptyable' => false];
    }

    public static function userDeleteCleanInit(\XF\Service\User\DeleteCleanUp $deleteService, array &$deletes)
    {
        $deletes['xf_nf_tickets_ticket_read'] = 'user_id = ?';
        $deletes['xf_nf_tickets_ticket_watch'] = 'user_id = ?';
        $deletes['xf_nf_tickets_user_watch'] = 'user_id = ?';
        $deletes['xf_nf_tickets_category_read'] = 'user_id = ?';
        $deletes['xf_nf_tickets_category_watch'] = 'user_id = ?';
    }

    public static function userMergeCombine(
        \XF\Entity\User $target, \XF\Entity\User $source, \XF\Service\User\Merge $mergeService
    )
    {
        /** @var User $target */
        /** @var User $source */
        $target->nf_tickets_count += $source->nf_tickets_count;
    }

    public static function importImporterClasses(\XF\SubContainer\Import $container, Container $parentContainer, array &$importers)
    {
        $importers = \array_merge(
            $importers, \XF\Import\Manager::getImporterShortNamesForType('NF/Tickets')
        );
    }

	public static function addOnPostRebuild(
		\XF\AddOn\AddOn $addOn, \XF\Entity\AddOn $installedAddOn, array $json
	): void
	{
		if (empty(Setup::$supportedAddOns[$addOn->getAddOnId()]))
		{
			return;
		}

		$setup = new Setup($addOn, \XF::app());
		$setup->applySchemaUpdates();
	}

	public static function addOnPostInstall(
		\XF\AddOn\AddOn $addOn, \XF\Entity\AddOn $installedAddOn, array $json, array &$stateChanges
	): void
	{
		if (empty(Setup::$supportedAddOns[$addOn->getAddOnId()]))
		{
			return;
		}

		$setup = new Setup($addOn, \XF::app());
		$setup->applySchemaUpdates();
	}

    /**
     * @return \NF\Tickets\XF\Entity\User
     */
    public static function visitor()
    {
        /** @var \NF\Tickets\XF\Entity\User $visitor */
        $visitor = \XF::visitor();
        return $visitor;
    }
}