<?php

namespace NF\Tickets\Notifier\Message;

use NF\Tickets\Mail\Mail;
use XF\Notifier\AbstractNotifier;

use function in_array;

/**
 * Class AbstractWatch
 *
 * @package NF\Tickets\Notifier\Message
 */
abstract class AbstractWatch extends AbstractNotifier
{
    /**
     * @var \NF\Tickets\Entity\Message
     */
    protected $message;

    protected $actionType;
    protected $isApplicable;

    protected $userReadDates    = [];
    protected $previousMessages = null;

    /**
     * @return array
     */
    abstract protected function getDefaultWatchNotifyData();

    /**
     * @return array
     */
    abstract protected function getApplicableActionTypes();

    /**
     * @return string
     */
    abstract protected function getWatchEmailTemplateName();

    /**
     * AbstractWatch constructor.
     *
     * @param \XF\App                    $app
     * @param \NF\Tickets\Entity\Message $message
     * @param                            $actionType
     */
    public function __construct(\XF\App $app, \NF\Tickets\Entity\Message $message, $actionType)
    {
        parent::__construct($app);

        $this->message = $message;
        $this->actionType = $actionType;
        $this->isApplicable = $this->isApplicable();
    }

    /**
     * @return bool
     */
    protected function isApplicable()
    {
        if (!in_array($this->actionType, $this->getApplicableActionTypes(), true))
        {
            return false;
        }

        if (!$this->message->isVisible())
        {
            return false;
        }

        return true;
    }

    /**
     * @param \XF\Entity\User $user
     * @return bool
     */
    public function canNotify(\XF\Entity\User $user): bool
    {
        if (!$this->isApplicable)
        {
            return false;
        }

        $message = $this->message;

        if ($user->user_id === $message->user_id)
        {
            return false;
        }

        return true;
    }

    /**
     * @param \XF\Entity\User $user
     * @return bool
     */
    public function sendAlert(\XF\Entity\User $user)
    {
        return \XF::asVisitor($user, function () use ($user) {
            $message = $this->message;
            $fromUser = $message->getAnonymousUser(true);

            return $this->basicAlert(
                $user,
                $fromUser->user_id,
                $fromUser->username,
                'nf_tickets_message',
                $message->message_id,
                'insert',
                ['depends_on_addon_id' => 'NF/Tickets']
            );
        });
    }

    /**
     * @param \XF\Entity\User $user
     * @return bool
     */
    public function sendEmail(\XF\Entity\User $user)
    {
        if (!$user->email || $user->user_state !== 'valid')
        {
            return false;
        }

        $message = $this->message;

        $params = [
            'message'  => $message,
            'ticket'   => $message->Ticket,
            'category' => $message->Ticket->Category,
            'receiver' => $user,
        ];

        $template = $this->getWatchEmailTemplateName();

        $mailer = $this->app()->mailer();
        $mailer->setMailClass('NF\Tickets\Mail\Mail');
        /** @var Mail $mail */
        $mail = $mailer->newMail();
        $mail->setTicket($this->message->Ticket)
             ->setToUser($user)
             ->setTemplate($template, $params)
             ->queue();

        return true;
    }

    /**
     * @return array
     */
    public function getDefaultNotifyData()
    {
        if (!$this->isApplicable)
        {
            return [];
        }

        return $this->getDefaultWatchNotifyData();
    }

    /**
     * @param array $userIds
     * @return \XF\Mvc\Entity\Entity[]
     */
    public function getUserData(array $userIds)
    {
        $users = parent::getUserData($userIds);
        $this->userReadDates = $this->getUserReadDates($userIds);

        return $users;
    }

    /**
     * @param array $userIds
     * @return array
     */
    protected function getUserReadDates(array $userIds)
    {
        if (!$userIds)
        {
            return [];
        }

        $autoReadDate = \XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400;
        $message = $this->message;

        $db = $this->app()->db();
        $readDates = $db->fetchPairs('
            SELECT user.user_id,
                GREATEST(
                    COALESCE(ticket_read.ticket_read_date, 0),
                    COALESCE(category_read.category_read_date, 0),
                    ?
                )
            FROM xf_user AS user
            LEFT JOIN xf_nf_tickets_ticket_read AS ticket_read ON
                (ticket_read.user_id = user.user_id AND ticket_read.ticket_id = ?)
            LEFT JOIN xf_nf_tickets_category_read AS category_read ON
                (category_read.user_id = user.user_id AND category_read.ticket_category_id = ?)
            WHERE user.user_id IN (' . $db->quote($userIds) . ')
        ', [$autoReadDate, $message->ticket_id, $message->Ticket->ticket_category_id]);

        foreach ($userIds AS $userId)
        {
            if (!isset($readDates[$userId]))
            {
                $readDates[$userId] = $autoReadDate;
            }
        }

        return $readDates;
    }

    /**
     * @return null|\XF\Mvc\Entity\AbstractCollection
     */
    protected function getPreviousMessages()
    {
        if ($this->previousMessages === null)
        {
            $autoReadDate = \XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400;

            $finder = $this->app()->finder('NF\Tickets:Message')
                           ->where('ticket_id', $this->message->ticket_id)
                           ->where('message_state', 'visible')
                           ->where('message_date', '<', $this->message->message_date)
                           ->where('message_date', '>=', $autoReadDate)
                           ->order('message_date', 'desc');

            $this->previousMessages = $finder->fetch(15);
        }

        return $this->previousMessages;
    }
}