<?php

namespace NF\Tickets\Pub\Controller;

use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\AbstractReply;
use XF\Mvc\Reply\View;

/**
 * Class Category
 *
 * @package NF\Tickets\Pub\Controller
 */
class Category extends AbstractTicketController
{
    /** @noinspection PhpParameterByRefIsNotUsedAsReferenceInspection */
    protected function postDispatchController($action, ParameterBag $params, AbstractReply &$reply)
	{
		if ($reply instanceof View)
		{
			$viewParams = $reply->getParams();
			$category = null;

			if (isset($viewParams['category']))
			{
				$category = $viewParams['category'];
			}
			if (isset($viewParams['ticket']))
			{
				$category = $viewParams['ticket']->Category;
			}
			if ($category)
			{
				$reply->setContainerKey('nfTicketsCategory-' . $category->ticket_category_id);
			}
		}
	}

	public function actionIndex(ParameterBag $params)
	{
		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($params->ticket_category_id, $this->getCategoryViewExtraWith());

		/** @var \NF\Tickets\ControllerPlugin\TicketList $ticketListPlugin */
		$ticketListPlugin = $this->plugin('NF\Tickets:TicketList');

		$categoryParams = $ticketListPlugin->getCategoryListData();

		/** @var \XF\Tree $categoryTree */
		$categoryTree = $categoryParams['categoryTree'];
		$descendants = $categoryTree->getDescendants($category->ticket_category_id);

		$sourceCategoryIds = \array_keys($descendants);
		$sourceCategoryIds [] = $category->ticket_category_id;

		$category->cacheViewableDescendants($descendants);

		$listParams = $ticketListPlugin->getCoreListData($sourceCategoryIds, $category);

		$this->assertValidPage(
			$listParams['page'],
			$listParams['perPage'],
			$listParams['total'],
			'tickets/categories',
			$category
		);
		$this->assertCanonicalUrl($this->buildLink(
			'tickets/categories',
			$category,
			['page' => $listParams['page']]
		));

		$viewParams = [
			'category' => $category,
		];
		$viewParams += $categoryParams + $listParams;

		return $this->view('NF\Tickets:Category\View', 'nf_tickets_category_view', $viewParams);
	}

	protected function getCategoryViewExtraWith()
	{
		$extraWith = [];
		$userId = \XF::visitor()->user_id;
		if ($userId)
		{
			$extraWith[] = 'Watch|' . $userId;
		}

		return $extraWith;
	}

	public function actionFilters(ParameterBag $params)
	{
		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($params->ticket_category_id);

		/** @var \NF\Tickets\ControllerPlugin\TicketList $ticketListPlugin */
		$ticketListPlugin = $this->plugin('NF\Tickets:TicketList');

		return $ticketListPlugin->actionFilters($category);
	}

	public function actionWatch(ParameterBag $params)
	{
		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($params->ticket_category_id);
		if (!$category->canWatch($error))
		{
			return $this->noPermission($error);
		}

		$visitor = \XF::visitor();

		if ($this->isPost())
		{
			if ($this->filter('stop', 'bool'))
			{
				$action = 'delete';
				$config = [];
			}
			else
			{
				$action = 'watch';
				$config = [
					'notify_on' => $this->filter('notify_on', 'str'),
					'send_alert' => $this->filter('send_alert', 'bool'),
					'send_email' => $this->filter('send_email', 'bool'),
					'include_children' => $this->filter('include_children', 'bool'),
				];
			}

			/** @var \NF\Tickets\Repository\CategoryWatch $watchRepo */
			$watchRepo = $this->repository('NF\Tickets:CategoryWatch');
			$watchRepo->setWatchState($category, $visitor, $action, $config);

			$redirect = $this->redirect($this->buildLink('tickets/categories', $category));
			$redirect->setJsonParam('switchKey', $action === 'delete' ? 'watch' : 'unwatch');
			return $redirect;
		}
		else
		{
			$viewParams = [
				'category' => $category,
				'isWatched' => !empty($category->Watch[$visitor->user_id]),
			];
			return $this->view('NF\Tickets:Category\Watch', 'nf_tickets_category_watch', $viewParams);
		}
	}

	public function actionMarkRead(ParameterBag $params)
	{
		$visitor = \XF::visitor();
		if (!$visitor->user_id)
		{
			return $this->noPermission();
		}
		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($params->ticket_category_id);

		$markDate = $this->filter('date', 'uint');
		if (!$markDate)
		{
			$markDate = \XF::$time;
		}

		if ($this->isPost())
		{
			$categoryRepo = $this->getCategoryRepo();
			$categoryList = $categoryRepo->getViewableCategories($category);
			$categoryRepo->markCategoriesReadByVisitor($categoryList, $markDate);

			return $this->redirect(
				$this->getDynamicRedirect('tickets/categories', $category),
				\XF::phrase('nf_tickets_all_tickets_marked_as_read')
			);
		}
		else
		{
			$viewParams = [
				'category' => $category,
				'date' => $markDate,
			];
			return $this->view('NF\Tickets:Ticket\MarkRead', 'nf_tickets_tickets_category_mark_read', $viewParams);
		}
	}

	/**
	 * @param \NF\Tickets\Entity\Category $category
	 *
	 * @return \NF\Tickets\Service\Ticket\Creator
	 */
	protected function setupCreate(\NF\Tickets\Entity\Category $category)
	{
		$title = $this->filter('title', 'str');

		/** @var \XF\ControllerPlugin\Editor $editorPlugin */
		$editorPlugin = $this->plugin('XF:Editor');
		$message = $editorPlugin->fromInput('message');

		/** @var \NF\Tickets\Service\Ticket\Creator $creator */
		$creator = $this->service('NF\Tickets:Ticket\Creator', $category);
		$creator->setContent($title, $message);

		if ($category->canOpenTicketForAnyone())
		{
			$createFor = $this->filter('create_for', 'array-str');
			switch ($createFor['type'])
			{
				case 'member':
					/** @var \XF\Entity\User $member */
					$member = $this->em()->findOne('XF:User', ['username' => $createFor['name']]);

					if (!$member)
					{
						throw $this->exception($this->error(\XF::phrase('request_user_not_found')));
					}

					$creator->createForMember($member);
					break;
				case 'guest':
					$creator->createForGuest($createFor['name'], $createFor['email']);
					break;
			}
		}
		if ($category->canActionAsAnonymized())
		{
			$anonymize = $this->filter('anonymized', 'bool');
			if ($anonymize)
			{
				$creator->setAnonymize(true);
			}
		}
		$prefixId = $this->filter('prefix_id', 'uint');
		if ($prefixId && $category->isPrefixUsable($prefixId))
		{
			$creator->setPrefix($prefixId);
		}

        if ($category->canSetPriorityOnTicketCreation())
        {
            $priorityId = $this->filter('priority', 'uint');
            $priority = $this->getPriorityRepo()->getActivePriority($priorityId);
            if ($priority)
            {
                $creator->setPriority($priority->priority_id);
            }
        }

		$visitor = \XF::visitor();

		if (!$visitor->user_id)
		{
			$name = $this->filter('user_name', 'str');
			$email = $this->filter('user_email', 'str');
			$creator->setGuestUser($name, $email);
		}

		if ($category->canUploadAndManageAttachments())
		{
			$creator->setAttachmentHash($this->filter('attachment_hash', 'str'));
		}

		$customFields = $this->filter('custom_fields', 'array');
		$creator->setCustomFields($customFields);

		$creator->setUserAgent($this->request()->getUserAgent());

		return $creator;
	}

	/**
	 * @param \NF\Tickets\Service\Ticket\Creator $creator
	 *
	 * @throws \XF\PrintableException
	 */
	protected function finalizeCreate(\NF\Tickets\Service\Ticket\Creator $creator)
	{
		$creator->sendNotifications();

		$category = $creator->getCategory();
		$ticket = $creator->getTicket();
		$visitor = \XF::visitor();

		$setOptions = $this->filter('_xfSet', 'array-bool');
		if ($ticket->canWatch())
		{
			/** @var \NF\Tickets\Repository\TicketWatch $ticketWatchRepo */
			$ticketWatchRepo = $this->repository('NF\Tickets:TicketWatch');

			if (isset($setOptions['watch_ticket']))
			{
				$watch = $this->filter('watch_ticket', 'bool');
				if ($watch)
				{
					$config = $this->filter([
						'email_subscribe' => 'bool',
					]);
					$ticketWatchRepo->setWatchState($ticket, $visitor, 'watch', $config);
				}
			}
			else
			{
				// use user preferences
				$ticketWatchRepo->autoWatchTicket($ticket, $visitor, true);
			}
		}

		if ($visitor->user_id)
		{
			$this->getTicketRepo()->markTicketReadByVisitor($ticket, $ticket->start_date);

			$category->draft_ticket->delete();
		}
		else
		{
			// Set guest access token
			$this->getTicketRepo()->updateGuestTicketAccess($ticket);
		}
	}

	/**
	 * @param ParameterBag $parameterBag
	 *
	 * @return AbstractReply
	 * @throws \XF\Mvc\Reply\Exception
	 * @throws \XF\PrintableException
	 */
	public function actionCreate(ParameterBag $parameterBag)
	{
		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($parameterBag->ticket_category_id);
		if (!$category->canCreateTicket($error))
		{
			return $this->noPermission($error);
		}

		if ($this->isPost())
		{
			if ($this->filter('no_captcha', 'bool')) // JS is disabled so user hasn't seen Captcha.
			{
				$this->request->set('requires_captcha', true);

				return $this->rerouteController(__CLASS__, 'create', $parameterBag);
			}

			if (!$this->captchaIsValid())
			{
				return $this->error(\XF::phrase('did_not_complete_the_captcha_verification_properly'));
			}

			// attempting to post search criteria while creating a ticket is done by a bot, just reject
			if ($this->request()->exists('c') || $this->request()->exists('_xfClientLoadTime'))
			{
				return $this->noPermission();
			}

			$creator = $this->setupCreate($category);
			$creator->checkForSpam();

			if (!$creator->validate($errors))
			{
				return $this->error($errors);
			}
			$this->assertNotFlooding('ticket', $this->app->options()->floodCheckLengthDiscussion ?: null);

			/** @var \NF\Tickets\Entity\Ticket $ticket */
			$ticket = $creator->save();
			$this->finalizeCreate($creator);

			return $this->redirect($this->buildLink('tickets', $ticket));
		}

		/** @var \XF\Repository\Attachment $attachmentRepo */
		$attachmentRepo = $this->repository('XF:Attachment');

		if ($this->request->exists('requires_captcha'))
		{
			$creator = $this->setupCreate($category);
			$ticket = $creator->getTicket();
			$draftCustomFields = null;
			$attachmentHash = $creator->getMessagePreparer()->getAttachmentHash();
		}
		else
		{
			$ticket = $category->getNewTicket();
			$draft = $category->draft_ticket;
			/** @noinspection PhpUndefinedFieldInspection */
			$ticket->title = $draft->title ?: '';
			/** @noinspection PhpUndefinedFieldInspection */
			$draftCustomFields = $draft->custom_fields;
			if ($draftCustomFields)
			{
                $customFields = $ticket->custom_fields;
				$customFields->bulkSet($draftCustomFields, null, 'user', true);
			}
			/** @noinspection PhpUndefinedFieldInspection */
			$attachmentHash = $draft->attachment_hash;
		}

		if ($category->canUploadAndManageAttachments())
		{
			$attachmentData = $attachmentRepo->getEditorData('nf_tickets_message', $category, $attachmentHash);
		}
		else
		{
			$attachmentData = null;
		}

		$viewParams = [
			'category' => $category,
			'ticket' => $ticket,
			'prefixes' => $category->getUsablePrefixes(),
			'priorities' => $this->getPriorityRepo()->findPrioritiesForList(true)->fetch(),

			'attachmentData' => $attachmentData,
		];
		return $this->view('NF\Tickets:Category\CreateTicket', 'nf_tickets_category_create_ticket', $viewParams);
	}

	public function actionDraft(ParameterBag $parameterBag)
	{
		$this->assertDraftsEnabled();

		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($parameterBag->ticket_category_id);
		if (!$category->canCreateTicket($error))
		{
			return $this->noPermission($error);
		}

		$this->assertPostOnly();

		$creator = $this->setupCreate($category);
		$ticket = $creator->getTicket();

		$extraData = [
			'prefix_id' => $ticket->prefix_id,
			'title' => $ticket->title,
			'attachment_hash' => $this->filter('attachment_hash', 'str'),
			'custom_fields' => $ticket->custom_fields->getFieldValues(),
		];

		/** @var \XF\ControllerPlugin\Draft $draftPlugin */
		$draftPlugin = $this->plugin('XF:Draft');
		return $draftPlugin->actionDraftMessage($category->draft_ticket, $extraData);
	}

	public function actionPreview(ParameterBag $params)
	{
		$this->assertPostOnly();

		/** @noinspection PhpUndefinedFieldInspection */
		$category = $this->assertViewableCategory($params->ticket_category_id);
		if (!$category->canCreateTicket($error))
		{
			return $this->noPermission($error);
		}

		$creator = $this->setupCreate($category);
		if (!$creator->validate($errors))
		{
			return $this->error($errors);
		}

		$ticket = $creator->getTicket();
		$message = $creator->getMessage();
		$attachments = null;

		$tempHash = $this->filter('attachment_hash', 'str');
		if ($tempHash && $ticket->Category->canUploadAndManageAttachments())
		{
			/** @var \XF\Repository\Attachment $attachRepo */
			$attachRepo = $this->repository('XF:Attachment');
			$attachments = $attachRepo->findAttachmentsByTempHash($tempHash)->fetch();
		}

		/** @var \XF\ControllerPlugin\BbCodePreview $plugin */
		$plugin = $this->plugin('XF:BbCodePreview');

		return $plugin->actionPreview(
			$message->message, 'nf_tickets_message', $message->User, $attachments, $ticket->canViewAttachments()
		);
	}

	/**
	 * @param int   $categoryId
	 * @param array $extraWith
	 *
	 * @return \NF\Tickets\Entity\Category
	 * @throws \XF\Mvc\Reply\Exception
	 */
	protected function assertViewableCategory(int $categoryId, array $extraWith = []): \NF\Tickets\Entity\Category
    {
		$visitor = \XF::visitor();

		$extraWith[] = 'AnonymousUser';
		$extraWith[] = 'Permissions|' . $visitor->permission_combination_id;

		$userId = $visitor->user_id;
		if ($userId)
		{
			$extraWith[] = 'Watch|' . $userId;
			$extraWith[] = 'Read|' . $userId;
		}

		/** @var \NF\Tickets\Entity\Category $category */
		$category = $this->em()->find('NF\Tickets:Category', $categoryId, $extraWith);
		if (!$category)
		{
			throw $this->exception($this->notFound(\XF::phrase('requested_category_not_found')));
		}

		if (!$category->canView($error))
		{
			throw $this->exception($this->noPermission($error));
		}

		return $category;
	}
}