<?php

namespace NF\Tickets\Pub\Controller;

use NF\Tickets\Entity\KBArticle;
use NF\Tickets\Entity\KBCategory;
use NF\Tickets\XF\Entity\User;
use XF\Mvc\Entity\AbstractCollection;
use XF\Mvc\FormAction;
use XF\Mvc\ParameterBag;
use XF\Pub\Controller\AbstractController;

class KB extends AbstractController
{
	protected function preDispatchController($action, ParameterBag $params)
	{
		/** @var User $visitor */
		$visitor = \XF::visitor();

		if (!$visitor->canViewKB($error))
		{
			throw $this->exception($this->noPermission($error));
		}
	}

	public function actionIndex(ParameterBag $params)
    {
        if ($params->kb_article_id)
        {
            return $this->rerouteController(__CLASS__, 'view', $params);
        }

        /** @var \NF\Tickets\ControllerPLugin\TicketList $overviewPlugin */
        $overviewPlugin = $this->plugin('NF\Tickets:TicketList');

        $categoryParams = $overviewPlugin->getCategoryListData();

        /** @var AbstractCollection $categories */
        $categories = $categoryParams['categories'];
        $viewableCategoryIds = $categories->keys();

        $listParams = $overviewPlugin->getCoreListData($viewableCategoryIds);

        $this->assertValidPage($listParams['page'], $listParams['perPage'], $listParams['total'], 'knowledge-base');
        $this->assertCanonicalUrl($this->buildLink('knowledge-base', null, ['page' => $listParams['page']]));

        $viewParams = $categoryParams + $listParams;
        $viewParams += [
            'kbCategories' => $this->finder('NF\Tickets:KBCategory')->order('display_order', 'asc')->fetch(),
            'kbArticlesGrouped' => $this->finder('NF\Tickets:KBArticle')->fetch()->groupBy('kb_category_id'),
        ];

        return $this->view(
        	'NF\Tickets:KB\Index',
			'nf_tickets_kb_index',
			$viewParams
		);
    }

    public function actionView(ParameterBag $params)
    {
        $article = $this->assertViewableArticle($params->kb_article_id);

        $articleRepo = $this->getKBArticleRepo();
        $articleRepo->logArticleView($article);

        $viewParams = [
            'article' => $article
        ];

        return $this->view(
            'NF\Tickets:KB\Article\View',
            'nf_tickets_kb_article_view',
            $viewParams
        );
    }

	public function actionCreate(/** @noinspection PhpUnusedParameterInspection */ ParameterBag $parameterBag)
	{
		/** @var User $visitor */
		$visitor = \XF::visitor();
		if (!$visitor->canEditKBArticles($error))
		{
			return $this->noPermission($error);
		}

		$this->assertCanonicalUrl($this->buildLink('knowledge-base/create'));

		$categoryRepo = $this->getKBCategoryRepo();

		$categories = $categoryRepo->getViewableCategories();

		if (!$categories->count())
		{
			return $this->noPermission(\XF::phrase('nf_tickets_there_no_categories_to_create_knowledge_base_articles_within'));
		}

		$categoryTree = $categoryRepo->createCategoryTree($categories);

		$categoryExtras = $categoryRepo->getCategoryListExtras($categoryTree);

		$viewParams = [
			'categoryTree'   => $categoryTree,
			'categoryExtras' => $categoryExtras,
		];

		return $this->view('NF\Tickets:KB\CategoryChooser', 'nf_tickets_kb_category_chooser', $viewParams);
	}

	public function actionCategory()
	{
		return $this->redirect('knowledge-base'); // TODO: back this out in to own controller with filtering
	}

	public function actionCategoryAdd(ParameterBag $params)
	{
		$category = $this->assertViewableCategory($params->kb_category_id);

		/** @var User $visitor */
		$visitor = \XF::visitor();
		if (!$visitor->canEditKBArticles())
		{
			return $this->noPermission();
		}

		if ($this->isPost())
		{
			$creator = $this->setupCreate($category);

			if (!$creator->validate($errors))
			{
				return $this->error($errors);
			}
			$article = $creator->save();

			$this->finalizeCreate($creator);

			return $this->redirect(
				$this->buildLink('knowledge-base', $article)
			);
		}
		else
		{
			$article = $category->getNewArticle();

			if ($category->canUploadAndManageAttachments())
			{
				/** @var \XF\Repository\Attachment $attachmentRepo */
				$attachmentRepo = $this->repository('XF:Attachment');
				$attachmentData = $attachmentRepo->getEditorData('nf_tickets_kb_article', $article, null, [
					'kb_category_id' => $article->kb_category_id
				]);
			}
			else
			{
				$attachmentData = null;
			}

			$viewParams = [
				'article' => $article,
				'category' => $category,
				'attachmentData' => $attachmentData,
			];
			return $this->view('NF\Tickets:KB\Category\Create', 'nf_tickets_kb_article_create', $viewParams);
		}
	}

	protected function setupCreate(KBCategory $category)
	{
		$title = $this->filter('title', 'str');
        /** @var \XF\ControllerPlugin\Editor $editorPlugin */
        $editorPlugin = $this->plugin('XF:Editor');
		$content = $editorPlugin->fromInput('content');

		/** @var \NF\Tickets\Service\KBArticle\Creator $creator */
		$creator = $this->service('NF\Tickets:KBArticle\Creator', $category);
		$creator->setContent($title, $content);

		if ($category->canUploadAndManageAttachments())
		{
			$creator->setAttachmentHash($this->filter('attachment_hash', 'str'));
		}

		return $creator;
	}

	protected function finalizeCreate(\NF\Tickets\Service\KBArticle\Creator $creator)
	{
	}

    /**
     * @param ParameterBag $parameterBag
     * @return \XF\Mvc\Reply\AbstractReply
     * @throws \XF\Mvc\Reply\Exception
     */
    public function actionEdit(ParameterBag $parameterBag)
    {
	    $article = $this->assertViewableArticle($parameterBag->kb_article_id);
	    if (!$article->canEdit($error))
	    {
	    	return $this->noPermission($error);
	    }

	    if ($this->isPost())
	    {
	    	$editor = $this->setupEdit($article);

	    	if (!$editor->validate($errors))
		    {
		    	return $this->error($errors);
		    }
		    $article = $editor->save();

		    $this->finalizeEdit($editor);

		    return $this->redirect(
			    $this->buildLink('knowledge-base', $article)
		    );
	    }
	    else
	    {
		    $category = $article->Category;

		    if ($category && $category->canUploadAndManageAttachments())
		    {
			    /** @var \XF\Repository\Attachment $attachmentRepo */
			    $attachmentRepo = $this->repository('XF:Attachment');
			    $attachmentData = $attachmentRepo->getEditorData('nf_tickets_kb_article', $article, null, [
			    	'kb_category_id' => $article->kb_category_id
			    ]);
		    }
		    else
		    {
			    $attachmentData = null;
		    }

		    $viewParams = [
			    'article' => $article,
			    'attachmentData' => $attachmentData
		    ];
		    return $this->view('NF\Tickets:KB\Category\Edit', 'nf_tickets_kb_article_edit', $viewParams);
	    }
    }

	protected function setupEdit(KBArticle $article)
	{
		$title = $this->filter('title', 'str');
        /** @var \XF\ControllerPlugin\Editor $editorPlugin */
        $editorPlugin = $this->plugin('XF:Editor');
        $content = $editorPlugin->fromInput('content');

		/** @var \NF\Tickets\Service\KBArticle\Editor $editor */
		$editor = $this->service('NF\Tickets:KBArticle\Editor', $article);
		$editor->setContent($title, $content);

		if ($article->Category->canUploadAndManageAttachments())
		{
			$editor->setAttachmentHash($this->filter('attachment_hash', 'str'));
		}

		return $editor;
	}

	protected function finalizeEdit(\NF\Tickets\Service\KBArticle\Editor $editor)
	{
	}

    public function actionAddPreview(ParameterBag $params)
    {
        $this->assertPostOnly();

        /** @var KBArticle $article */
        $article = $this->em()->create('NF\Tickets:KBArticle');
        /** @var \XF\ControllerPlugin\Editor $editorPlugin */
        $editorPlugin = $this->plugin('XF:Editor');
        $article->content = $editorPlugin->fromInput('content');
//        $attachments = null;
//
//        $tempHash = $this->filter('attachment_hash', 'str');
//        if ($tempHash && $thread->Forum->canUploadAndManageAttachments())
//        {
//            $attachRepo = $this->repository('XF:Attachment');
//            $attachments = $attachRepo->findAttachmentsByTempHash($tempHash)->fetch();
//        }

        /** @var \XF\ControllerPlugin\BbCodePreview $plugin */
        $plugin = $this->plugin('XF:BbCodePreview');

        return $plugin->actionPreview(
            $article->content, 'nf_tickets_kb_article', \XF::visitor()//, $attachments, $thread->canViewAttachments()
        );
    }

    /**
     * @param KBArticle $article
     * @return FormAction
     */
    protected function articleSaveProcess(KBArticle $article)
    {
        $entityInput = $this->filter([
            'title' => 'str',
            'kb_category_id' => 'uint',
        ]);

        /** @var \XF\ControllerPlugin\Editor $editorPlugin */
        $editorPlugin = $this->plugin('XF:Editor');
        $entityInput['content'] = $editorPlugin->fromInput('content');

        $form = $this->formAction();
        $form->basicEntitySave($article, $entityInput);

        return $form;
    }

    /**
     * @param ParameterBag $parameterBag
     *
     * @return \XF\Mvc\Reply\AbstractReply
     * @throws \XF\Mvc\Reply\Exception
     * @throws \XF\PrintableException
     */
    public function actionSave(ParameterBag $parameterBag)
    {
        $this->assertPostOnly();

	    /** @var User $visitor */
	    $visitor = \XF::visitor();
	    if (!$visitor->canEditKBArticles())
	    {
		    return $this->noPermission();
	    }

        if ($parameterBag->kb_article_id)
        {
            $article = $this->assertViewableArticle($parameterBag->kb_article_id);
        }
        else if ($parameterBag->kb_category_id)
        {
            $article = $this->assertViewableCategory($parameterBag->kb_category_id)->getNewArticle();
        }
        else
        {
            /** @var KBArticle $article */
            $article = $this->em()->create('NF\Tickets:KBArticle');
            $article->kb_category_id = $this->filter('kb_category_id', 'int');
        }

        $this->articleSaveProcess($article)->run();

        if ($article->Category->canUploadAndManageAttachments())
        {
            $attachmentHash = $this->filter('attachment_hash', 'str');
            /** @var \XF\Service\Attachment\Preparer $inserter */
            $inserter = $this->service('XF:Attachment\Preparer');
            $associated = $inserter->associateAttachmentsWithContent($attachmentHash, 'nf_tickets_kb_article', $article->kb_article_id);
            if ($associated)
            {
                $article->fastUpdate('attach_count', $article->attach_count + $associated);
            }
        }

        return $this->redirect(
            $this->buildLink('knowledge-base', $article)
        );
    }

    public function actionDelete(ParameterBag $params)
    {
	    /** @var User $visitor */
	    $visitor = \XF::visitor();
	    if (!$visitor->canDeleteKBArticles())
	    {
		    return $this->noPermission();
	    }

        $article = $this->assertViewableArticle($params->kb_article_id);
        if ($this->isPost())
        {
            $article->delete();
            return $this->redirect($this->getDynamicRedirect(
                $this->buildLink('knowledge-base'),
                false
            ));
        }

        $viewParams = [
            'article' => $article
        ];
        return $this->view(
            'NF\Tickets:KB\Article\Delete',
            'nf_tickets_kb_article_delete',
            $viewParams
        );
    }

    /**
     * @param int   $articleId
     * @param array $extraWith
     * @return KBArticle
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function assertViewableArticle(int $articleId, array $extraWith = []): KBArticle
    {
        $extraWith[] = 'Category';

        /** @var KBArticle $article */
        $article = $this->em()->find('NF\Tickets:KBArticle', $articleId, $extraWith);
        if (!$article)
        {
            throw $this->exception($this->notFound(\XF::phrase('nf_tickets_requested_article_not_found')));
        }

        if (!$article->canView($error))
        {
            throw $this->exception($this->noPermission($error));
        }

        return $article;
    }
    /**
     * @param int   $categoryId
     * @param array $extraWith
     * @return KBCategory
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function assertViewableCategory(int $categoryId, array $extraWith = []): KBCategory
    {
        /** @var KBCategory $category */
        $category = $this->em()->find('NF\Tickets:KBCategory', $categoryId, $extraWith);
        if (!$category)
        {
            throw $this->exception($this->notFound(\XF::phrase('nf_tickets_requested_article_not_found')));
        }

        if (!$category->canView($error))
        {
            throw $this->exception($this->noPermission($error));
        }

        return $category;
    }

    protected function getKBCategoryRepo(): \NF\Tickets\Repository\KBCategory
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:KBCategory');
    }

    protected function getKBArticleRepo(): \NF\Tickets\Repository\KBArticle
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:KBArticle');
    }
}