<?php

namespace NF\Tickets\Pub\Controller;

use XF\Mvc\ParameterBag;

class Message extends AbstractTicketController
{
    public function actionIndex(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);
        $ticket = $message->Ticket;

        $page = \floor($message->position / $this->options()->messagesPerPage) + 1;

        $link = 'tickets';

        if ($page > 1)
        {
            $params = [
                'page' => $page,
            ];
        }
        else
        {
            $params = [];
        }

        return $this->redirectPermanently(
            $this->buildLink($link, $ticket, $params) . '#message-' . $message->message_id
        );
    }

    public function actionPreview(ParameterBag $params)
    {
        $this->assertPostOnly();

        $message = $this->assertViewableMessage($params->message_id);
        if (!$message->canEdit($error))
        {
            return $this->noPermission($error);
        }

        $ticket = $message->Ticket;

        $editor = $this->setupEdit($message);

        if (!$editor->validate($errors))
        {
            return $this->error($errors);
        }

        $attachments = [];
        $tempHash = $this->filter('attachment_hash', 'str');

        if ($ticket->Category->canUploadAndManageAttachments())
        {
            /** @var \XF\Repository\Attachment $attachmentRepo */
            $attachmentRepo = $this->repository('XF:Attachment');
            $attachmentData = $attachmentRepo->getEditorData('nf_tickets_message', $message, $tempHash);
            $attachments = $attachmentData['attachments'];
        }
        /** @var \XF\ControllerPlugin\BbCodePreview $plugin */
        $plugin = $this->plugin('XF:BbCodePreview');

        return $plugin->actionPreview(
            $message->message, 'post', $message->User, $attachments, $ticket->canViewAttachments()
        );
    }

    public function actionEdit(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);
        if (!$message->canEdit($error))
        {
            return $this->noPermission($error);
        }

        $noInlineMod = $this->filter('_xfNoInlineMod', 'bool');

        if ($this->isPost())
        {
            $editor = $this->setupEdit($message);
            $editor->checkForSpam();

            if (!$editor->validate($errors))
            {
                return $this->error($errors);
            }
            $editor->save();

            $this->finalizeEdit($editor);

            if ($this->filter('_xfWithData', 'bool') && $this->filter('_xfInlineEdit', 'bool'))
            {
                /** @var \XF\Repository\Attachment $attachmentRepo */
                $attachmentRepo = $this->repository('XF:Attachment');
                $attachmentRepo->addAttachmentsToContent([$message->message_id => $message], 'nf_tickets_message');

                $viewParams = [
                    'message' => $message,

                    'noInlineMod' => $noInlineMod,
                ];
                $reply = $this->view('NF\Tickets:Message\EditNewMessage', 'nf_tickets_message_edit_new_message', $viewParams);
                $reply->setJsonParam('message', \XF::phrase('your_changes_have_been_saved'));

                return $reply;
            }
            else
            {
                return $this->redirect($this->buildLink('tickets/messages', $message));
            }
        }
        else
        {
            $category = $message->Ticket->Category;
            if ($category->canUploadAndManageAttachments())
            {
                /** @var \XF\Repository\Attachment $attachmentRepo */
                $attachmentRepo = $this->repository('XF:Attachment');
                $attachmentData = $attachmentRepo->getEditorData('nf_tickets_message', $message);
            }
            else
            {
                $attachmentData = null;
            }

            $viewParams = [
                'message' => $message,

                'attachmentData' => $attachmentData,
                'quickEdit'   => $this->filter('_xfWithData', 'bool'),
                'noInlineMod' => $noInlineMod,
            ];

            return $this->view('NF\Tickets:Message\Edit', 'nf_tickets_message_edit', $viewParams);
        }
    }

    public function actionDelete(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);
        if (!$message->canDelete('soft', $error))
        {
            return $this->noPermission($error);
        }

        if ($this->isPost())
        {
            $type = $this->filter('hard_delete', 'bool') ? 'hard' : 'soft';
            $reason = $this->filter('reason', 'str');

            if (!$message->canDelete($type, $error))
            {
                return $this->noPermission($error);
            }

            /** @var \NF\Tickets\Service\Message\Deleter $deleter */
            $deleter = $this->service('NF\Tickets:Message\Deleter', $message);

            if ($this->filter('author_alert', 'bool') && $message->canSendModeratorActionAlert())
            {
                $deleter->setSendAlert(true, $this->filter('author_alert_reason', 'str'));
            }

            $deleter->delete($type, $reason);

            /** @var \XF\ControllerPlugin\InlineMod $plugin */
            $plugin = $this->plugin('XF:InlineMod');
            $plugin->clearIdFromCookie('nf_tickets_message', $message->message_id);

            if ($deleter->wasTicketDeleted())
            {
                $plugin->clearIdFromCookie('nf_tickets', $message->ticket_id);

                return $this->redirect(
                    $message->Ticket && $message->Ticket->Category
                        ? $this->buildLink('tickets/categories', $message->Ticket->Category)
                        : $this->buildLink('tickets/queue')
                );
            }

            return $this->redirect(
                $this->getDynamicRedirect($this->buildLink('tickets', $message->Ticket), false)
            );
        }
        else
        {
            $viewParams = [
                'message' => $message,
            ];

            return $this->view('NF\Tickets:Message\Delete', 'nf_tickets_message_delete', $viewParams);
        }
    }

    public function actionUndelete(ParameterBag $params)
    {
        /** @noinspection PhpUndefinedFieldInspection */
        $message = $this->assertViewableMessage($params->message_id);

        /** @var \XF\ControllerPlugin\Undelete $plugin */
        $plugin = $this->plugin('XF:Undelete');
        return $plugin->actionUndelete(
            $message,
            $this->buildLink('tickets/messages/undelete', $message),
            $message->getContentUrl(),
            $message->getContentTitle('undelete'),
            'message_state'
        );
    }

    public function actionIp(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);
        $breadcrumbs = $this->getMessageBreadcrumbs($message);

        /** @var \XF\ControllerPlugin\Ip $ipPlugin */
        $ipPlugin = $this->plugin('XF:Ip');

        return $ipPlugin->actionIp($message, $breadcrumbs);
    }

    public function actionReport(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);
        if (!$message->canReport($error))
        {
            return $this->noPermission($error);
        }

        /** @var \XF\ControllerPlugin\Report $reportPlugin */
        $reportPlugin = $this->plugin('XF:Report');

        return $reportPlugin->actionReport(
            'nf_tickets_message', $message,
            $this->buildLink('tickets/messages/report', $message),
            $this->buildLink('tickets/messages', $message)
        );
    }

    public function actionReact(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);

        /** @var \XF\ControllerPlugin\Reaction $reactionPlugin */
        $reactionPlugin = $this->plugin('XF:Reaction');

        return $reactionPlugin->actionReactSimple($message, 'ticket-message');
    }

    public function actionReactions(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);

        $breadcrumbs = $this->getMessageBreadcrumbs($message);

        /** @var \XF\ControllerPlugin\Reaction $reactionPlugin */
        $reactionPlugin = $this->plugin('XF:Reaction');

        return $reactionPlugin->actionReactions(
            $message,
            'ticket-message/reactions',
            null, $breadcrumbs
        );
    }

    public function actionQuote(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);
        if (!$message->Ticket->canReply())
        {
            return $this->noPermission();
        }

        /** @var \XF\ControllerPlugin\Quote $quotePlugin */
        $quotePlugin = $this->plugin('XF:Quote');

        return $quotePlugin->actionQuote($message, 'nf_tickets_message');
    }

    public function actionBookmark(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);

        /** @var \XF\ControllerPlugin\Bookmark $bookmarkPlugin */
        $bookmarkPlugin = $this->plugin('XF:Bookmark');

        return $bookmarkPlugin->actionBookmark(
            $message, $this->buildLink('tickets/messages/bookmark', $message)
        );
    }

    public function actionWarn(ParameterBag $params)
    {
        $message = $this->assertViewableMessage($params->message_id);

        if (!$message->canWarn($error))
        {
            return $this->noPermission($error);
        }

        $breadcrumbs = $this->getMessageBreadcrumbs($message);

        /** @var \XF\ControllerPlugin\Warn $warnPlugin */
        $warnPlugin = $this->plugin('XF:Warn');

        return $warnPlugin->actionWarn(
            'nf_tickets_message', $message,
            $this->buildLink('tickets/messages/warn', $message),
            $breadcrumbs
        );
    }

    protected function setupEdit(\NF\Tickets\Entity\Message $message)
    {
        /** @var \XF\ControllerPlugin\Editor $editorPlugin */
        $editorPlugin = $this->plugin('XF:Editor');
        $messageText = $editorPlugin->fromInput('message');

        /** @var \NF\Tickets\Service\Message\Editor $editor */
        $editor = $this->service('NF\Tickets:Message\Editor', $message);
        $editor->setMessageText($messageText);

        if ($message->Ticket->canActionAsAnonymized())
        {
            $anonymize = $this->filter('anonymized', 'bool');
            $editor->setAnonymize($anonymize);
        }

        if ($message->Ticket->Category->canUploadAndManageAttachments())
        {
            $editor->setAttachmentHash($this->filter('attachment_hash', 'str'));
        }

        if ($this->filter('author_alert', 'bool') && $message->canSendModeratorActionAlert())
        {
            $editor->setSendAlert(true, $this->filter('author_alert_reason', 'str'));
        }

        return $editor;
    }

    protected function finalizeEdit(\NF\Tickets\Service\Message\Editor $editor)
    {
    }

    protected function getMessageBreadcrumbs(\NF\Tickets\Entity\Message $message)
    {
        $breadcrumbs = [
            [
                'href'  => $this->buildLink('tickets', $message->Ticket),
                'value' => $message->Ticket->title,
            ],
        ];

        return $breadcrumbs;
    }

    /**
     * @param int   $messageId
     * @param array $extraWith
     * @return \NF\Tickets\Entity\Message
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function assertViewableMessage(int $messageId, array $extraWith = []): \NF\Tickets\Entity\Message
    {
        $extraWith[] = 'User';
        $extraWith = \array_unique($extraWith);

        /** @var \NF\Tickets\Entity\Message $message */
        $message = $this->em()->find('NF\Tickets:Message', $messageId, $extraWith);
        if (!$message)
        {
            throw $this->exception($this->notFound(\XF::phrase('requested_message_not_found')));
        }
        if (!$message->canView($error))
        {
            throw $this->exception($this->noPermission($error));
        }

        return $message;
    }

    public static function getActivityDetails(array $activities)
    {
        return \XF::phrase('nf_tickets_viewing_tickets');
    }
}