<?php

namespace NF\Tickets\Repository;

use NF\Tickets\Repository\Response as ResponseRepo;
use NF\Tickets\Service\Ticket\Replier as TicketReplierSvc;
use XF\Mvc\Entity\Repository;
use XF\Phrase;
use XF\Job\Manager as JobManager;
use NF\Tickets\Entity\Ticket as TicketEntity;

use function array_key_exists;

class Escalation extends Repository
{
    /**
     * @param bool $activeOnly
     * @return \NF\Tickets\Finder\Escalation
     */
    public function findEscalationsForList($activeOnly = true)
    {
        /** @var \NF\Tickets\Finder\Escalation $finder */
        $finder = $this->finder('NF\Tickets:Escalation')
                       ->setDefaultOrder('execution_order');

        if ($activeOnly)
        {
            $finder->isActive();
        }

        return $finder;
    }

    public function quickEscalate(TicketEntity $ticket)
    {
        if ($ticket->isShadowBanned())
        {
            return;
        }

        $activeEscalations = $this->findEscalationsForList()->fetch();

        if ($activeEscalations->count() > 0)
        {
            $user = $ticket->User;
            if (!$user)
            {
                /** @var \XF\Repository\User $userRepo */
                $userRepo = $this->repository('XF:User');
                $user = $userRepo->getGuestUser($ticket->user_name);
                $user->setAsSaved('email', $ticket->user_email);
            }

            foreach ($activeEscalations AS $escalation)
            {
                $ticketCriteria = $this->app()->criteria('NF\Tickets:Ticket', $escalation->ticket_criteria, $ticket);
                $ticketCriteria->setMatchOnEmpty(false);
                if ($ticketCriteria->isMatched($user))
                {
                    $this->escalate($escalation, $ticket);
                }
            }
        }
    }

    /**
     * @throws \Exception
     */
    protected function escalate(\NF\Tickets\Entity\Escalation $escalation, TicketEntity $ticket)
    {
        $actions = $escalation->actions;
        if (!$actions)
        {
            return true;
        }

        $visitor = \XF::visitor();
        $actions = \array_merge([
            'send_message' => false,
            'send_alert' => false
        ], $actions);
        $updates = [];

        if (array_key_exists('category_id', $actions) && $actions['category_id'] > 0)
        {
            $updates['ticket_category_id'] = $actions['category_id'];
        }

        if (array_key_exists('status_id', $actions))
        {
            $updates['status_id'] = $actions['status_id'];
        }

        if (array_key_exists('priority', $actions))
        {
            $updates['priority'] = $actions['priority'];
        }

        if (array_key_exists('lock', $actions))
        {
            $updates['ticket_locked'] = true;
        }

        if (array_key_exists('unlock', $actions))
        {
            $updates['ticket_locked'] = false;
        }

        if ($updates)
        {
            $ticket->fastUpdate($updates);
        }

        $this->db()->insert('xf_nf_tickets_escalation_trigger_log', [
            'ticket_id'     => $ticket->ticket_id,
            'escalation_id' => $escalation->escalation_id,
            'trigger_date'  => \XF::$time,
            'changes'       => json_encode($updates),
        ]);

        if ($actions['send_alert'])
        {
            $actor = $escalation->AlertSender ?: $visitor;
            $this->jobManager()->enqueue('NF\Tickets:EscalationUserAlert', [
                'ticket_id' => $ticket->ticket_id,
                'escalation_id' => $escalation->escalation_id,
                'actor_id' => $actor->user_id
            ]);
        }

        if ($actions['send_message'])
        {
            $actor = $escalation->AlertSender ?: $visitor;
            \XF::asVisitor($actor, function () use($ticket, $escalation)
            {
                $ticketReplierSvc = $this->getTicketReplierSvc($ticket);
                $ticketReplierSvc->setIsAutomated();

                $message = \strtr(
                    $escalation->message_body,
                    $this->getResponseRepo()->getReplacementValues($ticket)
                );
                $ticketReplierSvc->setMessage($message);

                if ($ticketReplierSvc->validate($errors))
                {
                    $ticketReplierSvc->save();
                }

                foreach ($errors AS $error)
                {
                    $tmpError = $error;
                    if ($tmpError instanceof Phrase)
                    {
                        $tmpError = $tmpError->render('raw');
                    }

                    \XF::logError('[NF] Tickets - ' . $tmpError);
                }
            });
        }

        return $ticket;
    }

    protected function jobManager() : JobManager
    {
        return $this->app()->jobManager();
    }

    protected function getResponseRepo() : ResponseRepo
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Response');
    }

    protected function getTicketReplierSvc(TicketEntity $ticket) : TicketReplierSvc
    {
        return $this->app()->service('NF\Tickets:Ticket\Replier', $ticket);
    }
}