<?php

namespace NF\Tickets\Repository;

use XF\Entity\User;
use XF\Mvc\Entity\Repository;
use XF\Util\Arr;
use XF\Validator\Email;

use function count, intval, is_array, is_string;

/**
 * Class Ticket
 *
 * @package NF\Tickets\Repository
 */
class Ticket extends Repository
{
    /**
     * @param int|null $length
     * @return mixed|null|string|string[]
     */
    public function generateReference($length = null)
    {
        if ($length === null)
        {
            $length = $this->options()->ticketRefLength;
        }

        return \utf8_strtoupper(\XF::generateRandomString($length));
    }

    /**
     * @param null $length
     * @return string
     */
    public function generatePassword($length = null)
    {
        if ($length === null)
        {
            $length = 8;
        }

        $password = \XF::generateRandomString($length);

        return $password;
    }

    /**
     * @param array|null $viewableCategoryIds
     * @param bool $includeCategory
     * @param bool $allowOwnPending
     * @return \NF\Tickets\Finder\Ticket
     */
    public function findTicketsForOverviewList(array $viewableCategoryIds = null, $includeCategory = true, $allowOwnPending = false)
    {
        /** @var \NF\Tickets\Finder\Ticket $ticketFinder */
        $ticketFinder = $this->finder('NF\Tickets:Ticket');

        if (is_array($viewableCategoryIds))
        {
            $ticketFinder->applyCategoryVisibilityChecks($viewableCategoryIds, $allowOwnPending);
        }
        else
        {
            $ticketFinder->with('Category.Permissions|' . \XF::visitor()->permission_combination_id);
            $ticketFinder->applyGlobalVisibilityChecks($allowOwnPending);
        }

        $ticketFinder
            ->forFullView($includeCategory)
            ->useDefaultOrder();

        return $ticketFinder;
    }

    public function findTicketsForWatchedList($categoryIds = null, $userId = null, array $limits = [])
    {
        $finder = $this->findTicketsForOverviewList($categoryIds, null, true);

        if ($userId === null)
        {
            $userId = \XF::visitor()->user_id;
        }
        $userId = intval($userId);

        $finder
            ->with('Watch|' . $userId, true)
            ->with('Read|' . $userId)
            ->with('LastMessage')
            ->with('LastMessageUser')
            ->where('discussion_state', 'visible')
            ->setDefaultOrder('last_modified_date', 'DESC');

        return $finder;
    }

    public function findTicketsForApi(\NF\Tickets\Entity\Category $withinCategory = null)
	{
		/** @var \NF\Tickets\Finder\Ticket $finder */
		$finder = $this->finder('NF\Tickets:Ticket')
			->with('api');

        /** @var Category $repo */
        $repo = $this->repository('NF\Tickets:Category');

		if (\XF::isApiCheckingPermissions())
		{
			$categoryIds = $repo->getViewableCategoryIds($withinCategory);
			$finder->where('ticket_category_id', $categoryIds);
			$finder->applyGlobalVisibilityChecks();
		}
		else if ($withinCategory)
		{
			$categoryIds = $repo->getCategoryIds($withinCategory);
			$finder->where('ticket_category_id', $categoryIds);
		}

		$finder->useDefaultOrder();

		return $finder;
	}

    /**
     * @return float|int
     */
    public function getReadMarkingCutOff()
    {
        return \XF::$time - $this->options()->readMarkingDataLifetime * 86400;
    }

    /**
     * @param \NF\Tickets\Entity\Ticket $ticket
     * @param null $newRead
     * @return bool
     */
    public function markTicketReadByVisitor(\NF\Tickets\Entity\Ticket $ticket, $newRead = null)
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
            return false;
        }

        if ($newRead === null && $ticket->LastMessage)
        {
            $newRead = \max(\XF::$time, $ticket->LastMessage->message_date);
        }

        $cutOff = $this->getReadMarkingCutOff();
        if ($newRead <= $cutOff)
        {
            return false;
        }

        $readDate = $ticket->getVisitorReadDate();
        if ($newRead <= $readDate)
        {
            return false;
        }

        $this->db()->insert('xf_nf_tickets_ticket_read', [
            'ticket_id' => $ticket->ticket_id,
            'user_id' => $visitor->user_id,
            'ticket_read_date' => $newRead,
        ], false, 'ticket_read_date = VALUES(ticket_read_date)');

        if ($newRead < $ticket->last_message_date)
        {
            // ticket no fully viewed
            return false;
        }

        if ($ticket->Category && !$this->countUnreadTicketsInCategory($ticket->Category))
        {
            /** @var Category $categoryRepo */
            $categoryRepo = $this->repository('NF\Tickets:Category');
            $categoryRepo->markCategoryReadByVisitor($ticket->Category);
        }

        return true;
    }

    /**
     * @param \NF\Tickets\Entity\Category $category
     * @return int
     */
    public function countUnreadTicketsInCategory(\NF\Tickets\Entity\Category $category)
    {
        $visitor = \XF::visitor();
        $userId = $visitor->user_id;
        if (!$userId)
        {
            return 0;
        }

        $read = $category->Read[$userId];
        $cutOff = $this->getReadMarkingCutOff();

        $readDate = $read ? \max($read->category_read_date, $cutOff) : $cutOff;

        /** @var \NF\Tickets\Finder\Ticket $finder */
        $finder = $this->finder('NF\Tickets:Ticket');
        $finder
            ->where('ticket_category_id', $category->ticket_category_id)
            ->where('last_message_date', '>', $readDate)
            ->where('discussion_state', 'visible')
            ->whereOr(
                ["Read|{$userId}.ticket_id", null],
                [$finder->expression('%s > %s', 'last_message_date', "Read|{$userId}.ticket_read_date")]
            )
            ->skipIgnored();

        return $finder->total();
    }

    public function getUserTicketCount($userId)
    {
        return $this->db()->fetchOne("
            SELECT COUNT(*)
            FROM xf_nf_tickets_ticket
            WHERE user_id = ?
                AND discussion_state = 'visible'
        ", $userId);
    }

    public function sendModeratorActionAlert(
        \NF\Tickets\Entity\Ticket $ticket, $action, $reason = '', array $extra = [], User $forceUser = null
    )
    {
        if (!$forceUser)
        {
            if (!$ticket->user_id || !$ticket->User)
            {
                return false;
            }

            $forceUser = $ticket->User;
        }

        $extra = \array_merge([
            'priority' => $ticket->Priority,
            'title' => $ticket->title,
            'status' => $ticket->Status->getTitle(),
            'link' => $this->app()->router('public')->buildLink('nopath:tickets', $ticket),
            'reason' => $reason,
            ['depends_on_addon_id' => 'NF/Tickets'],
        ], $extra);

        /** @var \XF\Repository\UserAlert $alertRepo */
        $alertRepo = $this->repository('XF:UserAlert');
        $alertRepo->alert(
            $forceUser,
            0, '',
            'user', $forceUser->user_id,
            "nf_tickets_{$action}", $extra
        );

        return true;
    }

    public function rebuildTicketCounts()
    {
        $db = $this->db();
        /** @var Status $statusRepo */
        $statusRepo = $this->repository('NF\Tickets:Status');
        $status = $statusRepo->findStatusesForList()
            ->isCounted()
            ->fetch()
            ->keys();
        $cache = [
            'total' => $status ? $db->fetchOne("SELECT COUNT(*) FROM xf_nf_tickets_ticket WHERE status_id in (" . $db->quote($status) . ") AND discussion_state = 'visible'") : 0,
            'lastModified' => time(),
        ];

        \XF::registry()->set('nfTicketCounts', $cache);

        return $cache;
    }

    public function rebuildTicketMessagePositions($ticketId)
    {
        $db = $this->db();
        $db->query('SET @position := -1');
        $db->query("
            UPDATE xf_nf_tickets_message
            SET position = (@position := IF(message_state = 'visible', @position + 1, GREATEST(@position, 0)))
            WHERE ticket_id = ?
            ORDER BY message_date
        ", $ticketId);
    }

    public function removeGuestTicketAccess(\NF\Tickets\Entity\Ticket $ticket)
    {
        $ticketAuth = \XF::session()->get('nfGuestTicketAuth') ?: [];
        unset($ticketAuth[$ticket->ticket_id]);
        \XF::session()->set('nfGuestTicketAuth', $ticketAuth);
    }

    public function updateGuestTicketAccess(\NF\Tickets\Entity\Ticket $ticket)
    {
        if (\XF::visitor()->user_id)
        {
            throw new \BadMethodCallException("Attempt to update guest ticket access for non-guest visitor.");
        }

        $ticketAuth = \XF::session()->get('nfGuestTicketAuth') ?: [];
        $ticketAuth[$ticket->ticket_id] = ['nonce' => $this->generatePasswordAuthNonce($ticket->password)];
        \XF::session()->set('nfGuestTicketAuth', $ticketAuth);
    }

    public function generatePasswordAuthNonce($password)
    {
        /** @var \Closure $validator */
        $validator = $this->app()['csrf.validator'];

        return \XF::$time . ',' . $validator($password, \XF::$time);
    }

    public function getValidatedParticipants(\NF\Tickets\Entity\Ticket $ticket, $participants, &$error = null)
    {
        $error = null;

        if (is_string($participants))
        {
            $participants = Arr::stringToArray($participants, '#\s*,\s*#');
        }
        else if ($participants instanceof User)
        {
            $participants = [$participants];
        }

        if (!count($participants))
        {
            return [];
        }

        if ($participants instanceof \XF\Mvc\Entity\AbstractCollection)
        {
            $first = $participants->first();
        }
        else
        {
            $first = \reset($participants);
        }

        if ($first instanceof User)
        {
            $type = 'user';
        }
        else
        {
            $type = 'name';
        }

        foreach ($participants AS $k => $recipient)
        {
            if ($type === 'user' && !($recipient instanceof User))
            {
                throw new \InvalidArgumentException("Participant at key $k must be a user entity");
            }
        }

        if ($type === 'name')
        {
            /** @var \XF\Repository\User $userRepo */
            $userRepo = $this->repository('XF:User');
            $users = $userRepo->getUsersByNames($participants, $notFound);

            if ($notFound)
            {
                $error = \XF::phraseDeferred('nf_tickets_following_invitees_not_found_x', ['notFound' => \implode(', ', $notFound)]);
            }
        }
        else
        {
            $users = $participants;
        }

        $newParticipants = [];

        foreach ($users AS $user)
        {
            if ($user->user_id === $ticket->user_id)
            {
                $error = \XF::phraseDeferred('nf_tickets_you_cannot_invite_ticket_starter_to_ticket');
                continue;
            }

            if ($user->user_id === \XF::visitor()->user_id)
            {
                $error = \XF::phraseDeferred('nf_tickets_you_cannot_invite_yourself_to_ticket');
                continue;
            }

            $newParticipants[$user->user_id] = $user;
        }

        return $newParticipants;
    }

    public function getValidatedGuestParticipant(\NF\Tickets\Entity\Ticket $ticket, $email, $username, &$error = null)
    {
        $error = null;

        /** @var Email $emailValidator */
        $emailValidator = $this->app()->validator('email');
        if (!$emailValidator->isValid($email))
        {
            $error = "Please enter a valid email address";
            return null;
        }

        /** @var \XF\Validator\Username $usernameValidator */
        $usernameValidator = $this->app()->validator('username');
        $username = $usernameValidator->coerceValue($username);
        if (!$usernameValidator->isValid($username, $errorKey))
        {
            $error = $usernameValidator->getPrintableErrorValue($errorKey);
            return null;
        }

        /** @var Banning $shadowBannedRepo */
        $shadowBannedRepo = $this->repository('NF\Tickets:Banning');
        if ($shadowBannedRepo->isBanned($email, false))
        {
            $error = "Please enter a valid email address";
            return null;
        }

        /** @var \XF\Repository\User $userRepo */
        $userRepo = $this->repository('XF:User');

        if ($userRepo->getUserByNameOrEmail($email))
        {
            $error = 'A user with that email address exists. Please invite them as a user.';
            return null;
        }

        if ($ticket->getRelationFinder('GuestParticipants')->where('email', $email)->total())
        {
            $error = 'That guest is already a participant of this ticket.';
            return null;
        }

        return ['email' => $email, 'username' => $username];
    }
    
    public function insertParticipants(\NF\Tickets\Entity\Ticket $ticket, array $participantUsers, User $from = null)
    {
        /** @var \NF\Tickets\Entity\TicketParticipant[] $existingParticipants */
        $existingParticipants = $ticket->Participants->populate();
        $insertedActiveUsers = [];
        $inserted = 0;
        $fromUserId = $from ? $from->user_id : null;

        $this->db()->beginTransaction();

        /** @var User $user */
        foreach ($participantUsers AS $user)
        {
            if (isset($existingParticipants[$user->user_id]))
            {
                $participant = $existingParticipants[$user->user_id];
            }
            else
            {
                $participant = $ticket->getNewParticipant($user);
            }

            if ($participant->isInsert())
            {
                $insertedActiveUsers[] = $user->user_id;
            }

            $participant->save(true, false);
        }

        $this->db()->commit();

        return $insertedActiveUsers;
    }

    public function insertGuestParticipant(\NF\Tickets\Entity\Ticket $ticket, array $guestParticipant, User $from = null, &$password = null)
    {
        $guestParticipantEntity = $ticket->getNewGuestParticipant($guestParticipant['email'], $guestParticipant['username'], $password);
        $guestParticipantEntity->save();

        return $guestParticipantEntity;
    }

    /**
     * @param int $ticketId
     * @param int $startDate
     * @param string $binaryIp
     * @param string $email
     * @param int $days
     * @param int $limit
     * @return int
     */
    public function copySpamTriggerLogs($ticketId, $startDate, $binaryIp, $email, $days, $limit = 5)
    {
        $binaryIp = $binaryIp ?: '';
        $dateUpper = $startDate;
        $dateUpper = $dateUpper ?: \XF::$time;
        $dateLower = $dateUpper - ($days * 86400);

        $db = $this->db();
        $limit = intval($limit);
        $limitSQL = $limit ? "LIMIT {$limit}" : '';

        $orClause = [];
        $args = [$ticketId, $dateLower, $dateUpper];
        if ($binaryIp)
        {
            $orClause[] = 'log.ip_address = ?';
            $args[] = $binaryIp;
        }

        if ($email)
        {
            $orClause[] = '`user`.email = ?';
            $orClause[] = 'log.details like ?';
            $args[] = $email;
            $args[] = '%' . $email . '%';
        }

        if (!$orClause)
        {
            return 0;
        }
        $orSQL = \implode(' OR ', $orClause);

        $addOnsCache = \XF::app()->container('addon.cache');
        if (isset($addOnsCache['SV/SignupAbuseBlocking']))
        {
            $stmt = $db->query(
                "
            INSERT INTO xf_nf_tickets_spam_log (ticket_id, log_date, user_id, username, ip_address, result, details, request_state)
            SELECT DISTINCT ?, log.log_date, log.user_id, coalesce(`user`.username, `log`.username), log.ip_address, log.result, log.details, log.request_state
            FROM xf_sv_user_registration_log AS log
            LEFT JOIN xf_user AS `user` ON (log.user_id = `user`.user_id)
            WHERE 
                log.log_date >= ? AND log.log_date <= ?
                AND ({$orSQL})
            ORDER BY log.log_date DESC
            {$limitSQL}",
                $args
            );

            $rowsCopied = $stmt->rowsAffected();
            if ($rowsCopied)
            {
                return $rowsCopied;
            }
        }

        $stmt = $db->query(
            "
            INSERT INTO xf_nf_tickets_spam_log (ticket_id, log_date, user_id, username, ip_address, result, details, request_state)
            SELECT DISTINCT ?, log.log_date, log.user_id, `user`.username, log.ip_address, log.result, log.details, log.request_state
            FROM xf_spam_trigger_log AS log
            LEFT JOIN xf_user AS `user` ON (log.user_id = `user`.user_id)
            WHERE log.content_type in ('user', 'user_rejected')
                AND log.log_date >= ? AND log.log_date <= ?
                AND ({$orSQL})
            ORDER BY log.log_date DESC
            {$limitSQL}",
            $args
        );

        return $stmt->rowsAffected();
    }

	public function pruneTicketFields(\NF\Tickets\Entity\Ticket $ticket)
	{
		if (empty($ticket->custom_fields_))
		{
			return true;
		}

		$subsetUpdate = [];

		$fieldSet = $ticket->custom_fields;

		$expiryPhrase = \XF::phrase('nf_tickets_field_contents_expired')->render();

		foreach ($ticket->custom_fields AS $fieldId => $value)
		{
		    /** @var \NF\Tickets\Entity\TicketField $field */
			$field = \XF::finder('NF\Tickets:TicketField')
				->whereId($fieldId)
				->fetchOne();

			if (!$field)
			{
				// field doesn't exist anymore, nuke its old value to prevent save errors
				$fieldSet->removeFieldValue($fieldId);
				continue;
			}

			if ($field->expiry_length_type === 'never')
			{
				continue;
			}

			$pruneTime = \strtotime('+' . $field->expiry_length_unit . ' ' .$field->expiry_length_type, $ticket->start_date);

			if (\XF::$time >= $pruneTime)
			{
				$subsetUpdate[$fieldId] = $expiryPhrase;
			}
		}

		if (!empty($subsetUpdate))
		{
			$fieldSet->bulkSet(\array_merge($ticket->custom_fields_, $subsetUpdate), \array_keys($ticket->custom_fields_), 'admin');
			$ticket->save();
		}

		return true;
    }
}