<?php

namespace NF\Tickets\Search\Data;

use NF\Tickets\XF\Entity\User;
use SV\StandardLib\Helper;
use XF\Mvc\Entity\Entity;
use XF\Search\Data\AbstractData;
use XF\Search\IndexRecord;
use XF\Search\MetadataStructure;
use XF\Search\Query\MetadataConstraint;

class Message extends AbstractData
{
    public function getEntityWith($forView = false): array
    {
        $get = ['User', 'Ticket', 'Ticket.Category'];
        if ($forView)
        {
            $get[] = 'Ticket.Status';
            $visitor = \XF::visitor();
            $get[] = 'Ticket.Category.Permissions|' . $visitor->permission_combination_id;
        }

        return $get;
    }

    public function getIndexData(Entity $entity): ?IndexRecord
    {
        /** @var \NF\Tickets\Entity\Message $entity */
        $ticket = $entity->Ticket;
        if (!$ticket || !$ticket->Category)
        {
            return null;
        }

        if ($entity->isFirstMessage())
        {
            return $this->searcher->handler('nf_tickets_ticket')->getIndexData($ticket);
        }

        $index = IndexRecord::create('nf_tickets_message', $entity->message_id, [
            'title'         => '',
            'message'       => $entity->message,
            'date'          => $entity->message_date,
            'user_id'       => $entity->getAnonymousUser(false)->user_id,
            'discussion_id' => $entity->ticket_id,
            'metadata'      => $this->getMetaData($entity),
        ]);

        if (!$entity->isVisible())
        {
            $index->setHidden();
        }

        return $index;
    }

    protected function getMetaData(\NF\Tickets\Entity\Message $entity): array
    {
        $ticket = $entity->Ticket;

        $metadata = [
            'ticketcat' => $ticket->ticket_category_id,
            'ticket'    => $ticket->ticket_id,
            'ticket_status' => $ticket->status_id,
        ];

        if ($ticket->assigner_user_id)
        {
            $metadata['assigner_user'] = $ticket->assigner_user_id;
        }

        if ($ticket->assigned_user_id)
        {
            $metadata['assigned_user'] = $ticket->assigned_user_id;
        }

        return $metadata;
    }

    public function setupMetadataStructure(MetadataStructure $structure): void
    {
        $structure->addField('ticketcat', MetadataStructure::INT);
        $structure->addField('ticket', MetadataStructure::INT);
        $structure->addField('ticket_status', MetadataStructure::INT);
        $structure->addField('assigned_user', MetadataStructure::INT);
        $structure->addField('assigner_user', MetadataStructure::INT);
    }

    public function getResultDate(Entity $entity): int
    {
        /** @var \NF\Tickets\Entity\Message $entity */
        return $entity->message_date;
    }

    public function getTemplateData(Entity $entity, array $options = []): array
    {
        /** @var \NF\Tickets\Entity\Message $entity */
        return [
            'message' => $entity,
            'ticket'  => $entity->Ticket,
            'options' => $options,
        ];
    }

    public function canUseInlineModeration(Entity $entity, &$error = null): bool
    {
        /** @var \NF\Tickets\Entity\Message $entity */
        return $entity->canUseInlineModeration($error);
    }

    public function applyTypeConstraintsFromInput(\XF\Search\Query\Query $query, \XF\Http\Request $request, array &$urlConstraints)
    {
        if ($this->canUseSearchImprovementsFeatures())
        {
            $constraints = $request->filter([
                'c.assigned' => 'str',
                'c.assigner' => 'str',
            ]);

            $repo = \SV\SearchImprovements\Globals::repo();
            $repo->applyUserConstraint($query, $constraints, $urlConstraints,
                'c.assigned', 'assigned_user'
            );
            $repo->applyUserConstraint($query, $constraints, $urlConstraints,
                'c.assigner', 'assigner_user'
            );
        }

        $statusIds = $request->filter('c.nfTickets.statuses', 'array-uint');
        $statusIds = array_unique($statusIds);
        if (count($statusIds) !== 0 && !in_array(0, $statusIds))
        {
            $statues = $this->getStatusRepo()
                            ->findStatusesForList()
                            ->whereId($statusIds)
                            ->fetch();
            if ($statues->count() > 0)
            {
                $query->withMetadata('ticket_status', $statues->keys());
                $urlConstraints['statuses'] = $statusIds;
            }
            else
            {
                unset($urlConstraints['nfTickets']['statuses']);
            }
        }
        else
        {
            unset($urlConstraints['nfTickets']['statuses']);
        }

        $ticketId = $request->filter('c.nfTickets.ticket', 'uint');
        if ($ticketId)
        {
            $query->withMetadata('ticket', $ticketId);

            if ($query instanceof \XF\Search\Query\KeywordQuery)
            {
                $query->inTitleOnly(false);
            }
        }
        else
        {
            unset($urlConstraints['nfTickets']['ticket']);

            $categoryIds = $request->filter('c.nfTickets.categories', 'array-uint');
            $categoryIds = array_values(array_unique($categoryIds));
            if ($categoryIds && reset($categoryIds))
            {
                if ($request->filter('c.nfTickets.child_categories', 'bool'))
                {
                    $categoryRepo = $this->getCategoryRepo();
                    $categories = $categoryRepo->getViewableCategories();
                    $categoryTree = $categoryRepo->createCategoryTree($categories);

                    $searchIds = array_fill_keys($categoryIds, true);
                    $categoryTree->traverse(function(int $id, \NF\Tickets\Entity\Category $node) use (&$searchIds)
                    {
                        if (isset($searchIds[$id]) || isset($searchIds[$node->parent_category_id]))
                        {
                            // if we're in the search node list, the user selected the node explicitly
                            // if the parent is in the list, then that node was selected via traversal so we're included too
                            $searchIds[$id] = true;
                        }

                        // we still need to traverse children though, as children may be selected
                    });

                    $categoryIds = array_unique(array_keys($searchIds));
                }
                else
                {
                    unset($urlConstraints['nfTickets']['child_categories']);
                }

                $query->withMetadata('ticketcat', $categoryIds);
            }
            else
            {
                unset($urlConstraints['nfTickets']['categories']);
                unset($urlConstraints['nfTickets']['child_categories']);
            }
        }
    }

    public function getTypePermissionConstraints(\XF\Search\Query\Query $query, $isOnlyType): array
    {
        $categoryRepo = $this->getCategoryRepo();
        $with = [];
        $with[] = 'Permissions|' . \XF::visitor()->permission_combination_id;
        $categories = $categoryRepo->findCategoryList(null, $with)->fetch();

        $skip = [];
        foreach ($categories AS $category)
        {
            /** @var \NF\Tickets\Entity\Category $category */
            if (!$category->canView())
            {
                $skip[] = $category->ticket_category_id;
            }
        }

        if ($skip)
        {
            return [
                new MetadataConstraint('ticketcat', $skip, MetadataConstraint::MATCH_NONE)
            ];
        }
        else
        {
            return [];
        }
    }

    public function getSearchableContentTypes(): array
    {
        return ['nf_tickets_message', 'nf_tickets_ticket'];
    }

    public function getGroupByType(): string
    {
        return 'nf_tickets_ticket';
    }

    public function getSearchFormTab(): ?array
    {
        /** @var User $visitor */
        $visitor = \XF::visitor();
        if (!\is_callable([$visitor, 'canViewTickets']) || !$visitor->canViewTickets() || !$visitor->user_id)
        {
            return null;
        }

        return [
            'title' => \XF::phrase('nf_tickets_search_tickets'),
            'order' => 5000,
        ];
    }

    public function getSearchFormData(): array
    {
        $categoryRepo = $this->getCategoryRepo();
        $categories = $categoryRepo->getViewableCategories();
        $categoryTree = $categoryRepo->createCategoryTree($categories);
        $categoryExtras = $categoryRepo->getCategoryListExtras($categoryTree);

        return [
            //'priorities' => $priorityRepo->findPrioritiesForList()->fetch(),
            'statuses' => $this->getStatusRepo()->findStatusesForList()->fetch(),
            'categories' => $categories,
            'categoryTree' => $categoryTree,
            'categoryExtras' => $categoryExtras,
            'canUseSearchImprovementsFeatures' => $this->canUseSearchImprovementsFeatures(),
        ];
    }

    protected function canUseSearchImprovementsFeatures(): bool
    {
        return Helper::repo()->hasDesiredAddOnVersion('SV/SearchImprovements', "2.9.0");
    }

    public function canIncludeInResults(Entity $entity, array $resultIds): bool
    {
        /** @var \NF\Tickets\Entity\Message $entity */
        $id = $entity->ticket_id;
        $key = 'nf_tickets_ticket-'.$id;
        if (isset($resultIds[$key]) && $entity->isFirstMessage())
        {
            return false;
        }

        return true;
    }

    public function getSectionContext(): string
    {
        return 'nfTickets';
    }

    protected function getCategoryRepo(): \NF\Tickets\Repository\Category
    {
        /** @var \NF\Tickets\Repository\Category $repo */
        $repo = \XF::repository('NF\Tickets:Category');
        return $repo;
    }

    protected function getStatusRepo(): \NF\Tickets\Repository\Status
    {
        /** @var \NF\Tickets\Repository\Status $repo */
        $repo = \XF::repository('NF\Tickets:Status');
        return $repo;
    }
}