<?php

namespace NF\Tickets\Searcher;

use XF\Mvc\Entity\Finder;
use XF\Searcher\AbstractSearcher;

class Ticket extends AbstractSearcher
{
	protected $allowedRelations = ['Category'];

	protected $formats = [
		'title' => 'like',
		'username' => 'like',
		'start_date' => 'date',
		'last_message_date' => 'date'
	];

	protected $arrayValueKeys = [
		'ticket_field'
	];

	protected $whitelistOrder = [
		'title' => true,
		'username' => true,
		'start_date' => true,
	];

	protected $order = [['last_message_date', 'desc']];

	protected function getEntityType()
	{
		return 'NF\Tickets:Ticket';
	}

	protected function getDefaultOrderOptions()
	{
		return [
			'last_message_date' => \XF::phrase('forum_sort.last_message_date'),
			'start_date' => \XF::phrase('start_date'),
			'title' => \XF::phrase('title'),
		];
	}

    /** @noinspection PhpParameterByRefIsNotUsedAsReferenceInspection */
    protected function validateSpecialCriteriaValueAfter($key, &$value, $column, $format, $relation)
	{
		if ($key == 'started_in_last' || $key == 'last_message_in_last')
		{
			if (!is_array($value) || !isset($value['value']) || $value['value'] <= 0)
			{
				return false;
			}
		}
		if ($key == 'prefix_id' && $value == -1)
		{
			return false;
		}
		if ($key == 'ticket_category_id')
		{
			if (
				$value == 0
				|| (is_array($value) && isset($value[0]) && $value[0] == 0)
			)
			{
				return false;
			}
		}
		if (
			($key == 'starter_user_group_id' || $key == 'starter_not_user_group_id')
			&& !$value
		)
		{
			return false;
		}

		return null;
	}

	protected function applySpecialCriteriaValue(Finder $finder, $key, $value, $column, $format, $relation)
	{
		if ($key == 'ticket_category_id')
		{
			if (!is_array($value))
			{
				$value = [$value];
			}

			if (isset($value['search_type']) && $value['search_type'] === 'exclude')
			{
				$matchInCategories = false;
			}
			else
			{
				$matchInCategories = true;
			}
			unset($value['search_type']);

			$finder->where('ticket_category_id', $matchInCategories ? '=' : '<>', $value);

			return true;
		}

		if ($key == 'started_in_last' || $key == 'last_message_in_last')
		{
			$cutOff = $this->convertRelativeTimeToCutoff(
				$value['value'],
				$value['unit']
			);
			if ($cutOff)
			{
				$column = $key == 'last_message_in_last' ? 'start_date' : 'last_message_date';

				$finder->where($column, '>=', $cutOff);
			}
			return true;
		}

		if ($key == 'not_discussion_type')
		{
			$finder->where('discussion_type', '<>', $value);
			return true;
		}

		if ($key == 'ticket_field')
		{
			$exactMatchFields = !empty($value['exact']) ? $value['exact'] : []; // used for multi-choice field searches
			$customFields = $value + $exactMatchFields;
			unset($customFields['exact']);

			foreach ($customFields AS $fieldId => $value)
			{
				if ($value === '' || (is_array($value) && !$value))
				{
					continue;
				}

				$finder->with('CustomFields|' . $fieldId);
				$isExact = !empty($exactMatchFields[$fieldId]);
				$conditions = [];

				foreach ((array)$value AS $possible)
				{
					$columnName = 'CustomFields|' . $fieldId . '.field_value';
					if ($isExact)
					{
						$conditions[] = [$columnName, '=', $possible];
					}
					else
					{
						$conditions[] = [$columnName, 'LIKE', $finder->escapeLike($possible, '%?%')];
					}
				}

				if ($conditions)
				{
					$finder->whereOr($conditions);
				}
			}
		}

		if ($key == 'owner_user_group_id' || $key == 'owner_not_user_group_id')
		{
			if (!is_array($value))
			{
				$value = [$value];
			}

			$finder->with('User');

			$userGroupIdColumn = $finder->columnSqlName('User.user_group_id');
			$secondaryGroupIdsColumn = $finder->columnSqlName('User.secondary_group_ids');
			$positiveMatch = ($key == 'owner_user_group_id');
			$parts = [];

			// for negative matches, we default to allowing guests, but if they say "not the guest"
			// group, then we'll disable it
			$orIsGuest = $positiveMatch ? false : true;

			foreach ($value AS $userGroupId)
			{
				$quotedGroupId = $finder->quote($userGroupId);
				if ($positiveMatch)
				{
					$parts[] = "$userGroupIdColumn = $quotedGroupId "
						. "OR FIND_IN_SET($quotedGroupId, $secondaryGroupIdsColumn)";

					if ($userGroupId == \XF\Entity\User::GROUP_GUEST)
					{
						// if explicitly selecting the guest group, allow guest threads
						// as they're hard to filter for otherwise
						$parts[] = $finder->columnSqlName('user_id') . ' = 0';
					}
				}
				else
				{
					$parts[] = "$userGroupIdColumn <> $quotedGroupId "
						. "AND FIND_IN_SET($quotedGroupId, $secondaryGroupIdsColumn) = 0";

					if ($userGroupId == \XF\Entity\User::GROUP_GUEST)
					{
						$orIsGuest = false;
					}
				}
			}
			if ($parts)
			{
				$joiner = $positiveMatch ? ' OR ' : ' AND ';
				$sql = implode($joiner, $parts);
				if ($orIsGuest)
				{
					$sql = "($sql) OR " . $finder->columnSqlName('user_id') . ' = 0';
				}
				$finder->whereSql($sql);
			}
			return true;
		}

		return false;
	}

	public function getFormData()
	{
		/** @var \NF\Tickets\Repository\TicketPrefix $prefixRepo */
		$prefixRepo = $this->em->getRepository('NF\Tickets:TicketPrefix');
		$prefixes = $prefixRepo->getPrefixListData();

		/** @var \NF\Tickets\Repository\Category $categoryRepo */
		$categoryRepo = $this->em->getRepository('NF\Tickets:Category');
		$categories = $categoryRepo->getCategoryOptionsData(false);

		/** @var \XF\Repository\UserGroup $userGroupRepo */
		$userGroupRepo = $this->em->getRepository('XF:UserGroup');
		$userGroups = $userGroupRepo->findUserGroupsForList()->fetch();

		return [
			'prefixes' => $prefixes,
			'categories' => $categories,
			'userGroups' => $userGroups,
		];
	}

	public function getFormDefaults()
	{
		return [
			'prefix_id' => -1,
			'ticket_category_id' => 0,

			'reply_count' => ['end' => -1],

			'discussion_state' => ['visible', 'moderated', 'deleted', 'shadow_ban'],
			'ticket_locked' => [1, 0],
		];
	}
}