<?php

namespace NF\Tickets\Service\Message;

use NF\Tickets\Entity\Message;
use XF\Service\AbstractNotifier;

/**
 * Class Notifier
 *
 * @package NF\Tickets\Service\Message
 */
class Notifier extends AbstractNotifier
{
    /**
     * @var Message
     */
    protected $message;

    protected $actionType;

    /**
     * Notifier constructor.
     *
     * @param \XF\App $app
     * @param Message $message
     * @param         $actionType
     */
    public function __construct(\XF\App $app, Message $message, $actionType)
    {
        parent::__construct($app);

        switch ($actionType)
        {
            case 'reply':
            case 'ticket':
                break;

            default:
                throw new \InvalidArgumentException("Unknown action type '$actionType'");
        }

        $this->actionType = $actionType;
        $this->message = $message;
    }

    /**
     * @param array $extraData
     * @return null|\XF\Service\AbstractService
     */
    public static function createForJob(array $extraData)
    {
        /** @var Message $message */
        $message = \XF::app()->find('NF\Tickets:Message', $extraData['messageId'], ['Ticket', 'Ticket.Category']);
        if (!$message)
        {
            return null;
        }

        return \XF::service('NF\Tickets:Message\Notifier', $message, $extraData['actionType']);
    }

    /**
     * @return array
     */
    protected function getExtraJobData()
    {
        return [
            'messageId'  => $this->message->message_id,
            'actionType' => $this->actionType,
        ];
    }

    protected function loadNotifiers(): array
    {
        $notifiers = [
            'quote' => $this->app->notifier('NF\Tickets:Message\Quote', $this->message),
            'mention' => $this->app->notifier('NF\Tickets:Message\Mention', $this->message),
	        'queueWatch' => $this->app->notifier('NF\Tickets:Message\QueueWatch', $this->message, $this->actionType),
            'categoryWatch' => $this->app->notifier('NF\Tickets:Message\CategoryWatch', $this->message, $this->actionType),
        ];

        // if this is not the last message, then another notification would have been triggered already
        if ($this->message->isLastMessage())
        {
            $notifiers['ticketWatch'] = $this->app->notifier('NF\Tickets:Message\TicketWatch', $this->message, $this->actionType);
        }

        return $notifiers;
    }

    /**
     * @param array $users
     */
    protected function loadExtraUserData(array $users)
    {
        $permCombinationIds = [];
        foreach ($users AS $user)
        {
            $id = $user->permission_combination_id;
            $permCombinationIds[$id] = $id;
        }

        $this->app->permissionCache()->cacheMultipleContentPermsForContent(
            $permCombinationIds,
            'nf_tickets_category',
            $this->message->Ticket->Category->ticket_category_id
        );
    }

    /**
     * @param \XF\Entity\User $user
     * @return mixed
     * @throws \Exception
     */
    protected function canUserViewContent(\XF\Entity\User $user)
    {
        return \XF::asVisitor(
            $user,
            function () { return $this->message->canView(); }
        );
    }

    /**
     * @param \NF\Tickets\Entity\Category $category
     */
    public function skipUsersWatchingCategory(\NF\Tickets\Entity\Category $category)
    {
        $watchers = $this->db()->fetchAll('
            SELECT user_id, send_alert, send_email
            FROM xf_nf_tickets_category_watch
            WHERE ticket_category_id = ?
                AND (send_alert = 1 OR send_email = 1)
        ', $category->ticket_category_id);

        foreach ($watchers AS $watcher)
        {
            if ($watcher['send_alert'])
            {
                $this->setUserAsAlerted($watcher['user_id']);
            }

            if ($watcher['send_email'])
            {
                $this->setUserAsEmailed($watcher['user_id']);
            }
        }
    }
}