<?php

namespace NF\Tickets\Service\Message;

use NF\Tickets\Entity\Message;
use NF\Tickets\Repository\Banning;
use XF\Service\AbstractService;

use function count, is_array;

/**
 * Class Preparer
 *
 * @package NF\Tickets\Service\Message
 */
class Preparer extends AbstractService
{
    /**
     * @var Message
     */
    protected $message;

    /**
     * @var string
     */
    protected $attachmentHash;

    /**
     * @var bool|null|string
     */
    protected $logIp = true;

    /**
     * @var array
     */
    protected $quotedMessages = [];

    /**
     * @var array
     */
    protected $mentionedUsers = [];

    /** @var int */
    protected $minimumWordCount = 0;
    /** @var string */
    protected $minimumWordCountAction = 'allowed';
    /** @var array|bool|null */
    protected $shadowBanChecked = null;

    /**
     * Preparer constructor.
     *
     * @param \XF\App $app
     * @param Message $message
     */
    public function __construct(\XF\App $app, Message $message)
    {
        parent::__construct($app);

        $this->message = $message;
    }

    /**
     * @return Message
     */
    public function getMessage()
    {
        return $this->message;
    }

    /**
     * @param $logIp
     */
    public function logIp($logIp)
    {
        $this->logIp = $logIp;
    }

    /**
     * @return array
     */
    public function getQuotedMessages()
    {
        return $this->quotedMessages;
    }

    /**
     * @return array
     */
    public function getQuotedUserIds()
    {
        if (!$this->quotedMessages)
        {
            return [];
        }

        $messageIds = \array_keys($this->quotedMessages);
        $quotedUserIds = [];

        $db = $this->db();
        $messageUserMap = $db->fetchPairs('
            SELECT message_id, user_id
            FROM xf_nf_tickets_message
            WHERE message_id IN (' . $db->quote($messageIds) . ')
        ');

        foreach ($messageUserMap AS $messageId => $userId)
        {
            if (!isset($this->getQuotedMessages()[$messageId]) || !$userId)
            {
                continue;
            }

            $quote = $this->getQuotedMessages()[$messageId];
            if (!isset($quote['member']) || $quote['member'] === $userId)
            {
                $quotedUserIds[] = $userId;
            }
        }

        return $quotedUserIds;
    }

    /**
     * @param bool $limitPermissions
     * @return array
     */
    public function getMentionedUsers($limitPermissions = true)
    {
        if ($limitPermissions && $this->message)
        {
            /** @var \XF\Repository\User $userRepo */
            $userRepo = $this->repository('XF:User');

            $user = $this->message->User ?: $userRepo->getGuestUser();

            return $user->getAllowedUserMentions($this->mentionedUsers);
        }

        return $this->mentionedUsers;
    }

    /**
     * @param bool $limitPermissions
     * @return array
     */
    public function getMentionedUserIds($limitPermissions = true)
    {
        return \array_keys($this->getMentionedUsers($limitPermissions));
    }

    /**
     * @param string $message
     * @param bool   $format
     * @param bool   $checkValidity
     * @return bool
     */
    public function setMessage($message, $format = true, $checkValidity = true)
    {
        $preparer = $this->getMessagePreparer($format);
        $this->message->message = $preparer->prepare($message, $checkValidity);
        $this->message->embed_metadata = $preparer->getEmbedMetadata();

        $this->quotedMessages = $preparer->getQuotesKeyed('ticketMessage');
        $this->mentionedUsers = $preparer->getMentionedUsers();

        return $preparer->pushEntityErrorIfInvalid($this->message);
    }

    public function getMinimumWordCount(): int
    {
        return $this->minimumWordCount;
    }

    public function setMinimumWordCount(int $minimumWordCount)
    {
        $this->minimumWordCount = $minimumWordCount;
    }

    public function getMinimumWordCountAction(): string
    {
        return $this->minimumWordCountAction;
    }

    public function setMinimumWordCountAction(string $minimumWordCountAction)
    {
        $this->minimumWordCountAction = $minimumWordCountAction;
    }

    /**
     * @param bool $format
     * @return \XF\Service\Message\Preparer
     */
    protected function getMessagePreparer($format = true)
    {
        $options = $this->app->options();

        /** @var \XF\Service\Message\Preparer $preparer */
        $preparer = $this->service('XF:Message\Preparer', 'nf_tickets_message', $this->message);
        $preparer->setConstraint('maxLength', $options->nfTicketsMaxMessageLength);

        if (!$format)
        {
            $preparer->disableAllFilters();
        }

        return $preparer;
    }

    /**
     * @param string $hash
     */
    public function setAttachmentHash($hash)
    {
        $this->attachmentHash = $hash;
    }

    protected function getMessageWordCount(string $string, bool $stripBbCode = true, bool $stripQuote = true): int
    {
        if ($stripBbCode)
        {
            $string = \XF::app()->stringFormatter()->stripBbCode($string, [
                'stripQuote' => $stripQuote,
            ]);
            // remove non-visible placeholders
            $string = \str_replace('[*]', ' ', $string);
        }

        // ref: http://php.net/manual/de/function.str-word-count.php#107363
        return count(\preg_split('~[^\p{L}\p{N}\']+~u', $string, -1, PREG_SPLIT_NO_EMPTY));
    }

    /**
     * @return string
     */
    public function getAttachmentHash()
    {
        return $this->attachmentHash;
    }

    public function checkForShadowBan()
    {
        if ($this->shadowBanChecked !== null)
        {
            return;
        }

        $message = $this->message;
        $ticket = $this->message->Ticket;

        /** @var Banning $shadowBannedRepo */
        $shadowBannedRepo = $this->repository('NF\Tickets:Banning');
        if ($message->user_email)
        {
            $ref = $shadowBannedRepo->getShadowBanReference($message->user_email);
            if ($ref)
            {
                if ($ref[1] ?? false)
                {
                    $ticket->discussion_state = 'shadow_ban';
                }
                else
                {
                    $ticket->error(\XF::phrase('email_address_you_entered_has_been_banned_by_administrator'), 'user_email');
                }
                $this->shadowBanChecked = $ref;

                return;
            }
        }

        $this->shadowBanChecked = false;
    }

    public function checkForSpam()
    {
        $message = $this->message;
        $ticket = $this->message->Ticket;

        /** @var \XF\Repository\User $userRepo */
        $userRepo = $this->repository('XF:User');

        $user = $message->User ?: $userRepo->getGuestUser($message->username);

        if ($message->isFirstMessage())
        {
            $contentMessage = $ticket->title . "\n" . $message->message;
            $contentType = 'ticket';
        }
        else
        {
            $contentMessage = $message->message;
            $contentType = 'nf_tickets_message';
        }

        $checker = $this->app->spam()->contentChecker();
        $checker->check($user, $contentMessage, [
            'permalink'    => $this->app->router('public')->buildLink('canonical:tickets', $ticket),
            'content_type' => $contentType,
        ]);

        $this->checkForShadowBan();
        $isShadowBanned = $ticket->isShadowBanned();
        if ($isShadowBanned)
        {
            $checker->logDecision('ticketShadowBannedFromEmail', 'denied');
            $checker->logDetail('ticketShadowBannedFromEmail', 'nf_tickets_email_banned_from_ticket', [
                'email' => $message->user_email,
                'match' => is_array($this->shadowBanChecked) ? ($this->shadowBanChecked[0] ?? '*') : '*',
            ]);
        }

        if ($message->isFirstMessage())
        {
            if ($this->minimumWordCount > 0)
            {
                $wordCount = $this->getMessageWordCount($message->message);
                if ($wordCount < $this->minimumWordCount)
                {
                    $action = $this->getMinimumWordCountAction();
                    if ($action === 'shadow_ban')
                    {
                        $ticket->discussion_state = 'shadow_ban';
                        $checker->logDecision('ticketWordCount', 'denied');
                    }
                    else
                    {
                        $checker->logDecision('ticketWordCount', $action);
                    }
                    $checker->logDetail('ticketWordCount', 'nf_tickets_minimum_ticket_word_count', [
                        'WordCount'        => $wordCount,
                        'MinimumWordCount' => $this->minimumWordCount,
                    ]);
                }
            }

            if ($isShadowBanned)
            {
                $checker->logDecision('ticketShadowBannedFromEmail', 'denied');
                $checker->logDetail('ticketShadowBannedFromEmail', 'nf_tickets_email_banned_from_ticket', [
                    'email' => $message->user_email,
                    'match' => $ref[0] ?? '*',
                ]);
            }
        }
        else if ($isShadowBanned && \XF::options()->nftSpamRejectRepliesToShadowBannedTickets)
        {
            $checker->logDecision('shadowBannedTicket', 'denied');
            $checker->logDetail('shadowBannedTicket', 'nf_tickets_reject_reply_to_shadow_banned_ticket');
        }

        $decision = $checker->getFinalDecision();
        switch ($decision)
        {
            case 'allow':
                if ($ticket->discussion_state === 'shadow_ban')
                {
                    // logSpamTrigger does not log if it is allowed, so force a denied and then continue
                    $checker->logDecision('ticketShadowBanned', 'denied');
                    $checker->logSpamTrigger( 'ticket', null);
                }
                break;

            case 'moderated':

                if ($message->isFirstMessage())
                {
                    $ticket->discussion_state = 'moderated';
                }
                else
                {
                    $message->message_state = 'hidden';
                }
                break;

            case 'denied':
                $checker->logSpamTrigger($message->isFirstMessage() ? 'ticket' : 'nf_tickets_message', null);
                if ($ticket->discussion_state !== 'shadow_ban')
                {
                    $message->error(\XF::phrase('your_content_cannot_be_submitted_try_later'));
                }
                break;
        }
    }

    public function afterInsert()
    {
        if ($this->attachmentHash)
        {
            $this->associateAttachments($this->attachmentHash);
        }

        if ($this->logIp)
        {
            $ip = ($this->logIp === true ? $this->app->request()->getIp() : $this->logIp);
            $this->writeIpLog($ip);
        }

        $message = $this->message;
        $ticket = $this->message->Ticket;
        $checker = $this->app->spam()->contentChecker();

        if ($message->isFirstMessage())
        {
            $contentType = 'ticket';
            $contentId = $ticket->ticket_id;
            $ticket->logSpamTriggerRecords();
        }
        else
        {
            $contentType = 'nf_tickets_message';
            $contentId = $message->message_id;
        }

        $checker->logContentSpamCheck($contentType, $contentId);
        $checker->logSpamTrigger($contentType, $contentId);
    }

    public function afterUpdate()
    {
        if ($this->attachmentHash)
        {
            $this->associateAttachments($this->attachmentHash);
        }

        $message = $this->message;
        $ticket = $this->message->Ticket;
        $checker = $this->app->spam()->contentChecker();

        if ($message->isFirstMessage())
        {
            $contentType = 'ticket';
            $contentId = $ticket->ticket_id;
        }
        else
        {
            $contentType = 'nf_tickets_message';
            $contentId = $message->message_id;
        }

        $checker->logSpamTrigger($contentType, $contentId);
    }

    /**
     * @param $hash
     */
    protected function associateAttachments($hash)
    {
        $message = $this->message;

        /** @var \XF\Service\Attachment\Preparer $inserter */
        $inserter = $this->service('XF:Attachment\Preparer');
        $associated = $inserter->associateAttachmentsWithContent($hash, 'nf_tickets_message', $message->message_id);
        if ($associated)
        {
            $message->fastUpdate('attach_count', $message->attach_count + $associated);
        }
    }

    /**
     * @param $ip
     */
    protected function writeIpLog($ip)
    {
        $message = $this->message;

        /** @var \XF\Repository\IP $ipRepo */
        $ipRepo = $this->repository('XF:Ip');
        $ipEnt = $ipRepo->logIp($message->user_id, $ip, 'nf_tickets_message', $message->message_id);
        if ($ipEnt)
        {
            $message->fastUpdate('ip_id', $ipEnt->ip_id);
        }
    }
}