<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Category;
use NF\Tickets\Entity\Message;
use XF\Entity\ConversationMaster;
use XF\Entity\ConversationMessage;
use XF\Entity\Ip;
use XF\Mvc\Entity\AbstractCollection;

class ConversationConverter extends Creator
{
    /** @var ConversationMaster */
    protected $conversation;

    public function __construct(\XF\App $app, ConversationMaster $conversation, Category $category)
    {
        parent::__construct($app, $category);

        $this->setConversation($conversation);
    }

    public function setConversation(ConversationMaster $conversation)
    {
        $this->ticket->first_message_date = $this->ticket->start_date = $conversation->FirstMessage->message_date;
        $this->conversation = $conversation;

        $this->setUser($conversation->Starter);

        $this->setContent($this->conversation->title, $this->conversation->FirstMessage->message);
        $this->copyAttachments($conversation->FirstMessage, $tempHash);
        $this->setAttachmentHash($tempHash);
    }

    protected function _save()
    {
        \XF::db()->beginTransaction();

        $response = parent::_save();

        $convMessage = $this->conversation->FirstMessage;
        $this->rewriteAttachmentBbCode($convMessage, $this->message);
        // First post reactions, needs a saved message.
        $this->copyReactions($convMessage, $this->message);
        $this->copyIp($convMessage, $this->message);

        /** @var \XF\Finder\ConversationMessage $finder */
        $finder = $this->finder('XF:ConversationMessage');

        $conversationReplies = $finder->inConversation($this->conversation)->where('message_id', '!=', $this->conversation->first_message_id)->fetch();
        $this->addReplies($conversationReplies);

        \XF::db()->commit();

        return $response;
    }

    /**
     * @param AbstractCollection|ConversationMessage[] $messages
     */
    protected function addReplies($messages)
    {
        foreach ($messages AS $post)
        {
            $this->addReply($post);
        }
    }

    protected function addReply(ConversationMessage $conversationMessage)
    {
        $this->copyAttachments($conversationMessage, $tempHash);

        \XF::asVisitor($conversationMessage->User, function () use ($conversationMessage, $tempHash) {
            /** @var Replier $replier */
            $replier = $this->service('NF\Tickets:Ticket\Replier', $this->ticket);
            $replier->logIp(false);
            $replier->setMessage($conversationMessage->message);
            $replier->setAttachmentHash($tempHash);

            if (!$conversationMessage->user_id && !$this->ticket->user_id && $this->ticket->username === $conversationMessage->username)
            {
                $replier->setGuestUser($conversationMessage->username, $this->ticket->user_email);
            }

            // TODO: Conversation messages don't have a position... make sure this doesn't break ticket.

            $message = $replier->getMessage();
            $message->message_date = $conversationMessage->message_date;

            $replier->save();

            $this->rewriteAttachmentBbCode($conversationMessage, $message);
            $message->fastUpdate('message_date', $conversationMessage->message_date); // ffs 3
            $this->copyReactions($conversationMessage, $message);
            $this->copyIp($conversationMessage, $message);

            return $message;
        });
    }

    /**
     * @param ConversationMessage $conversationMessage
     * @param                     $tempHash
     * @throws \XF\PrintableException
     */
    protected function copyAttachments(ConversationMessage $conversationMessage, &$tempHash = null)
    {
        $tempHash = $tempHash ?: \XF::generateRandomString(32);

        foreach ($conversationMessage->Attachments AS $attachment)
        {
            /** @var \XF\Entity\Attachment $newAttachment */
            $newAttachment = $this->em()->create('XF:Attachment');
            $newAttachment->data_id = $attachment->data_id;
            $newAttachment->content_type = 'nf_tickets_message';
            $newAttachment->temp_hash = $tempHash;
            $newAttachment->save();
        }
    }

    protected function rewriteAttachmentBbCode(ConversationMessage $source, Message $target)
    {
        if (!$target->attach_count)
        {
            return;
        }

        // build a mapping between old & new based on data_id that should be shared
        $oldAttachment = [];
        foreach($source->Attachments as $attachment)
        {
            $oldAttachment[$attachment->data_id] = $attachment->attachment_id;
        }

        $message = $target->message;

        $linkToMetaData = false;
        $embedMetadata = $target->embed_metadata ?: [];
        foreach($target->Attachments as $attachment)
        {
            if (empty($oldAttachment[$attachment->data_id]))
            {
                continue;
            }
            $newAttachmentId = $attachment->attachment_id;
            $oldAttachmentId = $oldAttachment[$attachment->data_id];

            /** @noinspection RegExpRedundantEscape */
            $message = \preg_replace_callback(
                '#(\[attach[^\]]*\])' . $oldAttachmentId . '(\[/attach\])#i',
                function ($matches) use ($newAttachmentId, &$linkToMetaData) {
                    $linkToMetaData = true;

                    return "{$matches[1]}{$newAttachmentId}{$matches[2]}";
                },
                $message
            );

            unset($embedMetadata['attachments'][$oldAttachmentId]);
            if ($linkToMetaData)
            {
                $embedMetadata['attachments'][$newAttachmentId] = $newAttachmentId;
            }
        }

        if ($message !== $target->message)
        {
            if ($embedMetadata)
            {
                $target->fastUpdate('embed_metadata', $embedMetadata);
            }
            $target->fastUpdate('message', $message);
        }
    }

    protected function copyReactions(ConversationMessage $source, Message $target)
    {
        $oldTime = \XF::$time;
        try
        {
            foreach ($source->Reactions AS $reactionContent)
            {
                \XF::$time = $reactionContent->reaction_date;
                \XF::asVisitor($reactionContent->ReactionUser, function () use ($reactionContent, $target) {
                    if ($this->message->canReact())
                    {
                        /** @var \XF\Repository\Reaction $repo */
                        $repo = $this->repository('XF:Reaction');
                        $repo->reactToContent($reactionContent->reaction_id, $target->getEntityContentType(), $target->getEntityContentTypeId(), \XF::visitor());
                    }
                });
            }
        }
        finally
        {
            \XF::$time = $oldTime;
        }
    }

    protected function copyIp(ConversationMessage $source, Message $target)
    {
        if ($source->ip_id)
        {
            /** @var Ip $existingIpLog */
            $existingIpLog = $this->em()->find('XF:Ip', $source->ip_id);
            if ($existingIpLog)
            {
                /** @var \XF\Repository\Ip $ipRepo */
                $ipRepo = $this->repository('XF:Ip');
                $ipLog = $ipRepo->logIp($source->user_id, $existingIpLog->ip, $target->getEntityContentType(), $target->getEntityContentTypeId());

                if ($ipLog)
                {
                    $ipLog->fastUpdate('log_date', $target->message_date);
                    $target->fastUpdate('ip_id', $ipLog->ip_id);
                }
            }
        }
    }

    public function deleteConversation()
    {
        // normally conversations are deleted when all users have left,
        // this allows overriding of when a conversation is deleted
        foreach ($this->conversation->Recipients as $recipient)
        {
            $recipient->recipient_state = 'deleted';
            $recipient->saveIfChanged();
        }
    }
}