<?php

namespace NF\Tickets\Service\Ticket\Converter;

use NF\Tickets\Entity\Message;
use NF\Tickets\Entity\Ticket;
use XF\Entity\Forum;
use XF\Entity\Ip;
use XF\Entity\Post;
use XF\Mvc\Entity\AbstractCollection;
use XF\Service\Thread\Creator;

class ToThread extends Creator
{
    /** @var Ticket */
    protected $sourceTicket;

    public function __construct(\XF\App $app, Ticket $ticket, Forum $destination)
    {
        parent::__construct($app, $destination);

        $this->setSourceTicket($ticket);
    }

    public function setSourceTicket(Ticket $ticket)
    {
        $this->post->post_date = $this->thread->post_date = $ticket->start_date;
        $this->sourceTicket = $ticket;

        $this->setIsAutomated();

        $user = $this->sourceTicket->User ?? null;

        if ($user)
        {
            $this->setUser($this->sourceTicket->User);
        }
        else
        {
            /** @var \XF\Repository\User $userRepo */
            $userRepo = $this->repository('XF:User');
            $this->setUser($userRepo->getGuestUser($this->sourceTicket->username));
        }

        $this->setContent($ticket->title, $ticket->FirstMessage->message);
        $this->copyAttachments($ticket->FirstMessage, $tempHash);
        $this->setAttachmentHash($tempHash);
    }

    protected function finalSetup()
    {
        $this->thread->last_post_date = $this->sourceTicket->last_message_date;
        $this->thread->last_post_user_id = $this->sourceTicket->last_message_user_id;
        $this->thread->last_post_username = $this->sourceTicket->last_message_username;

        $this->post->position = 0;
    }

    protected function _save()
    {
        \XF::db()->beginTransaction();

        $response = parent::_save();

        $firstMessage = $this->sourceTicket->FirstMessage;
        $this->rewriteAttachmentBbCode($firstMessage, $this->post);
        // First post reactions, needs a saved message.
        $this->copyReactions($firstMessage, $this->post);
        $this->copyIp($firstMessage, $this->post);

        /** @var \NF\Tickets\Finder\Message $ticketMessageFinder */
        $ticketMessageFinder = $this->finder('NF\Tickets:Message');
        $ticketReplies = $ticketMessageFinder->inTicket($this->sourceTicket)
                                             ->where('message_id', '!=', $this->sourceTicket->first_message_id)->fetch();
        $this->addReplies($ticketReplies);

        \XF::db()->commit();

        return $response;
    }

    /**
     * @param AbstractCollection|Message[] $messages
     */
    public function addReplies($messages)
    {
        \XF::db()->beginTransaction();

        foreach ($messages AS $message)
        {
            $this->addReply($message);
        }

        \XF::db()->commit();
    }

    public function addReply(Message $message)
    {
        \XF::db()->beginTransaction();

        $this->copyAttachments($message, $tempHash);

        \XF::asVisitor($message->User, function () use ($message, $tempHash) {
            /** @var \XF\Service\Thread\Replier $replier */
            $replier = $this->service('XF:Thread\Replier', $this->thread);
            $replier->setIsAutomated();
            $replier->setMessage($message->message);
            $replier->setAttachmentHash($tempHash);

            $post = $replier->getPost();
            $post->position = $message->position;

            $replier->save();

            $this->rewriteAttachmentBbCode($message, $post);
            $post->fastUpdate('post_date', $message->message_date); // ffs
            $this->copyReactions($message, $post);
            $this->copyIp($message, $post);

            return $post;
        });

        \XF::db()->commit();
    }

    protected function copyIp(Message $source, Post $target)
    {
        if ($source->ip_id)
        {
            /** @var Ip $existingIpLog */
            $existingIpLog = $this->em()->find('XF:Ip', $source->ip_id);
            if ($existingIpLog)
            {
                /** @var \XF\Repository\Ip $ipRepo */
                $ipRepo = $this->repository('XF:Ip');
                $ipLog = $ipRepo->logIp($source->user_id, $existingIpLog->ip, $target->getEntityContentType(), $target->post_id);

                if ($ipLog)
                {
                    $ipLog->fastUpdate('log_date', $target->post_date);
                    $target->fastUpdate('ip_id', $ipLog->ip_id);
                }
            }
        }
    }

    protected function copyAttachments(Message $ticketMessage, &$tempHash = null)
    {
        $tempHash = $tempHash ?: \XF::generateRandomString(32);

        foreach ($ticketMessage->Attachments AS $attachment)
        {
            /** @var \XF\Entity\Attachment $newAttachment */
            $newAttachment = $this->em()->create('XF:Attachment');
            $newAttachment->data_id = $attachment->data_id;
            $newAttachment->content_type = 'post';
            $newAttachment->temp_hash = $tempHash;
            $newAttachment->save();
        }
    }

    protected function rewriteAttachmentBbCode(Message $source, Post $target)
    {
        if (!$target->attach_count)
        {
            return;
        }

        // build a mapping between old & new based on data_id that should be shared
        $oldAttachment = [];
        foreach ($source->Attachments as $attachment)
        {
            $oldAttachment[$attachment->data_id] = $attachment->attachment_id;
        }

        $message = $target->message;

        $linkToMetaData = false;
        $embedMetadata = $target->embed_metadata ?: [];
        foreach ($target->Attachments as $attachment)
        {
            if (empty($oldAttachment[$attachment->data_id]))
            {
                continue;
            }
            $newAttachmentId = $attachment->attachment_id;
            $oldAttachmentId = $oldAttachment[$attachment->data_id];

            /** @noinspection RegExpRedundantEscape */
            $message = \preg_replace_callback(
                '#(\[attach[^\]]*\])' . $oldAttachmentId . '(\[/attach\])#i',
                function ($matches) use ($newAttachmentId, &$linkToMetaData) {
                    $linkToMetaData = true;

                    return "{$matches[1]}{$newAttachmentId}{$matches[2]}";
                },
                $message
            );

            unset($embedMetadata['attachments'][$oldAttachmentId]);
            if ($linkToMetaData)
            {
                $embedMetadata['attachments'][$newAttachmentId] = $newAttachmentId;
            }
        }

        if ($message !== $target->message)
        {
            if ($embedMetadata)
            {
                $target->fastUpdate('embed_metadata', $embedMetadata);
            }
            $target->fastUpdate('message', $message);
        }
    }

    /**
     * @param Message $source
     * @param Post    $target
     * @throws \Exception
     */
    protected function copyReactions(Message $source, Post $target)
    {
        foreach ($source->Reactions AS $reactionContent)
        {
            \XF::asVisitor($reactionContent->ReactionUser, function () use ($reactionContent, $target) {
                if ($target->canReact())
                {
                    /** @var \XF\Repository\Reaction $reactionRepo */
                    $reactionRepo = $this->repository('XF:Reaction');
                    $reactionRepo->reactToContent($reactionContent->reaction_id, $target->getEntityContentType(), $target->getEntityContentTypeId(), \XF::visitor());
                }
            });
        }
    }

    public function deleteTicket()
    {
        $this->sourceTicket->delete();
    }
}