<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Category;
use NF\Tickets\Entity\Message;
use NF\Tickets\Entity\Priority;
use NF\Tickets\Entity\Ticket;
use NF\Tickets\Mail\Mail;
use XF\Entity\User;
use XF\Service\AbstractService;
use XF\Entity\Thread AS ThreadEntity;
use XF\Service\Thread\Creator AS ThreadCreator;
use XF\Service\ValidateAndSavableTrait;

use function strlen;

/**
 * Class Creator
 *
 * @package NF\Tickets\Service\Ticket
 */
class Creator extends AbstractService
{
    use ValidateAndSavableTrait;

    /**
     * @var Category
     */
    protected $category;

    /**
     * @var Ticket
     */
    protected $ticket;

    /**
     * @var Message
     */
    protected $message;

    /**
     * @var \NF\Tickets\XF\Entity\User
     */
    protected $user;

    protected $proxyTicket = false;

    /**
     * @var \NF\Tickets\Service\Message\Preparer
     */
    protected $messagePreparer;

    /**
     * @var ThreadCreator
     */
    protected $threadCreator;

    protected $performValidations = true;

    protected $password;

    /** @var bool */
    protected $anonymized;

    /**
     * @var User[]
     */
    protected $notifyUsers = [];

    /**
     * Creator constructor.
     *
     * @param \XF\App $app
     * @param Category $category
     */
    public function __construct(\XF\App $app, Category $category)
    {
        parent::__construct($app);
        $this->category = $category;
        $this->setupDefaults();
    }

    public function setupDefaults()
    {
        $this->ticket = $this->category->getNewTicket();
        $this->message = $this->ticket->getNewMessage();

        $this->messagePreparer = $this->service('NF\Tickets:Message\Preparer', $this->message);
        $this->messagePreparer->setMinimumWordCount(\XF::options()->nftMinimumWordCount ?? 0);
        $this->messagePreparer->setMinimumWordCountAction(\XF::options()->nftMinimumWordCountAction ?? 'allowed');

        $this->ticket->addCascadedSave($this->message);
        $this->message->hydrateRelation('Ticket', $this->ticket);

        $this->setUser(\XF::visitor());
        $this->setPriority((\XF::options()->nfTicketsDefaultPriority ?? null) ?: null);
        $this->setStatus(\XF::options()->nfTicketsDefaultStatus ?? 1);
    }

    public function setPrefix($prefixId)
    {
        $this->ticket->prefix_id = $prefixId;
    }

    public function createForMember(User $user)
    {
        $this->setTicketUser($user);

        $this->proxyTicket = true;
    }

    public function createForGuest($name, $email)
    {
        $this->setTicketGuestUser($name, $email);

        $this->proxyTicket = true;
    }

    /**
     * @return Category
     */
    public function getCategory()
    {
        return $this->category;
    }

    /**
     * @return Ticket
     */
    public function getTicket()
    {
        return $this->ticket;
    }

    /**
     * @return Message
     */
    public function getMessage()
    {
        return $this->message;
    }

    /**
     * @return \NF\Tickets\XF\Entity\User
     */
    public function getUser()
    {
        return $this->user;
    }

    /**
     * @return \NF\Tickets\Service\Message\Preparer
     */
    public function getMessagePreparer()
    {
        return $this->messagePreparer;
    }

    /**
     * @param User $user
     */
    protected function setUser(User $user)
    {
        $this->user = $user;

        $this->ticket->starter_user_id = $user->user_id;
        $this->ticket->user_id = $user->user_id;
        $this->ticket->username = $user->username;
        $this->ticket->user_name = $user->username;

        $this->message->user_id = $user->user_id;
        $this->message->username = $user->username;
        $this->message->user_name = $user->username;

        if ($user->email)
        {
            $this->ticket->user_email = $user->email;
            $this->message->user_email = $user->email;
        }
    }

    /**
     * @param User $user
     */
    protected function setTicketUser(User $user)
    {
        $this->user = $user;

        $this->ticket->starter_user_id = \XF::visitor()->user_id;
        $this->ticket->user_id = $user->user_id;
        $this->ticket->username = $user->username;
        $this->ticket->user_name = $user->username;

        if ($user->email)
        {
            $this->ticket->user_email = $user->email;
        }
    }

    protected function validateGuest($name, $email)
    {
        /** @var \XF\Validator\Username $usernameValidator */
        $usernameValidator = \XF::app()->validator('Username');
        $usernameValidator->setOption('admin_edit', true);
        $usernameValidator->setOption('check_unique', false);
        $name = $usernameValidator->coerceValue($name);
        if (!$usernameValidator->isValid($name, $errorKey))
        {
            $error = $usernameValidator->getPrintableErrorValue($errorKey);
            $this->ticket->error($error, 'username');
        }

        /** @var \XF\Validator\Email $emailValidator */
        $emailValidator = \XF::app()->validator('Email');
        // $emailValidator->setOption('banned', $this->app->container('bannedEmails'));
        $email = $emailValidator->coerceValue($email);
        if (!$email || !$emailValidator->isValid($email))
        {
            $error = "Please enter a valid email address";
            $this->ticket->error($error, 'email');
        }
    }

    public function setGuestUser($name, $email)
    {
        $this->ticket->username = $name;
        $this->ticket->user_name = $name;

        $this->message->username = $name;
        $this->message->user_name = $name;

        if ($email)
        {
            $this->ticket->user_email = $email;
            $this->message->user_email = $email;
        }

        $this->validateGuest($name, $email);
    }

    public function setTicketGuestUser($name, $email)
    {
        $this->ticket->starter_user_id = \XF::visitor()->user_id;
        $this->ticket->user_id = 0;
        $this->ticket->username = $name;
        $this->ticket->user_name = $name;

        if ($email)
        {
            $this->ticket->user_email = $email;
        }

        $this->validateGuest($name, $email);
    }

    /**
     * @param bool $perform
     */
    public function setPerformValidations($perform)
    {
        $this->performValidations = (bool)$perform;
    }

    /**
     * @return bool
     */
    public function getPerformValidations()
    {
        return $this->performValidations;
    }

    public function setIsAutomated()
    {
        $this->logIp(false);
        $this->setPerformValidations(false);
    }

    /**
     * @param $logIp
     */
    public function logIp($logIp)
    {
        $this->messagePreparer->logIp($logIp);
    }

	public function setUserAgent($userAgent): void
	{
		$this->message->user_agent = $userAgent;
	}

    /**
     * @param string $title
     * @param string $message
     * @param bool $format
     *
     * @return bool
     */
    public function setContent($title, $message, $format = true)
    {
        $this->ticket->set('title', $title,
            ['forceConstraint' => $this->getPerformValidations()]
        );

        return $this->messagePreparer->setMessage($message, $format, $this->getPerformValidations());
    }

    /**
     * @param string $hash
     */
    public function setAttachmentHash($hash)
    {
        $this->messagePreparer->setAttachmentHash($hash);
    }

    public function setPriority($priority)
    {
        if ($priority instanceof Priority)
        {
            $this->ticket->priority = $priority->priority_id;
        }
        else
        {
            $this->ticket->priority = $priority;
        }
    }

    public function setStatus($status)
    {
        $this->ticket->status_id = $status;
    }

    /**
     * @param bool $locked
     */
    public function setLocked($locked)
    {
        $this->ticket->ticket_locked = (bool)$locked;
    }

    /**
     * @param array $customFields
     */
    public function setCustomFields(array $customFields)
    {
        $ticket = $this->ticket;

        $fieldSet = $ticket->custom_fields;
        $fieldDefinition = $fieldSet->getDefinitionSet()
            ->filterEditable($fieldSet, 'user')
            ->filterOnly($this->category->field_cache);

        $customFieldsShown = \array_keys($fieldDefinition->getFieldDefinitions());

        if ($customFieldsShown)
        {
            $fieldSet->bulkSet($customFields, $customFieldsShown);
        }
    }

    /**
     * @param bool $anonymize
     */
    public function setAnonymize($anonymize)
    {
        $this->anonymized = (bool)$anonymize;
        $this->message->anonymized = (bool)$anonymize;
    }

    /**
     * @return bool
     */
    public function isAnonymize()
    {
        return $this->anonymized;
    }

    public function checkForSpam()
    {
        $visitor = \XF::visitor();
        $this->messagePreparer->checkForShadowBan();
        if ($visitor->isSpamCheckRequired())
        {
            $this->messagePreparer->checkForSpam();
        }
    }

    protected function finalSetup()
    {
        if ($this->performValidations)
        {
            $this->ticket->start_date = \XF::$time;
            $this->ticket->first_message_date = \XF::$time;
            $this->ticket->last_message_date = \XF::$time;
            $user = $this->message->getAnonymousUser(false);
            $this->ticket->last_message_user_id = $user->user_id;
            $this->ticket->last_message_username = $user->username;

            $this->message->message_date = \XF::$time;
        }

        $this->message->position = 0;

        if (!$this->password)
        {
            $this->password = $this->getTicketRepo()->generatePassword();
            $this->ticket->set('password', password_hash($this->password, PASSWORD_DEFAULT));
        }
    }

    /**
     * @return array
     */
    protected function _validate()
    {
        $this->finalSetup();

        $ticket = $this->ticket;

        if (!$ticket->user_id)
        {
            /** @var \XF\Validator\Username $validator */
            $validator = $this->app->validator('Username');
            $validator->setOption('check_unique', false);
            $ticket->username = $validator->coerceValue($ticket->username);
            $this->message->username = $ticket->username;

            if ($this->performValidations && !$validator->isValid($ticket->username, $error))
            {
                return [
                    $validator->getPrintableErrorValue($error)
                ];
            }
        }

        $ticket->preSave();
        $errors = $ticket->getErrors();

        if ($this->performValidations)
        {
            if (!$ticket->prefix_id
                && $this->category->require_prefix
                && $this->category->getUsablePrefixes()
            )
            {
                $errors[] = \XF::phraseDeferred('please_select_a_prefix');
            }
        }

        return $errors;
    }

    /**
     * @return Ticket
     * @throws \XF\PrintableException
     */
    protected function _save()
    {
        $category = $this->getCategory();
        $ticket = $this->getTicket();
        $message = $this->getMessage();
        if (!strlen($this->user->username))
        {
            $this->user->setReadOnly(false);
            $this->user->setTrusted('username', $this->ticket->username);
            $this->user->setReadOnly(true);
        }

        $db = $this->db();
        $db->beginTransaction();

        $ticket->save(true, false);
        $isShadowBanned = $this->ticket->isShadowBanned();

        $ticket->fastUpdate([
            'first_message_id' => $message->message_id,
            'last_message_id' => $message->message_id
        ]);

        $this->messagePreparer->afterInsert();

        if (!$isShadowBanned && $category->thread_node_id && $category->ThreadForum)
        {
            /** @var ThreadCreator $creator */
            $creator = \XF::asVisitor($this->user, function() use ($category) {
                return $this->setupTicketThreadCreation($category->ThreadForum);
            });
            if ($creator && $creator->validate())
            {
                /** @var ThreadEntity $thread */
                $thread = $creator->save();
                $thread->fastUpdate('discussion_type', 'nf_tickets_ticket');
                $ticket->fastUpdate('thread_id', $thread->thread_id);
                $this->threadCreator = $creator;

                $this->afterTicketThreadCreated($thread);
            }
        }

        $db->commit();

        if ($this->proxyTicket && $this->ticket->user_id)
        {
            $this->notifyUsers[] = $this->ticket->user_id;
        }

        $ticket->quickEscalate();

        return $ticket;
    }

    protected function setupTicketThreadCreation(\XF\Entity\Forum $forum)
    {
        /** @var ThreadCreator $creator */
        $creator = $this->service('XF:Thread\Creator', $forum);
        $creator->setIsAutomated();

        $creator->setContent($this->ticket->getExpectedThreadTitle(), $this->getThreadMessage(), false);
        $creator->setPrefix($this->category->thread_prefix_id);

        $thread = $creator->getThread();
        $thread->hydrateRelation('Ticket', $this->ticket);
        $thread->discussion_type = 'nf_tickets_ticket';
        $thread->discussion_state = $this->ticket->discussion_state;

        return $creator;
    }

    protected function getThreadMessage()
    {
        $ticket = $this->ticket;

        $snippet = $this->app->bbCode()->render(
            $this->app->stringFormatter()->wholeWordTrim($this->ticket->FirstMessage->message, 500),
            'bbCodeClean',
            'nf_tickets_message',
            null
        );

        $phrase = \XF::phrase('nf_tickets_ticket_thread_create', [
            'title' => $ticket->title,
            'link' => $this->app->router('public')->buildLink('canonical:tickets', $this->ticket),
            'username' => $ticket->user_name,
            'snippet' => $snippet
        ]);

        return $phrase->render('raw');
    }

    protected function afterTicketThreadCreated(ThreadEntity $thread)
    {
        /** @var \XF\Repository\Thread $threadRepo */
        $threadRepo = $this->repository('XF:Thread');
        /** @var \XF\Repository\ThreadWatch $threadWatchRepo */
        $threadWatchRepo = $this->repository('XF:ThreadWatch');

        $threadRepo->markThreadReadByVisitor($thread);
        $threadWatchRepo->autoWatchThread($thread, \XF::visitor(), true);
    }

    public function sendNotifications()
    {
        if ($this->ticket->discussion_state === 'shadow_ban')
        {
            return;
        }

        // Standard notifications

        if ($this->ticket->isVisible())
        {
            /** @var \NF\Tickets\Service\Message\Notifier $notifier */
            $notifier = $this->service('NF\Tickets:Message\Notifier', $this->message, 'ticket');
            $notifier->setMentionedUserIds($this->messagePreparer->getMentionedUserIds());
            $notifier->setQuotedUserIds($this->messagePreparer->getQuotedUserIds());
            $notifier->notifyAndEnqueue(3);
        }

        if ($this->proxyTicket && $this->ticket->user_id)
        {
            /** @var Notifier $notifier */
            $notifier = $this->service('NF\Tickets:Ticket\Notifier', $this->ticket, $this->user, $this->anonymized, 'create');
            $notifier->addNotifications('create', $this->notifyUsers, true, true);
            $notifier->notifyAndEnqueue(3);
        }

        // Guest ticket email

        if (!$this->ticket->user_id)
        {
            $mailer = $this->app->mailer();
            $mailer->setMailClass('NF\Tickets\Mail\Mail');
            /** @var Mail $mail */
            $mail = $mailer->newMail();
            $mail->setTicket($this->ticket);
            $mail->setTo($this->ticket->user_email, $this->ticket->user_name);
            $mail->setTemplate($this->proxyTicket ? 'nf_tickets_ticket_opened_proxy' : 'nf_tickets_ticket_opened', [
                'ticket' => $this->ticket,
                'password' => $this->password
            ]);
            $mail->send();
        }

        if ($this->threadCreator && $this->threadCreator->getThread()->isVisible())
        {
            $this->threadCreator->sendNotifications();
        }
    }

    protected function getTicketRepo(): \NF\Tickets\Repository\Ticket
    {
        /** @noinspection PhpIncompatibleReturnTypeInspection */
        return $this->repository('NF\Tickets:Ticket');
    }
}