<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Category;
use NF\Tickets\Entity\Ticket;
use XF\Entity\User;

class Editor extends \XF\Service\AbstractService
{
    use \XF\Service\ValidateAndSavableTrait;

    /**
     * @var Ticket
     */
    protected $ticket;

    /**
     * @var User
     */
    protected $user;

    /**
     * @var \XF\Service\Tag\Changer
     */
    protected $tagChanger;

    protected $alert = false;

    protected $performValidations = true;
    /**
     * @var Move
     */
    protected $mover = null;
    /**
     * @var Category
     */
    protected $newCategory;

    public function __construct(\XF\App $app, Ticket $ticket)
    {
        parent::__construct($app);
        $this->ticket = $ticket;
        $this->user = \XF::visitor();
        //$this->tagChanger = $this->service('XF:Tag\Changer', 'nf_tickets_ticket', $ticket);
    }

    public function getTicket()
    {
        return $this->ticket;
    }

    public function getUser()
    {
        return $this->user;
    }

    public function setPerformValidations($perform)
    {
        $this->performValidations = (bool)$perform;
    }

    public function getPerformValidations()
    {
        return $this->performValidations;
    }

    public function setTitle($title)
    {
        $this->ticket->set('title', $title,
            ['forceConstraint' => !$this->performValidations]
        );
    }

    public function setPrefix($prefixId)
    {
        $this->ticket->prefix_id = $prefixId;
    }

    public function setPriority($priority)
    {
        $this->ticket->priority = $priority;
    }

    public function setStatus($status, $alert = false)
    {
        if ($status !== $this->ticket->status_id)
        {
            $this->ticket->status_id = $status;

            $this->alert = (bool)$alert;
        }
    }

    public function setDiscussionState($discussionState)
    {
        $this->ticket->discussion_state = $discussionState;
    }

    public function setCustomFields(array $customFields, $subsetUpdate = false)
    {
        $ticket = $this->ticket;

        $editMode = $ticket->getFieldEditMode();

        $fieldSet = $ticket->custom_fields;
        $fieldDefinition = $fieldSet->getDefinitionSet()
                                    ->filterEditable($fieldSet, $editMode)
                                    ->filterOnly($ticket->Category->field_cache);

        $customFieldsShown = \array_keys($fieldDefinition->getFieldDefinitions());

        if ($subsetUpdate)
        {
            // only updating the values passed through, so remove anything not present
            foreach ($customFieldsShown AS $k => $fieldName)
            {
                if (!isset($customFields[$fieldName]))
                {
                    unset($customFieldsShown[$k]);
                }
            }
        }

        if ($customFieldsShown)
        {
            $fieldSet->bulkSet($customFields, $customFieldsShown, $editMode);
        }
    }

    public function setTags($tags)
    {
        if ($this->tagChanger->canEdit())
        {
            $this->tagChanger->setEditableTags($tags);
        }
    }

    public function addTags($tags)
    {
        if ($this->tagChanger->canEdit())
        {
            $this->tagChanger->addTags($tags);
        }
    }

    public function removeTags($tags)
    {
        if ($this->tagChanger->canEdit())
        {
            $this->tagChanger->removeTags($tags);
        }
    }

    public function setLockState($state)
    {
        $this->ticket->ticket_locked = $state;
    }

    public function setMove(Move $mover, Category $ticketCategory = null)
    {
        $this->mover = $mover;
        $this->newCategory = $ticketCategory;
    }

    public function checkForSpam()
    {
        $ticket = $this->ticket;
        $message = $ticket->FirstMessage;

        if ($ticket->discussion_state === 'visible' && $this->user->isSpamCheckRequired())
        {
            $user = $this->user;

            $message = $ticket->title . "\n" . $message->message;

            $checker = $this->app->spam()->contentChecker();
            $checker->check($user, $message, [
                'permalink'    => $this->app->router('public')->buildLink('canonical:tickets', $ticket),
                'content_type' => 'nf_tickets_ticket',
            ]);

            $decision = $checker->getFinalDecision();
            switch ($decision)
            {
                case 'moderated':

                    $ticket->discussion_state = 'moderated';
                    break;

                case 'denied':
                    $checker->logSpamTrigger('nf_tickets_ticket', $ticket->ticket_id);
                    $ticket->error(\XF::phrase('your_content_cannot_be_submitted_try_later'));
                    break;
            }
        }
    }

    protected function finalSetup()
    {
    }

    protected function _validate()
    {
        $this->finalSetup();

        $ticket = $this->ticket;

        if ($this->performValidations)
        {
            $this->checkForSpam();
        }

        $ticket->preSave();
        $errors = $ticket->getErrors();

        if ($this->performValidations)
        {
            if (!$ticket->prefix_id
                && $ticket->Category->require_prefix
                && $ticket->Category->getUsablePrefixes()
                && !$ticket->canMove()
            )
            {
                $errors[] = \XF::phraseDeferred('please_select_a_prefix');
            }
            // the canMove check allows moderators to bypass this requirement when editing; they're likely editing
            // another user's thread so don't force them to add a prefix

            if ($ticket->canUpdateStatus() && $ticket->status_id)
            {
                /** @var \NF\Tickets\Entity\Status $status */
                $status = $this->em()->find('NF\Tickets:Status', $ticket->status_id);

                if (!$status || !$status->active)
                {
                    $errors[] = \XF::phraseDeferred('nf_tickets_please_select_valid_status');
                }
            }

            if ($this->mover)
            {
                if (!$this->newCategory || !$this->newCategory->canView())
                {
                    $errors['target_category_id'] = \XF::phrase('requested_category_not_found');
                }
            }
            /*if ($this->tagChanger->canEdit())
            {
                $tagErrors = $this->tagChanger->getErrors();
                if ($tagErrors)
                {
                    $errors = \array_merge($errors, $tagErrors);
                }
            }*/
        }

        return $errors;
    }

    protected function _save()
    {
        $ticket = $this->ticket;

        $db = $this->db();
        $db->beginTransaction();

        $ticket->save(true, false);

        //if ($this->tagChanger->canEdit())
        //{
        //    $this->tagChanger->save($this->performValidations);
        //}

        $db->commit();

        //$ticket->quickEscalate();

        if ($this->mover && $this->newCategory)
        {
            $this->mover->move($this->newCategory);
        }

        if ($this->alert && $ticket->canSendModeratorActionAlert() && !$this->ticket->isShadowBanned())
        {
            /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
            $ticketRepo = $this->repository('NF\Tickets:Ticket');
            $ticketRepo->sendModeratorActionAlert($ticket, 'status');
        }

        return $ticket;
    }
}