<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Ticket;
use NF\Tickets\Entity\TicketGuestParticipant;
use NF\Tickets\Mail\Mail;
use XF\Entity\User;
use XF\Service\AbstractService;
use XF\Service\ValidateAndSavableTrait;

class Invite extends AbstractService
{
    use ValidateAndSavableTrait;

    /** @var Ticket */
    protected $ticket;

    /** @var User[] */
    protected $inviteUsers = [];

    protected $inviteGuest = null;

    /** @var \NF\Tickets\Repository\Ticket */
    protected $ticketRepo;

    protected $errors = [];

    protected $from;

    /**
     * @var User[]
     */
    protected $notifyUsers = [];

    /** @var TicketGuestParticipant */
    protected $notifyGuest = [];

    protected $autoSendNotifications = false;

    /** @var bool */
    protected $anonymized;

    public function __construct(\XF\App $app, Ticket $ticket, User $from)
    {
        parent::__construct($app);

        $this->ticket = $ticket;
        $this->from = $from;
        $this->ticketRepo = $this->repository('NF\Tickets:Ticket');
    }

    /**
     * @return Ticket
     */
    public function getTicket()
    {
        return $this->ticket;
    }

    /**
     * @param bool $value
     */
    public function setAutoSendNotifications($value)
    {
        $this->autoSendNotifications = (bool)$value;
    }

    /**
     * @param      $invitees
     * @param bool $triggerErrors
     */
    public function setInvitees($invitees, $triggerErrors = true)
    {
        $this->inviteUsers = $this->ticketRepo->getValidatedParticipants($this->ticket, $invitees, $error);

        if ($triggerErrors && $error)
        {
            $this->errors = [$error];
        }
    }

    public function inviteGuest($email, $username, $triggerErrors = true)
    {
        $this->inviteGuest = $this->ticketRepo->getValidatedGuestParticipant($this->ticket, $email, $username, $error);

        if ($triggerErrors && $error)
        {
            $this->errors = [$error];
        }
    }

    /**
     * @param bool $anonymize
     */
    public function setAnonymize($anonymize)
    {
        $this->anonymized = (bool)$anonymize;
    }

    /**
     * @return bool
     */
    public function isAnonymize()
    {
        return $this->anonymized;
    }

    protected function _validate()
    {
        return $this->errors;
    }

    protected function _save()
    {
        if (!$this->inviteUsers && !$this->inviteGuest)
        {
            return true;
        }

        if (!empty($this->inviteUsers))
        {
            $this->notifyUsers = $this->ticketRepo->insertParticipants(
                $this->ticket, $this->inviteUsers, $this->from
            );
        }

        if ($this->inviteGuest !== null)
        {
            $this->notifyGuest = $this->ticketRepo->insertGuestParticipant(
                $this->ticket, $this->inviteGuest, $this->from, $password
            );

            $this->sendGuestEmail($password);
        }

        if ($this->autoSendNotifications)
        {
            $this->sendNotifications();
        }

        return true;
    }

    public function sendNotifications()
    {
        if ($this->ticket->isVisible())
        {
            /** @var Notifier $notifier */
            $notifier = $this->service('NF\Tickets:Ticket\Notifier', $this->ticket, $this->from, $this->anonymized, 'invite');
            $notifier->addNotifications('invite', $this->notifyUsers, true, true);
            $notifier->notifyAndEnqueue(3);
        }
    }

    public function sendGuestEmail($password)
    {
        if ($this->notifyGuest && !$this->ticket->isShadowBanned())
        {
            $mailer = $this->app->mailer();
            $mailer->setMailClass('NF\Tickets\Mail\Mail');

            /** @var Mail $mail */
            $mail = $mailer->newMail();
            $mail->setTicket($this->ticket);
            $mail->setTo($this->notifyGuest->email);
            $mail->setTemplate('nf_tickets_ticket_invite_ticket_guest', [
                'ticket' => $this->ticket,
                'password' => $password
            ]);
            $mail->send();
        }
    }
}