<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Message;
use NF\Tickets\Entity\Ticket;
use XF\Service\AbstractService;

use function is_array;

class Merger extends AbstractService
{
    /**
     * @var Ticket
     */
    protected $target;

    protected $log = true;

    protected $alert;
    protected $alertReason = '';

    /** @var Ticket[] */
    protected $sourceTickets = [];

    /** @var array[] */
    protected $sourceMessages = [];

    public function __construct(\XF\App $app, Ticket $target)
    {
        parent::__construct($app);

        $this->target = $target;
    }

    /**
     * @return Ticket
     */
    public function getTarget()
    {
        return $this->target;
    }

    public function setSendAlert($alert, $reason = null)
    {
        $this->alert = (bool)$alert;
        if ($reason !== null)
        {
            $this->alertReason = $reason;
        }
    }

    public function merge($sourceTicketsRaw)
    {
        if ($sourceTicketsRaw instanceof \XF\Mvc\Entity\AbstractCollection)
        {
            $sourceTicketsRaw = $sourceTicketsRaw->toArray();
        }
        else if ($sourceTicketsRaw instanceof Ticket)
        {
            $sourceTicketsRaw = [$sourceTicketsRaw];
        }
        else if (!is_array($sourceTicketsRaw))
        {
            throw new \InvalidArgumentException('Tickets must be provided as collection, array or entity');
        }

        if (!$sourceTicketsRaw)
        {
            return false;
        }

        $db = $this->db();

        /** @var Ticket $sourceTicket */
        $sourceTickets = [];
        foreach ($sourceTicketsRaw AS $sourceTicket)
        {
            $sourceTicket->setOption('log_moderator', false);
            $sourceTickets[$sourceTicket->ticket_id] = $sourceTicket;
        }

        $ticketMessages = $db->fetchAllKeyed("
            SELECT message_id, ticket_id, user_id, message_state, reactions
            FROM xf_nf_tickets_message
            WHERE ticket_id IN (" . $db->quote(\array_keys($sourceTickets)) . ")
        ", 'message_id');

        $this->sourceTickets = $sourceTickets;
        $this->sourceMessages = $ticketMessages;

        $target = $this->target;
        $target->setOption('log_moderator', false);

        $db->beginTransaction();

        $this->moveDataToTarget();
        $this->updateTargetData();
        $this->updateUserCounters();

        if ($this->alert)
        {
            $this->sendAlert();
        }

        foreach ($sourceTickets AS $sourceTicket)
        {
            $sourceTicket->delete();
        }

        $this->finalActions();

        $db->commit();

        return true;
    }

    protected function moveDataToTarget()
    {
        $db = $this->db();
        $target = $this->target;

        $sourceTickets = $this->sourceTickets;
        $sourceTicketIds = \array_keys($sourceTickets);
        $sourceIdsQuoted = $db->quote($sourceTicketIds);

        $db->update('xf_nf_tickets_message',
            ['ticket_id' => $target->ticket_id],
            "ticket_id IN ($sourceIdsQuoted)"
        );
        $db->update('xf_nf_tickets_ticket_watch',
            ['ticket_id' => $target->ticket_id],
            "ticket_id IN ($sourceIdsQuoted)",
            [], 'IGNORE'
        );
        $db->update('xf_nf_tickets_spam_log',
            ['ticket_id' => $target->ticket_id],
            "ticket_id IN ($sourceIdsQuoted)"
        );
    }

    protected function updateTargetData()
    {
        $db = $this->db();
        $target = $this->target;
        $sourceTickets = $this->sourceTickets;

        $target->rebuildCounters();
        $target->save();

        /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
        $ticketRepo = $this->repository('NF\Tickets:Ticket');
        $ticketRepo->rebuildTicketMessagePositions($target->ticket_id);
    }

    protected function updateUserCounters()
    {
        $target = $this->target;

        $targetMessagesCount = ($target->discussion_state === 'visible');
        $targetReactionsCount = ($target->discussion_state === 'visible');

        $sourcesMessagesCount = [];
        $sourcesReactionsCount = [];

        $userTicketCountAdjust = [];

        foreach ($this->sourceTickets AS $id => $sourceTicket)
        {
            $sourcesMessagesCount[$id] = ($sourceTicket['discussion_state'] === 'visible');
            $sourcesReactionsCount[$id] = ($sourceTicket['discussion_state'] === 'visible');

            $userId = $sourceTicket['user_id'];
            if ($userId)
            {
                if ($sourcesMessagesCount[$id] && !$targetMessagesCount)
                {
                    if (!isset($userTicketCountAdjust[$userId]))
                    {
                        $userTicketCountAdjust[$userId] = 0;
                    }
                    $userTicketCountAdjust[$userId]--;
                }
                else if (!$sourcesMessagesCount[$id] && $targetMessagesCount)
                {
                    if (!isset($userTicketCountAdjust[$userId]))
                    {
                        $userTicketCountAdjust[$userId] = 0;
                    }
                    $userTicketCountAdjust[$userId]++;
                }
            }
        }

        $reactionsEnable = [];
        $reactionsDisable = [];

        // Note: $message is an array, not an Entity!
        foreach ($this->sourceMessages AS $id => $message)
        {
            if ($message['message_state'] !== 'visible')
            {
                continue; // everything will stay the same in the new thread
            }

            /** @var Message $message */
            $sourceReactionsCount = $sourcesReactionsCount[$message['ticket_id']];

            if ($message['reactions'])
            {
                if ($sourceReactionsCount && !$targetReactionsCount)
                {
                    $reactionsDisable[] = $id;
                }
                else if (!$sourceReactionsCount && $targetReactionsCount)
                {
                    $reactionsEnable[] = $id;
                }
            }
        }

        if ($reactionsDisable)
        {
            /** @var \XF\Repository\Reaction $reactionRepo */
            $reactionRepo = $this->repository('XF:Reaction');
            $reactionRepo->fastUpdateReactionIsCounted('nf_tickets_message', $reactionsDisable, false);
        }
        if ($reactionsEnable)
        {
            /** @var \XF\Repository\Reaction $reactionRepo */
            $reactionRepo = $this->repository('XF:Reaction');
            $reactionRepo->fastUpdateReactionIsCounted('nf_tickets_message', $reactionsEnable, true);
        }
        foreach ($userTicketCountAdjust AS $userId => $adjust)
        {
            if ($adjust)
            {
                $this->db()->query("
                    UPDATE xf_user
                    SET nf_tickets_count = GREATEST(0, nf_tickets_count + ?)
                    WHERE user_id = ?
                ", [$adjust, $userId]);
            }
        }
    }

    protected function sendAlert()
    {
        $target = $this->target;
        $actor = \XF::visitor();

        /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
        $ticketRepo = $this->repository('NF\Tickets:Ticket');

        $alertExtras = [
            'targetTitle' => $target->title,
            'targetLink'  => $this->app->router('public')->buildLink('nopath:tickets', $target),
        ];

        foreach ($this->sourceTickets AS $sourceTicket)
        {
            if ($sourceTicket->discussion_state === 'visible'
                && $sourceTicket->user_id !== $actor->user_id
            )
            {
                $ticketRepo->sendModeratorActionAlert($sourceTicket, 'merge', $this->alertReason, $alertExtras);
            }
        }
    }

    protected function finalActions()
    {
        $target = $this->target;
        $sourceTickets = $this->sourceTickets;
        $sourceTicketIds = \array_keys($sourceTickets);
        $messageIds = \array_keys($this->sourceMessages);

        if ($messageIds)
        {
            $this->app->jobManager()->enqueue('XF:SearchIndex', [
                'content_type' => 'nf_tickets_message',
                'content_ids'  => $messageIds,
            ]);
        }

        if ($this->log)
        {
            $this->app->logger()->logModeratorAction('nf_tickets_ticket', $target, 'merge_target',
                ['ids' => \implode(', ', $sourceTicketIds)]
            );
        }

        /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
        $ticketRepo = $this->repository('NF\Tickets:Ticket');
        $ticketRepo->rebuildTicketCounts();
    }
}