<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Message;
use NF\Tickets\Entity\Ticket;
use NF\Tickets\Mail\Mail;
use XF\Entity\User;
use XF\Service\AbstractService;
use XF\Service\ValidateAndSavableTrait;

/**
 * Class Replier
 *
 * @package NF\Tickets\Service\Ticket
 */
class Replier extends AbstractService
{
    use ValidateAndSavableTrait;

    /**
     * @var Ticket
     */
    protected $ticket;

    /**
     * @var Message
     */
    protected $message;

    /**
     * @var User
     */
    protected $user;

    /**
     * @var \NF\Tickets\Service\Message\Preparer
     */
    protected $messagePreparer;

    /**
     * @var bool
     */
    protected $performValidations = true;

    /**
     * Replier constructor.
     *
     * @param \XF\App $app
     * @param Ticket  $ticket
     */
    public function __construct(\XF\App $app, Ticket $ticket)
    {
        parent::__construct($app);
        $this->setTicket($ticket);
        $this->setUser(\XF::visitor());
        $this->setMessageDefaults();
    }

    /**
     * @param Ticket $ticket
     */
    protected function setTicket(Ticket $ticket)
    {
        $this->ticket = $ticket;
        $this->message = $ticket->getNewMessage();
        $this->messagePreparer = $this->service('NF\Tickets:Message\Preparer', $this->message);
    }

    /**
     * @return Ticket
     */
    public function getTicket()
    {
        return $this->ticket;
    }

    /**
     * @return Message
     */
    public function getMessage()
    {
        return $this->message;
    }

    /**
     * @return \NF\Tickets\Service\Message\Preparer
     */
    public function getMessagePreparer()
    {
        return $this->messagePreparer;
    }

    /**
     * @param User $user
     */
    protected function setUser(User $user)
    {
        $this->user = $user;

        $this->message->user_id = $this->user->user_id;
        $this->message->username = $this->user->username;
        $this->message->user_name = $this->user->username;

        if ($user->email)
        {
            $this->message->user_email = $this->user->email;
        }
    }

    public function setGuestUser($name, $email)
    {
        $this->message->user_id = 0;
        $this->message->username = $name;
        $this->message->user_name = $name;
        $this->message->user_email = $email;
    }

    /**
     * @param bool $anonymize
     */
    public function setAnonymize($anonymize)
    {
        $this->message->anonymized = (bool)$anonymize;
    }

    /**
     * @param $logIp
     */
    public function logIp($logIp)
    {
        $this->messagePreparer->logIp($logIp);
    }

    /**
     * @param $perform
     */
    public function setPerformValidations($perform)
    {
        $this->performValidations = (bool)$perform;
    }

    /**
     * @return bool
     */
    public function getPerformValidations()
    {
        return $this->performValidations;
    }

    public function setIsAutomated()
    {
        $this->logIp(false);
        $this->setPerformValidations(false);
    }

    /**
     *
     */
    protected function setMessageDefaults()
    {
        $category = $this->ticket->Category;

        if (!$category)
        {
            throw new \LogicException('Ticket is not in a valid category');
        }

        $this->message->message_state = 'visible';
    }

    /**
     * @param      $message
     * @param bool $format
     * @return bool
     */
    public function setMessage($message, $format = true)
    {
        return $this->messagePreparer->setMessage($message, $format, $this->performValidations);
    }

    /**
     * @param $hash
     */
    public function setAttachmentHash($hash)
    {
        $this->messagePreparer->setAttachmentHash($hash);
    }

	public function setUserAgent($userAgent): void
	{
		$this->message->user_agent = $userAgent;
	}

    public function selfAssign()
    {
        $this->ticket->assigner_user_id = $this->user->user_id;
        $this->ticket->assigner_username = $this->user->username;
        $this->ticket->assigned_user_id = $this->user->user_id;
        $this->ticket->assigned_username = $this->user->username;
    }

    public function checkForSpam()
    {
        $this->messagePreparer->checkForShadowBan();
        if ($this->message->message_state === 'visible' && $this->user->isSpamCheckRequired())
        {
            $this->messagePreparer->checkForSpam();
        }
    }

    protected function finalSetup()
    {
        if ($this->performValidations)
        {
            $this->message->message_date = \XF::$time;
        }
    }

    /**
     * @return array
     */
    protected function _validate()
    {
        $this->finalSetup();

        $message = $this->message;

        if (!$message->user_id)
        {
            /** @var \XF\Validator\Username $validator */
            $validator = $this->app->validator('Username');
            $validator->setOption('check_unique', false);
            $message->username = $validator->coerceValue($message->username);
            if ($this->performValidations && !$validator->isValid($message->username, $error))
            {
                return [$validator->getPrintableErrorValue($error)];
            }
            $message->user_name = $message->username;
        }

        $message->preSave();

        return $message->getErrors();
    }

    /**
     * @return Message
     * @throws \XF\PrintableException
     */
    protected function _save()
    {
        $message = $this->message;

        $db = $this->db();
        $db->beginTransaction();

        $ticketLatest = $this->db()->fetchRow('
            SELECT *
            FROM xf_nf_tickets_ticket
            WHERE ticket_id = ?
            FOR UPDATE
        ', $this->ticket->ticket_id);

        $this->setMessagePosition($ticketLatest);

        $message->save(true, false);

        $this->messagePreparer->afterInsert();

        $db->commit();

        $message->Ticket->quickEscalate();

        return $message;
    }

    /**
     * @param array $ticketInfo
     */
    protected function setMessagePosition(array $ticketInfo)
    {
        $message = $this->message;

        if ($message->message_date < $ticketInfo['last_message_date'])
        {
            throw new \LogicException('Replier can only add messages at the end of a ticket');
        }

        if ($message->message_state === 'visible')
        {
            $position = $ticketInfo['reply_count'] + 1;
        }
        else
        {
            $position = $ticketInfo['reply_count'];
        }

        $message->set('position', $position, ['forceSet' => true]);
    }

    public function sendNotifications()
    {
        // standard notifications
        if ($this->message->isVisible())
        {
            /** @var \NF\Tickets\Service\Message\Notifier $notifier */
            $notifier = $this->service('NF\Tickets:Message\Notifier', $this->message, 'reply');
            $notifier->setMentionedUserIds($this->messagePreparer->getMentionedUserIds());
            $notifier->setQuotedUserIds($this->messagePreparer->getQuotedUserIds());
            $notifier->notifyAndEnqueue(3);
        }

        // Guest email
        if (!$this->ticket->isShadowBanned() &&
            !$this->ticket->user_id && $this->ticket->user_email && $this->message->user_email !== $this->ticket->user_email)
        {
            $mailer = $this->app->mailer();
            $mailer->setMailClass('NF\Tickets\Mail\Mail');
            /** @var Mail $mail */
            $mail = $mailer->newMail();
            $mail->setTicket($this->ticket);
            $mail->setTo($this->ticket->user_email, $this->ticket->user_name);
            $mail->setTemplate('nf_tickets_guest_ticket_reply', [
                'ticket'  => $this->ticket,
                'message' => $this->message,
            ]);
            $mail->send();
        }
    }
}