<?php

namespace NF\Tickets\Service\Ticket;

use NF\Tickets\Entity\Category;
use NF\Tickets\Entity\Message;
use XF\Entity\Ip;
use XF\Entity\Post;
use XF\Entity\Thread;
use XF\Mvc\Entity\AbstractCollection;
use XF\Service\ValidateAndSavableTrait;

class ThreadConverter extends Creator
{
    use ValidateAndSavableTrait;

    /**
     * @var Thread
     */
    protected $thread;

	protected $alert = false;
	protected $alertReason = '';

    public function __construct(\XF\App $app, Thread $thread, Category $category)
    {
        parent::__construct($app, $category);

        $this->setThread($thread);
    }

	public function setSendAlert($alert, $reason = null)
	{
		$this->alert = (bool)$alert;
		if ($reason !== null)
		{
			$this->alertReason = $reason;
		}
	}

    public function setThread(Thread $thread, $userEmail = null, $username = null)
    {
        $this->ticket->first_message_date = $this->ticket->start_date = $this->message->message_date = $thread->FirstPost->post_date;
        $this->thread = $thread;

        if ($thread->user_id && $thread->User)
        {
            $this->setUser($thread->User);
        }
        else
        {
            if (!$userEmail)
            {
                throw new \InvalidArgumentException("Must provide userEmail if converting guest thread to ticket");
            }

            $this->setGuestUser($username ?: $thread->username, $userEmail);
        }

        $this->setIsAutomated();

        $this->setContent($thread->title, $thread->FirstPost->message);
        $this->copyAttachments($thread->FirstPost, $tempHash);
        $this->setAttachmentHash($tempHash);
    }

    protected function finalSetup()
    {
	    $this->ticket->last_message_date = $this->thread->LastPost->post_date;
	    $this->ticket->last_message_user_id = $this->thread->LastPost->user_id;
	    $this->ticket->last_message_username = $this->thread->LastPost->username;

	    $this->message->position = 0;

	    if (!$this->password)
	    {
		    $this->password = $this->getTicketRepo()->generatePassword();
		    $this->ticket->set('password', password_hash($this->password, PASSWORD_DEFAULT));
	    }
    }

	protected function _save()
    {
        \XF::db()->beginTransaction();

        $response = parent::_save();

        $ticket = $this->ticket;
	    $actor = \XF::visitor();

        $post = $this->thread->FirstPost;
        $this->rewriteAttachmentBbCode($post, $this->message);
        // First post reactions, needs a saved message.
        $this->copyReactions($post, $this->message);
        $this->copyIp($post, $this->message);

        /** @var \XF\Finder\Post $finder */
        $finder = $this->finder('XF:Post');
        $threadReplies = $finder->with('User')->inThread($this->thread)->isNotFirstPost()->fetch();
        $this->addReplies($threadReplies);

        \XF::db()->commit();

	    if ($ticket->canView() && $this->alert && $ticket->user_id !== $actor->user_id)
	    {
		    /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
		    $ticketRepo = $this->repository('NF\Tickets:Ticket');
		    $ticketRepo->sendModeratorActionAlert($ticket, 'convert', $this->alertReason);
	    }

        return $response;
    }

    /**
     * @param AbstractCollection|Post[] $posts
     */
    protected function addReplies($posts)
    {
        foreach ($posts AS $post)
        {
            $this->addReply($post);
        }
    }

    protected function addReply(Post $post)
    {
        $this->copyAttachments($post, $tempHash);

	    \XF::asVisitor($post->User, function () use ($post, $tempHash)
	    {
		    /** @var Replier $replier */
		    $replier = $this->service('NF\Tickets:Ticket\Replier', $this->ticket);
		    $replier->setIsAutomated();
		    $replier->setMessage($post->message);
		    $replier->setAttachmentHash($tempHash);

		    if (!$post->user_id && !$this->ticket->user_id && $this->ticket->username === $post->username)
		    {
			    $replier->setGuestUser($post->username, $this->ticket->user_email);
		    }

		    $message = $replier->getMessage();
		    $message->position = $post->position;
		    $message->message_state = $post->message_state;

		    $replier->save();

		    $this->rewriteAttachmentBbCode($post, $message);
		    $message->fastUpdate('message_date', $post->post_date); // ffs 2
            $this->copyReactions($post, $message);
            $this->copyIp($post, $message);

		    return $message;
	    });
    }

    /**
     * @param Post $post
     * @param      $tempHash
     * @throws \XF\PrintableException
     */
    protected function copyAttachments(Post $post, &$tempHash = null)
    {
        $tempHash = $tempHash ?: \XF::generateRandomString(32);

        /** @var \XF\Entity\Attachment $newAttachment */
        foreach ($post->Attachments AS $attachment)
        {
            /** @var \XF\Entity\Attachment $newAttachment */
            $newAttachment = $this->em()->create('XF:Attachment');
            $newAttachment->data_id = $attachment->data_id;
            $newAttachment->content_type = 'nf_tickets_message';
            $newAttachment->temp_hash = $tempHash;
            $newAttachment->save();
        }
    }

    protected function rewriteAttachmentBbCode(Post $source, Message $target)
    {
        if (!$target->attach_count)
        {
            return;
        }

        // build a mapping between old & new based on data_id that should be shared
        $oldAttachment = [];
        foreach($source->Attachments as $attachment)
        {
            $oldAttachment[$attachment->data_id] = $attachment->attachment_id;
        }

        $message = $target->message;

        $linkToMetaData = false;
        $embedMetadata = $target->embed_metadata ?: [];
        foreach($target->Attachments as $attachment)
        {
            if (empty($oldAttachment[$attachment->data_id]))
            {
                continue;
            }
            $newAttachmentId = $attachment->attachment_id;
            $oldAttachmentId = $oldAttachment[$attachment->data_id];

            /** @noinspection RegExpRedundantEscape */
            $message = \preg_replace_callback(
                '#(\[attach[^\]]*\])' . $oldAttachmentId . '(\[/attach\])#i',
                function ($matches) use ($newAttachmentId, &$linkToMetaData) {
                    $linkToMetaData = true;

                    return "{$matches[1]}{$newAttachmentId}{$matches[2]}";
                },
                $message
            );

            unset($embedMetadata['attachments'][$oldAttachmentId]);
            if ($linkToMetaData)
            {
                $embedMetadata['attachments'][$newAttachmentId] = $newAttachmentId;
            }
        }

        if ($message !== $target->message)
        {
            if ($embedMetadata)
            {
                $target->fastUpdate('embed_metadata', $embedMetadata);
            }
            $target->fastUpdate('message', $message);
        }
    }

    protected function copyReactions(Post $source, Message $target)
    {
        $oldTime = \XF::$time;
        try
        {
            foreach ($source->Reactions AS $reactionContent)
            {
                \XF::$time = $reactionContent->reaction_date;
                \XF::asVisitor($reactionContent->ReactionUser, function () use ($reactionContent, $target) {
                    if ($this->message->canReact())
                    {
                        /** @var \XF\Repository\Reaction $reactionRepo */
                        $reactionRepo = $this->repository('XF:Reaction');
                        $reactionRepo->reactToContent($reactionContent->reaction_id, $target->getEntityContentType(), $target->getEntityContentTypeId(), \XF::visitor());
                    }
                });
            }
        }
        finally
        {
            \XF::$time = $oldTime;
        }
    }

    protected function copyIp(Post $source, Message $target)
    {
        if ($source->ip_id)
        {
            /** @var Ip $existingIpLog */
            $existingIpLog = $this->em()->find('XF:Ip', $source->ip_id);
            if ($existingIpLog)
            {
                /** @var \XF\Repository\Ip $ipRepo */
                $ipRepo = $this->repository('XF:Ip');
                $ipLog = $ipRepo->logIp($source->user_id, $existingIpLog->ip, $target->getEntityContentType(), $target->getEntityContentTypeId());

                if ($ipLog)
                {
                    $ipLog->fastUpdate('log_date', $target->message_date);
                    $target->fastUpdate('ip_id', $ipLog->ip_id);
                }
            }
        }
    }

    public function deleteThread()
    {
        $this->thread->delete();
    }
}