<?php

namespace NF\Tickets;

use NF\Tickets\Install\Data\MySql;
use SV\StandardLib\InstallerHelper;
use XF\AddOn\AbstractSetup;
use XF\AddOn\StepRunnerInstallTrait;
use XF\AddOn\StepRunnerUninstallTrait;
use XF\AddOn\StepRunnerUpgradeTrait;
use XF\Db\Schema\Alter;

/**
 * Class Setup
 *
 * @package NF\Tickets
 */
class Setup extends AbstractSetup
{
    use InstallerHelper;
    use StepRunnerInstallTrait;
    use StepRunnerUpgradeTrait;
    use StepRunnerUninstallTrait;

	public static $supportedAddOns = [
		'XFRM' => true,
	];

    public function installStep1(): void
    {
        $this->applySchemaUpdates();
    }

    public function installStep2(): void
    {
        $this->db()->query("
            REPLACE INTO `xf_nf_tickets_category` (
                `ticket_category_id`, `title`, `description`, `parent_category_id`, `depth`, `lft`, `rgt`, 
                `display_order`, `ticket_count`, `last_ticket`, `last_ticket_title`, `last_ticket_id`, `breadcrumb_data`, 
                `thread_node_id`, `thread_prefix_id`, `field_cache`, `priority_selection`, `empty_message_allowed`, 
                `hide_message_field`, `category_emails`, `prefix_cache`, `default_prefix_id`, `require_prefix`) 
            VALUES (
                '1', 'Example category', 'This is an example category for the Ticket system.', 
                '0', '0', '1', '2', '1', '0', '0', '', '0', 'a:0:{}', '0', '0', '[]', '1', '0', '0', '', '[]', '0', '0');
        ");

        foreach ($this->getData() AS $dataQuery)
        {
            $this->db()->query($dataQuery);
        }
    }

    public function installStep3(): void
    {
        foreach ($this->getPhrases() AS $title => $text)
        {
            /** @var \XF\Entity\Phrase $phrase */
            $phrase = \XF::finder('XF:Phrase')
                         ->where('title', '=', $title)
                         ->where('language_id', '=', 0)
                         ->fetchOne();
            if (!$phrase)
            {
                /** @var \XF\Entity\Phrase $phrase */
                $phrase = \XF::em()->create('XF:Phrase');
                $phrase->title = $title;
                $phrase->phrase_text = $text;
                $phrase->language_id = 0;
                $phrase->save();
            }
        }
    }

    public function installStep4(): void
    {
        $this->db()->query("
			REPLACE INTO `xf_route_filter`
				(`prefix`,`find_route`,`replace_route`,`enabled`,`url_to_route_only`)
			VALUES
				('tickets', 'tickets/', 'support/', 1, 0);
		");
    }

    public function postInstall(array &$stateChanges): void
    {
        /** @var \XF\Service\RebuildNestedSet $service */
        $service = \XF::service('XF:RebuildNestedSet', 'NF\Tickets:Category', [
            'parentField' => 'parent_category_id',
        ]);
        $service->rebuildNestedSetInfo();

        /** @var \NF\Tickets\Repository\TicketField $ticketFieldRepo */
        $ticketFieldRepo = \XF::repository('NF\Tickets:TicketField');
        $ticketFieldRepo->rebuildFieldCache();

        $this->app->jobManager()->enqueueUnique(
            'phraseRebuild',
            'XF:PhraseRebuild',
            [],
            false
        );

        \XF::repository('XF:RouteFilter')->rebuildRouteFilterCache();
    }

	public function applySchemaUpdates(): void
	{
		$sm = $this->schemaManager();

		foreach ($this->getTables() as $tableName => $callback)
		{
			$sm->createTable($tableName, $callback);
			$sm->alterTable($tableName, $callback);
		}

		foreach ($this->getAlterTables() as $tableName => $callback)
		{
			if ($sm->tableExists($tableName))
			{
				$sm->alterTable($tableName, $callback);
			}
		}
	}

    public function upgrade1040370Step1()
    {
        $this->schemaManager()->alterTable('xf_nixfifty_ticket', function (Alter $table) {
            $this->addOrChangeColumn($table, 'assigned_username', 'varchar', 50);
            $this->addOrChangeColumn($table, 'thread_id', 'int', 10)->setDefault(0);
            $this->addOrChangeColumn($table, 'ticket_state')->resetDefinition()->type('enum')
                 ->values(['open', 'locked', 'assigned', 'closed', 'resolved'])
                 ->after('ticket_status_id')
                 ->setDefault('open');
        });
    }

    public function upgrade1040541Step1()
    {
        $this->schemaManager()->alterTable('xf_nixfifty_ticket', function (Alter $table) {
            $this->addOrChangeColumn($table, 'user_name', 'varchar', 50);
            $this->addOrChangeColumn($table, 'user_email', 'varchar', 150);
            $this->addOrChangeColumn($table, 'password', 'varchar', 255);
        });

        $this->schemaManager()->alterTable('xf_nixfifty_ticket_message', function (Alter $table) {
            $this->addOrChangeColumn($table, 'user_name', 'varchar', 50);
            $this->addOrChangeColumn($table, 'user_email', 'varchar', 150);
        });
    }

    public function upgrade1040730Step1()
    {
        $this->schemaManager()->alterTable('xf_nixfifty_ticket_message', function (Alter $table) {
            $this->addOrChangeColumn($table, 'ticket_locked', 'tinyint', 3)->setDefault(0)->after('ticket_state');
            $this->addOrChangeColumn($table, 'ticket_state')->resetDefinition()->type('enum')
                 ->values(['open', 'locked', 'assigned', 'closed', 'resolved', 'completed'])
                 ->after('ticket_status_id')
                 ->setDefault('open');
        });
        // TODO: locked status to ticket_locked flag
    }

    public function upgrade1060470Step1()
    {
        $this->schemaManager()->alterTable('xf_nixfifty_ticket_escalation', function (Alter $table) {
            $table->dropColumns(['execute_once', 'execute_type']);
        });

        $this->schemaManager()->alterTable('xf_nixfifty_ticket', function (Alter $table) {
            $this->addOrChangeColumn($table, 'starter_user_id', 'int', 10);
            $this->addOrChangeColumn($table, 'starter_username', 'varchar', 50);
        });

        // indexes get added later
    }

    public function upgrade1060571Step1()
    {
        $this->schemaManager()->alterTable('xf_nixfifty_ticket_field', function (Alter $table) {
            $table->dropColumns('allowed_user_group_ids');

            $this->addOrChangeColumn($table, 'allow_use_field_user_group_ids', 'blob');
            $this->addOrChangeColumn($table, 'allow_view_field_user_group_ids', 'blob');
            $this->addOrChangeColumn($table, 'allow_view_field_owner_in_user_group_ids', 'blob');
        });
    }

    public function upgrade1060870Step1()
    {
        $this->schemaManager()->alterTable('xf_nixfifty_ticket_category', function (Alter $table) {
            $this->addOrChangeColumn($table, 'category_emails', 'text');
        });
    }

    public function upgrade2000010Step1()
    {
        $renameTables = [
            'xf_nixfifty_ticket_category'               => 'xf_nf_tickets_category',
            'xf_nixfifty_ticket_field_category'         => 'xf_nf_tickets_category_field',
            'xf_nixfifty_ticket_category_read'          => 'xf_nf_tickets_category_read',
            'xf_nixfifty_ticket_category_watch'         => 'xf_nf_tickets_category_watch',
            'xf_nixfifty_ticket_escalation'             => 'xf_nf_tickets_escalation',
            'xf_nixfifty_ticket_escalation_trigger_log' => 'xf_nf_tickets_escalation_trigger_log',
            'xf_nixfifty_ticket_message'                => 'xf_nf_tickets_message',
            'xf_nixfifty_ticket_redirect'               => 'xf_nf_tickets_redirect',
            'xf_nixfifty_ticket_response'               => 'xf_nf_tickets_response',
            'xf_nixfifty_ticket_response_category'      => 'xf_nf_tickets_response_category',
            'xf_nixfifty_spam_log'                      => 'xf_nf_tickets_spam_log',
            'xf_nixfifty_ticket'                        => 'xf_nf_tickets_ticket',
            'xf_nixfifty_ticket_field'                  => 'xf_nf_tickets_ticket_field',
            'xf_nixfifty_ticket_field_value'            => 'xf_nf_tickets_ticket_field_value',
            'xf_nixfifty_ticket_read'                   => 'xf_nf_tickets_ticket_read',
            'xf_nixfifty_ticket_watch'                  => 'xf_nf_tickets_ticket_watch',
            'xf_nixfifty_ticket_user_watch'             => 'xf_nf_tickets_ticket_user_watch',
        ];

        foreach ($renameTables AS $from => $to)
        {
            $this->migrateTable($from, $to, false);
        }
    }

    public function upgrade2000010Step2()
    {
        $sm = $this->schemaManager();

        $sm->alterTable('xf_user', function (Alter $table) {
            $table->renameColumn('nixfifty_ticket_count', 'nf_tickets_count');
            $table->renameColumn('nixfifty_tickets_unread', 'nf_tickets_unread');
        });
    }

    public function upgrade2000010Step3()
    {
        $db = $this->db();
        $sm = $this->schemaManager();

        $sm->alterTable('xf_nf_tickets_category', function (Alter $table) {
            $table->renameColumn('category_id', 'ticket_category_id');
            $table->renameColumn('category_title', 'title');
            $table->renameColumn('category_description', 'description');
            $table->renameColumn('category_breadcrumb', 'breadcrumb_data');
        });

        $sm->alterTable('xf_nf_tickets_category_field', function (Alter $table) {
            $table->renameColumn('category_id', 'ticket_category_id');
        });

        $sm->alterTable('xf_nf_tickets_ticket', function (Alter $table) {
            $table->renameColumn('category_id', 'ticket_category_id');
            $table->renameColumn('custom_ticket_fields', 'custom_fields');
            $table->dropColumns('ticket_status_id');
            $this->addOrChangeColumn($table, 'status_id', 'int')->after('priority');
            $this->addOrChangeColumn($table, 'first_message_reaction_score', 'int')->unsigned(false)->setDefault(0);
            $this->addOrChangeColumn($table, 'first_message_reactions', 'blob')->nullable()->after('first_message_reaction_score');
        });

        $sm->alterTable('xf_nf_tickets_message', function (Alter $table) {
            $table->dropColumns('is_staff_reply');
            $this->addOrChangeColumn($table,'embed_metadata', 'blob')->nullable();
        });

        $sm->alterTable('xf_nf_tickets_ticket_watch', function (Alter $table) {
            $table->dropColumns('watch_key');
        });

        $sm->alterTable('xf_nf_tickets_ticket_field', function (Alter $table) {
            $this->addOrChangeColumn($table, 'display_template', 'text');

            $this->addOrChangeColumn($table,'field_id')->resetDefinition()->type('varbinary', 25)->setDefault('none');
            $this->addOrChangeColumn($table,'field_type')->resetDefinition()->type('varchar', 25)->setDefault('textbox');
            $this->addOrChangeColumn($table,'match_type')->resetDefinition()->type('varchar', 25)->setDefault('none');

            $table->renameColumn('allow_use_field_user_group_ids', 'usable_user_group_ids');
            $table->renameColumn('allow_view_field_user_group_ids', 'viewable_user_group_ids');
            $table->renameColumn('allow_view_field_owner_in_user_group_ids', 'viewable_owner_user_group_ids');

            $this->addOrChangeColumn($table,'match_params', 'blob')->after('match_type');
            $this->addOrChangeColumn($table,'display_list', 'tinyint')->setDefault(0)->after('display_template');
        });

        /** @noinspection SqlResolve */
        foreach ($db->fetchAllKeyed('SELECT * FROM xf_nf_tickets_ticket_field', 'field_id') AS $fieldId => $field)
        {
            if (!isset($field['match_regex']))
            {
                continue;
            }

            $update = [];
            $matchParams = [];

            switch ($field['match_type'])
            {
                case 'regex':
                    if ($field['match_regex'])
                    {
                        $matchParams['regex'] = $field['match_regex'];
                    }
                    break;

                case 'callback':
                    if ($field['match_callback_class'] && $field['match_callback_method'])
                    {
                        $matchParams['callback_class'] = $field['match_callback_class'];
                        $matchParams['callback_method'] = $field['match_callback_method'];
                    }
                    break;
                case 'none':
                case 'number':
                case 'alphanumeric':
                case 'email':
                case 'url':
                    break;
                default: // unknown, enum would break so reset to none
                    $field['match_type'] = 'none';
                    break;
            }

            if (!empty($matchParams))
            {
                $update['match_params'] = \json_encode($matchParams);
            }

            if ($field['field_choices'] && $fieldChoices = @\unserialize($field['field_choices']))
            {
                $update['field_choices'] = \json_encode($fieldChoices);
            }

            if (!empty($update))
            {
                $db->update('xf_nf_tickets_ticket_field', $update, 'field_id = ?', $fieldId);
            }
        }

        $sm->alterTable('xf_nf_tickets_ticket_field', function (Alter $table) {
            $table->dropColumns(['match_regex', 'match_callback_class', 'match_callback_method']);
            $this->addOrChangeColumn($table,'match_type')->resetDefinition()->type('enum')
                  ->values(['none', 'number', 'alphanumeric', 'email', 'url', 'regex', 'callback'])
                  ->setDefault('none');
        });

        $sm->alterTable('xf_nf_tickets_ticket_field_value', function (Alter $table) {
            $this->addOrChangeColumn($table,'field_id')->resetDefinition()->type('varbinary', 25)->setDefault('none');
        });

        $sm->alterTable('xf_nf_tickets_category_field', function (Alter $table) {
            $this->addOrChangeColumn($table,'field_id')->resetDefinition()->type('varbinary', 25)->setDefault('none');
            $this->addOrChangeColumn($table,'ticket_category_id')->length(10)->unsigned();
        });

        /** @noinspection SqlResolve */
        $db->query("
            UPDATE xf_nf_tickets_ticket_field
            SET field_type = 'stars'
            WHERE field_type = 'rating'
        ");

        /** @noinspection SqlResolve */
        $db->query("
            UPDATE xf_nf_tickets_ticket_field
            SET field_type = 'textbox'
            WHERE field_type = 'datepicker'
        ");
    }

    public function upgrade2000010Step4()
    {
        /** @noinspection SqlResolve */
        $this->db()->query("UPDATE xf_nf_tickets_ticket SET status_id = 1 WHERE ticket_state = 'open'");
    }

    public function upgrade2000010Step5()
    {
        /** @noinspection SqlResolve */
        $this->db()->query("UPDATE xf_nf_tickets_ticket SET status_id = 2 WHERE ticket_state = 'pending'");
    }

    public function upgrade2000010Step6()
    {
        /** @noinspection SqlResolve */
        $this->db()->query("UPDATE xf_nf_tickets_ticket SET status_id = 3 WHERE ticket_state = 'resolved'");
    }

    public function upgrade2000010Step7()
    {
        /** @noinspection SqlResolve */
        $this->db()->query("UPDATE xf_nf_tickets_ticket SET priority = 3 WHERE priority = 1");
    }

    public function upgrade2000010Step8()
    {
        /** @noinspection SqlResolve */
        $this->db()->query("UPDATE xf_nf_tickets_ticket SET priority = 2 WHERE priority = 2");
    }

    public function upgrade2000010Step9()
    {
        /** @noinspection SqlResolve */
        $this->db()->query("UPDATE xf_nf_tickets_ticket SET priority = 1 WHERE priority = 3");
    }

    public function upgrade2000013Step1(array $stepParams)
    {
        $position = $stepParams[0] ?? 0;
        $stepData = $stepParams[2] ?? [];

        return $this->entityColumnsToJson(
            'NF\Tickets:Category', ['field_cache', 'breadcrumb_data'], $position, $stepData
        );
    }

    public function upgrade2000013Step2(array $stepParams)
    {
        $position = $stepParams[0] ?? 0;
        $stepData = $stepParams[2] ?? [];

        return $this->entityColumnsToJson(
            'NF\Tickets:Ticket', ['custom_fields'], $position, $stepData
        );
    }

    public function upgrade2000013Step3(array $stepParams)
    {
        $position = $stepParams[0] ?? 0;
        $stepData = $stepParams[2] ?? [];

        return $this->entityColumnsToJson(
            'NF\Tickets:Message', ['like_users'], $position, $stepData
        );
    }

    public function upgrade2000013Step4()
    {
        $this->migrateTableToReactions('xf_nf_tickets_message');
    }

    public function upgrade2000013Step5()
    {
        $this->renameLikeAlertOptionsToReactions(['nf_tickets_message']);
    }

    public function upgrade2000013Step6()
    {
        $this->renameLikeAlertsToReactions(['nf_tickets_message']);
    }

    public function upgrade2000013Step7()
    {
        $this->renameLikePermissionsToReactions([
            'ticket' => true,
        ], 'like');

        $this->renameLikeStatsToReactions(['nf_tickets_message']);
    }

    public function upgrade2000014Step1()
    {
        $sm = $this->schemaManager();

        $sm->alterTable('xf_nf_tickets_ticket', function (Alter $table) {
            if ($table->getColumnDefinition('category_id'))
            {
                $table->renameColumn('category_id', 'ticket_category_id');
            }
        });

        $sm->alterTable('xf_nf_tickets_category_watch', function (Alter $table) {
            if ($table->getColumnDefinition('category_id'))
            {
                $table->renameColumn('category_id', 'ticket_category_id');
            }
        });

        $sm->alterTable('xf_nf_tickets_category_read', function (Alter $table) {
            if ($table->getColumnDefinition('category_id'))
            {
                $table->renameColumn('category_id', 'ticket_category_id');
            }
            $table->dropIndexes('user_id_category_id');
        });

        $sm->alterTable('xf_nf_tickets_ticket_field', function (Alter $table) {
            if ($table->getColumnDefinition('category_id'))
            {
                $table->renameColumn('category_id', 'ticket_category_id');
            }
            $table->dropIndexes('user_id_category_id');
        });
    }

    public function upgrade2000033Step1()
    {
        // migrate old alerts
        $db = $this->db();
        $db->query("
            UPDATE xf_user_alert
            SET content_type = 'nf_tickets_message', action = 'insert'
            WHERE content_type = 'ticket_message' AND action = 'insert_watch'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET content_type = 'nf_tickets_message'
            WHERE content_type = 'ticket_message' AND action = 'quote'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET content_type = 'nf_tickets_message', action = 'reaction'
            WHERE content_type = 'ticket_message' AND action = 'like'
        ");

        $db->query("
            UPDATE xf_user_alert
            SET action = 'nf_tickets_move'
            WHERE content_type = 'user' AND action = 'ticket_move'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET action = 'nf_tickets_message_delete'
            WHERE content_type = 'user' AND action = 'ticket_message_delete'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET action = 'nf_tickets_move'
            WHERE content_type = 'user' AND action = 'ticket_move'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET action = 'nf_tickets_delete'
            WHERE content_type = 'user' AND action = 'ticket_delete'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET content_type = 'nf_tickets', action = 'foreign_open'
            WHERE content_type = 'ticket' AND action = 'foreign_open'
        ");
        $db->query("
            UPDATE xf_user_alert
            SET content_type = 'nf_tickets', action = 'reassign_to'
            WHERE content_type = 'ticket' AND action = 'reassign'
        ");

        $db->query("
            UPDATE xf_user_alert
            SET depends_on_addon_id = 'NF/Tickets'
            WHERE content_type IN ('nf_tickets','ticket','nf_tickets_message','ticket_message') AND depends_on_addon_id = ''
        ");
    }

    public function upgrade2000037Step1()
    {
        // make sure indexes & default values are fixed up
        $this->installStep1();
    }

    public function upgrade2000037Step2()
    {
        $this->installStep2();
    }

    public function upgrade2000037Step3()
    {
        $data = $this->getData();

        $this->db()->query($data['xf_nf_tickets_priority']);
        $this->db()->query($data['xf_nf_tickets_status']);

        $this->db()->query("UPDATE xf_nf_tickets_status SET include_in_counts = 1, include_in_defaults = 1, is_close = 0, alert_on_transition = 1 WHERE status_id = 1");
        $this->db()->query("UPDATE xf_nf_tickets_status SET include_in_counts = 0, include_in_defaults = 1, is_close = 0, alert_on_transition = 0 WHERE status_id = 2");
        $this->db()->query("UPDATE xf_nf_tickets_status SET include_in_counts = 0, include_in_defaults = 0, is_close = 1, alert_on_transition = 1 WHERE status_id = 3");
    }

    public function upgrade2000037Step4()
    {
        $this->db()->query('UPDATE xf_nf_tickets_ticket SET priority = NULL WHERE priority = 0');
    }

    public function upgrade2000041Step1()
    {
        $this->installStep1();
    }

    public function upgrade2000041Step2()
    {
        $this->installStep3();
    }

    public function upgrade2000041Step3()
    {
        $this->db()->query('UPDATE xf_nf_tickets_status SET css_class = \'label label--primary\' WHERE css_class = \'\'');
    }

    public function upgrade2000041Step4(array $stepParams)
    {
        $position = $stepParams[0] ?? 0;
        $stepData = $stepParams[2] ?? [];

        return $this->entityColumnsToJson(
            'NF\Tickets:Escalation', ['user_criteria', 'ticket_criteria', 'actions'], $position, $stepData
        );
    }

    public function upgrade2000043Step1()
    {
        $map = [
            'nixfifty_tickets_field_*_choice_*' => 'nf_tickets_ticket_field_choice.$1_$2',
            'nixfifty_tickets_field_*_desc'     => 'nf_tickets_ticket_field_desc.*',
            'nixfifty_tickets_field_*'          => 'nf_tickets_ticket_field_title.*',
        ];

        $this->renamePhrases($map);
    }

    public function upgrade2000045Step1()
    {
        $this->installStep1();
    }

    public function upgrade2000045Step2(array $stepParams)
    {
        $position = $stepParams[0] ?? 0;
        $stepData = $stepParams[2] ?? [];

        return $this->entityColumnsToJson(
            'NF\Tickets:TicketSpamLog', ['details', 'request_state'], $position, $stepData
        );
    }

    public function upgrade2000045Step3()
    {
        // the xf_nf_tickets_spam_log table needs to have phrases renamed from the details
        $phraseMap = [
            'aed_detectspamreg_is_banned'          => 'sv_reg_log.multi_account_is_banned',
            'aed_detectspamreg_group_membership'   => 'sv_reg_log.multi_account_group',
            'aed_detectspamreg_accept'             => 'sv_reg_log.multi_account_accept',
            'aed_detectspamreg_moderate'           => 'sv_reg_log.multi_account_moderate',
            'aed_detectspamreg_reject'             => 'sv_reg_log.multi_account_reject',
            'tpu_detectspamreg_checking'           => 'sv_reg_log.checking',
            'tpu_detectspamreg_honeypot_fail'      => 'sv_reg_log.honeypot_fail',
            'tpu_detectspamreg_honeypot_pass'      => 'sv_reg_log.honeypot_pass',
            'tpu_detectspamreg_honeypot_ok'        => 'sv_reg_log.honeypot_ok',
            'tpu_detectspamreg_sfs_ok'             => 'sv_reg_log.sfs_ok',
            'tpu_detectspamreg_sfs_fail'           => 'sv_reg_log.sfs_fail',
            'tpu_detectspamreg_country_detected'   => 'sv_reg_log.country_ok',
            'tpu_detectspamreg_country_fail'       => 'sv_reg_log.country_fail',
            'tpu_detectspamreg_country_ok'         => 'sv_reg_log.country_ok',
            'tpu_detectspamreg_as_detected'        => 'sv_reg_log.as_ok',
            'tpu_detectspamreg_as_fail'            => 'sv_reg_log.as_fail',
            'tpu_detectspamreg_as_ok'              => 'sv_reg_log.as_ok',
            'tpu_detectspamreg_tor_fail'           => 'sv_reg_log.tor_fail',
            'tpu_detectspamreg_username_fail'      => 'sv_reg_log.username_fail',
            'tpu_detectspamreg_email_fail'         => 'sv_reg_log.email_fail',
            'tpu_detectspamreg_hostname_ok'        => 'sv_reg_log.hostname_ok',
            'tpu_detectspamreg_hostname_detected'  => 'sv_reg_log.hostname_ok',
            'tpu_detectspamreg_hostname_fail'      => 'sv_reg_log.hostname_fail',
            'tpu_detectspamreg_js_fail'            => 'sv_reg_log.js_fail',
            'tpu_detectspamreg_formtimer_fail'     => 'sv_reg_log.js_timer_fail',
            'tpu_detectspamreg_formtimer_detected' => 'sv_reg_log.js_timer_ok',
            'tpu_detectspamreg_totalscore'         => 'sv_reg_log.total_score',
            'tpu_detectspamreg_port_fail'          => 'sv_reg_log.port_scan_fail',
            'tpu_detectspamreg_fail_mod'           => 'sv_reg_log.fail_score_moderated',
            'tpu_detectspamreg_fail_modposts'      => 'sv_reg_log.fail_score_posts_moderated',
            'tpu_detectspamreg_fail_rej'           => 'sv_reg_log.fail_score_rejected',
            'tpu_detectspamreg_emaillen_fail'      => 'sv_reg_log.email_length_fail',

            'Rejected. Direct rule selection triggered'            => 'sv_reg_log.direct_rule_reject',
            'Moderated. Direct rule selection triggered'           => 'sv_reg_log.direct_rule_moderated',
            'New Posts Moderated. Direct rule selection triggered' => 'sv_reg_log.direct_rule_posts_moderated',
        ];

        if ($this->db()->fetchOne("
          SELECT trigger_log_id
          FROM xf_nf_tickets_spam_log 
          WHERE (details LIKE '%tpu%' OR details LIKE '%aed%')
          LIMIT 1
        "))
        {
            $this->app->jobManager()->enqueueUnique('migrate-ticket-spam-logs', 'NF\Tickets:MigrateTicketSpamLogs', [
                'phrases' => $phraseMap,
            ]);
        }
    }

    public function upgrade2000303Step1()
    {
        $this->installStep1();
    }

    public function upgrade2000303Step2()
    {
        $this->db()->query('UPDATE xf_nf_tickets_ticket SET starter_user_id = user_id  WHERE starter_user_id = 0 AND user_id <> 0');
    }

    public function upgrade2000303Step3()
    {
        $this->applyGlobalPermission('ticket', 'viewModerated', 'ticket', 'viewDeleted');
        $this->applyGlobalPermission('ticket', 'approveUnapprove', 'ticket', 'viewDeleted');
        $this->applyGlobalPermission('general', 'accessQueue', 'ticket', 'viewOthers');
    }

    public function upgrade2000303Step4()
    {
        $this->db()->query('UPDATE xf_nf_tickets_message SET message_state = ?  WHERE message_state = ?', ['hidden', 'moderated']);
    }

    public function upgrade2000303Step5()
    {
        $this->db()->query('
            INSERT IGNORE INTO xf_approval_queue (content_type, content_id, content_date)
            SELECT ?, ticket_id, start_date 
            FROM xf_nf_tickets_ticket
            WHERE discussion_state = ?
        ', ['nf_tickets_ticket', 'moderated']);

        $this->db()->query('
            DELETE FROM xf_approval_queue WHERE content_type = ? AND content_id NOT IN (SELECT ticket_id FROM xf_nf_tickets_ticket WHERE discussion_state = ?)
        ', ['nf_tickets_ticket', 'moderated']);
    }

    public function upgrade2000303Step6()
    {
        $this->db()->query('
            INSERT IGNORE INTO xf_approval_queue (content_type, content_id, content_date)
            SELECT ?, message_id, message_date 
            FROM xf_nf_tickets_message
            WHERE message_state = ?
        ', ['nf_tickets_message', 'hidden']);

        $this->db()->query('
            DELETE FROM xf_approval_queue WHERE content_type = ? AND content_id NOT IN (SELECT message_id FROM xf_nf_tickets_message WHERE message_state = ?)
        ', ['nf_tickets_message', 'hidden']);
    }

    public function upgrade2000303Step7()
    {
        $this->db()->query('UPDATE xf_deletion_log SET content_type = ?  WHERE content_type = ?', ['nf_tickets_ticket', 'ticket']);
    }

    public function upgrade2000303Step8()
    {
        $this->db()->query('UPDATE xf_nf_tickets_ticket_field SET display_group = ?  WHERE display_group = ?', ['below_message', 'inside_message']);
    }

	public function upgrade2010070Step1()
	{
		$this->applyGlobalPermission('ticket', 'viewEmail', 'ticket', 'manageAnyTicket');
	}

	public function upgrade2010070Step3()
	{
		$this->installStep1();
	}

	public function upgrade2010170Step1()
	{
		$sm = $this->schemaManager();

		$sm->alterTable('xf_nf_tickets_ticket_field_value', function (Alter $table)
		{
			$table->dropColumns(['is_expired']);
		});
	}

    public function upgrade2010570Step1()
    {
        $this->applyGlobalPermission('ticket', 'viewOwnPriority', 'ticket', 'view');
        $this->applyContentPermission('ticket', 'viewOwnPriority', 'ticket', 'view');
    }

    public function upgrade2030070Step1()
    {
    	$this->installStep1();
    }

    public function upgrade2030070Step2()
    {
        $this->applyGlobalPermission('ticket', 'warn', 'ticket', 'editAnyMessage');
        $this->applyContentPermission('ticket', 'warn', 'ticket', 'editAnyMessage');
    }

    public function upgrade2030105Step1()
    {
        $this->db()->query('
            UPDATE xf_thread as thread
            INNER JOIN xf_nf_tickets_ticket as ticket on ticket.thread_id = thread.thread_id
            SET thread.discussion_type = ? 
            WHERE thread.discussion_type = ? or thread.discussion_type = ?
        ', ['nf_tickets_ticket', '', 'discussion']);
    }

    public function upgrade2030105Step2()
    {
        $this->db()->query('
            UPDATE xf_thread 
            SET discussion_type = ? 
            WHERE discussion_type = ?
        ', ['nf_tickets_ticket', 'tickets']);
    }

    public function upgrade2030105Step3()
    {
        $this->db()->query('
            UPDATE xf_thread as thread
            LEFT JOIN xf_nf_tickets_ticket as ticket on ticket.thread_id = thread.thread_id
            SET thread.discussion_type = ? 
            WHERE thread.discussion_type = ? AND ticket.discussion_type is null
        ', ['discussion', 'nf_tickets_ticket']);
    }

	public function upgrade2040070Step1()
	{
		$params = [
			'rebuild_types' => [
				'nf_tickets_ticket',
				'nf_tickets_message',
				'nf_tickets_kb_article'
			],
		];

		$this->app->jobManager()->enqueueUnique(
			'nfTicketsSearchIndexRebuild',
			'XF:SearchRebuild',
			$params,
			false
		);
	}

	public function upgrade2060000Step1()
	{
		$this->installStep1();
	}

    public function postUpgrade($previousVersion, array &$stateChanges)
    {
        $jobList = [];

        if ($previousVersion && $previousVersion < 2000070)
        {
            /** @var \XF\Service\RebuildNestedSet $service */
            $service = \XF::service('XF:RebuildNestedSet', 'NF\Tickets:Category', [
                'parentField' => 'parent_category_id',
            ]);
            $service->rebuildNestedSetInfo();

            if ($previousVersion < 2000000)
            {
                $this->app->jobManager()->enqueueUnique(
                    'nfTicketsUpgradeLikeIsCountedRebuild_nf_tickets',
                    'XF:LikeIsCounted',
                    ['type' => 'ticket_message'],
                    false
                );
            }

	        $jobList[] = ['NF\Tickets:Ticket', []];
	        $jobList[] = ['NF\Tickets:Category', []];
        }

        if ($previousVersion > 2000000 && $previousVersion < 2000038)
        {
            $jobList[] = 'XF:PermissionRebuild';
        }

        if ($previousVersion && $previousVersion < 2000302)
        {
            $jobList[] = ['NF\Tickets:UserTicketCount', []];
        }

        if ($previousVersion > 2000302 && $previousVersion < 2010370)
        {
            $jobList[] = ['NF\Tickets:UserTicketCount', []];
            $jobList[] = ['NF\Tickets:Category', []];
        }

        if ($previousVersion < 2020570)
        {
            $jobList[] = ['NF\Tickets:Ticket', ['anonymized' => true]];
        }

        if ($jobList)
        {
            $this->app->jobManager()->enqueueUnique('nfTicketsPostUpgrade', 'XF:Atomic', [
                'execute' => $jobList,
            ]);
        }

        if ($previousVersion && $previousVersion < 2000301)
        {
            /** @var \XF\Repository\ApprovalQueue $approvalQueueRepo */
            $approvalQueueRepo = \XF::repository('XF:ApprovalQueue');
            $approvalQueueRepo->rebuildUnapprovedCounts();
        }

        /** @var \NF\Tickets\Repository\TicketField $ticketFieldRepo */
        $ticketFieldRepo = \XF::repository('NF\Tickets:TicketField');
        $ticketFieldRepo->rebuildFieldCache();

        /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
        $ticketRepo = \XF::repository('NF\Tickets:Ticket');
        $ticketRepo->rebuildTicketCounts();
    }

    public function uninstallStep1()
    {
        $sm = $this->schemaManager();

        foreach (\array_keys($this->getTables()) AS $tableName)
        {
            $sm->dropTable($tableName);
        }
    }

    public function uninstallStep2()
    {
        $sm = $this->schemaManager();

        foreach ($this->getRemoveAlterTables() AS $tableName => $callback)
        {
            if ($sm->tableExists($tableName))
            {
                $sm->alterTable($tableName, $callback);
            }
        }
    }

    public function uninstallStep3()
    {
        $db = $this->db();

        $contentTypes = ['nf_tickets_category'];

        $this->uninstallContentTypeData($contentTypes);

        $db->beginTransaction();

        $db->delete('xf_admin_permission_entry', "admin_permission_id = 'nfTickets'");
        $db->delete('xf_permission_cache_content', "content_type = 'nf_tickets_category'");
        $db->delete('xf_permission_entry', "permission_group_id = 'ticket'"); // TODO: rename and prefix?
        $db->delete('xf_permission_entry_content', "permission_group_id = 'ticket'"); // TODO: rename and prefix?

        $db->commit();
    }

    public function uninstallStep4()
    {
        $this->deletePhrases([
            'nf_tickets_ticket_prefix.%',
            'nf_tickets_ticket_prefix_group.%',
            'nf_tickets_status.%',
            'nf_tickets_priority.%',
        ]);
    }

    public function uninstallStep5()
    {
        $this->query("
			DELETE FROM xf_route_filter
			WHERE prefix = 'tickets'
			AND find_route = 'tickets/'
		");

        $this->app->repository('XF:RouteFilter')->rebuildRouteFilterCache();
    }

    protected function getTables()
    {
        $data = new MySql();

        return $data->getTables();
    }

    protected function getAlterTables()
    {
        $data = new MySql();

        return $data->getAlterTables();
    }

    protected function getRemoveAlterTables()
    {
        $data = new MySql();

        return $data->getRemoveAlterTables();
    }

    protected function getData()
    {
        $data = new MySql();

        return $data->getData();
    }

    protected function getPhrases()
    {
        $data = new MySql();

        return $data->getPhrases();
    }
}
