<?php

namespace NF\Tickets\XF\Entity;

use NF\Tickets\Entity\QueueWatch as TicketQueueWatch;
use XF\Mvc\Entity\Structure;

use function is_array;

/**
 * Extends \XF\Entity\User
 *
 * @property int                    $nf_tickets_count
 * @property int                    $nf_tickets_unread
 * @property-read array             $nf_ticket_stats
 * @property-read string|\XF\Phrase $nf_formatted_user_state
 * @property-read ?TicketQueueWatch $TicketQueueWatch
 */
class User extends XFCP_User
{
    /** @var int  */
    protected $ticketStatsCacheLifeTime = 604800; // 1 week

    /**
     * @param null $error
     * @return bool
     */
    public function canViewTickets(&$error = null)
    {
        return (bool)$this->hasPermission('ticket', 'view');
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canAccessTicketQueue(&$error = null)
    {
        if (!$this->user_id)
        {
            return false;
        }

        if (!$this->canViewTickets())
        {
            return false;
        }

        return (bool)$this->hasPermission('general', 'accessQueue');
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canViewTicketsOthers(&$error = null)
    {
        if (!$this->user_id)
        {
            return false;
        }

        return (bool)$this->hasPermission('ticket', 'viewOthers');
    }

    /**
     * @param null $error
     * @return bool
     */
    public function canCreateTicket(&$error = null)
    {
        return (bool)$this->hasPermission('ticket', 'create');
    }

	/**
	 * @param null $error
	 * @return bool
	 */
    public function canViewKB(&$error = null)
	{
		return (bool)$this->hasPermission('nfTicketsKB', 'view');
	}

	/**
	 * @param null $error
	 * @return bool
	 */
	public function canEditKBArticles(&$error = null)
	{
		return (bool)$this->hasPermission('nfTicketsKB', 'edit');
	}

	/**
	 * @param null $error
	 * @return bool
	 */
	public function canDeleteKBArticles(&$error = null)
	{
		return (bool)$this->hasPermission('nfTicketsKB', 'delete');
	}

    public function isShownCaptcha()
    {
        $forceCaptcha = $this->getOption('nf_force_captcha');
        if ($forceCaptcha !== null)
        {
            return (bool)$forceCaptcha;
        }

        return parent::isShownCaptcha();
    }

    /**
     * @param string $contentId
     * @param string $permission
     * @return bool|int
     */
    public function hasTicketCategoryPermission(string $contentId, string $permission)
    {
        return $this->PermissionSet->hasContentPermission('nf_tickets_category', $contentId, $permission);
    }

    /**
     * @param string $contentId
     * @param string $permission
     * @return bool|int
     */
    public function hasTicketKBCategoryPermission(string $contentId, string $permission)
    {
        return $this->hasPermission('nfTicketsKB', 'view');
    }

    /**
     * @param array|null $categoryIds
     */
    public function cacheTicketCategoryPermissions(array $categoryIds = null)
    {
        if (is_array($categoryIds))
        {
            \XF::permissionCache()->cacheContentPermsByIds($this->permission_combination_id, 'nf_tickets_category', $categoryIds);
        }
        else
        {
            \XF::permissionCache()->cacheAllContentPerms($this->permission_combination_id, 'nf_tickets_category');
        }
    }

    protected function getTicketStatCacheKey(): string
    {
        return 'nfTickets.userStats.' . $this->user_id;
    }

    public function invalidateTicketsStatsCache()
    {
        $cache = \XF::app()->cache();

        $session = \XF::session();
        if ($session->isStarted() &&
            $session->offsetExists('user_id') &&
            (int)$session->offsetGet('user_id') === \XF::visitor()->user_id)
        {
            $session->offsetUnset('nfTicketCounts'); // backup if there is no caching provider
        }
        $cacheKey = $this->getTicketStatCacheKey();
        if ($cache)
        {
            $cache->delete($cacheKey);
        }
    }

    public function updateTicketLastRead()
    {
        $stats = $this->nf_ticket_stats;
        $stats['lastRead'] = \XF::$time;
        $this->updateTicketStatsCache($stats);
    }

    protected function getNfTicketStats(): array
    {
        /** @var array $ticketCounts */
        $ticketCounts = \XF::app()->container('nfTicketCounts') ?: [];
        $lastBuilt = $ticketCounts['lastModified'] ?? 0;
        $stats = $this->readTicketStatsCache();

        if (($stats['lastBuilt'] ?? 0) >= $lastBuilt)
        {
            return $stats;
        }
        $lastRead = $stats['lastRead'] ?? 0;

        $stats = $this->rebuildUserTicketStats($lastBuilt, $lastRead);

        $this->updateTicketStatsCache($stats);

        return $stats ?? [];
    }

    protected function readTicketStatsCache(): array
    {
        $app = \XF::app();

        $session = $app->session();
        $cache = $app->cache();
        if ($cache)
        {
            $cacheKey = $this->getTicketStatCacheKey();
            $stats = $cache->fetch($cacheKey);
        }
        else if ($session->isStarted())
        {
            $stats = $session->get('nfTicketCounts');
        }
        else
        {
            $stats =  [];
        }

        return is_array($stats) ? $stats : [];
    }

    protected function updateTicketStatsCache(array $stats)
    {
        $app = \XF::app();
        $session = $app->session();
        $cache = $app->cache();
        if ($cache)
        {
            $cacheKey = $this->getTicketStatCacheKey();
            $cache->save($cacheKey, $stats, $this->ticketStatsCacheLifeTime);
        }
        else if ($session->isStarted())
        {
            $app->session()->set('nfTicketCounts', $stats);
        }

        $this->_getterCache['nf_ticket_stats'] = $stats;
    }

    protected function rebuildUserTicketStats(int $lastModified, int $lastRead): array
    {
        /** @var \NF\Tickets\Repository\Category $categoryRepo */
        $categoryRepo = $this->repository('NF\Tickets:Category');
        $categories = $categoryRepo->getViewableCategories();

        /** @var \NF\Tickets\Finder\Ticket $ticketsFinder */
        $ticketsFinder = $this->app()->finder('NF\Tickets:Ticket');
        $tickets = $ticketsFinder->isCounted()
                                 ->applyCategoryVisibilityChecks($categories->keys())
                                 ->fetch()
                                 ->filterViewable();

        $total = 0;
        $assigned = 0;
        /** @var \NF\Tickets\Entity\Ticket $ticket */
        foreach ($tickets as $ticket)
        {
            $total++;
            if ($ticket->assigned_user_id === $this->user_id)
            {
                $assigned++;
            }
        }

        return [
            'total'     => $total,
            'assigned'  => $assigned,
            'lastBuilt' => $lastModified,
            'lastRead'  => $lastRead,
        ];
    }

	public function getNfFormattedUserState(User $user = null)
	{
		if ($user === null)
		{
			$user = $this;
		}

		switch ($user->user_state)
		{
			case 'valid': return \XF::phrase('valid');
			case 'email_confirm': return \XF::phrase('awaiting_email_confirmation');
			case 'email_confirm_edit': return \XF::phrase('awaiting_email_confirmation_from_edit');
			case 'email_bounce': return \XF::phrase('email_invalid_bounced');
			case 'moderated': return \XF::phrase('awaiting_approval');
			case 'rejected': return \XF::phrase('rejected');
			case 'disabled': return \XF::phrase('disabled');
			default: return $user->user_state;
		}
	}

    protected function _postSave()
    {
        parent::_postSave();

        if ($this->isChanged(['user_group_id', 'secondary_group_ids', 'is_moderator', 'is_admin', 'is_staff']))
        {
            $this->invalidateTicketsStatsCache();
        }
    }

    protected function _postDelete()
    {
        parent::_postDelete();

        $this->invalidateTicketsStatsCache();
    }

    /**
     * @param Structure $structure
     * @return Structure
     */
    public static function getStructure(Structure $structure)
    {
        $structure = parent::getStructure($structure);

		$structure->relations['TicketQueueWatch'] = [
		    'entity' => 'NF\Tickets:QueueWatch',
		    'type' => self::TO_ONE,
		    'conditions' => 'user_id',
			'primary' => true
	    ];

        $structure->options['nf_is_banned'] = null;
        $structure->options['nf_force_captcha'] = null;
		$structure->options['nfTickets'] = true;
		$structure->columns['nf_tickets_count'] = ['type' => self::UINT, 'default' => 0, 'forced' => true, 'changeLog' => false];
		$structure->columns['nf_tickets_unread'] = ['type' => self::UINT, 'default' => 0, 'forced' => true, 'changeLog' => false];
		$structure->getters['nf_ticket_stats'] = ['getter' => 'getNfTicketStats', 'cache' => true];
		$structure->getters['nf_formatted_user_state'] = ['getter' => 'getNfFormattedUserState', 'cache' => true];

        return $structure;
    }
}