<?php

namespace NF\Tickets\XF\Pub\Controller;

use XF\Mvc\ParameterBag;
use XF\Mvc\Reply\AbstractReply;

/**
 * Extends \XF\Pub\Controller\Watched
 */
class Watched extends XFCP_Watched
{
    protected $ticketAction = false;
    protected function preDispatchType($action, ParameterBag $params)
    {
        $actionL = \preg_replace('#[^a-z0-9]#i', ' ', $action);
        $actionL = \str_replace(' ', '', \strtolower($actionL));
        if (\preg_match('#^ticket#', $actionL))
        {
            $this->ticketAction = true;
            /** @var \NF\Tickets\ControllerPlugin\Ticket $plugin */
            $plugin = $this->plugin('NF\Tickets:Ticket');
            $plugin->preDispatchForTicket($this->isDiscouraged(), $action, $params);
        }

        parent::preDispatchType($action, $params);
    }

    protected function postDispatchType($action, ParameterBag $params, AbstractReply &$reply)
    {
        if ($this->ticketAction)
        {
            /** @var \NF\Tickets\ControllerPlugin\Ticket $plugin */
            $plugin = $this->plugin('NF\Tickets:Ticket');
            $plugin->postDispatchForTicket($action, $params, $reply);
        }

        parent::postDispatchType($action, $params, $reply);
    }

    /** @var null|bool */
    protected $isDiscouraged = null;

    protected function isDiscouraged()
    {
        if ($this->isDiscouraged !== null)
        {
            return $this->isDiscouraged;
        }
        $this->isDiscouraged = parent::isDiscouraged();

        return $this->isDiscouraged;
    }

    /**
     * @return void
     * @throws \XF\Mvc\Reply\Exception
     */
    protected function assertCanViewTickets()
    {
        if (!$this->ticketAction)
        {
            throw $this->exception($this->noPermission());
        }
    }

    public function actionTicketCategories()
    {
        $this->assertCanViewTickets();
        $this->setSectionContext('nfTickets');

        $watchedFinder = $this->finder('NF\Tickets:CategoryWatch');
        $watchedCategories = $watchedFinder->where('user_id', \XF::visitor()->user_id)
                                           ->keyedBy('ticket_category_id')
                                           ->fetch();

        /** @var \NF\Tickets\Repository\Category $categoryRepo */
        $categoryRepo = $this->repository('NF\Tickets:Category');
        $categories = $categoryRepo->getViewableCategories();
        $categoryTree = $categoryRepo->createCategoryTree($categories);
        $categoryExtras = $categoryRepo->getCategoryListExtras($categoryTree);

        $viewParams = [
            'watchedCategories' => $watchedCategories,

            'categoryTree'   => $categoryTree,
            'categoryExtras' => $categoryExtras,
        ];

        return $this->view(
            'NF\Tickets:Watched\Categories',
            'nf_tickets_watched_ticket_categories',
            $viewParams
        );
    }

    public function actionTicketCategoriesUpdate()
    {
        $this->assertCanViewTickets();
        $this->assertPostOnly();
        $this->setSectionContext('nfTickets');

        /** @var \NF\Tickets\Repository\CategoryWatch $watchRepo */
        $watchRepo = $this->repository('NF\Tickets:CategoryWatch');

        $inputAction = $this->filter('watch_action', 'str');
        $action = $this->getCategoryWatchActionConfig($inputAction, $config);

        if ($action)
        {
            $visitor = \XF::visitor();

            $ids = $this->filter('ids', 'array-uint');
            $categories = $this->em()->findByIds('NF\Tickets:Category', $ids);

            /** @var \NF\Tickets\Entity\Category $category */
            foreach ($categories AS $category)
            {
                $watchRepo->setWatchState($category, $visitor, $action, $config);
            }
        }

        return $this->redirect(
            $this->getDynamicRedirect($this->buildLink('watched/ticket-categories'))
        );
    }

    protected function getCategoryWatchActionConfig($inputAction, array &$config = null)
    {
        $config = [];

        $parts = \explode(':', $inputAction, 2);

        $inputAction = $parts[0] ?? '';
        $boolSwitch = ($parts[1] ?? '') === 'on';

        switch ($inputAction)
        {
            case 'send_email':
            case 'send_alert':
            case 'include_children':
                $config = [$inputAction => $boolSwitch];

                return 'update';

            case 'delete':
                return 'delete';

            default:
                return null;
        }
    }

    public function actionTicketUsers()
    {
        $this->assertCanViewTickets();
        $this->setSectionContext('nfTickets');
    }

    public function actionTickets()
    {
        $this->assertCanViewTickets();
        $this->setSectionContext('nfTickets');

        $page = $this->filterPage();
        $perPage = 20;

        /** @var \NF\Tickets\ControllerPlugin\TicketList $ticketListPlugin */
        $ticketListPlugin = $this->plugin('NF\Tickets:TicketList');
        $categoryParams = $ticketListPlugin->getCategoryListData();

        $categoryIds = $categoryParams['categories']->keys();

        /** @var \NF\Tickets\Repository\Ticket $ticketRepo */
        $ticketRepo = $this->repository('NF\Tickets:Ticket');
        $finder = $ticketRepo->findTicketsForWatchedList($categoryIds);

        $total = $finder->total();
        $tickets = $finder->limitByPage($page, $perPage)->fetch();

        $viewParams = [
            'page'    => $page,
            'perPage' => $perPage,
            'total'   => $total,
            'tickets' => $tickets->filterViewable(),
        ];

        return $this->view(
            'NF\Tickets:Watched\Tickets',
            'nf_tickets_watched_tickets',
            $viewParams
        );
    }

    public function actionTicketsManage()
    {
        $this->assertCanViewTickets();
        $this->setSectionContext('nfTickets');

        if (!$state = $this->filter('state', 'str'))
        {
            return $this->redirect($this->buildLink('watched/tickets'));
        }

        if ($this->isPost())
        {
            /** @var \NF\Tickets\Repository\TicketWatch $ticketWatchRepo */
            $ticketWatchRepo = $this->repository('NF\Tickets:TicketWatch');

            if ($action = $this->getTicketWatchActionConfig($state, $updates))
            {
                $ticketWatchRepo->setWatchStateForAll(\XF::visitor(), $action, $updates);
            }

            return $this->redirect($this->buildLink('watched/tickets'));
        }
        else
        {
            $viewParams = [
                'state' => $state,
            ];

            return $this->view('NF\Tickets:Watched\TicketsManage', 'nf_tickets_watched_tickets_manage', $viewParams);
        }
    }

    public function actionTicketsUpdate()
    {
        $this->assertCanViewTickets();
        $this->assertPostOnly();
        $this->setSectionContext('nfTickets');

        /** @var \NF\Tickets\Repository\TicketWatch $watchRepo */
        $watchRepo = $this->repository('NF\Tickets:TicketWatch');

        $inputAction = $this->filter('watch_action', 'str');
        $action = $this->getTicketWatchActionConfig($inputAction, $config);

        if ($action)
        {
            $ids = $this->filter('ids', 'array-uint');
            $tickets = $this->em()->findByIds('NF\Tickets:Ticket', $ids);
            $visitor = \XF::visitor();

            /** @var \NF\Tickets\Entity\Ticket $ticket */
            foreach ($tickets AS $ticket)
            {
                $watchRepo->setWatchState($ticket, $visitor, $action, $config);
            }
        }

        return $this->redirect(
            $this->getDynamicRedirect($this->buildLink('watched/tickets'))
        );
    }

    protected function getTicketWatchActionConfig($inputAction, array &$config = null)
    {
        $config = [];

        switch ($inputAction)
        {
            case 'email_subscribe:on':
                $config = ['email_subscribe' => 1];

                return 'update';

            case 'email_subscribe:off':
                $config = ['email_subscribe' => 0];

                return 'update';

            case 'delete':
                return 'delete';

            default:
                return null;
        }
    }
}